# SpectrumProfile.R
#
# The spectral image is assumed to be in a .png image file.
#
# K. Myneni, 2013-12-1
#
# Revisions:
#   2013-12-16  km  factored code into several functions.

## We need to use the R package, png, to read image files.
require("png")

## Sum the total color intensity for a range of rows in an
##   image array
sum_rows <- function( img, start, stop ) {
	size = dim(img)
	nrows = size[1]
	ncols = size[2]
	ndepth = size[3]
	rsum = rep(0, ncols)
	for (i in start:stop) {
	  rred   = img[i, 1:ncols, 1]
	  rgreen = img[i, 1:ncols, 2]
	  rblue  = img[i, 1:ncols, 3]
	  rsum = rsum + rred + rgreen + rblue
	}
	return( rsum )
}

## Return the spectral line profile for a set of rows from a
##   PNG image file.
GetLineProfile <- function( filename, start_row, stop_row ) {
	img <- readPNG( filename )    ## Read the image file
	return( sum_rows( img, start_row, stop_row ) )
}

## Create a calibrated spectrum from the line profile.
## The slope is in units of nm/pixel, and yint has units of nm
MakeCalibratedSpectrum <- function( row_sum, slope, yint ) {
	np = length( row_sum )
	xscale=seq(1:np)*slope + yint  ## make wavelength scale
	d <- array(c(xscale, row_sum), c(np, 2))  ## make 2-D spectrum array
	return( d )
}

## Plot spectrum
PlotSpectrum <- function( spectrum ) {
	plot(spectrum[,1], spectrum[,2], type="l", col="blue",
	  xlab="Wavelength", ylab="Intensity")
}

## Save spectrum to a file
SaveSpectrum <- function( spectrum, filename ) { 
	write( t(spectrum), ncolumns=2, file=filename )
}


## Example of using the above functions:
## Assume the image of the spectrum is stored in the file sodium.png,
## and we want to compute, plot, and save the spectral profile obtained
## by summing rows 100 to 300. For now, we want the horizontal axis to
## be just the pixel number (column number).

#   p = GetLineProfile( "sodium.png", 100, 300 )
#   s = MakeCalibratedSpectrum( p, 1, 0 )   # x scale is pixel number
#   PlotSpectrum( s )
#   SaveSpectrum( s, "sodium_r100-300.dat" )





