# free-wavepacket.R
#
# Compute the time evolution of a free Gaussian wavepacket
# by numerically solving the 1D time-dependent wave equation,
# using finite difference equations.
#
# K. Myneni, 2013-02-27
#
# Notes:
#
#  The wavefunction at t=0 is given by,
#
#     psi(x,0) = exp(i*k0*x)*exp(-x^2/(2*sig0^2))
#
hbar = 1.0545715964e-34    ## Planck's const over 2pi
m_e = 9.10938188e-31       ## mass of electron

D = 1i*hbar/(2*m_e)
	
x1 = 10
dx = 0.01
x = seq(-x1, x1, by = dx)
nx = length(x)

library(compiler)

# Normalize a wave function array so that the probability density
# integrates to 1
normalize <- function(psi) {
	p = dx*sum(abs(psi)^2)
	return( psi/sqrt(p) )
}

nrmlz <- cmpfun(normalize)

k0 = 2*pi/0.2
sig0 = 0.2

psi_0 = normalize( exp(1i*k0*x)*exp(-x^2/(2*sig0)^2) )
psi_i = psi_0

dt =  0.02
k  =  D*dt/dx^2

# Evolve the solution by n time steps, i.e. by an elapsed time
#   of n*dt

evolve <- function(n) {
  for( i in 1:n ) {
    psi_ip1 = psi_i
    for( j in 2:(nx-1) ) {
        psi_ip1[j] = k*( psi_i[j+1] - 2*psi_i[j] + psi_i[j-1] ) + psi_i[j]
    }

    psi_ip1[1]  = psi_ip1[2]
    psi_ip1[nx] = psi_ip1[nx-1]

    psi_i <<- normalize( psi_ip1 )
  }
}

evlv <- cmpfun(evolve)

plot(x, abs( psi_0 )^2, type="l", col="black", xlim=c(-2,4), ylim=c(-1.5,2))
lines(x, Re( psi_0 ), col="red")
evlv(15000)
lines(x, abs( psi_i )^2, col="blue")


	
	
