# jones.R
#
# Tools for the Jones Calculus
#
# Trace the polarization state of a beam of light, represented by a Jones vector, 
# through a series of optical elements (linear polarizers, half and quarter wave 
# plates, etc.). The output Jones vector is computed by multiplying the input vector 
# by the Jones matrix of each optical element, successively and in the order in which 
# the beam encounters the element.
#
# Copyright (c) 2008, K. Myneni
# Provided under the GNU General Public License
#
# Revisions:
#   2008-08-29  first version
#
# Notes and References:
#
# [1] This code was written for the free computing environment, R 
#     (see www.r-project.org ).
#
# [2] The Jones vector representation of the polarization of light is valid only
#     for completely polarized light. Partially polarized light requires the
#     Mueller calculus, which is not handled here.
#
# [3] The beam of light is implicitly assumed to be monochromatic, owing to the
#     fact that the effect on polarization by certain optical elements depends
#     strongly on the wavelength, e.g. while linear polarizers may work over a
#     broad wavelength band, wave plates in particular are designed to provide
#     a fixed retardation (half-wave, quarter-wave) at a specific wavelength.
#
# [4] The following coordinate system, and beam propagation direction, must be
#     used in specifying the input Jones vector, and for interpreting the output
#     Jones vector. The beam is assumed to be propagating in the +z direction.
#     The coordinate system is the standard, right-handed, three-dimensional
#     Cartesian system:
#
#                                ^  +y-axis
#                                |
#                                |
#                               / ------> +x-axis
#                              /
#                             / +z-axis
#
#     If the above coordinate system is not used, the orientation and helicity of 
#     the polarization state will be hard to interpret for the output vector. 
#
# [5] Report any bugs, problems, or suggestions by email to krishnamyneni@bellsouth.net
#
# [6] References:
#
#	a) http://en.wikipedia.org/wiki/Jones_calculus
#	b) Introduction to Optics, 2nd ed., F. Pedrotti and L. Pedrotti
#
# Examples:
#
#  1) Linearly-polarized light, polarized horizontally (along x), passes through
#     a linear polarizer with its transmission axis (T.A.) oriented at 60 degrees 
#     (counter clock-wise from the x-axis). Find the output Jones vector and the 
#     relative beam intensity.
#
#	J_out = M_lp( deg_to_rad(60) ) %*% J_lp_x   # calculate output vector
#       J_out                                       # print output vector
#       intensity( J_out )                          # fractional output intensity
#
#  2) Linearly-polarized light along x, passing through a half-wave plate with its
#     fast axis (F.A.) along 45 degrees:
#
#	M_hwp( deg_to_rad(45) ) %*% J_lp_x          
#
#     will give the Jones vector for light linearly-polarized vertically (along y)
#
#  3) Linearly-polarized light along x, passing through a quarter-wave plate with its
#     F.A. along 45 degrees:
#
#	M_qwp( deg_to_rad(45) ) %*% J_lp_x
#
#     will give left-circularly polarized light.
#
#  4) Vertically linearly-polarized light passes through a sequence of optical elements:
#     a) a quarter-wave plate with its F.A. along 45 degrees, b) a quarter-wave plate
#     with its F.A. along 135 degrees, followed by c) a linear polarizer with its T.A. in
#     the vertical position.
#
#	M_lp_y %*% M_qwp(deg_to_rad(135)) %*% M_qwp(deg_to_rad(45)) %*% J_lp_y
#
#     The light is transmitted, and has vertical linear polarization. Note the order
#     of multiplication of Jones matrices. Change the first quarter-wave plate F.A. to
#     135 degrees, and see what happens.


# Utility

deg_to_rad <- function( deg ) { return( pi*deg/180 ) }
intensity <- function( v ) { return( Re( t(Conj(v))%*%v)[1]) }

# Rotation matrix

R <- function( theta ) {
	m <- matrix( c( cos(theta), -sin(theta), 
			sin(theta),  cos(theta) ),
			nrow=2, byrow=TRUE )
	return( m )
}

# Rotate Jones vector
RotateJonesVector <- function( v, theta ) { return( R(theta)%*%v ) }

# Rotate Jones matrix
RotateJonesMatrix <- function( m, theta ) { return( R(-theta) %*% m %*% R(theta) ) }

## Standard Jones vectors

J_lp_x <- c(1, 0)             # linearly-polarized light along horizontal axis (x)
J_lp_y <- c(0, 1)             #   "                  "         vertical axis (y)
J_lp_pi4 <- c(1, 1)/sqrt(2)   #   "                  "         45 degrees between x and y

J_rc <- c(1, -1i)/sqrt(2)     # right-circularly polarized light
J_lc <- c(1,  1i)/sqrt(2)     # left-circularly polarized light

## Standard Jones matrices

# Linear polarizer
M_lp_x <- matrix( c(1, 0, 0, 0), nrow=2, byrow=TRUE)     # T.A. along x
M_lp_y <- matrix( c(0, 0, 0, 1), nrow=2, byrow=TRUE)     # T.A. along y
M_lp_pi4 <- matrix( c(1, 1, 1, 1)/2, nrow=2, byrow=TRUE) # T.A. along 45 deg

# Half-wave plate
M_hwp_x <- matrix( c(-1i, 0, 0, 1i), nrow=2, byrow=TRUE)  # F.A. along x
M_hwp_y <- matrix( c( 1i, 0, 0, -1i), nrow=2, byrow=TRUE) # F.A. along y

# Quarter-wave plate
M_qwp_x <- matrix( c(1, 0, 0, 1i)*(1 + 1i)/sqrt(2), nrow=2, byrow=TRUE)  # F.A. along x


## Jones matrices of optical elements at a specified angle

# Linear Polarizer
M_lp <- function( theta ) {
	m <- matrix( c( cos(theta)^2, cos(theta)*sin(theta),
			sin(theta)*cos(theta), sin(theta)^2 ),
		nrow=2, byrow=TRUE )
	return( m )
}

# Half-wave plate
M_hwp <- function( theta ) { return( RotateJonesMatrix( M_hwp_x, theta ) ) }

# Quarter-wave plate
M_qwp <- function( theta ) { return( RotateJonesMatrix( M_qwp_x, theta ) ) }

