# diffusion-fast.R
#
# Numerically solve the 1-D diffusion equation using the method
# of finite differences, with inline compiled Fortran code:
#
#    u_t = D*u_xx
#
# K. Myneni, 2013-02-26
#
# Notes:
#
#  1) The function at t=0 is given by,
#
#        u(x,0) = 1
#
#      The boundary conditions are:
#
#        u(10, t) = 20  for t > 0
#        u_x(0, t) = 0
#
#  2) This program requires the R package, "inline". Type,
#     "install.packages()" from within R to install extra
#     packages.
#
# References:
#
# 1. S.J. Farlow, Partial Differential Equations for Scientists
#    and Engineers, Dover Publications (1982); see Lesson 38.

D = 0.5      # Diffusion coefficient
u_ext = 20

dx = 0.01
dt = 4e-5
x = seq(0, 10, by=dx)
nx = length(x)

u_x_0 = rep(1, nx)
u_i = u_x_0

k = D*dt/dx^2  # choose dt so that k < 0.5

# Evolve the solution by n time steps, i.e. by an elapsed time
#   of n*dt

# The equivalent R code, which is much slower, since it is not
# compiled.
#
# evolve <- function(n) {
#  u_ip1 = u_i
#  for( i in 1:n ) {
#    for( j in 2:(nx-1) ) {
#	u_ip1[j]  = k*( u_i[j+1] - 2*u_i[j] + u_i[j-1] ) + u_i[j]
#    }
#    u_ip1[1]  = u_ip1[2]   # b.c.: u_x(0, t) = 0
#    u_ip1[nx] = u_ext      # b.c.: u(10, t) = u_ext
#
#    u_i <<- u_ip1
#  }
# }


library(inline)

code <- "
        double precision, pointer, dimension (:) :: u_ip1
        integer :: i, j, k, nn, nnx, nnx2
        nn = n(1)
        nnx = nx(1)
        nnx2 = nnx - 1

        allocate( u_ip1(nnx) )

        do k = 1, nnx
          u_ip1(k) = u_i(k)
        end do

        do i = 1, nn
          do j = 2, nnx2
            u_ip1(j) = kscale(1)*(u_i(j+1) - 2*u_i(j) + u_i(j-1))
            u_ip1(j) = u_ip1(j) + u_i(j)
          end do

          u_ip1(1) = u_ip1(2)
          u_ip1(nnx) = u_ext(1)

          do k = 1, nnx
            u_i(k) = u_ip1(k)
          end do

       end do

       deallocate( u_ip1 )
"

evlv <- cfunction( signature(n="integer", nx="integer", kscale="numeric",
	  u_ext="numeric", u_i="numeric"), code, language="F95",
          convention=".Fortran")

plot(x, u_x_0, type="l", col="blue", 
  main="Solution of Diffusion Equation",
  ylim=c(0,20)
)

# Compute the solution at t=1
n = 25000
u_i = evlv(n, nx, k, u_ext, u_i)$u_i
lines(x, u_i, col="red")




	
	
