# diffusion.R
#
# Numerically solve the 1-D diffusion equation using the method
# of finite differences:
#
#    u_t = D*u_xx
#
# K. Myneni, 2013-02-26
#
# Notes:
#
#  The function at t=0 is given by,
#
#     u(x,0) = 1
#
#  The boundary conditions are:
#
#     u(10, t) = 20  for t > 0
#     u_x(0, t) = 0
#
# References:
#
# 1. S.J. Farlow, Partial Differential Equations for Scientists
#    and Engineers, Dover Publications (1982); see Lesson 38.

D = 0.5      # Diffusion coefficient
u_ext = 20

dx = 0.01
dt = 4e-5
x = seq(0, 10, by=dx)
nx = length(x)

u_x_0 = rep(1, nx)
u_i = u_x_0

k = D*dt/dx^2  # choose dt so that k < 0.5

# Evolve the solution by n time steps, i.e. by an elapsed time
#   of n*dt

evolve <- function(n) {
  u_ip1 = u_i
  for( i in 1:n ) {
    for( j in 2:(nx-1) ) {
	u_ip1[j]  = k*( u_i[j+1] - 2*u_i[j] + u_i[j-1] ) + u_i[j]
    }
    u_ip1[1]  = u_ip1[2]   # b.c.: u_x(0, t) = 0
    u_ip1[nx] = u_ext      # b.c.: u(10, t) = u_ext

    u_i <<- u_ip1
  }
}

library(compiler)
evlv <- cmpfun(evolve)

plot(x, u_x_0, type="l", col="blue", 
  main="Solution of Diffusion Equation",
  ylim=c(0,20)
)

# Compute the solution at t=1
evlv(25000)
lines(x, u_i, col="red")




	
	
