# pde-example.R
#
# Numerically solve the partial differential equation,
#
#   C_t = D*C_xx
#
# where C_t is the partial derivative of C(x,t) with respect to
# time and C_xx is the second partial derivative of C(x,t)
# w.r.t. position. 
#
# K. Myneni, 2013-02-23
#
# Revisions:
#   2013-02-26  km; provide functions get_Cx() and plot_Cx()
#                   to retrieve and plot C(x, t=constant) for 
#                   one of the discrete time values in the solution.
# Notes:
# -----
#
# 1) This is the one-dimensional diffusion equation with
#    a constant "diffusion coefficient", D [1,2].
#
# 2) The function C(x,t) is defined over the interval x=[0,10],
#    with the following boundary conditions:
#
#       a) C_x (x=0, t) = 0
#
#       b) C (x=10, t) = 100
#
# 3) This R program requires the R package, "ReacTran" (and its
#    dependencies). To install the package(s), within R, type
#
#        install.packages()
#
# 4) This example is a simplified version of the parabolic pde
#    example given in section 2 of the documentation for the 
#    ReacTran package [3].
#
# References:
# ----------
#
# 1. Wikipedia article on Diffusion Equation:
#    http://en.wikipedia.org/wiki/Diffusion_equation
#
# 2. Wikipedia article on the Heat Equation:
#    http://en.wikipedia.org/wiki/Heat_equation
#
# 3. K. Soetaert and F. Meysman, "Solving partial differential 
#    equations, using R package ReacTran", 2010.
#    http://cran.r-project.org/web/packages/ReacTran/vignettes/PDE.pdf

require(ReacTran)

D = 0.5
Cext = 20   # C(x = 10, t)

# Setup a 1000 point grid over the interval [x=0,10]
Grid = setup.grid.1D(N=1000, L=10)

# Define the PDE and set its b.c.s 
# (the b.c., C_x(0, t) = 0, is used by default)

pde1D <- function(t, C, parms) {
  list( tran.1D(C=C, D=D, C.down=Cext, dx=Grid)$dC )
}

# Initial function at t=0: C(x,0) = 1
C_x_0 = rep(1, Grid$N)

# Discrete times at which to sample C(x,t)
times = seq(0,100,by=1)

# Solve the PDE at the discrete time values in "times".
# The C(x,t) values are stored in the matrix "out". Each row of
# "out" contains the C(x) profile for a given time, with row 1
# corresponding to t=0. The first column in the row contains a
# label.
out = ode.1D(y=C_x_0, times=times, func=pde1D, parms=NULL, nspec=1)

# Retrieve a vector containing C(x, t=const) at the specified time
# This function assumes the discrete times start at zero, and step by
# delta t = 1.
get_Cx <- function( t ) { return( out[round(t)+1, 2:Grid$N] ) }

# Plot the function C(x, t=const) at the specified time.
plot_Cx <- function( t ) {
	x = Grid$x.mid[2:Grid$N]
	plot(x, get_Cx(t), type="l")
}

# png(filename="pde-example.png")

image(out, grid=Grid$x.mid,
   xlab="time (min)", 
   ylab="Distance (cm)",
   main = "Diffusion Equation: C_t = D*C_xx",
   add.contour = TRUE )

# dev.off()


