# qm6.R
#
# The Pauli Spin Matrices, Spinors, and Measurement Probabilities
#
# Copyright (c) 2003--2004 Krishna Myneni, krishna.myneni@ccreweb.org
# Provided under the GNU General Public License
#
# Use the spin state of an electron (or any fermion) to numerically 
# illustrate the fundamental principles of quantum mechanics:
#
#     a) measurement of a physical quantity is represented by an operator.
#     b) state of the system (particle) can be represented as a vector in
#          an N-dimensional space[1].
#     c) eigenvalues of an operator represent possible measurement 
#          values for the measurement associated with that operator.
#     d) probability of measuring a particular eigenvalue is
#          related to the projection of the state vector onto the
#          eigenvector corresponding to the particular eigenvalue.
#
# [1]  In the simple case of fermion spin, N=2 because there are only two
#      possible outcomes for the measurement of the particle's spin along 
#      *any* axis. The spin state of a fermion is represented by an ordered
#      set of 2 complex numbers, which are the amplitudes along two basis
#      vectors. The basis vectors for fermion spin are *chosen* to be the 
#      two eigenvectors of the "sz" operator.
#
#
# Suggested reading:
#
#     1) Quantum Physics of Atoms, Molecules, Solids, Nuclei, and
#        Particles, by R. Eisberg and R. Resnick, 2nd ed., Wiley 1985.
#        (sections 8-1 through 8-3; college undergraduate level).
#
#     2) Introduction to Quantum Mechanics, by D. J. Griffiths,
#        Prentice Hall 1995. (section 4.4; advanced undergrad level).
#
#     3) Quantum Mechanics, by. E. Merzbacher, 3rd ed., Wiley 1998.
#        (chapter 16; graduate course level).
#
# -----------------------------------------------------------------------
#
# Examples:
#
#  sx, sy, and sz are the three predefined Pauli spin matrices (operators).
#  Note that the actual spin operators (S_x, S_y, and S_z) are the corresponding 
#  Pauli matrices (sx, sy, sz) multiplied by hbar/2. The factor of hbar/2 will
#  therefore be absent in the calculation of eigenvalues and expectation values.
#  "one" is the predefined identity matrix.
#
#  sx                   print the sigma x Pauli spin matrix
#  sy                                   y
#  sz                                   z
#  sx %*% sy            multiply sx and sy and print the result
#  adjoint( sx )        print the adjoint of sx 
#  commutator( sx, sy ) compute and print the commutator of sx and sy
#  eigenvalues( sz )    compute and print the eigenvalues of sz
#  eigenvectors( sz )   compute and print the eigenvectors of sz
#  x1 = c( 1, 0 )       define the spinor x1 with components 1+i0, 0+i0
#  x1                   print the spinor x1
#  adjoint(x1) %*% x1   compute and print inner product <x1|x1>
#  inner_product( x1, x1 )  same as above: <x1|x1>, except that it returns
#                           a complex number rather than 1x1 matrix.
#  sx %*% x1             apply sx operator to vector x1 and show new vector
#  adjoint(x1) %*% sx %*% x1   compute and print <x1|sx|x1>, the expectation
#                               value of operator sx for the state x1.
#  inner_product( x1,  sx %*% x1 )  same as above: <x1|sx|x1>, except it returns
#                                   a complex number.
#  expectation_value( sx, x1 ) same as above: <x1|sx|x1>, except return is a
#                              real number.
#  x2 = c( 1, 1i )           define spinor x2 with components 1+i0, 0+i1
#  normalize( x2 )           print the normalized spinor
#  x2 = normalize( x2 )     set x2 = normalized(x2)
#  inner_product( x1, x2 )  compute and print projection of x2 along x1
#  abs( inner_product( x1, x2 ) )^2   print the probability of finding x1 given x2
#  ev = eigenvectors( sx )   compute and store eigenvectors of sx
#  inner_product( ev[,1], ev[,2] ) verify orthogonality of the eigenvectors of sx
#  ev = eigenvectors( sz )  eigenvectors of sz
#  e_up = ev[,1]            eigenvector for spin up (along z)
#  e_down = ev[,2]          eigenvector for spin down (along z)
#  x3 = c( 1i*3, 4 )        define spinor x3 with components 0+i3, 4
#  x3 = normalize( x3 )     normalize spinor x3
#  abs( inner_product(e_down, x3) )^2  print probability of measuring
#                           spin-down along z-axis for particle in state x3
#  P_down( x3 )             same as above
#  abs( inner_product(e_up, x3) )^2    print probability of measuring
#                           spin-up along z-axis for particle in state x3
#  P_up( x3 )               same as above
#  P_up( x3 ) + P_down( x3 )  should be 1
#  s1 = (sx + sz)/sqrt(2)   create a new operator s1 
# ----------------------------------------------------------------------
#		
# Exercises:
#
# 1. Show that the Pauli spin matrices are unitary, i.e.
#
#	sx*sx = sy*sy = sz*sz = one
#
# 2. Show that the spin matrices are Hermitian, i.e.
#
#	adjoint(si) = si
#
#       where adjoint means take the complex conjugate, then the transpose, 
#       and i = x, y, or z
#
# 3. Verify that the sx, sy, and sz matrices do not commute with each other,
#    and that they satisfy the commutation relations for angular momentum:
#
#	[sx, sy] = c*i*sz
#	[sy, sz] = c*i*sx
#	[sz, sx] = c*i*sy
#
#    where c is a real constant.
#
# 4. Determine the eigenvectors and eigenvalues for sx, sy, and sz.
#
# 5. Given the spin state
#
#		      / \
#		     | 1 |
#		x1 = |   |
#		     | 0 |
#		      \ /
#
#    compute the probability of measuring +1 ("spin up") and the 
#    probability of measuring -1 ("spin down") along the x, y, 
#    and z axes.
#
#    Hint: Apply the appropriate measurement operator (sx, sy, or sz)
#          to the input state; then find the projection of the resulting
#          vector onto the spin up or spin down eigenvector of the
#          operator. 
#
#    Partial Answer: The probability of measuring spin down along the
#                    y-axis is 0.5, or 50%.
#
#
# 6. Normalize the following spin states and compute their spin up
#    and spin down measurement probabilities for sx, sy, and sz.
#
#	      / \     / \     /  \     /     \     /      \
#	     | 0 |   | 1 |   | -1 |   | 1 - i |   | -1 + i |
#	x1 = | 	 |   |   |   |    |   |       |   |        |
#	     | 1 |,  | 1 |,  |  1 |,  | 1 + i |,  |  1 + i | 
#	      \ /     \ /     \  /     \     /     \      /
#
#
# 7. Compute the expectation values for sx, sy, and sz for the
#    normalized spin states of exercises 5 and 6.
#
#
# 8. Compute the probability of measuring spin down for the input
#      state of problem 5, along an axis in the x-z plane which is
#      45 degrees from both the z and x axes.
#
#    Hint 1: The spin measurement operator along an axis in the x-z
#            plane is the following linear combination of sx and sz 
#            operators:
#
#               s_theta = sin(theta)*sx + cos(theta)*sz
#
#            where theta is the angle between the measurement axis and
#            the z-axis. For this problem, theta is 45 degrees (don't
#            forget to convert to radians when computing the new operator). 
#
#   Answer: The probability of observing the particle with spin down along the 
#           45 degree axis in the x-z plane is 14.6%. 
#
#  If you are able to arrive at the answer to exercise 8, 
#  then you have successfully applied the fundamental principles 
#  of quantum mechanics to solve a real-world problem.
#
# ------------------------------------------------------------------
# 
# Revisions:
#   
#	2004-02-19  first release version  km
#       2008-08-29  ported from Forth to R  km       
#
# Requires: the R computing enviroment ( http://www.r-project.org )
#
# --------------------------------------------------------------------

# Define the identity and Pauli matrices.

one <- matrix( c(1, 0, 0, 1), nrow=2, byrow=TRUE )
sx  <- matrix( c(0, 1, 1, 0), nrow=2, byrow=TRUE )
sy  <- matrix( c(0, -1i, 1i, 0), nrow=2, byrow=TRUE )
sz  <- matrix( c(1, 0, 0, -1), nrow=2, byrow=TRUE )


## Elementary matrix operations
eigenvalues <- function( op ) { return(eigen(op, EISPACK=TRUE)$values) } 

eigenvectors <- function( op ) { return(eigen(op, EISPACK=TRUE)$vectors) }

## Operator and Vector Manipulation

# Hermitian adjoint; argument may be a matrix operator or a vector (spinor)
adjoint <- function( op ) { return( t(Conj(op)) ) }

# Commutator of two matrix operators
commutator <- function( op1, op2 ) { return( op1%*%op2 - op2%*%op1 ) }

# Complex inner product of two spinors
inner_product <- function( v1, v2 ) { return( (adjoint(v1)%*%v2)[1,1] ) } 

# Return a normalized spinor
normalize <- function( v ) {
	length = Re( inner_product( v, v ) )
	return( v/sqrt(length) )
}

# Probability of measuring spin down (along z)
P_down <- function( v ) {
	ev_down <- c( 0, 1) # eigenvector for spin down along z 
	vn = normalize(v)  # ensure normalization
	return( abs( inner_product( ev_down, vn ) )^2 )
}

# Probability of measuring spin up (along z)
P_up <- function( v ) {
	ev_up <- c( 1, 0 )  # eigenvector for spin up along z
	vn = normalize(v)  # ensure normalization
	return( abs( inner_product( ev_up, vn ) )^2 )
}

# Return the expectation value of an operator, given a spinor
expectation_value <- function( op, v ) { return( inner_product( v, op%*%v ) )}

# Compute and print the polarization vector for a spin state
pol <- function( v ) { 
	print( c( Re(expectation_value(sx, v)), "  x^  + ", 
		  Re(expectation_value(sy, v)), "  y^  + ",
		  Re(expectation_value(sz, v)), "  z^" ) ) 
}

