\ runge4.seq    Runge-Kutta ODE Solver for systems of ODEs
\
\ Forth Scientific Library Algorithm #29

\ )runge_kutta4_init ( 'dsdt n -- )
\               Initialize to use function dsdt() for n equations,
\               its stack diagram is:
\                      dsdt() ( 'u 'dudt -- ) ( F: t -- )
\               (the values in the array dudt get changed)
\
\ runge_kutta4_integrate() ( 'u steps -- ) ( F: t dt -- t' )
\               Integrate the equations STEPS time steps of size DT,
\               starting at time T and ending at time T'.  U is the
\               initial condition on input and the new state of the 
\               system on output.

\ runge_kutta4_done  ( -- )
\               Release previously allocated space.

\ )rk4qc_init ( 'dsdt n 's -- ) ( F: maxstep eps -- )
\               Initialize to use function dsdt() for n equations.
\               The initial function values are in s{. The output is also
\		in s{. The result is computed with a 5th-order Runge-Kutta
\ 		routine with adaptive step size control. The step size 
\		controller tries to keep the fractional error of any s{ 
\		component below eps. The maximum step size is limited to 
\		maxstep .
\
\ rk4qc_done  ( -- )
\               Release previously allocated space.
\
\ rk4qc_step ( -- flag ) ( F: step t -- step' t' )
\		Do one Runge-Kutta step, using adaptive step size control.
\		The flag is FALSE if the routine succeeds, else the step size
\		has become too small. The current step size and time are one 
\		the fp stack and will be updated by the routine.

\ This is an ANS Forth program requiring:
\      1. The Floating-Point word set
\      2. Uses words 'Private:', 'Public:' and 'Reset_Search_Order'
\         to control visibility of internal code
\      3. The word 'v:' to define function vectors and the
\         immediate 'defines' to set them.
\      4. The immediate words 'use(' and '&' to get function addresses
\      5. The words 'DARRAY' and '&!' to alias arrays.
\      6. Uses '}malloc' and '}free' to allocate and release memory
\         for dynamic arrays ( 'DARRAY' ).
\      7. The compilation of the test code is controlled by the VALUE
\         TEST-CODE? and the conditional compilation words in the
\          Programming-Tools wordset.
\      8. To run the code the fp stack needs to be at least 5 deep.
\	  To run all examples you need 7 positions on the fp stack.
\
\     (c) Copyright 1994  Everett F. Carter.     Permission is granted
\     by the author to use this software for any application provided
\     this copyright notice is preserved.

\  Revisions:
\     ?           mh; adaptive code was contributed by Marcel Hendrix
\
\     2007-09-23  km; added automated testing and a few of the Enright
\                     and Pryce test cases (see notes in test sec.).
\     2007-09-24  km; changed some "}t" to "r}t" tests for consistency with
\                     unified stack systems.
\     2007-09-27  km; cleaned up the test code, using a consistent format for
\                     both the E&P and non E&P tests; renamed rksolve1 and
\                     rksolve2 to )rksolve1 and )rksolve2 for better syntax,
\                     and added )rk_fixed1 and )rk_fixed2 wrappers for the
\                     fixed step size tests.
\     2007-10-27  km; save base, switch to decimal, and restore base

0 [IF] Additional documentation 

(see Chapter 15. 'Integration of Ordinary Differential Equations', in
 'Numerical Recipes in Pascal, The Art of Scientific Computing', 
 by William H. Press, Brian P. Flannery, Saul A. Teukolsky and William 
 T. Vetterling)

Problems involving ordinary differential equations (ODEs) can always be
reduced to the study of sets of first order differential equations. This is
done by introducing new variables that are usually derivatives of each other
and of the original variable. Example:

      2
    dy         dy	             
    --- + q(x) -- = r(x)  can be rewritten to
      2        dx 
    dx

    dy/dx = z(x);  dz/dx = r(x) - q(x) z(x).

The formula for traditional Euler integration is:

    y[n+1] = y[n] + h*f'(x[n], y[n])

This advances a solution from x[n] to x[n+1] = x[n] + h. It only uses
information (the derivatives f') from the start of the interval. 

When a trial step to the midpoint of the interval is done we can do better:

    k1 = h*f'(x[n], y[n])
    k2 = h*f'(x[n]+h/2, y[n]+k1/2)
    y[n+1] = y[n] + k2 + O(h^3)

This is called the second-order Runge-Kutta or midpoint method (a method
is called of order n when its error term is of order O(h^n+1)).

By taking this idea a few steps further along we arrive at the fourth-order 
Runge-Kutta method:

    k1 = h*f'(x[n], y[n])
    k2 = h*f'(x[n]+h/2, y[n]+k1/2)
    k3 = h*f'(x[n]+h/2, y[n]+k2/2)
    k4 = h*f'(x[n]+h/2, y[n]+k3)
    y[n+1] = y[n] + k1/6 + k2/3 + k3/3 + k4/6 + O(h^5)

You'll find that the following Forth code is an exact implementation of these
5 formulas.

When is Runge-Kutta(4) useful? Of course only when the increased number
of function evaluations (f') is compensated for by the fact that larger
steps can be taken than for Euler or Runge-Kutta(2) integration. Usually
this is the case. 

In general, higher-order RK methods are not interesting because the number 
of evaluations increases faster than the gain produced by the larger 
stepsize they allow.

Press et al highly recommend RK(4) in combination with an adaptive stepsize
algorithm ("Runge-Kutta is for ploughing the fields"). For smooth functions 
they recommend the Bulirsch-Stoer method ("high-strung racehorse").

You should be aware of the fact that a RK(4) routine _without_ adaptive
stepsize control can be a waste of computer time. This is how you'd use it:
 1. integrate the ODEs between point x1 and x2 with stepsize h.
 2. integrate the ODEs between point x1 and x2 with stepsize h/2.
 3. compare the results. If they're "sufficiently" equal, you're done,
    else goto 2.
 
This approach is useful when you just want a low resolution function sketch.
(most people would even leave out steps 2 and 3).
The fixed grid is exactly what is needed for plotting, an adaptive stepsize
would require the use of an interpolation algorithm after RK(4) finishes!

[THEN]

CR .( RUNGE4            V1.2e         27 October  2007   EFC )
BASE @ DECIMAL

Private:

DEFER dsdt()                     \ pointer to user function t, u, dudt

FLOAT DARRAY dum{             \ scratch space
FLOAT DARRAY dut{
FLOAT DARRAY ut{
FLOAT DARRAY dudt{

FVARIABLE h

FLOAT DARRAY u{               \ pointer to user array

0 VALUE dim

: F2/    0.5E0 F* ;
: F2*    2.0E0 F* ;


Public:


: )runge_kutta4_init ( &dsdt n -- )
     TO dim
     is dsdt()


     & dum{ dim }malloc
     malloc-fail? ABORT" runge_init failure (1) "

     & dut{ dim }malloc
     malloc-fail? ABORT" runge_init failure (2) "

     & ut{ dim }malloc
     malloc-fail? ABORT" runge_init failure (3) "

     & dudt{ dim }malloc
     malloc-fail? ABORT" runge_init failure (4) "

;

: runge_kutta4_done ( -- )

     & dum{ }free
     & dut{ }free
     & ut{ }free
     & dudt{ }free
;

Private:


: runge4_step ( -- ) ( F: t -- t' )

     FDUP u{ dudt{ dsdt()

     h F@ F2/
     dim 0 DO
             dudt{ I } F@ FOVER F* u{ I } F@ F+
             ut{ I } F!
            LOOP

     FOVER F+ ut{ dut{ dsdt()

     h F@ F2/
     dim 0 DO
             dut{ I } F@ FOVER F* u{ I } F@ F+
             ut{ I } F!
            LOOP

     FOVER F+ ut{ dum{ dsdt()

     h F@
     dim 0 DO
             dum{ I } F@ FOVER F* u{ I } F@ F+
             ut{ I } F!

             dum{ I } DUP F@ dut{ I } F@ F+ F!             

           LOOP

     F+           \ tos is now t+dt

     FDUP ut{ dut{ dsdt()

     h F@ 6.0E0 F/
     dim 0 DO
              dudt{ I } F@ dut{ I } F@ F+
              dum{ I } F@ F2* F+
              FOVER F*
              u{ I } DUP F@ F+ F!
           LOOP

     FDROP
          
;

Public:

: runge_kutta4_integrate() ( &u steps -- ) ( F: t dt -- t' )

     SWAP & u{ &!
     h F!

     0 DO runge4_step LOOP

;


0 [IF]  Adaptive Step Size Control For Runge-Kutta

The technique we'll be using is step doubling. We take each step twice,
once as a full step, then, independently, as two half steps. This creates
only about 38% of overhead because we'll get the accuracy of the half step.
A nice trick described by Press et al is that the results of the 3 operations
can be combined in a way that gives us a truncation error of order 6, not 5
as for the standard Runge-Kutta(4) (This is where the fcor factor is needed
for).

Given the results of Ringe-Kutta with the half and the full steps a new step 
size is chosen:

   h0 = h1 * |wanted_accuracy / delta1|^0.2

Here h stands for step size and delta1 is the maximum error of any component
of the output vector y, for the present step. If the step failed we know how
to choose the new step size h0, if the step was ok, we use h0 as the initial
step size for the next integration phase.

It is difficult to interpret "wanted_accuracy" correctly. When the dependent
variables differ enormously in amplitude you'd want to use fractional errors.
With oscillatory functions passing through zero it would be better to set
the wanted_accuracy to some part of the amplitude of these functions. Press 
et al recommend an additional array uscal{ to do all this. One of the 
arguments of the routine will be the vector of dependent variables at the 
beginning of the proposed step. Call that u{ 1..n }. Let us require the user 
to specify for each step another, corresponding, vector argument 
uscal{ 1..n }, and also an overall tolerance level eps. Then the desired 
accuracy for the ith equation will be

 	wanted_accuracy = eps * uscal{ i }

When constant fractional errors are desired you should plug y into the uscal{
calling slot. When constant absolute errors relative to some maximum values 
are needed you set uscal{ i } equal to those values. A useful trick for getting
constant fractional errors except very near zero crossings is to set uscal{ i }
equal to |u{ i }| + |h * dsdt{ i }| . We'll use that in the code below.

Further refinements of the step size controller result in exponents being 0.2
or 0.25 and the use of a safety factor S of 0.9:
   h0 = Sh1 * |wanted_accuracy / delta1|^0.2    wanted_accuracy >= delta1
      = Sh1 * |wanted_accuracy / delta1|^0.25   wanted_accuracy < delta1

[THEN]

Private:
	 1E-30 	  FCONSTANT tiny
	-0.20E0    FCONSTANT pgrow
	-0.25E0    FCONSTANT pshrink
	1e 15E0 F/ FCONSTANT fcor
	  0.9E0    FCONSTANT safety

4E0 safety F/  
1E0 pgrow  F/ F**  FCONSTANT errcon

FVARIABLE eps
FVARIABLE step
FVARIABLE tstart
FVARIABLE maxstep

FLOAT DARRAY uorig{
FLOAT DARRAY u1{
FLOAT DARRAY u2{
FLOAT DARRAY uscal{


\ Find reasonable scaling values to decide when to shrink step size.
: scale'm ( -- )
	tstart F@ uorig{ uscal{ dsdt()	
	dim 0 DO uscal{ I } DUP F@ step F@ F* FABS
		 uorig{ I }     F@ FABS F+ tiny F+
		 F!		 
	    LOOP ;

\ With a trick the result of a step can be made accurate to 5th order.
: 4th->5th ( -- )
	dim 0 DO 		\ get 5th order truncation error
		 uorig{ I } DUP F@  
		 FDUP  u1{ I } F@ F-  fcor F* 
		 F+  F! 
	    LOOP ;

\ Test if the step size needs shrinking
: shrink? ( -- bool ) ( F: -- diff )
	0.0E0 ( errmax )
	dim 0 DO  
		uorig{ I } F@  u1{ I } F@  F-  
		uscal{ I } F@  F/  FABS FMAX  
	    LOOP  
	eps F@ F/  FDUP 1e F> ;

Public:

\ Initialize to use function dsdt() for n equations. The initial function 
\ values are in s{. The output is also in s{. The result is computed with a 
\ 5th-order Runge-Kutta routine with adaptive step size control. The step size 
\ controller tries to keep the fractional error of any s{ component below eps.
\ The maximum step size is limited to maxstep .
: )rk4qc_init ( 'dsdt n 'u -- ) ( F: maxstep eps -- )
	eps F! maxstep F!
	& uorig{ &! 
	)runge_kutta4_init
	& u1{    dim }malloc malloc-fail? ABORT" )rk4qc_init :: malloc (1)" 
	& u2{    dim }malloc malloc-fail? ABORT" )rk4qc_init :: malloc (2)" 
	& uscal{ dim }malloc malloc-fail? ABORT" )rk4qc_init :: malloc (3)" ;

\ Release previously allocated space.
: rk4qc_done  ( -- )
	runge_kutta4_done
	& u1{    }free 
	& u2{    }free 
	& uscal{ }free ;
\ Do one Runge-Kutta step, using adaptive step size control. The flag is 
\ FALSE if the routine succeeds, else the step size has become too small. 
\ The current step size and time are one the fp stack and will be updated 
\ by the routine.

: rk4qc_step ( -- flag ) ( F: step t -- step' t' )
	tstart F!  step F!  scale'm
	uorig{ u1{ dim }fcopy	\ we need a fresh start after a shrink
	uorig{ u2{ dim }fcopy
   BEGIN	
	uorig{ 2 tstart F@ step F@ F2/ runge_kutta4_integrate() FDROP
	u1{    1 tstart F@ step F@     runge_kutta4_integrate() ( F: -- t' )
	FDUP tstart F@ 0.0E0 F~ IF 0.0E0 FSWAP FALSE EXIT THEN
	shrink?			\ maximum difference between these two tries
   WHILE			\ too large, shrink step size
	FLN pshrink F* FEXP step F@ F* safety F* step F!  FDROP
	u2{ uorig{ dim }fcopy	\ a fresh start after a shrink...
	u2{ u1{    dim }fcopy

  REPEAT			\ ok, grow step size for next time
	FDUP errcon F< IF  FDROP step F@ 4e F* 
		     ELSE  FLN pgrow F* FEXP step F@ F* safety F*
		     THEN 
	maxstep F@ FMIN		\ but don't grow excessively!
	FSWAP TRUE 4th->5th ;



Reset_Search_Order
BASE !
    
TEST-CODE? [IF]     \ test code ==========================================
[undefined] T{      [IF]  include ttester.fs  [THEN]
BASE @ DECIMAL

\ Generic Test Wrappers for the Adaptive and Fixed Step Solvers

3 FLOAT ARRAY x{

FVARIABLE  _dt
FVARIABLE  t_end
FVARIABLE  t_final

\ Generic one equation ODE solver (adaptive step size)
: )rksolve1 ( 'dsdt -- ) ( F: tstart tend y0 -- r )
    x{ 0 } F! t_end F!
    1  x{  _dt F@ rel-near F@ )rk4qc_init
    _dt F@ FSWAP  BEGIN  rk4qc_step 0= FDUP t_end F@ F>= OR UNTIL
    t_final F! FDROP  rk4qc_done  x{ 0 } F@   
;

\ Generic two equations ODE solver (adaptive step size)
: )rksolve2 ( 'dsdt -- ) ( F: tstart tend x0 y0 -- r )
    x{ 1 } F! x{ 0 } F! t_end F!
    2  x{  _dt F@ rel-near F@ )rk4qc_init
    _dt F@ FSWAP  BEGIN  rk4qc_step 0= FDUP t_end F@ F>= OR UNTIL
    t_final F! FDROP  rk4qc_done  x{ 0 } F@   
;

\ Generic one equation ODE solver (fixed step size)
: )rk_fixed1 ( 'dsdt nsteps -- ) ( F: tstart y0 -- r )
    x{ 0 } F! 
    SWAP 1 )runge_kutta4_init
    x{ SWAP  _dt F@ runge_kutta4_integrate()
    t_final F! runge_kutta4_done  x{ 0 } F@   
;

\ Generic two equations ODE solver (fixed step size)
: )rk_fixed2 ( 'dsdt nsteps -- ) ( F: tstart x0 y0 -- r )
    x{ 1 } F! x{ 0 } F!
    SWAP 2 )runge_kutta4_init
    x{ SWAP  _dt F@ runge_kutta4_integrate()
    t_final F! runge_kutta4_done  x{ 0 } F@   
;


\ The test cases here were originally given by Wayne Enright and John Pryce,
\ Algorithm 648, ACM Transactions on Mathematical Software, volume 13, no. 1,
\ pp 28--34, 1987.
\
\ Also, see
\
\ http://people.scs.fsu.edu/~burkardt/f_src/test_ode/test_ode.html

\ E & P nonstiff problem #A1:
\ dy/dt = -y
\ y(0) = 1
\ Exact solution is y(t) = exp(-t)
: derivs-A1() ( 'u 'dudt -- ) ( F: t -- )
    FDROP SWAP 0 } F@  FNEGATE 0 } F! ;

: A1 ( F: t -- r ) FNEGATE FEXP ;

\ E & P nonstiff problem #A2:
\ dy/dt = -(y^3)/2
\ y(0) = 1
\ Exact solution is y(t) = 1 / sqrt(t + 1)
: derivs-A2() ( 'u 'dudt -- ) ( F: t -- )
    FDROP SWAP 0 } F@ FDUP FDUP F* F* FNEGATE 2e F/ 0 } F! ;

: A2 ( F: t -- r )
    1e F+ FSQRT 1e FSWAP F/ ;

\ E & P nonstiff problem #A3:
\ dy/dt = cos(t) * y
\ y(0) = 1
\ Exact solution is y(t) = exp( sin( t ) )
: derivs-A3() ( 'u 'dudt -- ) ( F: t -- )
    FCOS SWAP 0 } F@ F* 0 } F! ;

: A3 ( F: t -- r )
    FSIN FEXP ;

\ E & P nonstiff problem #A4:
\ dy/dt = y*(20 - y)/80
\ y(0) = 1
\ Exact solution is y(t) = 20 / ( 1 + 19*exp( -t / 4 ) )
: derivs-A4() ( 'u 'dudt -- ) ( F: t -- )
    FDROP SWAP 0 } F@ FDUP 20e FSWAP F- F* 80e F/ 0 } F! ;

: A4 ( F: t -- r )
    FNEGATE 4e F/ FEXP 19e F* 1e F+ 20e FSWAP F/ ;

\ E & P nonstiff problem #A5:
\ dy/dt = (y - t)/(y + t)
\ y(0) = 1
\ Exact solution is
\      r = sqrt ( t + y(t)**2 )
\      theta = atan ( y(t) / t )
\
\      r = 4 * exp ( pi/2 - theta )

: derivs-A5() ( 'u 'dudt -- ) ( F: t -- )
    FDUP OVER 0 } F@ FSWAP F- FSWAP SWAP 0 } F@ F+ F/ 0 } F! ;


: A5 ( F: t -- r )
    ;

\ ------- Other test cases (not from E & P) ------------
    
\ The RC discharge equation:
\ dVc/dt = (1/tau)*(Vin-Vc)
\ Vc(0) = 0
\ Exact solution is Vc(t) = Vin*( 1 - exp( -t/tau ) )
\ Comments: Don't use the long-time limit solution for accuracy
\   testing, since there is an attracting fixed-point at Vc = Vin.
100E-3  FCONSTANT tau ( tau = R*C; 100 ms for this example)
10E     FCONSTANT Vin ( charging voltage source is 10 Volts)	

: derivs-Vc() ( 'u 'dudt -- ) ( F: t -- )
    FDROP
    Vin  SWAP 0 } F@ F-   tau F/   0 } F! ;

: VC ( F: t -- v )
    1e  FSWAP FNEGATE tau F/ FEXP  F-  Vin F* ; 


\ Damped vibrations:
\ u'' + cm*u' + km*u = 0
\ or  du/dt = v,  dv/dt = -cm*v - km*u
\ u(0) = 1/{2*pi}
\ v(0) = -cm/{4*pi}
\ Exact solution for u(t) is:
\ u(t) = (1/2pi)*exp{-cm*t/2}*cos(sqrt{km - cm^2/4}*t)
\ Comments: Don't use the long-time limit solution for accuracy
\   testing, since there is an attracting fixed-point at u = 0.
1.0E0 FATAN 8.0E0 F* FCONSTANT PI*2

1.92E0  FCONSTANT cm
960.0E0 FCONSTANT km
    
1e PI*2 F/ FCONSTANT u0
cm FNEGATE F2/ PI*2 F/ FCONSTANT v0

: derivs-DV() ( 'u 'dudt -- ) ( F: t -- )
    FDROP     \ does not use t
    OVER 1 } F@ DUP 0 } F!
    OVER 1 } F@ cm  F*
    SWAP 0 } F@ km  F* F+ FNEGATE
         1 } F!   
;

: DV ( F: t -- a )             \ just the U value not V for damped vib.
    cm FOVER F* F2/ FNEGATE FEXP
    FSWAP
    cm cm F* 4.0E0 F/ FNEGATE km F+ FSQRT
    F* FCOS
    F* PI*2 F/
;

\ Lorenz equations for chaos:
\ dx/dt = sig * (y - x)
\ dy/dt = r * x - y - x * z
\ dz/dt = -bp * z + x * y
\ Exact solution: NONE
\ Comments: Since chaotic equations are extremely sensitive to
\   initial conditions, the result after integration will
\   depend very sensitively on the precision of the input values
\   x(t0), y(t0), and z(t0), and on the precision with which
\   floating point calculations are performed. Such sensitivity is
\   not useful for directly testing the accuracy of a portable ODE
\   solver; however, some invariant properties of the attractor can
\   be computed from the solution, and may possibly serve as suitable
\   tests of the ODE solver. 
16.0E0  FCONSTANT sig
45.92E0 FCONSTANT r
4.0E0   FCONSTANT bp

: derivs-LE() ( 'u 'dudt -- ) ( F: t -- )
    FDROP     \ does not use t

    OVER DUP 1 } F@ 0 } F@ F- sig F*
    DUP 0 } F!

    OVER DUP DUP 2 } F@ FNEGATE r F+
                 0 } F@ F*
                 1 } F@ F-
    DUP 1 } F!

    SWAP DUP DUP 0 } F@ 1 } F@ F* 2 } F@ bp F* F-
        2 } F!    
;


\ ----- Begin Testing --------------------

1.0E-2 _dt F!


1e-13 rel-near F!
1e-13 abs-near F!
set-near    

CR
TESTING E&P Non-Stiff ODE Problems A1 to A4 (Adaptive Step)
t{ use( derivs-A1() 0e 20e 1e )rksolve1  ->  t_final F@ A1  r}t
t{ use( derivs-A2() 0e 20e 1e )rksolve1  ->  t_final F@ A2  r}t
t{ use( derivs-A3() 0e 20e 1e )rksolve1  ->  t_final F@ A3  r}t
t{ use( derivs-A4() 0e 20e 1e )rksolve1  ->  t_final F@ A4  r}t
\ t{ use( derivs-A5() 0e 20e 1e )rksolve1  ->  t_final F@ A5  r}t

\ Non E&P Problems
TESTING Damped Vibration (Adaptive Step)
t{ use( derivs-DV() 0e 0.8e u0 v0  )rksolve2  ->  t_final F@ DV  r}t

TESTING Charging Capacitor (Adaptive Step)
t{ use(  derivs-Vc() 0e 200e-3 0e  )rksolve1  ->  t_final F@ Vc  r}t


\ Tests Using Fixed Step Solver
\ Set relatively low accuracy since our time step _dt is coarse.
1e-4 rel-near F!   
1e-4 abs-near F!

TESTING Damped Vibration (Fixed Step)
t{ use(  derivs-DV() 80 0e u0 v0  )rk_fixed2  ->  t_final F@ DV  r}t

TESTING Charging Capacitor (Fixed Step)
t{ use(  derivs-Vc() 4  0e 0e  )rk_fixed1  ->  t_final F@ Vc  r}t



0 [IF]
\ see comments for the Lorenz equations, above -- km 2007-09-27
    
: lorenz_test ( n -- )               \ n is the number of time steps to run

    0.0E0 x{ 0 } F!   1.0E0 x{ 1 } F!   0.0E0 x{ 2 } F!     
    use( derivs-LE() 3 )runge_kutta4_init
    0.0E0       \ initial time
          
    0 DO
	x{ 1 dt runge_kutta4_integrate()      
    LOOP

    FDROP runge_kutta4_done
;
[THEN]

BASE !
[THEN]


