\ h2XJ0.4th
\
\ Find the vibrational level energies, E(v, J=0), and dissociation 
\ energy, D0, of the H_2 molecule in its ground electronic state 
\ (X 1Sigma_g^+), by solving the radial Schroedinger equation with
\ a theoretical inter-atomic potential [1].
\
\ References:
\
\ 1. K. Pachucki, "Born-Oppenheimer potential for H_2", 
\    Phys. Rev. A, 82, 032509 (2010).
\
\ 2. W. Kolos and L. Wolniewicz, "Accurate Adiabatic Treatment of
\    the Ground State of the Hydrogen Molecule", J. Chem. Phys.,
\    41, 3663 (1964).
\
\ 3. L. Wolniewicz, "Relativistic energies of the ground state of
\    the hydrogen molecule", J. Chem. Phys., 99, 1851 (1993).
\
\ 4. L. Wolniewicz, "Nonadiabatic energies of the ground state of
\    the hydrogen molecule", J. Chem. Phys., 103, 1792 (1995).
\
\ 5. I. Dabrowski, Canadian J. Phys., 62, 1639 (1984).
\
\ 6. R. J. LeRoy, 2007, "Level 8.0, A Computer Program for 
\    Solving the Radial Schroedinger Equation for Bound and 
\    Quasibound Levels", University of Waterloo, report CP-663. 
\    Software and manual available from the web at 
\    http://scienide2.uwaterloo.ca/~rleroy/level/
\
\ Notes:
\
\ 1. The potential given in [1] is uncorrected for nuclear motion,
\    relativistic effects, and QED contributions. It is only a 
\    high precision potential in the clamped nuclei approximation
\    (Born-Oppenheimer model). Corrections due to relativistic
\    effects, nuclear motion, and QED effects have been computed
\    [2--4] and may be added to the potential from [1] to improve the
\    quantitative comparison to experimental values [5]. The
\    computed levels with the B-O potential match to within about 
\    5 cm^-1 of the experimental values.
\
\ 2. The output from this program has been compared to the output
\    from the Level program [6], using the same potential, and
\    the results from both agree to within 0.01 cm^-1.
\
\ 3. To do: smart adjustment of initial trial values for solve --
\    currently, the level finder misses the last bound level
\    (v = 14) near the top of the well; also provide interface
\    to output radial wavefunctions to a file.
\
\ Revisions:
\   2010-10-19 km  first version; misses the v=14 level; ported
\                  to Gforth, and fixed typos in comments.
\   2010-10-21 km  revise notes.
\
\ Copyright (c) 2010 Krishna Myneni
\
\ This code may be used for any purpose as long as the copyright
\ notice above is preserved.
\

\ -- number formatting words

: fstring ( f n -- a u | convert f to a formatted string with n decimal places )
    1 swap dup >r 0 ?do 10 * loop s>f f* fround f>d dup -rot dabs
    <# r> 0 ?do # loop [char] . hold #s rot sign #> ; 

: fprint ( f n width -- | print an fp number to n decimal places in width)
    >r fstring r> over - dup 0> IF spaces ELSE drop THEN type ;


include schr1d.fs

\ Read the theoretical Born-Oppenheimer potential [1], V(R), for 
\ the ground electronic state of H2 [H(1s) + H(1s) in the symmetric 
\ configuration]

s" H2-potnl-pac2010.dat" read-potnl 

\ Use minimum 1024 point interpolated potential, with 8th order
\ polynomial interpolation
1024 8 setup-mesh 

\ Set the effective mass of the particle in the potential well
m_p m_e F+ 2e F/  mu F!  \ reduced mass of H_2

-1e    fconstant V_inf   \ potential energy in Hartrees at R=infinity
                         \ (this is the energy of two separated H atoms)

\ Convert energy in Hartrees to cm^-1 and print with uncertainty value
: print-energy ( E dE -- )
    cm^-1/Eh F* fswap V_inf F- cm^-1/Eh F*
    3 12 fprint ."   +/- " 6 8 fprint ;

32 constant MAX-LEVELS
MAX-LEVELS FLOAT ARRAY Ev{
0 value v
0.001e fconstant start_dE

: find-levels ( -- ) 
    cr ." Vibrational Level Energies in cm^-1 relative to"
    cr ." V(inf) = " V_inf cm^-1/Eh F* 4 14 fprint ."  cm^-1" 
    MAX-LEVELS Ev{ }fzero

    \ Find the energy of the lowest level (v=0)
    0 to v
    Vmin F@ start_dE solve ABORT" find-levels: Unable to find the lowest level!" 
    f2dup cr v 4 .R 2 spaces print-energy fdrop
    fdup Ev{ v } F!
    fdup Vmin F@ F- F+ 
    BEGIN
      start_dE solve
      0= IF
        v 1+ to v  
        f2dup cr v 4 .R 2 spaces print-energy fdrop
        \ The following energy increment works well to determine
        \   the trial value for the next eigenvalue
        fdup Ev{ v } F! 
        fdup Ev{ v 1- } F@ F- 2e F/  
        F+
      ELSE
        fdrop
      THEN
      fdup V_inf F>
    UNTIL
    fdrop
    \ Dissociation energy, D0
    V_inf Ev{ 0 } F@ F- cm^-1/Eh F* 
    cr ." D0 (cm^-1) = " 4 14 fprint 
    end-solve
;
    





