(      Title:  Modular programming in Forth
        File:  modules.4th
     Version:  0.3.3
     Revised:  November 4, 2011

Description:
-----------

This library provides a facility for modular programming in 
Forth. It supports both named and unnamed modules, and allows
for private definitions in a module to be made accessible. 
The library uses design features from several sources: 
David N. Williams' root-module.fs [v 0.8.2], Bruce McFarling's 
suggested module facility for the Forth Scientific Library [FSL], 
the original module facility in the FSL's fsl-util auxiliary files, 
and Neal Bridges' anonymous modules code.


Design elements:
---------------

1. The layout of a module is determined by the following words:

      MODULE:  BEGIN-MODULE  PUBLIC:  PRIVATE:  END-MODULE

   Here, BEGIN-MODULE is where the module begins, and END-MODULE
   is where it ends.  The code area between the two is "inside" the
   module. The words PUBLIC: and PRIVATE: control the visibility
   of the module code to external code and to the user. Assigning
   a name to the module with MODULE: is optional, but highly 
   encouraged to avoid name clash problems.

2. A module contains two types of words: "public" and "private".
   The word PUBLIC: declares that subsequent definitions are
   public words, while the word PRIVATE: declares subsequent
   definitions to be private words. The public words of a
   module constitute the API or application programming 
   interface to the module. Private words are not intended
   to be used by the module user.


3. A module may or may not be named:

   a] For an unnamed module, the public definitions are compiled 
      into the current compilation wordlist.

   b] For a named module, the public definitions are compiled
      into a new wordlist, the module's public wordlist. Invoking
      the module name replaces the top wordlist id on the search
      order with the public wordlist of the module. To add the
      public wordlist to the search order, use

            ALSO <module_name>

   c] For both a named and an unnamed module, a new private
      wordlist is created. For named modules, the private
      wordlist id can be retrieved using >PRIVATE on the
      module xt. The private wordlist of an unnamed module
      is not accessible outside the module.

4. PRIVATE: and PUBLIC: should occur only inside a module, and may
   be invoked any number of times there, in any order.  Definitions
   inside a module usually go into its public or private
   wordlist.

5. The module layout is not nestable.

6. The search order depth and compilation wordlist just before
   BEGIN-MODULE are restored by END-MODULE. The part of the search 
   order present just before the beginning of a module may not be 
   changed inside the module.

   [This could be made safer by actually saving and restoring
   the initial search order.  This implementation does not do
   that.]

7. The module name reference operator, ∍ , allows direct 
   referencing of a word inside of a named module, using the
   syntax,

    ∍ <module_name> <word>

   The reference operator ∍ may be used in either compilation
   or interpretation state.
)


\ *** GENERAL USE

[UNDEFINED] drops [IF]                 \ BMcF
: drops ( +n -- ) 0 ?DO drop LOOP ; [THEN]

[UNDEFINED] order-drops [IF]
: order-drops  ( +n -- ) ( o: wid_n ... wid_1 -- )
  0 ?DO previous LOOP ; [THEN]

[UNDEFINED] order-depth [IF]
: order-depth  ( -- n )  get-order dup >r drops r> ; [THEN]

[UNDEFINED] >order [IF]      \ same as BMcF's ADD-WORDLIST
: >order  ( wid -- order: wid )
  >r get-order r> swap 1+ set-order ; [THEN]

[UNDEFINED] order> [IF]  \ not used in this module
: order>  ( o: wid -- ) ( -- wid )
  get-order swap >r 1- set-order r> ; [THEN]

[UNDEFINED] order@ [IF]  \ not used in this module
: order@  ( o: wid -- ) ( -- wid )
  get-order over >r drops r> ; [THEN]


[UNDEFINED] MODULE: [IF]

VARIABLE initial-defs 
VARIABLE initial-order-depth 
VARIABLE private-defs
VARIABLE public-defs

VARIABLE named-module         0 named-module !

\ Create the name for a named module and reserve space 
\ to store public and private wordlists
: MODULE:  ( "name" -- )
    CREATE  here 0 , 0 , named-module !
  DOES>  ( o: -- wid )  @ >r get-order nip r> swap set-order 
;

\ Return private wordlist id from the xt of a named module
: >PRIVATE ( xtmodule -- wid )  >body cell+ @ ;

\ *** PUBLIC DEFINITIONS

: PRIVATE:  ( -- )  private-defs @ set-current ;
: PUBLIC:   ( -- )  public-defs  @ set-current ; 

\ BEGIN-MODULE and END-MODULE are not to be nested.

: BEGIN-MODULE  ( -- ) ( o: -- public private )
(
Save information that allows END-MODULE to restore the initial
search order and compilation wordlist, assuming the initial
order has remained unchanged. Create the private wordlist.
Create the public wordlist for a named module. Push the public 
and private wordlists onto the search order, and set the 
private wordlist to be the current compilation wordlist.
)
  get-current  initial-defs !
  order-depth  initial-order-depth !
  wordlist      \ new private wordlist
  named-module @ if  
    wordlist    \ new public wordlist 
    2dup named-module @ 2! 
  else  
    get-current  \ use compilation wordlist
  then
  dup   >order     public-defs !
  dup   >order dup private-defs !
  set-current    \ default section is private
;

: END-MODULE  ( o: <extras> -- )
(
Restore the initial search order depth and compilation wordlist. An
ambiguous condition exists if the initial search order has been
clobbered. Reset the named-module flag.
)
  initial-defs @ set-current
  order-depth initial-order-depth @ - order-drops 
  0  named-module ! ;


\ Module targeted word reference operator, represented by
\ Unicode math symbol, "small element of": U+0x220d
: ∍
    state @ if bl word find 0= ABORT" Unknown Module!"
    else ' then also execute
    state @ if postpone ['] postpone execute else ' execute then
    previous
; immediate


[THEN]
\ --------------------------------------------

\ The examples below assume that module.fs is loaded.

\ EXAMPLE 1: Layout for anonymous module
0 [IF]

\ Add named modules and Forth-83 vocabularies to the 
\ search order on which this module depends.
\ GET-ORDER and SET-ORDER may be used to remove 
\ dependencies from the search order, as shown.

get-order
ALSO <mod_1> ... ALSO <mod_n>

BEGIN-MODULE

\ Initial definitions are private, and placed in a hidden
\   wordlist accessible only to this module.

<initial private definitions>

PUBLIC:  
\ public definitions go into the current compilation wordlist

\ any number of subsequent PRIVATE: and PUBLIC: sections follow,
\ in any order.

END-MODULE

set-order

[THEN]  \ END EXAMPLE 1 LAYOUT
\ --------------------------------------------

\ EXAMPLE 2: Layout for a named module
0 [IF]

\ Make a vocabulary for the public definitions of a
\ module named Foo.

Module: Foo

get-order
ALSO <mod_1> ... ALSO <mod_n>

BEGIN-MODULE

PUBLIC:
\ public definitions go into the wordlist, Foo.

\ any number of subsequent private and public sections follow,
\ in any order.

END-MODULE
set-order

[THEN]  \ END EXAMPLE 2 LAYOUT
\ --------------------------------------------

\ To access the private definitions of Foo from another
\ module, retrieve the private wordlist of Foo, and add it 
\ to the search order, e.g.
\
\   ' Foo >private >order


