/*
pde1.c

Equivalent C code for benchmarking Forth version of pde1 partial differential
eqn. solver.
*/
#include<math.h>

#define true  1
#define false 0
#define GRIDSIZE 64

double grid [GRIDSIZE] [GRIDSIZE];
double last_grid [GRIDSIZE] [GRIDSIZE];

double tol = 1e-3;

// Rectangular Region Boundary Values

#define TOP_EDGE     100.0
#define RIGHT_EDGE     0.0
#define BOTTOM_EDGE    0.0
#define LEFT_EDGE     50.0

int inside_rectangle ( int row, int col)
{
    /* Are we inside the rectangular boundary? */

    return ((row > 0) && (row < GRIDSIZE-1))&& ((col > 0) && (col < GRIDSIZE-1));
}

void set_rectangular_bvs ()
{
    /* Setup the rectangular boundary values */
    int i;
    for (i = 0; i < GRIDSIZE; i++) grid[0][i] = TOP_EDGE;
    for (i = 0; i < GRIDSIZE; i++) grid[i][GRIDSIZE-1] = RIGHT_EDGE;
    for (i = 0; i < GRIDSIZE; i++) grid[GRIDSIZE-1][i] = BOTTOM_EDGE;
    for (i = 0; i < GRIDSIZE; i++) grid[i][0] = LEFT_EDGE;
}

void init_rectangular_grid ()
{
    /* Set up the starting grid values */
    int i, j;
    double mean;
    set_rectangular_bvs();
    mean = (TOP_EDGE + BOTTOM_EDGE + RIGHT_EDGE + LEFT_EDGE)/4.0;
    for (i = 0; i < GRIDSIZE; i++)
	for (j = 0; j < GRIDSIZE; j++)
	    if (inside_rectangle(i,j)) grid[i][j] = mean;
}

void nearest ( int row, int col, double nns[])
{
    int i;
    i = col - 1; if (i < 0) i = 0;
    nns[0] = grid[row][i];          // left nearest neighbor
    i = col + 1; if (i >= GRIDSIZE) i = GRIDSIZE - 1;
    nns[1] = grid[row][i];          // right nearest neighbor
    i = row - 1; if (i < 0) i = 0;
    nns[2] = grid[i][col];          // up nearest neighbor
    i = row + 1; if (i >= GRIDSIZE) i = GRIDSIZE - 1;
    nns[3] = grid[i][col];          // down nearest neighbor
}

/*
Apply the mean value theorem once to each of the interior grid values:
Replace each grid value with the average of the four nearest neighbor values.
*/

void iterate ()
{
    int i, j;
    double nns[4];

    for (i = 0; i < GRIDSIZE; i++)
    {
	for (j = 0; j < GRIDSIZE; j++)
	{
	    if (inside_rectangle(i,j))
	    {
		nearest(i, j, nns);
		grid[i][j] = (nns[0] + nns[1] + nns[2] + nns[3])/4.;
	    }
	}
    }
}


int converged ()
{
    /* test for convergence between current and last grid */

    int i,j;
    for (i = 0; i < GRIDSIZE; i++)
    {
	for (j = 0; j < GRIDSIZE; j++)
	{
	    if (inside_rectangle(i,j))
	    {
		if (fabs(grid[i][j] - last_grid[i][j]) > tol)
		    return false;
	    }
	}
    }
    return true;
}


void copy_to_last()
{
    /* copy current grid to last grid */

    int i, j;
    for (i = 0; i < GRIDSIZE; i++)
	for (j = 0; j < GRIDSIZE; j++)
	    last_grid[i][j] = grid[i][j];
}

int main()
{
    init_rectangular_grid();

    do
    {
	copy_to_last();
	iterate();
    } while (!converged());

    return 0;
}

    
