\ lkspec.4th
\
\ Calculate the time-averaged optical spectrum for a Lang-Kobayashi system.
\
\ Copyright (c) 2002 Krishna Myneni
\ 
\ Requires:
\
\	ans-words
\	complex
\	fsl-util
\       dynmem
\       runge4
\	sl.4th
\	lk.4th
\	four1.4th
\
\ Revisions:
\    2002-11-15  km; first version created.
\    2005-10-27  km; redefine params. to print all relevant parameters
\    2007-10-23  km; revised to use new versions of sl.4th and lk.4th,
\                    based on FSL arrays. Use new FFT module four1.4th.
\    2007-11-01  km; finished new version, and revised comments.
\    2007-11-10  km; revised Notes.
\
\ Notes:
\
\  1. To compute the spectrum with the default L-K parameters
\     (see lk.4th), execute lkspec. The spectrum is stored
\     in the matrix spec{{, a two-column matrix containing
\     (frequency shift, power) pairs. The frequency shift is
\     relative to the optical frequency of the solitary diode
\     laser, which has been factored out of the calculations.
\     The frequency shift is in units of GHz, and the 
\     and the spectral power is an arbitrary, relative scale.
\
\     For comparison with experimental spectra of a diode laser
\     under various feedback strengths, see
\
\     K. Myneni, et. al., 2001, "High-precision ranging using
\     a chaotic laser pulse train", Applied Physics Letters,
\     vol. 78, pp. 1496-.
\     
\
\  2. To print the spectrum, type 
\
\         NPTS 2 spec{{ }}fprint
\
\     To calculate the spectrum, and direct the output to a file,
\     e.g. spec.dat, type
\
\       true to verbose?
\	>file spec.dat lkspec console
\
\  3. To increase the frequency resolution, increase NPTS.
\     Always use a power of 2 (1024, 2048, etc.) for NPTS.
\     The frequency range is controlled by the time step, ds.
\
\  4. You may also increase the number of averages (nave) to 
\     see the results of measurement for different integration times.
\

include lk
include four1

512 constant NPTS
  
NPTS COMPLEX ARRAY E{	        \ array for recent E
NPTS 2* FLOAT ARRAY Eft{        \ array for FFT of E
NPTS 2 FLOAT MATRIX spec{{	\ averaged power spectrum
NPTS 2/ FLOAT ARRAY tmp{        \ use for reordering spectrum

fvariable min_freq	\ minumum of the spectrum frequency
fvariable freq_step	\ frequency step for the spectrum
fvariable dt            \ time steps in seconds
ds F@ t_p F@ F* dt F!

500 value Nave		\ number of averages
0 value sidx
0 value vflag

: }}fzero ( 'array n m -- | zero the n x m element real array)
    * FLOATS erase ;

: lkspec ( -- )
    E{ NPTS }zzero
    lk_init
    1.9e 1.1e 0.1e  3 sv{ }fput      \ initialize the state vector
    sv{ 0 } z@ !Epast
    use( derivs-lk() 3 )runge_kutta4_init
   
    \ Compute 100000 normalized time steps to go past the transient;

    verbose? TO vflag  false to verbose?
    0e 100000 lksteps  
    vflag TO verbose?
    
    \ Compute NAVE sets of NPTS time steps and the average power spectrum

    spec{{ NPTS 2 }}fzero

    Nave 0 DO
	0 TO sidx
	NPTS 10 * 0 DO
	    ds F@ sv{ 1 runge_kutta4_integrate()
	    sv{ 0 } z@ !Epast
	    I 10 mod 0= IF		\ sample at 10*ds 
		sv{ 0 } z@  E{ sidx } z!
		sidx 1+ TO sidx
	    THEN
	LOOP

	\ Compute the complex fft of the set of NPTS points.
	\ The inverse FFT is used because of our choice of sign in
	\   factoring the optical frequency out of the laser equations.

	E{ Eft{ NPTS 2* FLOATS MOVE
	NPTS -1 Eft{ }four1

	\ Accumulate time averaged spectrum

	NPTS 0 DO 
	    Eft{ I 2* } z@ |z|^2  spec{{ I 1 }} F@ F+  spec{{ I 1 }} F! 	
	LOOP
    LOOP

    \ Re-order the spectrum from min frequency to max frequency

    NPTS 2/ 0       DO  spec{{ I 1 }} F@  tmp{ I } F!                 LOOP
    NPTS NPTS 2/    DO  spec{{ I 1 }} F@  spec{{ I NPTS 2/ - 1 }} F!  LOOP  
    NPTS 2/ 0       DO  tmp{ I }      F@  spec{{ I NPTS 2/ + 1 }} F!  LOOP 

    \ Compute a frequency scale and print out the spectrum

    dt F@ 10e F* 2e F* -1e fswap F/ min_freq F!
    dt F@ 10e F* NPTS s>f F* 1e fswap F/ freq_step F!

    min_freq F@ freq_step F@ F-
    NPTS 0 DO 
	freq_step F@ F+ fdup 1e9 F/ spec{{ I 0 }} F!      \ freq in GHz
	spec{{ I 1 }} F@ Nave s>f F/  spec{{ I 1 }} F!  \ power is arbitrary
    LOOP FDROP

    verbose? IF  NPTS 2 spec{{ }}fprint  THEN
;

: params.  ( -- | output all relevant parameters )
    lk_init params.
    separator cr
    ." Spectral calculation parameters: " cr
    separator cr
    ." Number of averages [Nave] = " tab Nave . cr
    ." Time segment              = " tab 
    dt F@ 10e F* NPTS s>f F* 1e-9 F/ F. ."  ns" cr
    ." Sampling interval         = " tab dt F@ 10e F* 1e-9 F/ F. ." ns" cr
    cr ;

params.
