\ gamma.4th
\
\ Evaluate the real gamma function.
\
\ Posted to comp.lang.forth by Andrew P. Haley, 03 Nov 2010,
\
\ Message ID:  <4qqdnVv1nNXWFkzRnZ2dnUVZ8uKdnZ2d@supernews.com>
\ with subsequent revisions to the Chebyshev coefficients,
\ and the algorithm.
\
\ Minor revs. by K. Myneni to use with FSL and ttester.
\ Requires fsl-util.4th and ttester.4th for automated tests.
\
\ Revisions:
\   2010-11-21  km  use coefficients 0--16 as default.
\
\ Evaluate a series in Chebyshev polynomials of the first kind.
\
\ Clenshaw's recurrence is:
\
\    b[r] = 2x*b[r+1] - b[r+2] + a[r]
\
\    f[x] = 1/2 (b[0] - b[2])
\
\ http://en.wikipedia.org/wiki/Clenshaw_algorithm
\

BASE @
DECIMAL

\ Single to float
[undefined] s>f [IF] : s>f ( n - F: n)   s>d d>f ; [THEN]

\ The fractional part of x
: frac ( F: x - y)   fdup floor f- ;

\ a0 is address of the zeroth coefficient, an the last.  The zeroth
\ coefficient in the table should be divided by two.

Private:

fvariable 2x
0 ptr a0
0 ptr aN

Public:

: chebev ( x a0 aN -- y)   \  ( a0 aN -- ) ( F: x - y)
    to aN  to a0
    2.0e0 f* 2x f!
    0.0e fdup
    a0 1 floats +  aN do
	( b[r+2] b[r+1] )
	fdup  2x f@ f*
	frot f-
	i f@ f+
    -1 floats +loop
    2x f@ 2.0e0 f/ f*  fswap f-  a0 f@ f+
;

Private:

\ Coefficients for gamma(1+x) in T* shifted Chebyshev series
\
\ 0 <= x <= 1
\
\ The last few coefficients for gamma are removed because they don't
\ contribute much useful accuracy.  If you're using IEEE double
\ precision, which has at best 16.25 decimal digits, you can remove
\ the last few coefficients without significant loss of accuracy.
22 FLOAT ARRAY gamma_coeffs{
     +2.82756448404890450051e+00 2.0e f/  ( 0)  \ 0x2.d3db441a5237716df4p+0
     +4.91728480089007251438e-01  ( 1)      \ 0x7.de1eaec7e7b96b5987p-4
     +8.53245417060960622435e-02  ( 2)      \ 0x1.5d7d4442c82c9229c9p-4
     +8.21455815467368275021e-03  ( 3)      \ 0x2.1a596aa01b1aab88edp-8
     +8.87927809469078194810e-04  ( 4)      \ 0x3.a30f4e7230d7614382p-12
     +5.67656791518861034089e-05  ( 5)      \ 0x3.b85ebc493de0d2ff07p-16
     +5.26407619053145198095e-06  ( 6)      \ 0x5.85108fb22099d43e84p-20
     +1.96127848487387649900e-07  ( 7)      \ 0x3.4a5cd995d61f7b7a75p-24
     +2.42546591996493612691e-08  ( 8)      \ 0x6.82c47a222fa09c06ddp-28
     -4.34058439984653663469e-11  ( 9)      \ -0x2.fb9a8af8b0579cfe0fp-36
     +1.28135422415311174339e-10  ( 10)     \ 0x8.ce2ea401268c9db9c9p-36
     -7.18515874473529316329e-12  ( 11)     \ -0x7.e671407e3c559c3973p-40
     +9.87886216808062770462e-13  ( 12)     \ 0x1.1610b43724aebafd2bp-40
     -8.93524894238542347525e-14  ( 13)     \ -0x1.9268681375728c2e12p-44
     +9.47267335939156879211e-15  ( 14)     \ 0x2.aa93fb2cf1b1862b9cp-48
     -9.42668601023366544621e-16  ( 15)     \ -0x4.3ed2a9b1dfd7e7a986p-52
     +9.59100989075118037123e-17  ( 16)     \ 0x6.e93aa2ee5b82fdc3a8p-56
\     -9.67795804241081532202e-18  ( 17)     \ -0xb.286dd5bce5b0a7d95fp-60
\     +9.78898946842480322954e-19  ( 18)     \ 0x1.20eb8362d7a6dcdda3p-60
\     -9.89123297962332627184e-20  ( 19)     \ -0x1.d319ab5bc2aecc09d9p-64
\     +9.99586299858095279825e-21  ( 20)     \ 0x2.f3441c21f346743a00p-68
\     -1.00999275547836639649e-21  ( 21)     \ -0x4.c5019e2b5541938d52p-72
17 constant ncoeffs
ncoeffs gamma_coeffs{ }fput
gamma_coeffs{  0 }  ptr  gamma_coeffs
gamma_coeffs{ ncoeffs 1- }  ptr  /gamma_coeff

Public:

\ gamma(x+2), 0 <= x <= 1
: gamma(x+2) ( F: x - y)
   2.0e0 f*  1.0e0 f-
   gamma_coeffs /gamma_coeff chebev ;

\ gamma(x), valid for all real x
\ The method basically follows Hart, p133
\ Computer Approximations, John F. Hart et al., Wiley 1968
\
\   X = frac(x)
\   N = x - X - 2
\
\   gamma(x) = prod(k=0..N-1) (X+2+k).gamma(X+2)   N > 0
\   gamma(x) = prod(k=1..-N) gamma(X+2)/(X+2-k)  N < 0
\

\ prod(k=0..n-1) (x+2+k).gamma(x+2)
: gamma1 \ ( n F: x gamma(x+2) - y)
    >r
    fswap  2.0e0 f+  fswap
    r> 0 do  fover f*  fswap 1.0e0 f+ fswap  loop
    fswap fdrop ;

\ prod(k=1..-n) gamma(x+2)/(x+2-k)
: gamma2 \ ( n F: x gamma(x+2) - y)
    >r
    fswap  1.0e0
    r> negate 1- -1 do  fover i s>f f-  f*  loop
    fswap fdrop  f/ ;

\ gamma(x), -1 <= x <= 0
: gamma3 ( F: x - y)
   fdup  1.0e0 f+  fdup gamma(x+2) -2 gamma2
   fswap f/ ;

\ gamma(x)
: gamma ( F: x - y)
   \ Avoid loss of significance for small -ve x
   fdup f0< if
      fdup -1.0e0 f> if  gamma3  exit  then then

   fdup frac  fswap fover
   f- f>d drop 2 - >r    \ R: int(x)-2
   fdup gamma(x+2)       \ F: frac(x)  gamma(x+2)

   r@ 0 > if  r@ gamma1
      else r@ 0< if  r@ gamma2
         else  fswap fdrop        
       then then
   r> drop ; 


Reset-Search-Order
BASE !

TEST-CODE? [IF] \ -----------------------------------------------
[UNDEFINED] T{ [IF] s" ttester" included [THEN]
BASE @
DECIMAL

0e ABS-NEAR F!
6e-16 REL-NEAR F!
SET-NEAR

cr
TESTING GAMMA
t{ -5.50000000000000000000e0   gamma ->  1.091265478190986298673234429377905644050e-2 r}t
t{ -1.50000000000000000000e0   gamma ->  2.363271801207354703064223311121526910397e0  r}t
t{ -5.00000000000000000000e-1  gamma -> -3.544907701811032054596334966682290365595e0  r}t
t{ -9.96093750000000000000e-2  gamma -> -1.072503239536597514958102042245354802713e1  r}t
t{ -9.76562500000000000000e-3  gamma -> -1.029869618821304653530065997279497010271e2  r}t
t{ -9.76562500000000000000e-4  gamma -> -1.024578182406251657399893334832135395808e3  r}t
t{ -9.15527343750000000000e-5  gamma -> -1.092324397288995617696983265948265923149e4  r}t
t{ -9.53674316406250000000e-6  gamma -> -1.048581772250973570707723687558376661871e5  r}t
t{ -9.53674316406250000000e-7  gamma -> -1.048576577216608139658440777112175594556e6  r}t
t{ -8.94069671630859375000e-8  gamma -> -1.118481124388241999670367804518639747770e7  r}t
t{ -9.31322574615478515625e-9  gamma -> -1.073741829772156741128346993949966485455e8  r}t
t{ -9.31322574615478515625e-10 gamma -> -1.073741824577215665822663037401346809860e9  r}t
t{  9.31322574615478515625e-10 gamma ->  1.073741823422784336019597314614097372460e9  r}t
t{  9.31322574615478515625e-9  gamma ->  1.073741818227843443097688207594467448551e8  r}t
t{  8.94069671630859375000e-8  gamma ->  1.118481008945109019362344877219153706402e7  r}t
t{  9.53674316406250000000e-7  gamma ->  1.048575422785278334942024740755654594258e6  r}t
t{  9.53674316406250000000e-6  gamma ->  1.048570227937673889355688064142857546799e5  r}t
t{  9.15527343750000000000e-5  gamma ->  1.092208954154494030761882107122382678852e4  r}t
t{  9.76562500000000000000e-4  gamma ->  1.023423749345567830369498718239930270889e3  r}t
t{  9.76562500000000000000e-3  gamma ->  1.018323574465657785492625352485558252969e2  r}t
t{  9.96093750000000000000e-2  gamma ->  9.552397687164723022417629289987808228824e0  r}t
t{  1.99707031250000000000e-1  gamma ->  4.597968034013318226695456341842933379706e0  r}t
t{  2.99804687500000000000e-1  gamma ->  2.993616880053459168661436986823960210170e0  r}t
t{  3.99902343750000000000e-1  gamma ->  2.218714529939859677739331379384482879923e0  r}t
t{  5.00000000000000000000e-1  gamma ->  1.772453850905516027298167483341145182798e0  r}t
t{  5.99609375000000000000e-1  gamma ->  1.490089134484006114176588874284823585028e0  r}t
t{  6.99707031250000000000e-1  gamma ->  1.298519535868670737632132738963089866089e0  r}t
t{  7.99804687500000000000e-1  gamma ->  1.164449217446022582886308246575328249578e0  r}t
t{  8.99902343750000000000e-1  gamma ->  1.068707497692992543225759611255141289229e0  r}t
t{  1.00000000000000000000e0   gamma ->  1.000000000000000000000000000000000000000e0  r}t
t{  1.50000000000000000000e0   gamma ->  8.862269254527580136490837416705725913988e-1 r}t
t{  2.00000000000000000000e0   gamma ->  1.000000000000000000000000000000000000000e0  r}t
t{  2.50000000000000000000e0   gamma ->  1.329340388179137020473625612505858887098e0  r}t
t{  3.00000000000000000000e0   gamma ->  2.000000000000000000000000000000000000000e0  r}t
t{  5.50000000000000000000e0   gamma ->  5.234277778455352018114900849241819367949e1  r}t
t{  2.00000000000000000000e1   gamma ->  1.216451004088320000000000000000000000000e17 r}t
t{  1.00000000000000000000e2   gamma ->  9.332621544394415268169923885626670049072e155 r}t

BASE !

[THEN]

