\ henon.4th
\
\ Compute the Henon map, its fixed point(s), the Jacobian,
\   and eigenvalues and eigenvectors. Use analytic methods
\   and numerical methods.
\
\ Copyright (c) 2002--2010, Krishna Myneni
\
\ Permission is granted to use this code for any purpose, as long
\ as the copyright notice above is preserved.
\ 
\ Requires:
\
\	fsl-util.4th
\       complex.4th
\	quadratic.4th
\       array-utils0.4th
\       eigen22.4th
\
\ Revisions:
\	2002-03-18 km  added storage for evolution of initial conditions
\       2010-10-27 km  revised to use FSL matrices and FSL-compatible 
\                      utilities

(

The Henon map is a two-dimensional map which can exhibit chaos.
It is given by:

	x_n+1 = B*y_n - {x_n}^2 + A
	y_n+1 = x_n 

A and B are parameters, which for certain selections, give rise
to a chaotic sequence of values for x_n, y_n.

Here, we use A = 1.4, B = 0.3, which give chaos. The user
may experiment with other values of A and B as well. To observe 
the chaotic sequence, select an initial point such as 
x = 0.5, y = 0.5; then apply the map repeatedly.


Example:	0.5e 0.5e H 

	Applies the map once and returns the next x and y on the stack.
	The new x and y values may be printed by typing:

		fswap f. f.

-------
Example:	0.5e 0.5e 10 times

	applies the map 10 times and shows the new point at each iteration.
	To see the chaotic "attractor", iterate the map a large number of
	times and redirect the output to a file: 

		>file henon.dat 0.5e 0.5e 1000 times console

	The x, y pairs may then be plotted using a program such as xyplot.
	Plot the data as discrete points, rather than connected by line
	segments, to see the attractor properly.
 
-------
Example:	solve_fixed_point

	finds the point xf, yf at which the sequence remains stationary,
        i.e.

                x_n+1 = x_n
	        y_n+1 = y_n

	Applying these conditions to the Henon map leads to the
	quadratic equation:

		{xf}^2 + {1 - B}*xf - A = 0

	One may see that the Henon map stays at the fixed point over
	several iterations by doing,

		solve_fixed_point set_fixed_point
		xf f@ yf f@ 20 times

	However, typing

		xf f@ yf f@ 200 times

	shows that the sequence become chaotic once again. This is because the
	fixed point is "unstable". Small displacements about the actual fixed 
	point cause the sequence to move away rapidly from the fixed point.
	Since our solution of the fixed point is only to finite precision,
	we have in effect a displacement from the actual fixed point.

	'solve_fixed_point' finds the analytic solution of the fixed point,
	specific to the Henon map parameters. A numerical method for finding 
	the fixed point may also be used:

		<x> <y> search_fixed_point

	where x and y are an initial starting point for the search
	routine. As the map is iterated, the same point must be seen
	successively 5 times, to within a specified tolerance, to be
	considered the fixed point. The search may take a long time,
	depending on the initial point which is specified.

-------
Example:	solve_fixed_point set_fixed_point
	        xf f@ yf f@ setup-ball
		evolve-ball

	A ball of points around the fixed point are stored, and
        then one iteration of the map is executed for each of the points.
        It can be seen that the ball of "initial conditions" transforms
        to an ellipse, contracting in one direction and expanding along 
        another direction. After executing the above instructions, type

		>file ellipse.dat 360 2 ellipse{{ }}fprint console

	to write the evolved x,y pairs to a file. Plot the data
	to observe the ellipse.

-------
Example:	solve_fixed_point set_fixed_point
		xf f@ yf f@ Jacobian
		eigenvalues
		eigenvectors
		2 2 e{{ }}fprint

	computes and prints the stable and unstable eigenvectors
	at the fixed point. The eigenvectors give the directions
	along which the fixed point is stable and unstable. First
	column is the stable direction, second column is unstable
	direction.
)

include ans-words
include fsl/fsl-util
include fsl/complex
include fsl/quadratic
include fsl/extras/array-utils0
include fsl/extras/eigen22

[undefined] fsquare [IF] 
: fsquare postpone fdup postpone f* ; immediate
[THEN]

\ The following parameters for the Henon map are chosen to
\   produce chaos
fvariable A	1.4e A f!
fvariable B     0.3e B f!

\ Iterate the Henon map once
: H ( x y -- xn yn)
    B F@ F*  fover fsquare F-    \ B*y_n - x_n^2
    A F@ F+  fswap ;


\ Iterate the Henon map n times starting from the given point
\   (initial condition) and print the x and y values at each
\   iteration.

: times ( x0 y0 n -- x_n y_n )
    0 ?DO  fover F. 2 spaces fdup F. cr  H  LOOP ;

\ Fixed point methods

fvariable xf
fvariable yf

: set_fixed_point ( x y -- )
    yf F! xf F! ;

\ ====== Analytical solution of fixed point for the Henon map

\ The equation to solve is -x^2 + (B-1)*x + A = 0

: solve_fixed_point ( -- x y )
    -1e  B f@ 1e f- A f@ solve_quadratic
    fdrop  \ assume real root and take only the positive root
    fdup F0< IF f2drop ELSE fnip fnip THEN fdup 
;

\ ======= Numerical method for finding the fixed point.

\ Starting with initial condition x0 and y0, iterate the Henon map
\   to find the unstable fixed point.

1e-4 fconstant tolerance

fvariable x 
fvariable y
variable seen

: some_where ( -- )   0 seen ! ;
: seen_again ( -- )   1 seen +! ;

: remember_this_point ( x y -- x y ) fdup y F! fover x F! ;

: distance_to_last ( x y -- x y dist)
    f2dup y F@ F- fsquare fswap x F@ F- fsquare F+ fsqrt ;

: same_point? ( x y -- flag | true if same as last point)
    distance_to_last tolerance F< ;

: search_fixed_point ( x0 y0 -- xf yf )
    some_where
    BEGIN
	remember_this_point	
	H			\ apply the map once
	same_point?
	IF seen_again ELSE some_where THEN
	seen @ 5 =
    UNTIL ;


\ ======= Analytical methods for determining stability of the fixed point

2 2 FLOAT matrix DM{{

: Jacobian ( x y -- | compute the Jacobian at point x,y)
    fdrop          ( Jacobian for Henon Map does not depend on y or A)
    -2e F* DM{{ 0 0 }} F!
    B f@   DM{{ 0 1 }} F!
    1e     DM{{ 1 0 }} F!
    0e     DM{{ 1 1 }} F!  ;


\ Find eigenvalues of 2x2 Henon map Jacobian matrix.
\ The eigenvalues are real for the Henon map. Order the 
\ eigenvalues so that lambda1 has smaller magnitude.
\ When the Jacobian is evaluated at the fixed point, 
\ lambda1 will be the stable eigenvalue and lambda2 
\ will be the unstable eigenvalue

: eigenvalues ( -- lambda1 lambda2)
    DM{{ }}feigenvalues22 fdrop fnip
    f2dup F> IF fswap THEN ;

\ eigenvector matrix
2 2 FLOAT matrix e{{

: eigenvectors ( -- )
    DM{{ e{{ 2 2 }}fcopy
    e{{ }}feigen22  f2drop f2drop
;

\ ======== Calculations on a ball of initial conditions

: dist ( x1 y1 x2 y2 -- d | distance between two points)
    frot f- fsquare 2>r f- fsquare 2r> f+ fsqrt ;


fvariable theta
1e-2 fconstant ball_radius  ( chosen specifically for this map )

360 2 FLOAT MATRIX ellipse{{

: setup-ball ( x y -- | create a ball of initial conditions about a point )
	\ store the initial conditions in the matrix ellipse
	360 0 do
	  f2dup
	  I s>f deg>rad theta f!
	  fswap theta f@ fcos ball_radius f* f+	\ starting x
	  fswap theta f@ fsin ball_radius f* f+	\ starting y
	  \ store the evolved point
	  ellipse{{ I 1 }} F!  ellipse{{ I 0 }} F!
	loop 
	f2drop ;
 
: evolve-ball ( -- | evolve the ball of initial conditions by one iteration )
        360 0 do  
	  ellipse{{ I 0 }} f@  ellipse{{ I 1 }} f@ 
	  H
	  ellipse{{ I 1 }} f!  ellipse{{ I 0 }} f!
        loop
;


: major_axis? ( -- theta_major flength | determine major axis of ellipse)
	0 0e
	180 0 do
	  \ i ellipse frow@ drop
	  ellipse{{ I  0 }} F@  ellipse{{ I  1 }} F@
	  I 180 + >r 
	  ellipse{{ r@ 0 }} F@  ellipse{{ r> 1 }} F@ 
	  dist f2dup
	  f< if fnip rot drop I 1- -rot else fdrop then
	loop
	rot s>f deg>rad fswap ;


: minor_axis? ( -- theta_minor flength | determine minor axis of ellipse)
	0 1e100
	180 0 do
	  \ i ellipse frow@ drop
	  ellipse{{ I 0 }} F@  ellipse{{ I 1 }} F@
	  I 180 + >r 
	  ellipse{{ r@ 0 }} F@  ellipse{{ r> 1 }} F@
	  dist f2dup
	  f> if fnip rot drop i 1- -rot else fdrop then
	loop
	rot s>f deg>rad fswap ;  



