\ quadratic.4th
\
\ Solve for the two complex solutions of the quadratic equation:
\
\   a*x^2 + b*x + c = 0
\
\ for real coefficients a, b, and c. The two complex roots are 
\ z1 = x1r + i*x1i and z2 = x2r + i*x2i
\
\ Environmental dependencies:
\   1. requires FLOATING and FLOATING EXT wordsets
\   2. supports unified or separate floating point stack
\
\ Provides:
\   Q_DISCRIMINANT
\   SOLVE_QUADRATIC
\
\  Copyright 2002--2009, K. Myneni, krishna.myneni@ccreweb.org
\  Permission is granted to modify and use this code
\  for any application, provided this notice is preserved.
\
\ Notes:
\
\   1. Ensure that argument "a" is not zero, or an infinity will result;
\      the correct solution of the simple linear equation will not be
\      given.
\  
\   2. The complex number roots, returned by solve_quadratic, are 
\      ordered as follows: 
\   
\         a) For real roots, r1 and r2, with r1 < r2, the return
\            order is ( F: -- z2 z1 ), where z1 = r1 +i0, z2 = r2 + i0
\
\         b) For complex conjugate roots, z1 = e - i*f, z2 = e + i*f,
\            (f > 0), the ordering on the stack is ( F: -- z2 z1 ).
\
\ Revisions:
\
\   2002-11-02  km; first version.
\   2003-10-25  Christopher Brannon; fixed problem with calculation
\                 of complex roots.
\   2007-11-04  km; revised comments; added test code; save and restore base. 
\   2009-08-05  km; revised to preserve accuracy when the product a*c is
\                   much less than b^2; see [1]. Added new test case.
\   2009-08-11  km; added test case for purely imaginary roots.
\   2009-08-18  km; factored the code with modification of example posted
\                   to comp.lang.forth by "humptydumpty", and used words from
\                   the jvn-dnw complex library; revised comments; V1.2
\   2009-08-19  km; additional comments, change "zconj" to "conjg"; V1.2b
\   2009-08-26  km; revised q_discriminant and added (q_discriminant). Also
\                   added q_check to test for b=0 AND c=0 case, and a test
\                   for this case.
\ References:
\
\ 1. W.H. Press, et. al., Numerical Recipes in C, 2nd ed., pp. 183--184,
\    eqns. 5.6.4 and 5.6.5.

CR .( QUADRATIC         V1.2b         19 August    2009   KM )
BASE @ DECIMAL

[UNDEFINED] fsquare [IF] : fsquare fdup f* ;      [THEN]
[UNDEFINED] f2dup   [IF] : f2dup   fover fover ;  [THEN]

\ Useful words from the complex library
[UNDEFINED] zdup  [IF] : zdup  fover fover ; [THEN]
[UNDEFINED] conjg [IF] : conjg fnegate ;     [THEN]

: q_discriminant ( F: a b c -- d )
     frot f* -4e f* fswap fsquare f+ ;

\ For thread-safety, make the following variables per thread 
\ variables if using this code in a multi-threaded Forth 
\ environment.
fvariable qa
fvariable 2qa
fvariable qb
fvariable qc

: (q_discriminant) ( F: -- d )
    qb f@ fsquare 4e qa f@ f* qc f@ f* f- ;  
    
: q_real_roots ( F: d -- r1 r2 )
    fsqrt qb f@ fdup f0< IF f- ELSE f+ fnegate THEN 2e f/ 
    qc f@ fover f/ fswap qa f@ f/ ;

: q_complex_roots ( F: d -- z1 z2 )
    fabs  fsqrt   2qa f@  f/           \ imaginary part 
    qb f@ fnegate 2qa f@  f/           \ real part
    fswap zdup conjg                   \ complex conjugate
;

: q_check ( F: b c -- flag )
    f0= >r f0= r> and ;

: solve_quadratic ( F: a b c -- z1 z2 )
    f2dup q_check IF f2drop fdrop 0e 0e zdup EXIT THEN  
    qc f! qb f! fdup qa f! 2e f* 2qa f!
    (q_discriminant)
    fdup f0<  IF                       \ complex conjugate roots
       q_complex_roots 
    ELSE                               \ two real roots
       q_real_roots                    \ F: -- r1 r2
       0e fswap 0e                     \ promote to z1 z2
    THEN
;

BASE !

TEST-CODE? [IF]   \ test code ==============================================
[undefined] T{      [IF]  include ttester  [THEN]    
BASE @ DECIMAL

1e-15  rel-near F!
1e-256 abs-near F!
set-near

\ Examples from:
\
\    http://www.purplemath.com/modules/quadform.htm

-2e 3e F/       fconstant -2/3
-3e 2e F/       fconstant -3/2
 5e fsqrt       fconstant sqrt{5}
 2e fsqrt 3e F/ fconstant sqrt{2}/3
 3e fsqrt 2e F/ fconstant sqrt{3}/2
10e fsqrt 2e F/ fconstant sqrt{10}/2

: zz}t  rrrr}t ;

CR
TESTING SOLVE_QUADRATIC
t{ 1e  0e  0e solve_quadratic ->  0e               0e    0e               0e   zz}t
t{ 1e  3e -4e solve_quadratic ->  1e               0e   -4e               0e   zz}t
t{ 2e -4e -3e solve_quadratic ->  1e sqrt{10}/2 F- 0e    1e sqrt{10}/2 F+ 0e   zz}t
t{ 1e -2e -4e solve_quadratic ->  1e sqrt{5} F-    0e    1e sqrt{5} F+    0e   zz}t
t{ 9e 12e  4e solve_quadratic ->  -2/3             0e   -2/3              0e   zz}t
t{ 3e  4e  2e solve_quadratic ->  -2/3      sqrt{2}/3   -2/3 sqrt{2}/3 fnegate zz}t
t{ 1e  3e  3e solve_quadratic ->  -3/2      sqrt{3}/2   -3/2 sqrt{3}/2 fnegate zz}t
t{ 1e  0e  1e solve_quadratic ->  0e               1e    0e              -1e   zz}t

\ Test case which loses accuracy with ordinary quadratic formula:
\ 
\        x^2 + x + c = 0
\
\ when c << 1, the approximate solution is  x1 = -c,  x2 = -1 + c
\
t{ 1e 1e 1e-17 solve_quadratic -> -1e-17           0e    -1e 1e-17 f+     0e   zz}t


BASE !
[THEN]

