\ simple-fonts-x11.4th
\
\ Simplified font loading and selection for X11 windows.
\
\ 
\ Copyright (c) 2012 Krishna Myneni, Creative Consulting for
\   Research and Education, http://ccreweb.org
\
\ This code may be used for any purpose, provided the copyright
\ notice above is included.
\
\ Provides:
\
\   Font weight constants:  medium  bold
\   Font slant  constants:  regular  italic  oblique
\
\   Font Tables
\
\     FontTable      ( "name" -- )         \ create a font table
\     >font-index    ( weight slant ptsize -- index )
\     !font-entry    ( afontstruct font_id index atable -- )
\     @font-entry    ( index atable -- afontstruct font_id)
\
\   Loading and Setting Fonts and Font Tables
\
\     load-common-font  ( wgt slnt ptsize family -- afstruct font_id flag )
\     load-font-table   ( family atable -- )  \ load specific fonts in family
\     unload-font-table ( atable -- )        \ unload all fonts in a table
\     select-font       ( weight slant ptsize atable -- )
\     get-current-font  ( -- afontstruct )
\
\   Font Geometry
\
\     get-font-height  ( afontstruct -- height )  \ height in pixels of a font
\     current-font-height ( -- height )           \ in pixels
\     get-string-box ( caddr u -- width height  ) \ width and height in pixels
\                                                 \  of string in current font      
\  
\ Revisions:
\   2012-04-28  km  created.
\   2012-04-30  km  added GET-CURRENT-FONT and Font Geometry words
\

Module: simple-fonts-x11

Also font-strings-x11
Also simple-graphics-x11

Begin-Module

Public:

\ Simple names for font weights and slants
FONT_WEIGHT_MEDIUM constant  medium
FONT_WEIGHT_BOLD   constant  bold
FONT_SLANT_REGULAR constant  regular
FONT_SLANT_ITALIC  constant  italic
FONT_SLANT_OBLIQUE constant  oblique

 5 constant SUPPORTED_POINTSIZES
20 constant MAX_FAMILY_FONTS
MAX_FAMILY_FONTS cells 2* constant FONT_TABLE_SIZE

Private:

: pointsize>index ( pointsize -- index )
   100 max 240 min   \ enforce bounds on pointsize
   100 - 20 /        \ 0, 1, 2, 4, 7
   dup 3 7 within if drop 3 then
   dup 3 > if drop 4 then
;

: slant>index  ( slant -- index )  1- 0 max 1 min ;
: weight>index ( weight -- index ) 1- 0 max 1 min ;

Public:
   
: >font-index ( weight slant pointsize -- fontindex )
    pointsize>index >r  slant>index >r  weight>index
    SUPPORTED_POINTSIZES 2* * r> SUPPORTED_POINTSIZES * + r> +
;


: load-common-font ( weight slant pointsize family -- afontstruct font_id flag)
    get-common-font-xlfd
    dup 0= IF  2drop 0 0 false  ELSE  load-font  THEN
;

\ Store information about a loaded font into the specified font table
: !font-entry ( afontstruct  font_id  index  atable --  )  
    over 0 MAX_FAMILY_FONTS within 
    IF >r cells 2* r> + 2! ELSE  2drop 2drop THEN ;

: @font-entry ( index atable -- afontstruct font_id )
    >r cells 2* r> + dup >r cell+ a@ r> @ ;

: ?add-font ( weight slant pointsize family atable -- )
    >r 2over 2over drop >font-index >r
    load-common-font
    IF    r> r> !font-entry  
    ELSE  ." load-common-font failed! font index = " r> .
          r> drop 2drop 
    THEN ;

: FontTable  create FONT_TABLE_SIZE allot ;

\ Loads 20 specific fonts from a font family and stores
\ the loaded font information into the specified table.
\ Fonts from the font family are loaded at two different
\ weights (medium, bold), two different slants (regular,
\ italic), and five different point sizes: 
\ 100, 120, 140, 180, 240.

: load-font-table ( family atable -- )
    2>r 
    medium regular 100  2r@  ?add-font
    medium regular 120  2r@  ?add-font
    medium regular 140  2r@  ?add-font
    medium regular 180  2r@  ?add-font
    medium regular 240  2r@  ?add-font
    medium italic  100  2r@  ?add-font
    medium italic  120  2r@  ?add-font
    medium italic  140  2r@  ?add-font
    medium italic  180  2r@  ?add-font
    medium italic  240  2r@  ?add-font
    bold   regular 100  2r@  ?add-font
    bold   regular 120  2r@  ?add-font
    bold   regular 140  2r@  ?add-font
    bold   regular 180  2r@  ?add-font
    bold   regular 240  2r@  ?add-font
    bold   italic  100  2r@  ?add-font
    bold   italic  120  2r@  ?add-font
    bold   italic  140  2r@  ?add-font
    bold   italic  180  2r@  ?add-font
    bold   italic  240  2r@  ?add-font    
    2r> 2drop
;
 
: unload-font-table ( atable -- )
    MAX_FAMILY_FONTS 0 do
      I over @font-entry 
      dup IF unload-font ELSE drop THEN drop
    loop
    drop
;

Private:

variable pCurrentFont       \ pointer to XFontstruct for current font
variable CurrentFontHeight

Public:

\ Return the font height in pixels for a specified font
: get-font-height ( afontstruct -- height )
    dup IF
      dup >r  XFontStruct->ascent @  r>  XFontStruct->descent @ + 
    ELSE  drop 0
    THEN ;

\ Select a font from the specified table, by specifying
\ its weight (medium or bold), slant (regular or italic)
\ and its point size. If the point size is not one of
\ standard point sizes, the closest match will be used.

: select-font ( weight slant pointsize atable -- )
    >r >font-index r> @font-entry  set-font  
    dup pCurrentFont !
    \ store the font height 
    get-font-height CurrentFontHeight !
;

\ Return the pointer to the x11 font structure for current font
: get-current-font ( -- afontstruct ) pCurrentFont a@ ;

\ Return the current font height in pixels
: current-font-height ( -- height ) CurrentFontHeight @ ;

\ Return the width and height in pixels of a string in the current font
: get-string-box ( caddr u -- width height )
    strpck count 2>r pCurrentFont a@ 2r> XTextWidth 
    CurrentFontHeight @ 
;
        
End-Module

