\ ogy.4th
\
\ Demonstrate Ott-Grebogi-Yorke chaos control method for a fixed point
\   on the Henon map, following OGYs original paper [1].
\
\   
\ Copyright (c) 2002--2012, Krishna Myneni, Creative Consulting for
\   Research and Education, http://ccreweb.org
\
\ Requires:
\
\	henon.4th
\
\ Notes:
\
\ 0) This program demonstrates that a point iterated on the Henon
\    map can be captured and held at the fixed point by means of
\    very tiny perturbations to the map. In this demonstration,
\    the parameter A in the Henon map is changed by no more than
\    a maximum value of p*. When the iterated point wanders close
\    to the fixed point, then the Ott-Grebogi-Yorke (OGY) control
\    method[1] is used to compute the control signal, p, required
\    to cause the next iterate to fall on the stable axis of
\    the fixed point. The control signal will quickly fall off,
\    and the value of the control signal required to hold the 
\    iterated point on the fixed point will approach the numerical
\    "noise" in the system, due to the finite machine precision of
\    the calculations.
\
\ 1) For the default parameters used for the Henon map, the
\    fixed point lies at xf = yf = 0.883896. Try the following,
\
\         0.8e 0.8e 150 ogy
\
\    You may start from other nearby points, even x = y = -1e,
\    but the number of iterations required for the point to
\    come near the fixed point may be large, i.e. you will need
\    to increase the number of iterations from 150 to a larger
\    number.
\
\ References:
\
\  1. E. Ott, C. Grebogi, and J.A. Yorke, "Controlling Chaos",
\     Physical Review Letters, vol 64, p 1196 (1990)
\
\
\ Revisions:
\   2002-01-20  km  first version
\   2012-04-22  km  revised to use fsl matrices and be compatible
\                   with latest version of henon.4th; added notes.
\   2012-04-24  km  revised notes.
\
include henon

fvariable p	0e p f!		\ control parameter
fvariable p*    1e-2 p* f!      \ max control parameter
fvariable A0    A f@ A0 f!

: set_control ( f -- | set control parameter )
	fdup p f! A0 f@ f+ A f! ;

: g (  -- dx dy | determine the control vector g near the fixed point)
	0e set_control          \ start with zero control parameter
	solve_fixed_point       \ determine the fixed pt starting somewhere 
	set_fixed_point		\ save the fixed point
	1e-3 set_control	\ change the control parameter by small amount 
	solve_fixed_point       \ find the new fixed point
	yf f@ f- p f@ f/ fswap 
	xf f@ f- p f@ f/ fswap  \ compute vector g
	0e set_control ;

\ OGY's f vectors. These vectors are defined by the relations:
\
\	fs * es = 1    fs * eu = 0
\       fu * es = 0    fu * eu = 1
\
\ where es and eu are the eigenvectors for the stable and unstable
\ directions and '*' indicates a dot product. The f vectors are 
\ ordered such that f_s is in 1st col, f_u is in 2nd col

2 2 float matrix f{{  \ OGY's f vectors

: fvectors ( -- | determine the f vectors)

	\ Compute fs

	e{{ 0 1 }} f@  e{{ 1 0 }} f@  f*  ( eux*esy )
	e{{ 0 0 }} f@  e{{ 1 1 }} f@  f*  ( esx*euy )
	f- 2>r
	e{{ 1 1 }} f@ fnegate 2r@ f/ f{{ 0 0 }} f!    ( fsx )
	e{{ 0 1 }} f@         2r@ f/ f{{ 1 0 }} f!    ( fsy )

	\ Compute fu

	2r> fnegate 2>r
	e{{ 1 0 }} f@ fnegate 2r@ f/  f{{ 0 1 }} f!   ( fux )
	e{{ 0 0 }} f@         2r> f/  f{{ 1 1 }} f!   ( fuy )	
;

: fu ( -- fux fuy | return the components of fu)
	f{{ 0 1 }} f@  f{{ 1 1 }} f@ ;

: dot ( x1 y1 x2 y2 -- s | compute dot product of two vectors)
	frot f* 2>r f* 2r> f+ ;
 
fvariable control_scale
fvariable lambda2
fvariable lambda1

: compute_scale_factor ( -- | precompute the control parameter scale factor)
	lambda2 f@ fdup 1e f- f/
	g fu dot f/ 
	control_scale f! ;

: compute_control ( x y -- p | compute the control parameter based on state)
	yf f@ f- fswap xf f@ f- fswap  ( subtract fixed point vector) 
	fu dot 
	control_scale f@ f* ;

\ Starting with the given initial state, apply the OGY control technique
\   to the Henon map and display the state upon each iteration, up to n times

: ogy ( x y n -- | apply the OGY control method )

	\ Preliminaries -- compute relevant quantities of Henon map,
	\   the f vectors, the control vector g, and the control scale factor

	A f@ A0 f!  	\ keep copy of uncontrolled A parameter
	solve_fixed_point 
	Jacobian 
	eigenvalues   lambda2 f! lambda1 f!
	eigenvectors
	fvectors
	compute_scale_factor

	\ Now we are ready to iterate the map and apply the control

	0 do
	  fover f. 2 spaces fdup f. \ show current x y
	  fover fover compute_control 
	  fdup fabs p* f@ f> if fdrop 0e then 
	  fdup 2 spaces f. cr       \ show parameter perturbation p 
	  set_control
	  H			    \ iterate the map 
	loop
	fdrop fdrop
	A0 f@ A f!                  \ restore original parameter
;

