\ schr1d.4th
\
\ Solve the eigenvalue problem for the radial Schroedinger 
\ equation, obtaining both a bound state eigenvalue and the 
\ corresponding eigenfunction, given the potential energy
\ function and an initial trial energy.
\
\ The one-dimensional radial Schroedinger equation may be
\ cast in the form,
\
\   P''(r) = 2*mu*(V(r) - E)*P(r)
\
\ The Numerov integration method can be used to solve this
\ equation for the unknowns, E and P(r) [1--4]. The discrete 
\ solutions for the eigenvalues, E, and the corresponding
\ radial eigenfunctions, P(r), may be found for a given 
\ potential energy function, V(r).
\
\ Q(r) is defined to be 2*mu*(V(r) - E), so that the equation 
\ to be solved, for both E and P(r), is,
\
\               P''(r) = Q(r)*P(r)
\
\ Units for the above equations are atomic units.
\
\ References:
\
\ 1. Anders Blom, 2002, "Computing algorithms for solving the 
\    Schroedinger and Poisson equations", available on the web at
\    http://www.teorfys.lu.se/personal/Anders.Blom/useful/scr.pdf
\
\ 2. Tao Pang, 1997, "An Introduction to Computational Physics",
\    Cambridge University Press, see program code3d.f.
\    Also on the web at 
\    http://www.physics.unlv.edu/~pang/comp/code3d.f
\
\ 3. R. J. LeRoy, 2007, "Level 8.0, A Computer Program for 
\    Solving the Radial Schroedinger Equation for Bound and 
\    Quasibound Levels", University of Waterloo, report CP-663. 
\    Software and manual available from the web at 
\    http://scienide2.uwaterloo.ca/~rleroy/level/
\    
\ 4. J. W. Cooley, 1961, "An Improved Eigenvalue Corrector Formula
\    for Solving the Schroedinger Equation for Central Fields",
\    Mathematics of Computation, 15, pp. 363--374.
\
\
\ Revisions:
\   2010-10-18  km  first version; analytic potential functions
\                   not yet supported.
\   2010-10-19  km  revise comments.
\   2011-08-15  km  fixed SOLVE to avoid infinite loop condition
\                     by adding divide by two to energy increment
\                     when reversing direction away from zero crossing.
\
\ Copyright (c) 2010 Krishna Myneni
\
\ This code may be used for any purpose, as long as the copyright
\ notice above is preserved.
\
include ans-words
include fsl/fsl-util
include fsl/dynmem
include fsl/polrat
include fsl/extras/numerov
include fsl/extras/find
include fsl/extras/read_xyfile.4th
include fsl/extras/array-utils0.4th
include fsl/extras/array-utils1.4th

[undefined] fsquare [IF] 
: fsquare postpone fdup postpone f* ; immediate
[THEN]

219474.6314e    fconstant  cm^-1/Eh  \ Hartree to wavenumber conversion
0.5291772083e   fconstant  A/a0  \ Bohr radius to Angstroms conversion

1e              fconstant  m_e   \ mass of electron in atomic units
1836.15267247e  fconstant  m_p   \ mass of proton in atomic units


\ NPOTNL is the number of points in the potential energy 
\ function, V(r). If NPOTNL < 0, use an analytic function, 
\ referred to by PPOTNL, to compute V(r). 
variable Npotnl  
Defer ppotnl

\ Scale factors to convert raw distance and potential energy
\   data to atomic units.
fvariable rscale
fvariable Vscale
1e rscale F!
1e Vscale F!


\ Inputs are the position and potential energy arrays

4096 constant MAX_POTNL
MAX_POTNL FLOAT ARRAY r{
MAX_POTNL FLOAT ARRAY V{
fvariable Rmin
fvariable Rmax
fvariable Vmin
fvariable Vmax
fvariable Rdel_min
 
\ Read the potential function data from a file, into the 1D 
\ arrays, r{, and V{, and find its characteristics.
  
: read-potnl ( a u -- )
    r{ V{ 2swap read_xyfile ABORT" Unable to read input file!"
    dup Npotnl !
    dup r{ }fmin  Rmin  F!
    dup r{ }fmax  Rmax  F!
    dup V{ }fmin  Vmin  F!
    dup V{ }fmax  Vmax  F!
    \ Find minimum increment for r{ values
    >r
    r{ 1 } F@  r{ 0 } F@ F- fabs
    r> 2 DO  r{ I } F@  r{ I 1- } F@ F- fabs fmin  LOOP
    Rdel_min F!
;

12 constant MAX_INTERP
MAX_INTERP FLOAT ARRAY rI{
MAX_INTERP FLOAT ARRAY VI{

variable interp_order
8 interp_order !  \ default polynomial order

\ Interpolate the value of V[r] for a given r, Rmin <= r <= Rmax
\ Assumes r{ and V{ arrays are in ascending order for r
: InterpolateV ( r -- V[r] )
    fdup Rmin F@ F<= IF fdrop V{ 0 } F@ EXIT THEN  \ return left end point
    fdup Rmax F@ F>= IF fdrop V{ Npotnl @ 1- } F@ EXIT THEN \ return right end point
    fdup Npotnl @ r{ }ffind >R
    fdup r{ R@ } F@ F= IF  fdrop V{ R> } F@ EXIT  THEN  \ exact match
    R> interp_order @ 2/ - 0 max
    dup interp_order @ + Npotnl @ > 
    IF drop Npotnl @ interp_order @ - THEN
    interp_order @ 0 DO  
      r{ over } F@  rI{ I } F!  
      V{ over } F@  VI{ I } F!
      1+
    LOOP drop 
    rI{  VI{ interp_order @ polint ( ratint) fdrop
;

\ Set up arrays for position and potential energy on
\ the fixed length mesh.
variable Nmesh   \ mesh size for integrator
fvariable h      \ step size on r_mesh{
FLOAT DARRAY r_mesh{
FLOAT DARRAY V_mesh{
FLOAT DARRAY Pl{
FLOAT DARRAY Pr{
FLOAT DARRAY Ql{
FLOAT DARRAY Qr{

: pad2 ( n1 -- n2 ) 
   2 BEGIN  1 LSHIFT 2DUP < UNTIL NIP ;

\ Set up the interpolated mesh for the potential
\ Nmin is the minimum number of points in the mesh, and
\ norder is the order of the interpolation polynomial
: setup-mesh ( Nmin norder -- )
   dup MAX_INTERP > ABORT" setup-mesh: interpolation order too large!"
   interp_order !
   Rmax F@ Rmin F@ F- Rdel_min F@ F/ 
   ftrunc>s pad2 max Nmesh !
   Rmax F@ Rmin F@ F- Nmesh @ 1- s>f F/ h F!
   & r_mesh{ Nmesh @ }malloc
   & V_mesh{ Nmesh @ }malloc
   & Pl{ Nmesh @ }malloc
   & Pr{ Nmesh @ }malloc
   & Ql{ Nmesh @ }malloc
   & Qr{ Nmesh @ }malloc   
   malloc-fail? ABORT" setup-mesh: Can't allocate mem!"
   Rmin F@  Nmesh @ 0 DO fdup r_mesh{ I } F!  h F@ F+  LOOP   
   fdrop
   Nmesh @ 0 DO  r_mesh{ I } F@ InterpolateV  V_mesh{ I } F!  LOOP
;


variable  mi
variable  mo

\ Normalize the integral of P(r)^2 using trapezoid integration
fvariable Pnorm
: normalize-Ps ( -- )
    Pl{ 0 } F@ fsquare 0.5e F*   
    mo @ 1 DO  Pl{ I } F@ fsquare F+  LOOP  
    Pl{ mo @ } F@ fsquare 0.5e F* F+ 
    Pr{ 0 } F@ fsquare 0.5e F*   
    mi @ 1 DO  Pr{ I } F@ fsquare F+  LOOP  
    Pr{ mi @ } F@ fsquare 0.5e F* F+
    F+ h F@ F*  fsqrt fdup Pnorm F!
    mo @ 2+ 0 DO  Pl{ I } F@ fover F/  Pl{ I } F!  LOOP
    mi @ 2+ 0 DO  Pr{ I } F@ fover F/  Pr{ I } F!  LOOP
    fdrop
;

\ Rescale Pl to equal Pr at the matching point
: match-Ps ( -- )
    Pr{ mi @ } F@  Pl{ mo @ } F@  F/  mo @ 2+ Pl{ }fscale ;

\ Join the inward solution, Pr, to the outward solution, Pl.
\ The contiguous radial function will be stored in Pl.
: join-lr ( -- )
    Nmesh @ mo @ 1+ DO  
      Pr{ Nmesh @ I - 1- } F@  Pl{ I } F!  
    LOOP  ;


\ Integrate the radial Schroedinger equation, inward and outward,
\   for a given energy, and return the difference in slopes at the
\   matching point.

fvariable mu   \ reduced mass in Schr. Eqn. (in atomic units)

: schr-integrate ( E -- delSlope nerror)
    \ Initialize Ql and Qr
    Nmesh @ 0 DO  V_mesh{ I } F@ fover F-   mu F@ 2e F* F* Ql{ I } F!  LOOP
    fdrop   \ -- 
    Nmesh @ 0 DO  Ql{ Nmesh @ I - 1- } F@   Qr{ I } F!  LOOP

    \ Find rightmost zero crossing of Ql{}; this will be the
    \   matching point for the outward and inward solutions
    0 
    1 Nmesh @ 1- DO  
      Ql{ I }    F@  0e F>  Ql{ I 1- } F@  0e F<=  and
      IF  drop I  leave THEN
    -1 +LOOP
    dup 0= IF drop 0e 1 EXIT THEN  \ ERROR: Unable to find zero crossing!
    mo !
    Nmesh @ mo @ - 1- mi !   \ indices mo and mi correspond to same r

    \ Outward integration to one step past zero crossing
    Pl{  Ql{  mo @ 2+ h F@ numerov_integrate

    \ Inward integration to one step before zero crossing
    Pr{  Qr{  mi @ 2+ h F@ numerov_integrate

    match-Ps
    normalize-Ps

    \ Determine slope difference between outward and inward solutions
    Pl{ mo @ 1+ } F@  Pl{ mo @ 1- } F@ F- h F@ 2e F* F/   \ outward slope
    Pr{ mi @ 1- } F@  Pr{ mi @ 1+ } F@ F- h F@ 2e F* F/   \ inward  slope
    F- 
    0  \ no error
;



\ Given a trial energy eigenvalue, find the nearest actual eigenvalue
\   to within the absolute error, Etol.

fvariable dE
fvariable sd1
fvariable sd2
fvariable Etol
1e-9 Etol F!   \ Default absolute error in eigenvalue (in Hartrees)

false value verbose?

: solve ( E dE -- E' dE' nerror )
    dE F!
    \ Sanity check for starting value of E
    fdup Vmin F@ F<= IF  fdrop Vmin F@ dE F@ fabs 1.5e F* F+  THEN
 
    Nmesh @ Pl{ }fzero  Nmesh @ Pr{ }fzero

    \ Set initial values for Pl{ 0 }, Pl{ 1 }, Pr{ 0 }, and Pr{ 1 }
    0e   Pl{ 0 } F!   1e-6 Pl{ 1 } F!
    1e-6 Pr{ 0 } F!   2e-6 Pr{ 1 } F!
 
    verbose? IF 
      cr ." Max Error in E = " Etol F@ F. ." Eh"
    THEN

    \ Find the solution slope differences at E 
    fdup  schr-integrate  
    dup IF  >r fdrop dE F@ r>  EXIT  ELSE  drop  THEN
    sd1 F!
    dE f@ F+  \ -- E+dE

    BEGIN
      fdup  schr-integrate
      dup IF  >r fdrop dE F@ r> EXIT  ELSE  drop  THEN
      sd2 F!

      verbose? IF
        cr ." E = " fdup 15 f>string count s"  Eh" strcat type
        cr ."   Energy inc. (Eh)  = "  dE F@ F.
        cr ."   Matching r (a.u.) = "  r_mesh{ mo @ } F@ F.
        cr ."   Prev Slope Del  =   "  sd1 F@ F.
        cr ."   Curr Slope Del  =   "  sd2 F@ F.
      THEN

      \ Error > Tolerance ?
      dE F@ fabs Etol F@  F>
    WHILE
      sd1 F@ sd2 F@ f2dup F* 0e F< IF
        \ slope difference crossed zero; reduce dE and reverse sign
        sd1 F! fdrop
        dE F@ 2e F/  fnegate dE F!
      ELSE
        fabs fswap fabs fswap  \ -- E  |sd1|  |sd2|
        F<  >R  dE F@  R> IF fnegate 2e F/ THEN  dE F!
        sd2 F@  sd1 F!
      THEN
      dE F@ F+  \ Apply correction to E
    REPEAT
    dE F@  0
;

\ For debug. Write the Pl{r} and Pr{r} solutions to a file.
\ col 1 = r, col 2 = Pl, col 3 = Pr, col 4 = P (contiguous
\ solution).
: write-soln ( -- )
   s" soln.dat" delete-file drop
   s" >file soln.dat" evaluate
   Nmesh @ 0 DO
     r_mesh{ I } F@ F. 2 spaces  Pl{ I } F@ fdup F. 2 spaces
     Pr{ Nmesh @ I - 1- } F@ fdup F. 2 spaces
     I mo @ > IF  fswap THEN  fdrop F. cr
   LOOP
   console
;

\ Clean up remaining dynamically allocated memory
: end-solve ( -- )
    & Qr{ }free
    & Ql{ }free
    & Pr{ }free
    & Pl{ }free
    & V_mesh{ }free
    & r_mesh{ }free
;


