\ Adaptive integration using trapezoidal rule
\ with Richardson extrapolation
\ Integrate a real function from xa to xb

\ Forth Scientific Library Algorithm #19

\ V1.1         @(#)adaptint.seq	1.1 08:53:03 11/3/94


\ Usage:  USE( fn.name xa xb err )INTEGRAL
\ Examples:

\ USE( FSQRT  0e 1e 1.E-3 )INTEGRAL FE. 6.666659e-1 ok
\ USE( FSQRT  0e 2e 1.E-4 )INTEGRAL FE. 1.885618e0 ok

\ : F1     FDUP FSQRT F*  ;  ok
\ USE( F1  0e 1e 1.E-3 )INTEGRAL FE. 4.000005e-1 ok
\ USE( F1  0e 2e 1.E-4 )INTEGRAL FE. 2.262741e0 ok

\ Programmed by J.V. Noble (from "Scientific FORTH" by JVN)
\ ANS Standard Program  -- version of  10/5/1994

\ This is an ANS Forth program requiring:
\      The FLOAT and FLOAT EXT word sets
\ Environmental dependencies:
\       Assumes independent floating point stack
\       Default: 32-bit floating point precision
\ Non STANDARD words:
\      S>F  : S>F   S>D  D>F   ; ( n --) ( F: -- n)
\      F=0  ( puts 0 on fpstack)
\ function vectoring
\ : USE(    '   ;

\     (c) Copyright 1994  Julian V. Noble.     Permission is granted
\     by the author to use this software for any application provided
\     the copyright notice is preserved.

\ Revisions:
\   2007-10-25  km; revised the test code; fixed problem in }DOWN (needed CELL-)
\   2007-10-27  km; save base, switch to decimal, and restore base.

CR .( ADAPTINT          V1.1c         27 October  2007   JVN )
BASE @ DECIMAL

[undefined] s>f [IF] : s>f s>d d>f ;  [THEN]

\ Data structures
0 S>F  FCONSTANT F=0
4 S>F  3 S>F  F/  FCONSTANT F=4/3

20  FLOAT  ARRAY X{
20  FLOAT  ARRAY E{
20  FLOAT  ARRAY F{
20  FLOAT  ARRAY I{

0 VALUE  N

FVARIABLE  OLD.I
FVARIABLE  FINAL.I

\ Begin definitions proper
: )INT  ( n --)                           \ trapezoidal rule
\   F" ( F(N) + F(N-1) ) * ( X(N) - X(N-1) ) / 2  "
    X{ OVER } F@   X{ OVER 1- }  F@
        F-  F2/
    F{ OVER } F@   F{ OVER 1- }  F@
        F+  F*
    I{ SWAP 1- } F!  ;

Defer DUMMY                                  \ dummy function name

: INITIALIZE  ( xt --)  ( F: xa xb eps -- integral)
     IS  DUMMY
     1 TO N
     E{ 0 } F!   X{ 1 } F!   X{ 0 } F!
     X{ 0 } F@   DUMMY   F{ 0 } F!      \ F" f(0) = dummy( x(0) ) "
     X{ 1 } F@   DUMMY   F{ 1 } F!      \ F" f(1) = dummy( x(1) ) "
     1 )INT
     F=0  FINAL.I F! ;

: CHECK.N       N  19 >   ABORT" Too many subdivisions!"  ;
: E/2   E{ N  1- }  DUP   F@   F2/    F! ;
: }DOWN    ( adr n --)
        OVER cell- @  >R   }   DUP   R@ +   R>   MOVE  ;

: MOVE.DOWN    E{ N  1-       }DOWN
               X{ N           }DOWN
               F{ N           }DOWN  ;

: X'   \  F" X(N) = ( X(N) + X(N-1) ) / 2 "
       \  F" F(N) = DUMMY(X(N)) "
       X{ N }  F@   X{ N 1- }  F@    F+  F2/
       FDUP  X{ N }  F!    DUMMY  F{ N }  F!   ;

: N+1   N 1+   TO N  ;
: N-2   N 2 -  TO N  ;

: SUBDIVIDE    CHECK.N     E/2   MOVE.DOWN
        I{ N 1- }  F@  OLD.I  F!
        X'   N )INT   N 1+ )INT    ;

: CONVERGED?   ( F: -- I[N]+I'[N-1]-I[N-1])  ( -- f)
\       F" I(N) + IP(N-1) - I(N-1) "
        I{ N } F@  I{ N 1- } F@   F+  OLD.I F@  F-
        FDUP   FABS
        E{ N 1- } F@   F2*    F<  ;

: INTERPOLATE  ( F: I[N]+I'[N-1]-I[N-1] -- )

\     F" FINAL.I = ( I(N)+I'(N-1) - OLD.I ) * (4/3) + OLD.I + FINAL.I "
      F=4/3   F*   OLD.I F@   F+
      FINAL.I F@   F+                    \ accumulate
      FINAL.I F!  ;                      \ store it


: )INTEGRAL    ( F: A B ERR -- I[A,B]) ( xt --)
     INITIALIZE
     BEGIN   N 0>   WHILE
        SUBDIVIDE
        CONVERGED?    N+1
        IF    INTERPOLATE  N-2
        ELSE  FDROP    THEN
     REPEAT   FINAL.I  F@  ;

BASE !
 
TEST-CODE? [IF]   \ test code ==============================================
[undefined] T{      [IF]  include ttester.fs  [THEN]
BASE @ DECIMAL

1e-8 rel-near F!     \ higher accuracy may be specified, but integrator
1e-8 abs-near F!     \   becomes sluggish.
set-near

CR
TESTING )INTEGRAL
t{ use( FSQRT  0e  1e  rel-near F@ )integral  -> 2e 3e F/        r}t 
t{ use( FSQRT  0e  2e  rel-near F@ )integral  -> 32e 9e F/ FSQRT r}t

: f1     FDUP FSQRT F*  ;

t{ use( f1  0e  1e  rel-near F@ )integral  ->  2e 5e F/  r}t
t{ use( f1  0e  2e  rel-near F@ )integral  ->  128e 25e F/ FSQRT r}t

BASE !
[THEN]

