\ quadratic.fs
\
\ Solve for the two complex solutions of the quadratic equation:
\
\   a*x^2 + b*x + c = 0
\
\ Assume a, b, and c are real. The two complex roots are 
\ z1 = x1r + i*x1i and z2 = x2r + i*x2i
\
\  2002, K. Myneni, krishna.myneni@ccreweb.org
\  Permission is granted to modify and use this code
\  for any application, provided this notice is preserved.
\
\ Notes:
\
\   1. Ensure that argument "a" is not zero, or an infinity will result;
\      the correct solution of the simple linear equation will not be
\      given.
\  
\   2. The returned roots are unordered. 
\   
\ Revisions:
\
\   2002-11-02  km; first version.
\   2003-10-25  Christopher Brannon; fixed problem with calculation
\                 of complex roots.
\   2007-11-04  km; revised comments; added test code; save and restore base. 
\   2009-08-05  km; revised to preserve accuracy when the product a*c is
\                   much less than b^2; see [1]. Added new test case.
\   2009-08-11  km; added test case for purely imaginary roots.
\
\ References:
\
\ 1. W.H. Press, et. al., Numerical Recipes in C, 2nd ed., pp. 183--184,
\    eqns. 5.6.4 and 5.6.5.

CR .( QUADRATIC         V1.1b         11 August    2009   KM )
BASE @ DECIMAL

fvariable qa
fvariable 2qa
fvariable qb
fvariable qc

: solve_quadratic ( F: a b c -- x1r x1i x2r x2i )
                  ( F: a b c -- z1  z2 )
    qc f! qb f! fdup qa f! 2e f* 2qa f!
    qb f@ fdup f* 4e qa f@ f* qc f@ f* f-  \ discriminant
    fdup f0<  IF                           \ complex conjugate roots
	fabs  fsqrt   2qa f@  f/           \ imaginary part 
        qb f@ fnegate 2qa f@  f/           \ real part
	fswap 
        fover fover fnegate                \ complex conjugate
    ELSE                                   \ two real roots
	fsqrt qb f@ fdup f0< IF f- ELSE f+ fnegate THEN 2e f/ 
	fdup  qc f@ fswap f/ 0e   frot qa f@ f/ 0e
    THEN
;

BASE !

TEST-CODE? [IF]   \ test code ==============================================
[undefined] T{      [IF]  include ttester.fs  [THEN]    
BASE @ DECIMAL

1e-15  rel-near F!
1e-256 abs-near F!
set-near

\ Examples from:
\
\    http://www.purplemath.com/modules/quadform.htm

-2e 3e F/       fconstant -2/3
-3e 2e F/       fconstant -3/2
 5e fsqrt       fconstant sqrt{5}
 2e fsqrt 3e F/ fconstant sqrt{2}/3
 3e fsqrt 2e F/ fconstant sqrt{3}/2
10e fsqrt 2e F/ fconstant sqrt{10}/2

: zz}t  rrrr}t ;

CR
TESTING SOLVE_QUADRATIC
t{ 1e  3e -4e solve_quadratic ->  1e               0e   -4e               0e   zz}t
t{ 2e -4e -3e solve_quadratic ->  1e sqrt{10}/2 F- 0e    1e sqrt{10}/2 F+ 0e   zz}t
t{ 1e -2e -4e solve_quadratic ->  1e sqrt{5} F-    0e    1e sqrt{5} F+    0e   zz}t
t{ 9e 12e  4e solve_quadratic ->  -2/3             0e   -2/3              0e   zz}t
t{ 3e  4e  2e solve_quadratic ->  -2/3      sqrt{2}/3   -2/3 sqrt{2}/3 fnegate zz}t
t{ 1e  3e  3e solve_quadratic ->  -3/2      sqrt{3}/2   -3/2 sqrt{3}/2 fnegate zz}t
t{ 1e  0e  1e solve_quadratic ->  0e               1e    0e              -1e   zz}t

\ Test case which loses accuracy with ordinary quadratic formula:
\ 
\        x^2 + x + c = 0
\
\ when c << 1, the approximate solution is  x1 = -c,  x2 = -1 + c
\
t{ 1e 1e 1e-17 solve_quadratic -> -1e-17           0e    -1e 1e-17 f+     0e   zz}t


BASE !
[THEN]

