\ quadratic.fs
\
\ Solve for the two complex solutions of the quadratic equation:
\
\   a*x^2 + b*x + c = 0
\
\ for real coefficients a, b, and c. The two complex roots are 
\ z1 = x1r + i*x1i and z2 = x2r + i*x2i
\
\ Environmental dependencies:
\   1. requires FLOATING and FLOATING EXT wordsets
\   2. supports unified or separate floating point stack
\   3. words provided by this module should be safe to use in 
\      multi-threaded Forth environments (not tested)
\
\ Provides:
\
\   Q_DISCRIMINANT
\   SOLVE_QUADRATIC
\
\ Compiled by K. Myneni, krishna.myneni@ccreweb.org, based
\ on messages posted to comp.lang.forth.
\
\ This code may be considered public domain.
\ 
\ Notes:
\
\   1. Ensure that argument "a" is not zero, or an infinity will result;
\      the correct solution of the simple linear equation will not be
\      given.
\  
\   2. The roots returned by solve_quadratic, are ordered as follows: 
\   
\       F:  --  -b/2a-sqrt(d)/2a   -b/2a+sqrt(d)2a
\
\
\ Revisions:
\
\   2002-11-02  km; first version.
\   2003-10-25  Christopher Brannon; fixed problem with calculation
\                 of complex roots.
\   2007-11-04  km; revised comments; added test code; save and restore base. 
\   2009-08-05  km; revised to preserve accuracy when the product a*c is
\                   much less than b^2; see [1]. Added new test case.
\   2009-08-11  km; added test case for purely imaginary roots.
\   2009-08-18  km; factored the code with modification of example posted
\                   to comp.lang.forth by "humptydumpty", and used words from
\                   the jvn-dnw complex library; revised comments; V1.2
\   2009-08-19  km; additional comments, change "zconj" to "conjg"; V1.2b
\   2009-08-26  km; removed locally used variables for thread-saftey;
\                   made use of code posted to comp.lang.forth by
\                   "humptydumpty" on 15 August 2009; added q_check to test 
\                   for b=0 AND c=0 case;
\
\ References:
\
\ 1. W.H. Press, et. al., Numerical Recipes in C, 2nd ed., pp. 183--184,
\    eqns. 5.6.4 and 5.6.5.

CR .( QUADRATIC         V1.3           26 August    2009   HD, KM )
BASE @ DECIMAL

[UNDEFINED] fsquare [IF] : fsquare fdup f* ;      [THEN]
[UNDEFINED] f2dup   [IF] : f2dup   fover fover ;  [THEN]
[UNDEFINED] f2drop  [IF] : f2drop  fdrop fdrop ;  [THEN]
[UNDEFINED] ftuck   [IF] : ftuck   fswap fover ;  [THEN]
[UNDEFINED] -frot   [IF] : -frot   frot frot ;    [THEN]

\ Useful words from the complex library
[UNDEFINED] zdup    [IF] : zdup  fover fover ;    [THEN]
[UNDEFINED] conjg   [IF] : conjg fnegate ;        [THEN]

: q_discriminant ( F: a b c -- d )
     frot f* -4e f* fswap fsquare f+ ;

\ Transformation of coefficients from a,b,c to q,p: 
\ 
\   q = c/a,  p = -b/2a
\
: abc_to_qp ( F: a b c -- q p )
    frot ftuck f/ -frot 
    2e f* f/ fnegate ;

\ Scaled discriminant:  d' = d/(4a^2)
: (q_discriminant) ( F: q p -- d' )
     fsquare fswap f- ;

: q_real_roots ( F: q p d' -- r1 r2 )
    fsqrt fswap 
    fdup f0< IF fswap fnegate THEN 
    f+ ftuck f/ ;

: q_complex_roots ( F: p d' -- z1 z2 )
    fabs fsqrt
    zdup conjg
;

\ Return true if b=0 and c=0
: q_check ( F: b c -- ) ( -- f)
    f0= >r f0= r> and ;

: solve_quadratic ( F: a b c -- z1 z2 )
    f2dup q_check IF f2drop fdrop 0e 0e zdup EXIT THEN 
    abc_to_qp                          \ transform coefficients to q,p
    f2dup (q_discriminant)             \ F: -- q p d'
    fdup f0<  IF                       \ complex conjugate roots
       frot fdrop
       q_complex_roots 
    ELSE                               \ two real roots
       q_real_roots                    \ F: -- r1 r2
       0e ftuck                        \ promote to z1 z2
    THEN
;

BASE !

TEST-CODE? [IF]   \ test code ==============================================
[undefined] T{      [IF]  include ttester.fs  [THEN]    
BASE @ DECIMAL

1e-15  rel-near F!
1e-256 abs-near F!
set-near

\ Examples from:
\
\    http://www.purplemath.com/modules/quadform.htm

-2e 3e F/       fconstant -2/3
-3e 2e F/       fconstant -3/2
 5e fsqrt       fconstant sqrt{5}
 2e fsqrt 3e F/ fconstant sqrt{2}/3
 3e fsqrt 2e F/ fconstant sqrt{3}/2
10e fsqrt 2e F/ fconstant sqrt{10}/2

: zz}t  rrrr}t ;

CR
TESTING SOLVE_QUADRATIC
t{ 1e  0e  0e solve_quadratic ->  0e               0e    0e               0e   zz}t
t{ 1e  3e -4e solve_quadratic -> -4e               0e    1e               0e   zz}t
t{ 2e -4e -3e solve_quadratic ->  1e sqrt{10}/2 F+ 0e    1e sqrt{10}/2 F- 0e   zz}t
t{ 1e -2e -4e solve_quadratic ->  1e sqrt{5} F+    0e    1e sqrt{5} F-    0e   zz}t
t{ 9e 12e  4e solve_quadratic ->  -2/3             0e   -2/3              0e   zz}t
t{ 3e  4e  2e solve_quadratic ->  -2/3      sqrt{2}/3   -2/3 sqrt{2}/3 fnegate zz}t
t{ 1e  3e  3e solve_quadratic ->  -3/2      sqrt{3}/2   -3/2 sqrt{3}/2 fnegate zz}t
t{ 1e  0e  1e solve_quadratic ->  0e               1e    0e              -1e   zz}t

\ Test case which loses accuracy with ordinary quadratic formula:
\ 
\        x^2 + x + c = 0
\
\ when c << 1, the approximate solution is  x1 = -c,  x2 = -1 + c
\
t{ 1e 1e 1e-17 solve_quadratic -> -1e 1e-17 f+     0e   -1e-17           0e    zz}t


BASE !
[THEN]

