\      Algorithm 680, Collected Algorithms from ACM.
\      This work published in Transactions on Mathematical Software,
\      vol. 16, no. 1, pp. 47.
\
\  Given a complex number z = (xi, yi), the word zwofz computes
\  the value of the Faddeeva-function 
\
\ 	w(z) = exp(-z**2)*erfc(-i*z)
\
\  where erfc is the complex complementary error-function and i
\  means sqrt(-1).
\
\  The accuracy of the algorithm for z in the 1st and 2nd quadrant
\  is 14 significant digits; in the 3rd and 4th it is 13 significant
\  digits outside a circular region with radius 0.126 around a zero
\  of the function.
\
\
\  The code contains a few compiler-dependent parameters :
\     RMAXREAL = the maximum value of RMAXREAL equals the root of
\                RMAX = the largest number which can still be 
\                implemented on the computer in double precision
\                floating-point arithmetic
\     RMAXEXP  = ln(RMAX) - ln(2)
\     RMAXGONI = the largest possible argument of a double precision
\                goniometric function (FCOS, FSIN, ...)
\  The reason why these parameters are needed as they are defined will
\  be explained in the code by means of comments
\
\
\  Parameter List
\     xi     = real      part of z
\     yi     = imaginary part of z
\     u      = real      part of w(z)
\     v      = imaginary part of w(z)
\     flag   = an error flag indicating whether overflow will
\              occur or not; TYPE LOGICAL;
\              The values of this variable have the following
\              meaning :
\              flag = FALSE : no error condition
\              falg = TRUE  : overflow will occur, the routine
\                             becomes inactive
\  xi, yi      are the input-parameters
\  u, v, flag  are the output-parameters
\
\  The routine is not underflow-protected but any variable can be
\  put to 0 upon underflow;
\
\  Reference - GPM Poppe, CMJ Wijers; More Efficient Computation of
\  the Complex Error-Function, ACM Trans. Math. Software.
\
\  Ported to Forth-94 by Krishna Myneni, krishna.myneni@ccreweb.org
\  31 July 2009
\
\  Revisions:
\    2009-08-04  km  revised comments, include link to comprehensive
\                    table of reference values. 

[UNDEFINED] s>f [IF] : s>f  s>d  d>f  ;  [THEN]
: idnint ( F: r -- ) ( -- n )  fround f>d d>s ;
: idint  ( F: r -- ) ( -- n ) f>d d>s ;
BASE @
DECIMAL

1.12837916709551257388E0 fconstant FACTOR     \ 2/sqrt(pi)
0.5E154                  fconstant RMAXREAL
708.503061461606E0       fconstant RMAXEXP
3.53711887601422E15      fconstant RMAXGONI

variable aflag
variable bflag
variable n
variable np1
variable nu
variable Jvar
variable kapn

fvariable u
fvariable v
fvariable u1
fvariable v1
fvariable u2
fvariable v2
fvariable xabs
fvariable yabs
fvariable x
fvariable y
fvariable xi
fvariable yi
fvariable qrh0
fvariable xabsq
fvariable xquad
fvariable yquad
fvariable xsum
fvariable ysum
fvariable daux
fvariable xaux
fvariable cc
fvariable h1
fvariable h2
fvariable qlambda
fvariable rx
fvariable ry
fvariable sx
fvariable sy
fvariable tx
fvariable ty
fvariable w1

: zwofz ( F: xi yi  -- u v ) ( -- flag )  \ Faddeeva function
\       ( F:     z  -- zw  ) ( -- flag )      
      yi f! xi f! 
      yi f@ fabs yabs f!
      xi f@ fabs xabs f! 
      xabs f@ 6.3e f/ x f!
      yabs f@ 4.4e f/ y f! 
\
\     The following IF-statement protects
\     QRHO = (X**2 + Y**2) against overflow
\
      xabs f@ RMAXREAL f>  yabs f@ RMAXREAL f>  or IF
	0e 0e true exit
      THEN

      x f@ fdup f*  y f@ fdup f*  f+  qrh0 f!

      xabs f@ fdup f*  xabsq f!
      xabsq f@  yabs f@ fdup f* f-  xquad f! 
      2e xabs f@ f* yabs f@ f*  yquad f!

      qrh0 f@  0.085264e0 f<  aflag !

      aflag @  IF
\
\  If (QRHO <  0.085264E0) then the Faddeeva-function is evaluated
\  using a power-series (Abramowitz/Stegun, equation (7.1.5), p.297)
\  N is the minimum number of terms needed to obtain the required
\  accuracy
\
        1e  0.85e y f@ f* f- qrh0 f@ fsqrt f*  qrh0 f!
        6e 72e qrh0 f@ f* f+  idnint  n !
        n @ 2 *  1+ Jvar !
        1.0e Jvar @ s>f f/ xsum f! 
        0.0e0 ysum f!
        1 n @ DO                      \ 10 I=N, 1, -1
          -2 Jvar +!
          xsum f@ xquad f@ f*  ysum f@ yquad f@ f* f- I s>f f/ xaux f!
          xsum f@ yquad f@ f*  ysum f@ xquad f@ f* f+ I s>f f/ ysum f!
          xaux f@  1.0e Jvar @ s>f f/ f+ xsum f!
        -1 +LOOP

        xsum f@ yabs f@ f*  ysum f@ xabs f@ f* f+ FACTOR f* fnegate 1.0e f+ u1 f!
        xsum f@ xabs f@ f*  ysum f@ yabs f@ f* f- FACTOR f*  v1 f!
        xquad f@ fnegate fexp daux f!
        yquad f@ fcos daux f@ f* u2 f!
        yquad f@ fsin daux f@ f* fnegate v2 f!

        u1 f@ u2 f@ f*  v1 f@ v2 f@ f*  f-  u f!
        u1 f@ v2 f@ f*  v1 f@ u2 f@ f*  f+  v f!

      ELSE
\
\  If (QRHO > 1.O) then w(z) is evaluated using the Laplace
\  continued fraction
\  nu is the minimum number of terms needed to obtain the required
\  accuracy
\
\  If ((QRHO > 0.085264D0) and (QRHO < 1.0)) then w(z) is evaluated
\  by a truncated Taylor expansion, where the Laplace continued fraction
\  is used to calculate the derivatives of w(z)
\  kapn is the minimum number of terms in the Taylor expansion needed
\  to obtain the required accuracy
\  nu is the minimum number of terms of the continued fraction needed
\  to calculate the derivatives with the required accuracy
\

        qrh0 f@ 1.0e f> IF
          0.0e0 h1 f!
          0 kapn !
          qrh0 f@ fsqrt qrh0 f!
          3e  1442e  26e qrh0 f@ f* 77e f+  f/  f+ idint nu !
        ELSE
          1e y f@ f-  1e qrh0 f@ f- fsqrt f*  qrh0 f!
          1.88e qrh0 f@ f*  h1 f!
          2e h1 f@ f* h2 f!
          7e  34e qrh0 f@ f*  f+  idnint  kapn !
          16e 26e qrh0 f@ f*  f+  idnint  nu !
        THEN

        h1 f@ 0.0e f>  bflag !

        bflag @  IF  h2 f@  kapn @ s>f f**  qlambda f!  THEN

        0.0e  rx  f!
        0.0e  ry  f!
        0.0e  sx  f!
        0.0e  sy  f!

        0 nu @ DO  \ N=NU, 0, -1
          I 1+ np1 !
          yabs f@ h1 f@ f+ np1 @ s>f rx f@ f* f+  tx f!
          xabs f@ np1 @ s>f ry f@ f* f-  ty f!
          0.5e tx f@ fdup f* ty f@ fdup f* f+ f/  cc f!
          cc f@ tx f@ f*  rx f!
          cc f@ ty f@ f*  ry f!
          bflag @   I kapn @ <=   and IF
            qlambda f@  sx f@ f+  tx f!
            rx f@ tx f@ f*  ry f@ sy f@ f* f-  sx f!
            ry f@ tx f@ f*  rx f@ sy f@ f* f+  sy f! 
            qlambda f@ h2 f@ f/  qlambda f!
          THEN
        -1 +LOOP

        h1 f@ f0=  IF
          FACTOR rx f@ f*
          FACTOR ry f@ f*
        ELSE
          FACTOR sx f@ f* 
          FACTOR sy f@ f*
        THEN
	v f!  u f!

        yabs f@ f0=  IF  xabs f@ fdup f* fnegate fexp u f!  THEN

      THEN

\
\  Evaluation of w(z) in the other quadrants
\

      yi f@  0.0e f<  IF

        aflag @ IF 
          u2 f@ 2e f* u2 f!
          v2 f@ 2e f* v2 f!
        ELSE
          xquad f@ fnegate xquad f!
\
\         The following IF-statement protects 2*exp(-z**2)
\         against overflow
\
          yquad f@ RMAXGONI f>   xquad f@ RMAXEXP f>  or IF
		0e 0e true EXIT
	  THEN

          xquad f@ fexp 2e f*  w1 f!
          yquad f@ fcos w1 f@ f*  u2 f!
          yquad f@ fsin w1 f@ f* fnegate v2 f!
        THEN

        u2 f@ u f@ f- u f!
        v2 f@ v f@ f- v f!
        xi f@  0.0e  f>  IF v f@ fnegate v f! THEN
      ELSE
        xi f@  0.0e  f<  IF v f@ fnegate v f! THEN
      THEN

      u f@ v f@ 0 
;

BASE !


TEST-CODE? [IF]
[UNDEFINED] T{  [IF] include ttester.fs  [THEN]

BASE @
DECIMAL

\ Reference values are taken from the Matpack special functions implementation
\ test routines in funtest.cpp, by Berndt M. Gammel. See http://www.matpack.de
\ For a more extensive table of values, reproduced from the same source, see
\ ftp://ccreweb.org/software/fsl/extras/zwofz-reference-values.txt

1e-256 abs-near f!
2e-15  rel-near f!   
set-near

CR
TESTING ZWOFZ
t{ -2.0e  -2.0e zwofz ->  -0.43895282712924287596643907240606330819e     -2.10989621033098140920150385396220394331e  0 rrx}t
t{ -2.0e  -1.5e zwofz ->   0.183289715319317036760070295301264235039e    -0.073260876796080792095198822633956123101e 0 rrx}t
t{ -2.0e  -1.0e zwofz ->  -0.205325580646587513283825721521229530107e    -0.146855485030167393064213599903333042823e 0 rrx}t
t{ -2.0e  -0.5e zwofz ->  -0.122932494822762374121292046660171680827e    -0.327555136333312587627222896684373565679e 0 rrx}t
t{ -2.0e   0.0e zwofz ->   0.01831563888873418029371802127324124221191e  -0.340026217066066201280467897123400351211e 0 rrx}t
t{ -2.0e   0.5e zwofz ->   0.103358823741366658953062349638804853052e    -0.28478588475009374558328290919841313745e  0 rrx}t
t{ -2.0e   1.0e zwofz ->   0.140239581366277943695959506114271778326e    -0.222213440179899102605794500435104617163e 0 rrx}t
t{ -2.0e   1.5e zwofz ->   0.150415438871039747617373178237581832522e    -0.170371142762476985627101574087970440296e 0 rrx}t
t{ -2.0e   2.0e zwofz ->   0.14795275951201582422875630874367394297e     -0.13117971708421785358525665747162659908e  0 rrx}t
t{ -1.5e  -2.0e zwofz ->  10.8674622395776211714967887828609726103e       3.0965521142997534229680658477024236488e   0 rrx}t
t{ -1.5e  -1.5e zwofz ->  -0.6227067163884116411051423068613715348e       1.79071165397990662077818231386062897937e  0 rrx}t
t{ -1.5e  -1.0e zwofz ->  -0.779111784231756692689265888094467986093e    -0.314034095888643736260374989580622706139e 0 rrx}t
t{ -1.5e  -0.5e zwofz ->  -0.17748955379745403236318048141809805694e     -0.607712851425209724486050937542131846141e 0 rrx}t
t{ -1.5e   0.0e zwofz ->   0.1053992245618643367832176892406980972685e   -0.483227330140769057926902291608470059127e 0 rrx}t
t{ -1.5e   0.5e zwofz ->   0.196636032243581962363885730243649142606e    -0.337720318346887945905938190305428231255e 0 rrx}t
t{ -1.5e   1.0e zwofz ->   0.211836585968510564256586582644861058736e    -0.233170977404442439968522175642150729022e 0 rrx}t
t{ -1.5e   1.5e zwofz ->   0.20111511752685222914417865727371074878e     -0.16434858135028749000008771513663057636e  0 rrx}t
t{ -1.5e   2.0e zwofz ->   0.1833347623811499753436522984018014989e      -0.1192982330062729387898603992894135238e   0 rrx}t
t{ -1.0e  -2.0e zwofz -> -26.476058778199206857477508766171844855e       30.3085711167433072583594590086268067272e   0 rrx}t
t{ -1.0e  -1.5e zwofz ->  -7.1679546161700898468604498767704492936e      -1.1203567295729395305616084222371854431e   0 rrx}t
t{ -1.0e  -1.0e zwofz ->  -1.13703787835119736645227530007111933967e     -2.02681379185419501807947707907820529935e  0 rrx}t
t{ -1.0e  -0.5e zwofz ->   0.155541142454331075901205660255323721008e    -1.137837215781686377738087088101075469584e 0 rrx}t
t{ -1.0e   0.0e zwofz ->   0.3678794411714423215955237701614608674458e   -0.607157705841393729115038235800744921161e 0 rrx}t
t{ -1.0e   0.5e zwofz ->   0.354900332867577883922445599634927294884e    -0.342871719131100716552337823750091427533e 0 rrx}t
t{ -1.0e   1.0e zwofz ->   0.30474420525691259245713884106959496013e     -0.20821893820283162728743734725471561394e  0 rrx}t
t{ -1.0e   1.5e zwofz ->   0.2571279392712283631817635118927714589e      -0.13524227699550782708878403838222364602e  0 rrx}t
t{ -1.0e   2.0e zwofz ->   0.21849261527489069682239847542358814e        -0.0929978093926018660475030380586688847e   0 rrx}t
t{ -0.5e  -2.0e zwofz -> -35.6353035120018890541413069531983298143e     -77.3801423753454349424382190735070680621e   0 rrx}t
t{ -0.5e  -1.5e zwofz ->   0.74200718289486356609832707083402721227e    -14.81894370296502187613346222409723237502e  0 rrx}t
t{ -0.5e  -1.0e zwofz ->   1.89640595954530034020825849466326552226e     -3.68999058851944924662944055512211701373e  0 rrx}t
t{ -0.5e  -0.5e zwofz ->   1.222008415868570518464334253164948182978e    -1.189339308592864409254253941565702334794e 0 rrx}t
t{ -0.5e   0.0e zwofz ->   0.7788007830714048682451702669783206472968e   -0.478925172901043472544937540717089342226e 0 rrx}t
t{ -0.5e   0.5e zwofz ->   0.533156707912174913768228912042711121005e    -0.23048823138445840870767807113455955863e  0 rrx}t
t{ -0.5e   1.0e zwofz ->   0.39123402145213608337190110677048354169e     -0.12720241088464801019460968171034158452e  0 rrx}t
t{ -0.5e   1.5e zwofz ->   0.30335511991319153439202737939137821881e     -0.07785087412615059529718940382149537324e  0 rrx}t
t{ -0.5e   2.0e zwofz ->   0.2452759902263585078573396831219042248e      -0.0515214783436358491099961278415586527e   0 rrx}t
t{  0.0e  -2.0e zwofz -> 108.9409043899779724123554338248132140423e       0.0e                                       0 rrx}t
t{  0.0e  -1.5e zwofz ->  18.65388625626273393874641550130021127851e      0.0e                                       0 rrx}t
t{  0.0e  -1.0e zwofz ->   5.00898008076228346630982459821480981469e      0.0e                                       0 rrx}t
t{  0.0e  -0.5e zwofz ->   1.95236048918255709327604771344113097989e      0.0e                                       0 rrx}t
t{  0.0e   0.0e zwofz ->   1.0e                                           0.0e                                       0 rrx}t
t{  0.0e   0.5e zwofz ->   0.615690344192925874870793422683741936782e     0.0e                                       0 rrx}t
t{  0.0e   1.0e zwofz ->   0.42758357615580700441075034449051518082e      0.0e                                       0 rrx}t
t{  0.0e   1.5e zwofz ->   0.32158541645431750235432258772326556903e      0.0e                                       0 rrx}t
t{  0.0e   2.0e zwofz ->   0.2553956763105057438650885809085427633e       0.0e                                       0 rrx}t
t{  0.5e  -2.0e zwofz -> -35.6353035120018890541413069531983298143e      77.3801423753454349424382190735070680621e   0 rrx}t
t{  0.5e  -1.5e zwofz ->   0.74200718289486356609832707083402721227e     14.81894370296502187613346222409723237502e  0 rrx}t
t{  0.5e  -1.0e zwofz ->   1.89640595954530034020825849466326552226e      3.68999058851944924662944055512211701373e  0 rrx}t
t{  0.5e  -0.5e zwofz ->   1.222008415868570518464334253164948182978e     1.189339308592864409254253941565702334794e 0 rrx}t
t{  0.5e   0.0e zwofz ->   0.7788007830714048682451702669783206472968e    0.478925172901043472544937540717089342226e 0 rrx}t
t{  0.5e   0.5e zwofz ->   0.533156707912174913768228912042711121005e     0.23048823138445840870767807113455955863e  0 rrx}t
t{  0.5e   1.0e zwofz ->   0.39123402145213608337190110677048354169e      0.12720241088464801019460968171034158452e  0 rrx}t
t{  0.5e   1.5e zwofz ->   0.30335511991319153439202737939137821881e      0.07785087412615059529718940382149537324e  0 rrx}t
t{  0.5e   2.0e zwofz ->   0.2452759902263585078573396831219042248e       0.0515214783436358491099961278415586527e   0 rrx}t
t{  1.0e  -2.0e zwofz -> -26.476058778199206857477508766171844855e      -30.3085711167433072583594590086268067272e   0 rrx}t
t{  1.0e  -1.5e zwofz ->  -7.1679546161700898468604498767704492936e       1.1203567295729395305616084222371854431e   0 rrx}t
t{  1.0e  -1.0e zwofz ->  -1.13703787835119736645227530007111933967e      2.02681379185419501807947707907820529935e  0 rrx}t
t{  1.0e  -0.5e zwofz ->   0.155541142454331075901205660255323721008e     1.137837215781686377738087088101075469584e 0 rrx}t
t{  1.0e   0.0e zwofz ->   0.3678794411714423215955237701614608674458e    0.607157705841393729115038235800744921161e 0 rrx}t
t{  1.0e   0.5e zwofz ->   0.354900332867577883922445599634927294884e     0.342871719131100716552337823750091427533e 0 rrx}t
t{  1.0e   1.0e zwofz ->   0.30474420525691259245713884106959496013e      0.20821893820283162728743734725471561394e  0 rrx}t
t{  1.0e   1.5e zwofz ->   0.2571279392712283631817635118927714589e       0.13524227699550782708878403838222364602e  0 rrx}t
t{  1.0e   2.0e zwofz ->   0.21849261527489069682239847542358814e         0.0929978093926018660475030380586688847e   0 rrx}t
t{  1.5e  -2.0e zwofz ->  10.8674622395776211714967887828609726103e      -3.0965521142997534229680658477024236488e   0 rrx}t
t{  1.5e  -1.5e zwofz ->  -0.6227067163884116411051423068613715348e      -1.79071165397990662077818231386062897937e  0 rrx}t
t{  1.5e  -1.0e zwofz ->  -0.779111784231756692689265888094467986093e     0.314034095888643736260374989580622706139e 0 rrx}t
t{  1.5e  -0.5e zwofz ->  -0.17748955379745403236318048141809805694e      0.607712851425209724486050937542131846141e 0 rrx}t
t{  1.5e   0.0e zwofz ->   0.1053992245618643367832176892406980972685e    0.483227330140769057926902291608470059127e 0 rrx}t
t{  1.5e   0.5e zwofz ->   0.196636032243581962363885730243649142606e     0.337720318346887945905938190305428231255e 0 rrx}t
t{  1.5e   1.0e zwofz ->   0.211836585968510564256586582644861058736e     0.233170977404442439968522175642150729022e 0 rrx}t
t{  1.5e   1.5e zwofz ->   0.20111511752685222914417865727371074878e      0.16434858135028749000008771513663057636e  0 rrx}t
t{  1.5e   2.0e zwofz ->   0.1833347623811499753436522984018014989e       0.1192982330062729387898603992894135238e   0 rrx}t
t{  2.0e  -2.0e zwofz ->  -0.43895282712924287596643907240606330819e      2.10989621033098140920150385396220394331e  0 rrx}t
t{  2.0e  -1.5e zwofz ->   0.183289715319317036760070295301264235039e     0.073260876796080792095198822633956123101e 0 rrx}t
t{  2.0e  -1.0e zwofz ->  -0.205325580646587513283825721521229530107e     0.146855485030167393064213599903333042823e 0 rrx}t
t{  2.0e  -0.5e zwofz ->  -0.122932494822762374121292046660171680827e     0.327555136333312587627222896684373565679e 0 rrx}t
t{  2.0e   0.0e zwofz ->   0.01831563888873418029371802127324124221191e   0.340026217066066201280467897123400351211e 0 rrx}t
t{  2.0e   0.5e zwofz ->   0.103358823741366658953062349638804853052e     0.28478588475009374558328290919841313745e  0 rrx}t
t{  2.0e   1.0e zwofz ->   0.140239581366277943695959506114271778326e     0.222213440179899102605794500435104617163e 0 rrx}t
t{  2.0e   1.5e zwofz ->   0.150415438871039747617373178237581832522e     0.170371142762476985627101574087970440296e 0 rrx}t
t{  2.0e   2.0e zwofz ->   0.14795275951201582422875630874367394297e      0.13117971708421785358525665747162659908e  0 rrx}t

BASE !
[THEN]

