\ HILBERT      Hilbert matrix routines

\ Forth Scientific Library Algorithm #25

\ HILBERT      Generates a finite segment of a Hilbert Matrix
\ HILBERT-INV  generates the inverse of a finite segment of a Hilbert
\              Matrix (ACM #50)
\ HILBERT-DET  calculates the determinant for a Hilbert matrix of a given order


\ These matrices provide severe test cases for matrix inverters and determinant
\ calculation routines.  They become numerically ill-conditioned even for
\ moderate sizes of 'n'.

\ This is an ANS Forth program requiring:
\      1. The Floating-Point word set ( *** This is the integrated data/fp stack version *** )
\      2. The word 'S>F' to convert an integer to a float
\      3. The word '}}' to dereference a two-dimensional array.
\      4. The test code Uses '}}malloc' and '}}free' to allocate and
\         release memory for dynamic matrices ( 'DMATRIX' ).
\      5. The compilation of the test code is controlled by the VALUE TEST-CODE?
\         and the conditional compilation words in the Programming-Tools wordset


\ Collected Algorithms from ACM, Volume 1 Algorithms 1-220,
\ 1980; Association for Computing Machinery Inc., New York,
\ ISBN 0-89791-017-6


\  (c) Copyright 1994 Everett F. Carter.  Permission is granted by the
\  author to use this software for any application provided this
\  copyright notice is preserved.


CR .( HILBERT           V1.1           13 October 1994   EFC )

: HILBERT ( &h n -- )             \ stores in array with address &h

    DUP 0 DO
	DUP 0 DO
	    OVER 1.0E0 I J + 1+ S>F F/
	    ROT I J }} F!
	LOOP
    LOOP
    2DROP
;


0 ptr hinv{{
0 value hn
: HILBERT-INV ( &s n -- )       \ stores inverse in array with address &s

    to hn  to hinv{{
    hn DUP * S>F FDUP hinv{{ F!  
    
    hn 1 ?DO                \ do diagonals
	hn DUP I + SWAP I - *   S>F
        I DUP *    S>F F/
	FDUP F* F*
	FDUP hinv{{ I I }} F!
    LOOP
    FDROP


    hn 1- 0 ?DO                \ do off-diagonals
	hn I 1+ DO
	    hn DUP  I +  SWAP I - * S>F
            I DUP * S>F F/ FNEGATE
            hinv{{ J I 1- }} F@ F*
            hinv{{ J I }} F!                      
	LOOP
    LOOP


    hn 1 ?DO                         \ normalize
	I 1+ 0 DO
	    I J + 1+ S>F
            hinv{{ I J }} DUP >R F@ FSWAP F/
            FDUP R> F!
            hinv{{ J I }} F!
	LOOP   
    LOOP
;

: HILBERT-DET ( n -- det )                \ calculates determinant of n order matrix
                                          \ the actual matrix is implicit
    to hn
    1.0E0  1.0E0
    hn 0 DO
	hn 0 DO
	    I J < IF   \ numerator accumulation
		I J - S>F FDUP F* FROT F* FSWAP
	    THEN

	    I J + 1+ S>F F*            \ denominator accumulation
	LOOP
    LOOP
    F/
;


TEST-CODE? [IF]   \ test code ==============================================


FLOAT DMATRIX s{{

: invresults-for-4 ( -- )

       CR
       ." For n = 4, the inverse is: " CR
       ."   16  -120   240  -140 " CR
       ." -120  1200 -2700  1680 " CR
       ."  240 -2700  6480 -4200 " CR
       ." -140  1680 -4200  2800 " CR
;

: results-for-4 ( -- )

       CR
       ." For n = 4, the matrix is: " CR
       ."  1      0.5      0.333333 0.25 " CR
       ."  0.5    0.333333 0.25     0.2 " CR
       ."  0.333333 0.25   0.2      0.166667 " CR
       ."  0.25   0.2      0.166667 0.142857 " CR

;

: hilbert-test ( n -- )

         CR

         & s{{ OVER DUP }}malloc

         malloc-fail? ABORT" malloc failure "
         
         s{{ OVER HILBERT

         ." The Hilbert Matrix: " CR
         DUP DUP s{{ }}fprint

         DUP 4 = IF results-for-4 THEN


         CR
         ." The INVERSE Hilbert Matrix: " CR
         s{{ OVER HILBERT-INV

         DUP DUP s{{ }}fprint

         DUP 4 = IF invresults-for-4 THEN

         & s{{ }}free

         HILBERT-DET         
         CR ." The determinant: " F. CR

;


\ 2 HILBERT-DET f.         0.0833333
\ 3 HILBERT-DET f.         0.000462963
\ 4 HILBERT-DET f.         1.65344e-7
\ 5 HILBERT-DET f.         3.7493E-12


[THEN]

