\ Horner          Evaluation of a polynomial by the Horner method

\ Forth Scientific Library Algorithm #3

\ This routine evaluates an Nth order polynomial Y(X) at point X
\ Y(X) = \sum_i=0^N a[i] x^i                  (NOTE: N+1 COEFFICIENTS)
\ by the Horner scheme.  This algorithm minimizes the number of multiplications
\ required to evaluate the polynomial.
\ The implementation demonstrates the use of array aliasing.

\ This code conforms with ANS requiring:
\      1. The Floating-Point word set
\      2. Test code uses 'ExpInt' for real exponential integrals

\ This algorithm is described in many places, e.g.,
\ Conte, S.D. and C. deBoor, 1972; Elementary Numerical Analysis, an algorithmic
\ approach, McGraw-Hill, New York, 396 pages

\ (c) Copyright 1994 Everett F. Carter.  Permission is granted by the
\ author to use this software for any application provided this
\ copyright notice is preserved.
\
\ Ported to kForth with minor revisions by K. Myneni, 2006-11-07
\
CR .( HORNER            V1.5b          31 October 1994   EFC )

0 ptr ha{

: }Horner ( fx 'a n -- fy)

    SWAP  to ha{ 
    >R  0.0e

    0 R> DO
	FOVER F*
	ha{ I } F@ F+  
    -1 +LOOP

    FSWAP FDROP
;


TEST-CODE? [IF]     \ test code =============================================

6 FLOAT ARRAY ArrayZ{
5 FLOAT ARRAY ArrayY{
5 FLOAT ARRAY ArrayW{

\ initialize with data for real exponential integral
: horner_test_init ( -- )

    -0.57721566e
     0.99999193e
    -0.24991055e
     0.05519968e
    -0.00976004e
     0.00107857e   6 ArrayZ{ }fput

     0.2677737343e
     8.6347608925e
    18.059016973e
     8.5733287401e
     1.0e          5 ArrayY{ }fput
    

     3.9584969228e
    21.0996530827e
    25.6329561486e
     9.5733223454e
     1.0e          5 ArrayW{ }fput

;

: local_exp ( fx -- expint[x] )

    FDUP
    1.0e F< IF
	FDUP ArrayZ{  5 }Horner
	FSWAP FLN F-
    ELSE
	FDUP  ArrayY{ 4 }Horner
	FOVER ArrayW{ 4 }Horner
	F/
	FOVER F/
	FSWAP -1.0e F* FEXP F*
    THEN
;

: TAB 9 emit ;

: do_both ( fx -- )

    FDUP FDUP
    ." X = " F. 2 spaces TAB
    ." Horner: " local_exp F. TAB
    ." ExpInt: " expint    F. TAB
    CR
;

\ compare ExpInt as coded in V1.0 against the general purpose
\ Horner routine

: horner_test ( -- )

    horner_test_init

    CR
    0.2e do_both
    0.5e do_both
    1.0e do_both
    2.0e do_both
    5.0e do_both
    10.0e do_both
    
;

[THEN]

