// vm.s
//
// The kForth Virtual Machine
//
// Copyright (c) 1998--2007 Krishna Myneni, Creative Consulting for
//   Research and Education
//
// This software is provided under the terms of the General Public License.
//
// Usage from C++
//
//       extern "C" int vm (byte* ip);
//       ecode = vm(ip);
//
// Originally written for the A386 assembler
// Revisions:
//	3-21-98
//	8-25-1998 ported to GNU assembler under Linux
//	9-8-1998  additional functions added: c@, c!, f=, f<>
//	9-10-1998 added: and, or, not, xor
//	9-11-1998 added: = , <>, <, >, <=, >=
//	9-15-1998 added: +!, fsqrt, fsin, fcos, -rot, sp@, rp@
//	9-17-1998 added: a@
//	9-27-1998 added: mod, 2*, 2/
//	10-1-1998 added: pick
//	10-4-1998 fixed signed integer division and mod; added /MOD
//	10-6-1998 added ?dup
//	10-14-1998 added count
//	10-16-1998 fixed L_div error
//	10-19-1998 added 0<, 0=, 0>, true, false
//	10-20-1998 added 2+, 2-
//	02-09-1999 added execute
//	03-01-1999 added open, lseek, close, read, write
//	03-02-1999 added ioctl
//	03-03-1999 added usleep	
//	03-07-1999 added fill, cmove
//	03-09-1999 interchanged meaning of execute and call
//		   to be consistent with ANS Forth
//	03-27-1999 added +loop, unloop
//	03-29-1999 added roll
//	03-31-1999 added cmove>, key
//	05-05-1999 fixed +loop
//	05-06-1999 added fround
//	05-15-1999 added floor
//	05-26-1999 added fatan2, lshift, rshift
//	05-27-1999 added u<, quit, base
//	06-02-1999 added */, */mod
//	06-09-1999 call CPP functions from vm
//	07-18-1999 added find
//	09-06-1999 added pTIB, word, tick
//	09-12-1999 added system
//	10-04-1999 added create, variable, fvariable as intrinsic words
//	10-06-1999 added constant, fconstant as intrinsic words
//	10-07-1999 added chdir
//      10-08-1999 added erase, brackettick
//	10-09-1999 added time&date, ms, question, bl
//      10-10-1999 added char, forget, cold
//      10-20-1999 added >file, console
//	10-28-1999 added key?
//	12-24-1999 added hooks for f0=, f0<, f0>, u>, s>d, d>f, f>d, 
//	              um*, um/mod, m*, m+, m/, m*/
//	12-25-1999 added cells, cell+, dfloats, dfloat+
//	12-27-1999 added bye
//	01-08-2000 fixed f0<, increased vm loop efficiency
//	01-13-1999 added ?allot, fixed f0=
//	01-22-2000 modified + to remove ordering sensitivity for address arithmetic
//	01-23-2000 added 0<>, [char], .r, u.r, changed opcodes for 
//	             relational operators
//	01-24-2000 added CPP_literal, CPP_cquote, CPP_squote, CPP_dotquote
//	02-04-2000 implemented m*, um*
//	02-26-2000 fixed fm/mod, added sm/rem, um/mod
//	03-02-2000 modified QUIT to clear the return stacks, added CPP_do
//	03-05-2000 added CPP_begin, CPP_while, CPP_repeat, CPP_until, CPP_again,
//	             CPP_leave, CPP_if, CPP_else, CPP_then, CPP_lparen
//	05-17-2000 added CPP_does
//	05-18-2000 fix L_plusloop for negative increments
//	06-04-2000 fix L_roll to roll the typestack as well
//	06-11-2000 added CPP_case, CPP_endcase, CPP_of, and CPP_endof
//	06-15-2000 added CPP_querydo, CPP_abortquote
//	09-05-2000 handled overflow condition on um/mod; added code for m/;
//		     added m+	;  hooks for CPP_lbracket, CPP_rbracket, CPP_ddot
//	11-30-2000 added L_dabs, L_dnegate, L_udmstar, L_mstarslash
//	01-26-2001 changed jumptable address for usleep to C_usec
//	04-01-2001 changed jmpl to jmp for L_fzerolt
//	04-22-2001 added L_dplus, L_dminus
//	04-24-2001 added L_twopush, L_twopop, L_tworfetch
//      05-07-2001 added L_definition
//	05-08-2001 improved vm execution efficiency by about 10%
//	05-13-2001 added CPP_dotparen, L_dlt, L_dzeroeq, L_deq
//	05-20-2001 added CPP_sharp, CPP_bracketsharp, CPP_sharpbracket, 
//	             CPP_sharps, CPP_sign, CPP_hold; fixed problem with L_dnegate
//	07-12-2001 fixed type-stack alignment problem with L_utmslash
//	08-26-2001 updated L_fval to avoid routing fp through NDP, changed all
//	             typestack settings of OP_FVAL to OP_IVAL
//	09-02-2001 added L_tobody
//      12-08-2001 added CPP_evaluate
//      07-29-2002 fixed L_ret to set GlobalRp and GlobalRtp to bottom upon
//                   return from vm.
//	07-31-2002 fixed L_cmovefrom for zero count argument; added
//		     CPP_backslash; fixed bug in L_fmslashmod for divisor < 0
//	08-01-2002 added C_msfetch, C_search, C_compare
//      09-11-2002 modified vm and L_ret to make vm re-entrant
//	09-29-2002 restored L_usleep since it provides a different function 
//	             than L_ms, added CPP_immediate
//	04-11-2003 renamed L_ftos to L_froundtos
//	04-15-2003 added L_ftrunc and L_ftrunctos
//	04-18-2003 fixed L_ftrunctos typestack pointer update
//	06-15-2003 recoded L_rot, L_swap, and L_over for increased efficiency
//	01-31-2004 extended JumpTable past 256 entries, 
//                   added CPP_include, CPP_source, CPP_refill, and
//                   CPP_nondeferred, CPP_state, CPP_colon, CPP_semicolon
//	02-09-2004 added CPP_allocate and CPP_free
//	03-18-2004 added L_fsincos, C_fsinh, C_fcosh, C_ftanh, C_fasinh,
//                   C_facosh, C_fatanh
//      03-21-2004 implemented L_fsincos
//      03-30-2004 modified L_ret to remove unneeded instructions
//      04-08-2004 added CPP_postpone and CPP_compilecomma
//      04-09-2004 added CPP_semicolon
//      04-18-2004 completed recoding with macros for hybrid threading.
//	04-26-2004 exported L_udmstar and L_utmslash for use by ForthVM.cpp
//      06-20-2004 modified L_call to prevent loss of return error code
//      09-05-2004 added C_forth_signal, C_raise, C_setitimer, C_getitimer
//      09-10-2004 moved CPP_spstore and CPP_rpstore to opcodes below 256
//      09-28-2005 changed some comment markers to // to prevent GNU as from quirky errors
//	02-10-2006 The EBP register is now used to hold GlobalIp for the jump-threaded portion
//                   of the VM. This results in 10--20% speedup for typical apps.
//                   Any additions to code in vm.s which needs to make use of the EBP
//		     register should save and restore this register using "pushl %ebp" at 
//	             the beginning and "popl %ebp" before returning (ret) or using NEXT.
//	             Also, calls from vm.s to external words (C/C++) that reference
//		     GlobalIp should update GlobalIp prior to the call:	"movl %ebp, GlobalIp"
//	04-01-2006 fixed bug in L_mslash
//      04-11-2006 added stubs for L_dult, L_utsslashmod, L_stsslashrem, L_udmstar, L_cputest,
//                   L_dsstar, L_dmax, and L_dmin 
//      04-19-2006 added C_syscall
//	04-21-2006 recoded relational operators L_eq, L_neq, L_ult, L_ugt so that they don't 
//                   require conditional branching! Added macro EQ
//      04-22-2006 finished recoding dyadic single relational operators L_lt, L_gt, L_le, L_ge,
//                   L_zeroeq, L_zerone, L_zerolt, L_zerogt. Added macros REL_DYADIC and REL_ZERO
//                   and removed macro EQ.
//      04-23-2006 added macro LOGIC_DYADIC and used it for recoding _AND, _OR, and _XOR macros.
//      04-24-2006 recoded L_dequal, L_dlt, and L_dzeroeq to remove conditional branching.
//      04-27-2006 made macro STOD and fixed L_mplus by using STOD and DPLUS macros; coded L_dult,
//		     recoded L_dlt to fix problem caused by using DMINUS; simplified _NOT macro
//		     and used it in L_fne; recoded L_feq and L_flt to remove conditional branches.
//	04-29-2006 coded L_dmin and L_dmax; added following macros: DLT FDUP FSWAP FDROP 
//	04-30-2006 added macros _ABS and TNEG, and finished coding L_dsstar
//      05-02-2006 added macro FREL_DYADIC and used it to recode dyadic fp relational operators.
//	05-30-2006 rename L_cmove to L_move, and coded correct L_cmove
//      10-17-2006 recoded L_fzeroeq and L_fzerolt to remove conditional branching. Added
//                   L_fzerogt.
//      11-06-2006 fixed L_utmslash (now passes divtest.4th) and added division by zero
//                   and division overflow checks in this word ;  added proper checks and handling
//                   for division overflow and div by zero in L_umslashmod ;  added division by
//                   zero checks in L_fmslashmod and L_smslashrem
//      02-19-2007 division by zero error checks added for L_mod and L_slashmod
//      07-01-2007 implemented UTS/MOD and STS/REM
//	10-05-2007 replaced memcpy with memmove in L_move
.equ WSIZE,	4

.equ OP_ADDR,	65
.equ OP_FVAL,	70
.equ OP_IVAL,	73
.equ OP_RET,	238
.equ SIGN_MASK,	0x80000000
	
// Error Codes

.equ E_NOT_ADDR,	1
.equ E_DIV_ZERO,	4
.equ E_RET_STK_CORRUPT,	5
.equ E_UNKNOWN_OP,	6
.equ E_DIV_OVERFLOW,   20
	
.data
FCONST_180: .double 180.
_JumpTable: .int L_false, L_true, L_cells, L_cellplus # 0 -- 3
           .int L_dfloats, L_dfloatplus, _CPP_case, _CPP_endcase  # 4 -- 7
           .int _CPP_of, _CPP_endof, _C_open, _C_lseek     # 8 -- 11
           .int _C_close, _C_read, _C_write, _C_ioctl # 12 -- 15
           .int L_usleep, L_ms, _C_msfetch, _C_syscall  # 16 -- 19
           .int L_fill, L_cmove, L_cmovefrom, _CPP_dotparen # 20 -- 23
           .int _CPP_bracketsharp, _CPP_tofile, _CPP_console, _CPP_sharpbracket  # 24 -- 27
           .int _CPP_sharps, _CPP_squote, _CPP_cr, L_bl    # 28 -- 31
           .int _CPP_spaces, L_store, _CPP_cquote, _CPP_sharp # 32 -- 35
           .int _CPP_sign, L_mod, L_and, _L_tick       # 36 -- 39
           .int _CPP_lparen, _CPP_hold, L_mul, L_add  # 40 -- 43
           .int L_nop, L_sub, _CPP_dot, L_div  # 44 -- 47
           .int _L_dabs, L_dnegate, L_umstar, L_umslashmod   # 48 -- 51
           .int L_mstar, L_mplus, L_mslash, L_mstarslash # 52 -- 55
           .int L_fmslashmod, L_smslashrem, _CPP_colon, _CPP_semicolon # 56 -- 59
           .int L_lt, L_eq, L_gt, L_question      # 60 -- 63
           .int L_fetch, L_addr, L_base, L_call   # 64 -- 67
           .int L_definition, L_erase, L_fval, _CPP_forget # 68 -- 71
           .int L_tobody, L_ival, _CPP_evaluate, _C_key     # 72 -- 75
           .int L_lshift, L_slashmod, _C_numberquery, _CPP_dotr # 76 -- 79
           .int _CPP_ddot, _C_keyquery, L_rshift, _CPP_dots  # 80 -- 83
           .int _C_accept, _CPP_char, _CPP_bracketchar, _CPP_word  # 84 -- 87
           .int L_starslash, L_starslashmod, _CPP_udotr, _CPP_lbracket  # 88 -- 91
           .int _CPP_backslash, _CPP_rbracket, L_xor, _CPP_literal  # 92 -- 95
           .int _CPP_queryallot, _CPP_allot, L_binary, L_count # 96 -- 99
           .int L_decimal, _CPP_emit, _CPP_fdot, _CPP_cold # 100 -- 103
           .int L_hex, L_i, L_j, _CPP_brackettick         # 104 -- 107
           .int _CPP_fvariable, _C_timeanddate, _CPP_find, _CPP_constant # 108 -- 111
           .int _CPP_immediate, _CPP_fconstant, _CPP_create, _CPP_dotquote  # 112 -- 115
           .int _CPP_type, _CPP_udot, _CPP_variable, _CPP_words # 116 -- 119
           .int _CPP_does, _C_system, _C_chdir, _C_search   # 120 -- 123
           .int L_or, _C_compare, L_not, L_move    # 124 -- 127
           .int L_fsin, L_fcos, _C_ftan, _C_fasin   # 128 -- 131
           .int _C_facos, _C_fatan, _C_fexp, _C_fln   # 132 -- 135
           .int _C_flog, L_fatan2, L_ftrunc, L_ftrunctos    # 136 -- 139
           .int _C_fmin, _C_fmax, L_floor, L_fround # 140 -- 143
           .int L_dlt, L_dzeroeq, L_deq, L_twopush  # 144 -- 147
           .int L_twopop, L_tworfetch, L_stod, L_stof # 148 -- 151
           .int L_dtof, L_froundtos, L_ftod, L_degtorad  # 152 -- 155
           .int L_radtodeg, L_dplus, _L_dminus, L_dult   # 156 -- 159
           .int L_inc, L_dec, L_abs, L_neg        # 160 -- 163
           .int L_min, L_max, L_twostar, L_twodiv # 164 -- 167
           .int L_twoplus, L_twominus, L_cfetch, L_cstore  # 168 -- 171
           .int L_wfetch, L_wstore, L_dffetch, L_dfstore  # 172 -- 175
           .int L_sffetch, L_sfstore, L_spfetch, L_plusstore # 176 -- 179
           .int L_fadd, L_fsub, L_fmul, L_fdiv    # 180 -- 183
           .int L_fabs, L_fneg, _C_fpow, L_fsqrt   # 184 -- 187
           .int _CPP_spstore, _CPP_rpstore, L_feq, L_fne  # 188 -- 191
           .int L_flt, L_fgt, L_fle, L_fge        # 192 -- 195
           .int L_fzeroeq, L_fzerolt, L_fzerogt, L_nop # 196 -- 199
           .int L_drop, L_dup, L_swap, L_over     # 200 -- 203
           .int L_rot, L_minusrot, L_nip, L_tuck  # 204 -- 207
           .int L_pick, L_roll, L_2drop, L_2dup   # 208 -- 211
           .int L_2swap, L_2over, L_2rot, _L_depth # 212 -- 215
           .int L_querydup, _CPP_if, _CPP_else, _CPP_then # 216 -- 219
           .int L_push, L_pop, L_puship, L_rfetch # 220 -- 223
           .int L_rpfetch, L_afetch, _CPP_do, _CPP_leave # 224 -- 227
           .int _CPP_querydo, _CPP_abortquote, L_jz, L_jnz  # 228 -- 231
           .int L_jmp, L_loop, L_plusloop, L_unloop  # 232 -- 235
           .int L_execute, _CPP_recurse, _L_ret, _L_abort  # 236 -- 239
           .int L_quit, L_ge, L_le, L_ne          # 240 -- 243
           .int L_zeroeq, L_zerone, L_zerolt, L_zerogt # 244 -- 247
           .int L_ult, L_ugt, _CPP_begin, _CPP_while    # 248 -- 251
           .int _CPP_repeat, _CPP_until, _CPP_again, _CPP_bye  # 252 -- 255
	   .int _L_utmslash, L_utsslashmod, L_stsslashrem, _L_udmstar # 256 -- 259
	   .int _CPP_include, _CPP_source, _CPP_refill, _CPP_nondeferred # 260--263
	   .int _CPP_state, _CPP_allocate, _CPP_free, L_nop  # 264--267
	   .int L_cputest, L_dsstar, _CPP_compilecomma, L_nop     # 268--271
	   .int _CPP_postpone, L_nop, L_nop, _C_forth_signal # 272--275
	   .int _C_raise, _C_setitimer, _C_getitimer, L_nop   # 276--279
	   .int L_nop, L_fsincos, _C_facosh, _C_fasinh  # 280--283
	   .int _C_fatanh, _C_fcosh, _C_fsinh, _C_ftanh   # 284--287
	   .int L_dmax, L_dmin, L_nop, L_nop          # 288--291
.text
	.align 4
.global _JumpTable
.global _L_depth, _L_tick, _L_quit, _L_abort, _L_ret, _L_dabs
.global _L_dminus, _L_mstarslash, _L_udmstar, _L_utmslash

.macro NEXT
	incl %ebp		 # increment the Forth instruction ptr
	movl %ebp, _GlobalIp
	movb (%ebp), %al         # get the opcode
	movl _JumpTable(,%eax,4), %ebx	# machine code address of word
	xor %eax, %eax
	jmpl *%ebx		# call the word
.endm

.macro SWAP
	pushl %ebp
        movl _GlobalSp, %ebx
        addl $WSIZE, %ebx
	movl %ebx, %ebp
	movl (%ebp), %ecx
	addl $WSIZE, %ebx
	movl (%ebx), %edx
	movl %ecx, (%ebx)
	movl %edx, (%ebp)
        movl _GlobalTp, %ebx
        incl %ebx
	movl %ebx, %ebp
	movb (%ebp), %al
	incl %ebx
	movb (%ebx), %cl
	movb %al, (%ebx)
	movb %cl, (%ebp)
	popl %ebp
        xor %eax, %eax
.endm

.macro DUP
	movl _GlobalSp, %ebx
        movl WSIZE(%ebx), %eax
        movl %eax, (%ebx)
	subl $WSIZE, %ebx
	movl %ebx, _GlobalSp
        movl _GlobalTp, %ebx
        movb 1(%ebx), %al
        movb %al, (%ebx)
        decl _GlobalTp
        xor %eax, %eax
.endm 

.macro DROP
        addl $WSIZE, _GlobalSp
        incl _GlobalTp
.endm

.macro OVER
        movl _GlobalSp, %ebx
        movl 2*WSIZE(%ebx), %eax
        movl %eax, (%ebx)
	subl $WSIZE, %ebx
	movl %ebx, _GlobalSp
        movl _GlobalTp, %ebx
        movb 2(%ebx), %al
        movb %al, (%ebx)
        decl _GlobalTp
        xor %eax, %eax
.endm

.macro FDUP
	movl _GlobalSp, %ebx
	movl %ebx, %ecx
	addl $WSIZE, %ebx
	movl (%ebx), %edx
	addl $WSIZE, %ebx
	movl (%ebx), %eax
	movl %ecx, %ebx
	movl %eax, (%ebx)
	subl $WSIZE, %ebx
	movl %edx, (%ebx)
	subl $WSIZE, %ebx
	movl %ebx, _GlobalSp
	movl _GlobalTp, %ebx
	incl %ebx
	movw (%ebx), %ax
	subl $2, %ebx
	movw %ax, (%ebx)
	decl %ebx
	movl %ebx, _GlobalTp
	xor %eax, %eax
.endm
	
.macro FDROP
	movl $WSIZE, %eax
	addl %eax, %eax
	addl %eax, _GlobalSp
        addl $2, _GlobalTp
	xor %eax, %eax
.endm

.macro FSWAP
	movl $WSIZE, %ecx
	movl _GlobalSp, %ebx
	addl %ecx, %ebx
	movl (%ebx), %edx
	addl %ecx, %ebx
	movl (%ebx), %eax
	addl %ecx, %ebx
	xchgl %edx, (%ebx)
	addl %ecx, %ebx
	xchgl %eax, (%ebx)
	subl %ecx, %ebx
	subl %ecx, %ebx
	movl %eax, (%ebx)
	subl %ecx, %ebx
	movl %edx, (%ebx)
	movl _GlobalTp, %ebx
	incl %ebx
	movw (%ebx), %ax
	addl $2, %ebx
	xchgw %ax, (%ebx)
	subl $2, %ebx
	movw %ax, (%ebx)
	xor %eax, %eax
.endm
		
.macro FOVER
	movl _GlobalSp, %ebx
	movl %ebx, %ecx
	addl $3*WSIZE, %ebx
	movl (%ebx), %edx
	addl $WSIZE, %ebx
	movl (%ebx), %eax
	movl %ecx, %ebx
	movl %eax, (%ebx)
	subl $WSIZE, %ebx
	movl %edx, (%ebx)
	subl $WSIZE, %ebx
	movl %ebx, _GlobalSp
	movl _GlobalTp, %ebx
	movl %ebx, %ecx
	addl $3, %ebx
	movw (%ebx), %ax
	movl %ecx, %ebx
	decl %ebx
	movw %ax, (%ebx)
	decl %ebx
	movl %ebx, _GlobalTp
	xor %eax, %eax	
.endm
	
.macro _PUSH
	movl $WSIZE, %eax
	movl _GlobalSp, %ebx
	addl %eax, %ebx	
        movl %ebx, _GlobalSp
	movl (%ebx), %ecx
	movl _GlobalRp, %ebx
	movl %ecx, (%ebx)
	subl %eax, %ebx
	movl %ebx, _GlobalRp
	movl _GlobalTp, %ebx
	incl %ebx
	movl %ebx, _GlobalTp
	movb (%ebx), %al
	movl _GlobalRtp, %ebx
	movb %al, (%ebx)
	decl %ebx
	movl %ebx, _GlobalRtp
        xor %eax, %eax
.endm
	
.macro _POP
	movl $WSIZE, %eax
	movl _GlobalRp, %ebx
	addl %eax, %ebx
	movl %ebx, _GlobalRp
	movl (%ebx), %ecx
	movl _GlobalSp, %ebx
	movl %ecx, (%ebx)
	subl %eax, %ebx
	movl %ebx, _GlobalSp
	movl _GlobalRtp, %ebx
	incl %ebx
	movl %ebx, _GlobalRtp
	movb (%ebx), %al
	movl _GlobalTp, %ebx
	movb %al, (%ebx)
	decl %ebx
	movl %ebx, _GlobalTp
	xor %eax, %eax
.endm
	
// Dyadic Logic operators 
	
.macro LOGIC_DYADIC op
	movl $WSIZE, %ecx
	movl _GlobalSp, %ebx
	addl %ecx, %ebx
	movl %ebx, _GlobalSp
	movl (%ebx), %eax
	addl %ecx, %ebx
	\op (%ebx), %eax
	movl %eax, (%ebx)
	movl _GlobalTp, %eax
	incl %eax
	movl %eax, _GlobalTp
	movb $OP_IVAL, 1(%eax)	
	xorl %eax, %eax 
.endm
	
.macro _AND
	LOGIC_DYADIC andl
.endm
	
.macro _OR
	LOGIC_DYADIC orl
.endm
	
.macro _XOR
	LOGIC_DYADIC xorl
.endm
	
.macro _NOT
	movl _GlobalSp, %ebx
	notl WSIZE(%ebx)
.endm

// use algorithm from DNW's vm-osxppc.s
.macro _ABS	
	movl _GlobalSp, %ebx
	addl $WSIZE, %ebx
	movl (%ebx), %ecx
	xorl %eax, %eax
	cmpl %eax, %ecx
	setl %al
	negl %eax
	movl %eax, %edx
	xorl %ecx, %edx
	subl %eax, %edx
	movl %edx, (%ebx)
	xorl %eax, %eax
.endm
	
// Dyadic relational operators (single length numbers) 
	
.macro REL_DYADIC setx
	movl $WSIZE, %ecx
	movl _GlobalSp, %ebx
	addl %ecx, %ebx
	movl %ebx, _GlobalSp
	movl (%ebx), %eax
	addl %ecx, %ebx
	cmpl %eax, (%ebx)
	movl $0, %eax
	\setx %al
	negl %eax
	movl %eax, (%ebx)
	movl _GlobalTp, %eax
	incl %eax
	movl %eax, _GlobalTp
	movb $OP_IVAL, 1(%eax)
	xorl %eax, %eax
.endm

// Relational operators for zero (single length numbers)
	
.macro REL_ZERO setx
	movl _GlobalSp, %ebx
	addl $WSIZE, %ebx
	movl (%ebx), %eax
	cmpl $0, %eax
	movl $0, %eax
	\setx %al
	negl %eax
	movl %eax, (%ebx)
	movl _GlobalTp, %eax
	movb $OP_IVAL, 1(%eax)
	xorl %eax, %eax
.endm

.macro FREL_DYADIC logic arg set
	movl $WSIZE, %ecx
	movl _GlobalSp, %ebx
	addl %ecx, %ebx
	fldl (%ebx)
	addl %ecx, %ebx
	addl %ecx, %ebx
	movl %ebx, _GlobalSp
	fcompl (%ebx)
	fnstsw %ax
	andb $65, %ah
	\logic \arg, %ah
	movl $0, %eax
	\set %al
	negl %eax
	addl %ecx, %ebx
	movl %eax, (%ebx)
	movl _GlobalTp, %eax
	addl $3, %eax
	movl %eax, _GlobalTp
	movb $OP_IVAL, 1(%eax)
	xorl %eax, %eax
.endm
				
.macro STOD
	movl $WSIZE, %ecx
	movl _GlobalSp, %ebx
	movl WSIZE(%ebx), %eax
	cdq
	movl %edx, (%ebx)
	subl %ecx, %ebx
	movl %ebx, _GlobalSp
	mov _GlobalTp, %eax
	movb $OP_IVAL, (%eax)
	decl %eax
	movl %eax, _GlobalTp
	xor %eax, %eax
.endm


	# b = (d1.hi < d2.hi) OR ((d1.hi = d2.hi) AND (d1.lo u< d2.lo))
.macro DLT
	movl $WSIZE, %ecx
	xorl %edx, %edx
	movl _GlobalSp, %ebx
	addl %ecx, %ebx
	movl (%ebx), %eax
	cmpl %eax, 2*WSIZE(%ebx)
	sete %dl
	setl %dh
	addl %ecx, %ebx
	movl (%ebx), %eax
	addl %ecx, %ebx
	movl %ebx, _GlobalSp
	addl %ecx, %ebx
	cmpl %eax, (%ebx)
	setb %al
	andb %al, %dl
	orb  %dh, %dl
	xorl %eax, %eax
	movb %dl, %al
	negl %eax
	movl %eax, (%ebx)
	movl _GlobalTp, %eax
	addl $4, %eax
	movb $OP_IVAL, (%eax)
	decl %eax
	movl %eax, _GlobalTp	
	xorl %eax, %eax
.endm
			
.macro DNEGATE
	movl _GlobalSp, %ebx
	addl $WSIZE, %ebx
	movl %ebx, %ecx
	addl $WSIZE, %ebx
	movl (%ebx), %eax
	notl %eax
	clc
	addl $1, %eax
	movl %eax, (%ebx)
	movl %ecx, %ebx
	movl (%ebx), %eax
	notl %eax
	adcl $0, %eax
	movl %eax, (%ebx)
	xor %eax, %eax	
.endm
	
.macro DPLUS
	movl _GlobalSp, %ebx
	addl $2*WSIZE, %ebx
	movl (%ebx), %eax
	clc
	addl 2*WSIZE(%ebx), %eax
	movl %eax, 2*WSIZE(%ebx)
	movl WSIZE(%ebx), %eax
	adcl -WSIZE(%ebx), %eax
	movl %eax, WSIZE(%ebx)
	movl %ebx, _GlobalSp
	incl _GlobalTp
	incl _GlobalTp
	xor %eax, %eax
.endm
	
.macro DMINUS
	movl _GlobalSp, %ebx
	addl $2*WSIZE, %ebx
	movl 2*WSIZE(%ebx), %eax
	clc
	subl (%ebx), %eax
	movl %eax, 2*WSIZE(%ebx)
	movl WSIZE(%ebx), %eax
	sbbl -WSIZE(%ebx), %eax
	movl %eax, WSIZE(%ebx)
	movl %ebx, _GlobalSp
	incl _GlobalTp
	incl _GlobalTp
	xor %eax, %eax
.endm
	
.macro STARSLASH
	movl $WSIZE, %eax
	sall $1, %eax
        addl %eax, _GlobalSp
        movl _GlobalSp, %ebx
        movl WSIZE(%ebx), %eax
        imull (%ebx)
	idivl -WSIZE(%ebx)
	movl %eax, WSIZE(%ebx)
	incl _GlobalTp
	incl _GlobalTp
	xor %eax, %eax
.endm
	

.macro TNEG
	movl $WSIZE, %eax
	movl _GlobalSp, %ebx
	addl %eax, %ebx
	movl (%ebx), %edx
	addl %eax, %ebx
	movl (%ebx), %ecx
	addl %eax, %ebx
	movl (%ebx), %eax
	notl %eax
	notl %ecx
	notl %edx
	clc
	addl $1, %eax
	adcl $0, %ecx
	adcl $0, %edx
	movl %eax, (%ebx)
	movl $WSIZE, %eax
	subl %eax, %ebx
	movl %ecx, (%ebx)
	subl %eax, %ebx
	movl %edx, (%ebx)
	xor %eax, %eax	
.endm

// VIRTUAL MACHINE 
						
.global _vm
//	.type	vm,@function
_vm:	
        pushl %ebp
        pushl %ebx
        pushl %ecx
        pushl %edx
	pushl _GlobalIp
	pushl _vmEntryRp
        movl %esp, %ebp
        movl 28(%ebp), %ebp     # load the Forth instruction pointer
        movl %ebp, _GlobalIp
	movl _GlobalRp, %eax
	movl %eax, _vmEntryRp
	xor %eax, %eax
next:
        movb (%ebp), %al         # get the opcode
	movl _JumpTable(,%eax,4), %ebx	# machine code address of word
	xor %eax, %eax          # clear error code
	call *%ebx		# call the word
	movl _GlobalIp, %ebp
	incl %ebp		 # increment the Forth instruction ptr
	movl %ebp, _GlobalIp
	cmpb $0, %al		 # check for error
	jz next        
exitloop:
        cmpl $OP_RET, %eax         # return from vm?
        jnz vmexit
        xor %eax, %eax            # clear the error
vmexit:
	pop _vmEntryRp
	pop _GlobalIp
	pop %edx
        pop %ecx
        pop %ebx
        pop %ebp
        ret
L_cputest:
	ret
L_nop:
        movl $E_UNKNOWN_OP, %eax   # unknown operation
        ret
L_quit:
	movl _BottomOfReturnStack, %eax	# clear the return stacks
	movl %eax, _GlobalRp
	movl %eax, _vmEntryRp
	movl _BottomOfReturnTypeStack, %eax
	movl %eax, _GlobalRtp
	movl $8, %eax		# exit the virtual machine
	ret
_L_abort:
	movl _BottomOfStack, %eax
	movl %eax, _GlobalSp
	movl _BottomOfTypeStack, %eax
	movl %eax, _GlobalTp
	jmp L_quit
L_base:	
	movl _GlobalSp, %ebx
	movl $_Base, %eax
	movl %eax, (%ebx)
	subl $WSIZE, %ebx
	movl %ebx, _GlobalSp
	movl _GlobalTp, %ebx
	movb $OP_ADDR, (%ebx)
	decl _GlobalTp
	xor %eax, %eax
	NEXT
L_binary:
	movl $_Base, %ebx
	movl $2, (%ebx)
	NEXT
L_decimal:	
	movl $_Base, %ebx
	movl $10, (%ebx)
	NEXT
L_hex:	
	movl $_Base, %ebx
	movl $16, (%ebx)
	NEXT		
L_false:
	movl _GlobalSp, %ebx
	movl $0, (%ebx)
	subl $WSIZE, %ebx
	movl %ebx, _GlobalSp
	movl _GlobalTp, %ebx
	movb $OP_IVAL, (%ebx)
	decl _GlobalTp
	NEXT
L_true:
	movl _GlobalSp, %ebx
	movl $-1, (%ebx)
	subl $WSIZE, %ebx
	movl %ebx, _GlobalSp
	movl _GlobalTp, %ebx
	movb $OP_IVAL, (%ebx)
	decl _GlobalTp
	NEXT
L_cells:
	movl _GlobalSp, %ebx
	addl $WSIZE, %ebx
	movl (%ebx), %eax
	sall $2, %eax
	movl %eax, (%ebx)
	xor %eax, %eax
	NEXT
L_cellplus:
	movl _GlobalSp, %ebx
	addl $WSIZE, %ebx
	movl (%ebx), %eax
	addl $WSIZE, %eax
	movl %eax, (%ebx)
	xor %eax, %eax
	NEXT
L_dfloats:	
	movl _GlobalSp, %ebx
	addl $WSIZE, %ebx
	movl (%ebx), %eax
	sall $3, %eax
	movl %eax, (%ebx)
	xor %eax, %eax
	NEXT
L_dfloatplus:	
	movl _GlobalSp, %ebx
	addl $WSIZE, %ebx
	movl (%ebx), %eax
	addl $WSIZE, %eax
	addl $WSIZE, %eax
	movl %eax, (%ebx)
	xor %eax, %eax
	NEXT				
L_bl:
	movl _GlobalSp, %ebx
	movl $32, (%ebx)
	subl $WSIZE, %ebx
	movl %ebx, _GlobalSp
	movl _GlobalTp, %ebx
	movb $OP_IVAL, (%ebx)
	decl _GlobalTp	
	NEXT
_L_ret:
	movl _vmEntryRp, %eax		# Return Stack Ptr on entry to VM
	movl _GlobalRp, %ecx
	cmpl %eax, %ecx
	jl ret1
        movl $OP_RET, %eax             # exhausted the return stack so exit vm
        ret
ret1:
	addl $WSIZE, %ecx
        movl %ecx, _GlobalRp
        incl _GlobalRtp	
	movl _GlobalRtp, %ebx
	movb (%ebx), %al
        cmpb $OP_ADDR, %al
        jz ret2
        movl $E_RET_STK_CORRUPT, %eax   # indicate return stack corrupted
        jmp retexit
ret2:   movl (%ecx), %eax
	movl %eax, _GlobalIp		# reset the instruction ptr
        xor %eax, %eax
retexit:
        ret
_L_tick:
	movl _GlobalSp, %ebx
	movl $32, (%ebx)
	movl $WSIZE, %eax
	subl %eax, _GlobalSp
	decl _GlobalTp
	call _CPP_word
	call _CPP_find
	DROP
	ret
L_tobody:
	movl _GlobalSp, %ebx
	addl $WSIZE, %ebx
	movl (%ebx), %ecx	# code address
	incl %ecx		# the data address is offset by one
	movl (%ecx), %ecx
	movl %ecx, (%ebx)
	ret
#
# For precision delays, use MS instead of USLEEP
# Use USLEEP when task can be put to sleep and reawakened by OS
#
L_usleep:
	movl $WSIZE, %eax
	addl %eax, _GlobalSp
	incl _GlobalTp
	movl _GlobalSp, %ebx
	movl (%ebx), %eax
	pushl %eax
	call _usleep
	popl %eax
	xor %eax, %eax
	ret
L_ms:
	movl _GlobalSp, %ebx
	movl WSIZE(%ebx), %eax
	imull $1000, %eax
	movl %eax, WSIZE(%ebx)
	call _C_usec
	ret
L_fill:
	SWAP
	movl $WSIZE, %eax
	addl %eax, _GlobalSp
	incl _GlobalTp
	movl _GlobalSp, %ebx
	movl (%ebx), %ebx
	pushl %ebx
	addl %eax, _GlobalSp
	incl _GlobalTp
	movl _GlobalSp, %ebx
	movl (%ebx), %ebx
	pushl %ebx
	addl %eax, _GlobalSp
	incl _GlobalTp
	movl _GlobalTp, %ebx
	movb (%ebx), %al
	cmpb $OP_ADDR, %al
	jz fill2
	popl %ebx
	popl %ebx
	movl $E_NOT_ADDR, %eax
	jmp fillexit
fill2:	movl _GlobalSp, %ebx
	movl (%ebx), %ebx
	pushl %ebx
	call _memset
	addl $12, %esp
	xor %eax, %eax
fillexit:	
	ret
L_erase:
	movl _GlobalSp, %ebx
	movl $0, (%ebx)
	subl $WSIZE, %ebx
	movl %ebx, _GlobalSp
	decl _GlobalTp
	call L_fill
	ret	
L_move:
	movl $WSIZE, %eax
	addl %eax, _GlobalSp
	incl _GlobalTp
	movl _GlobalSp, %ebx
	movl (%ebx), %ebx
	pushl %ebx
	SWAP
	movl $WSIZE, %eax
	addl %eax, _GlobalSp
	incl _GlobalTp
	movl _GlobalTp, %ebx
	movb (%ebx), %al
	cmpb $OP_ADDR, %al
	jz move2
	popl %ebx
	movl $E_NOT_ADDR, %eax
	ret
move2:	movl _GlobalSp, %ebx
	movl (%ebx), %ebx
	pushl %ebx
	movl $WSIZE, %eax
	addl %eax, _GlobalSp
	incl _GlobalTp
	movl _GlobalTp, %ebx
	movb (%ebx), %al
	cmpb $OP_ADDR, %al
	jz move3
	popl %ebx
	popl %ebx
	movl $E_NOT_ADDR, %eax
	ret
move3:	movl _GlobalSp, %ebx
	movl (%ebx), %ebx
	pushl %ebx
	call _memmove
	addl $12, %esp
	xor %eax, %eax				
	ret
L_cmove:
	movl $WSIZE, %eax
	movl _GlobalSp, %ebx
	addl %eax, %ebx
	movl (%ebx), %ecx		# nbytes in ecx
	cmpl $0, %ecx
	jnz  cmove1
	addl $2*WSIZE, %ebx
	movl %ebx, _GlobalSp
	addl $3, _GlobalTp
	xorl %eax, %eax
	NEXT		
cmove1:	incl _GlobalTp
	addl %eax, %ebx
	movl (%ebx), %edx		# dest addr in edx
	movl %ebx, _GlobalSp
	incl _GlobalTp
	movl _GlobalTp, %ebx
	movb (%ebx), %al
	cmpb $OP_ADDR, %al
	jz cmove2
	movl $E_NOT_ADDR, %eax
	ret
cmove2:	movl _GlobalSp, %ebx
	movl $WSIZE, %eax
	addl %eax, %ebx
	movl (%ebx), %eax		# src addr in eax
	movl %ebx, _GlobalSp
	incl _GlobalTp
	movl _GlobalTp, %ebx
	movb (%ebx), %bl
	cmpb $OP_ADDR, %bl
	jz cmove3
	movl $E_NOT_ADDR, %eax
	ret
cmove3:	movl %eax, %ebx			# src addr in ebx
cmoveloop: movb (%ebx), %al
	movb %al, (%edx)
	incl %ebx
	incl %edx
	loop cmoveloop
	xor %eax, %eax				
	NEXT				
L_cmovefrom:
	movl $WSIZE, %eax
	addl %eax, _GlobalSp
	incl _GlobalTp
	movl _GlobalSp, %ebx
	movl (%ebx), %ecx	# load count register
	addl %eax, _GlobalSp
	incl _GlobalTp
	movl _GlobalTp, %ebx
	movb (%ebx), %al
	cmpb $OP_ADDR, %al
	jz cmovefrom2
	movl $E_NOT_ADDR, %eax						
	ret
cmovefrom2:
	movl _GlobalSp, %ebx
	movl (%ebx), %ebx
	movl %ecx, %eax
	decl %eax
	addl %eax, %ebx
	movl %ebx, %edx		# dest addr in %edx
	movl $WSIZE, %eax
	addl %eax, _GlobalSp
	incl _GlobalTp
	movl _GlobalTp, %ebx
	movb (%ebx), %al
	cmpb $OP_ADDR, %al
	jz cmovefrom3
	movl $E_NOT_ADDR, %eax
	ret
cmovefrom3:
	movl _GlobalSp, %ebx
	movl (%ebx), %ebx
	movl %ecx, %eax
	cmpl $0, %eax
	jnz cmovefrom4
	ret
cmovefrom4:	
	decl %eax
	addl %eax, %ebx		# src addr in %ebx
cmovefromloop:	
	movb (%ebx), %al
	decl %ebx
	xchgl %ebx, %edx
	movb %al, (%ebx)
	decl %ebx
	xchgl %ebx, %edx
	loop cmovefromloop	
	xor %eax, %eax
	ret
L_call:	
	movl $WSIZE, %eax
	addl %eax, _GlobalSp
	xor %eax, %eax
	incl _GlobalTp
	movl _GlobalSp, %ebx
	# call *(%ebx)
	jmpl *(%ebx)
	ret		
L_push:
	_PUSH
        NEXT
L_pop:
	_POP
	NEXT
L_twopush:
	movl _GlobalSp, %ebx
	addl $WSIZE, %ebx
	movl (%ebx), %edx
	addl $WSIZE, %ebx
	movl (%ebx), %eax
	movl %ebx, _GlobalSp
	movl _GlobalRp, %ebx
	movl %eax, (%ebx)
	subl $WSIZE, %ebx
	movl %edx, (%ebx)
	subl $WSIZE, %ebx
	movl %ebx, _GlobalRp
	movl _GlobalTp, %ebx
	incl %ebx
	movw (%ebx), %ax
	incl %ebx
	movl %ebx, _GlobalTp
	movl _GlobalRtp, %ebx
	decl %ebx
	movw %ax, (%ebx)
	decl %ebx
	movl %ebx, _GlobalRtp
	xor %eax, %eax
	NEXT
L_twopop:
	movl _GlobalRp, %ebx
	addl $WSIZE, %ebx
	movl (%ebx), %edx
	addl $WSIZE, %ebx
	movl (%ebx), %eax
	movl %ebx, _GlobalRp
	movl _GlobalSp, %ebx
	movl %eax, (%ebx)
	subl $WSIZE, %ebx
	movl %edx, (%ebx)
	subl $WSIZE, %ebx
	movl %ebx, _GlobalSp
	movl _GlobalRtp, %ebx
	incl %ebx
	movw (%ebx), %ax
	incl %ebx
	movl %ebx, _GlobalRtp
	movl _GlobalTp, %ebx
	decl %ebx
	movw %ax, (%ebx)
	decl %ebx
	movl %ebx, _GlobalTp
	xor %eax, %eax				
	NEXT
L_puship:
        movl %ebp, %eax
        movl _GlobalRp, %ebx
        mov %eax, (%ebx)
	movl $WSIZE, %eax
        subl %eax, _GlobalRp
        movl _GlobalRtp, %ebx
	movb $OP_ADDR, %al
        movb %al, (%ebx)
	decl _GlobalRtp
        xor %eax, %eax
        NEXT
L_execute:	
        movl %ebp, %ecx
        movl _GlobalRp, %ebx
        movl %ecx, (%ebx)
	movl $WSIZE, %eax 
        subl %eax, %ebx 
	movl %ebx, _GlobalRp
        movl _GlobalRtp, %ebx
	movb $OP_ADDR, (%ebx)
	decl %ebx
        movl %ebx, _GlobalRtp
	movl _GlobalSp, %ebx
        addl %eax, %ebx
	movl %ebx, _GlobalSp
        movl (%ebx), %eax
	decl %eax
	movl %eax, %ebp
	incl _GlobalTp
        xor %eax, %eax
        NEXT
L_definition:
        movl %ebp, %ebx
	movl $WSIZE, %eax
	incl %ebx
	movl (%ebx), %ecx # address to execute
	addl $3, %ebx
	movl %ebx, %edx
	movl _GlobalRp, %ebx
	movl %edx, (%ebx)
	subl %eax, %ebx
	movl %ebx, _GlobalRp
	movl _GlobalRtp, %ebx
	movb $OP_ADDR, (%ebx)
	decl %ebx
	movl %ebx, _GlobalRtp
	decl %ecx
	movl %ecx, %ebp
        xor %eax, %eax	
	NEXT
L_rfetch:
        movl _GlobalRp, %ebx
        addl $WSIZE, %ebx
        movl (%ebx), %eax
        movl _GlobalSp, %ebx
        movl %eax, (%ebx)
	movl $WSIZE, %eax
        subl %eax, _GlobalSp
        movl _GlobalRtp, %ebx
        incl %ebx
        movb (%ebx), %al
        movl _GlobalTp, %ebx
        movb %al, (%ebx)
        decl _GlobalTp
        xor %eax, %eax
	NEXT
L_tworfetch:
	movl _GlobalRp, %ebx
	addl $WSIZE, %ebx
	movl (%ebx), %edx
	addl $WSIZE, %ebx
	movl (%ebx), %eax
	movl _GlobalSp, %ebx
	movl %eax, (%ebx)
	subl $WSIZE, %ebx
	movl %edx, (%ebx)
	subl $WSIZE, %ebx
	movl %ebx, _GlobalSp
	movl _GlobalRtp, %ebx
	incl %ebx
	movw (%ebx), %ax
	incl %ebx
	movl _GlobalTp, %ebx
	decl %ebx
	movw %ax, (%ebx)
	decl %ebx
	movl %ebx, _GlobalTp
	xor %eax, %eax				
	NEXT
L_rpfetch:
	movl _GlobalRp, %eax
	addl $WSIZE, %eax
	movl _GlobalSp, %ebx
	movl %eax, (%ebx)
	subl $WSIZE, %ebx
	movl %ebx, _GlobalSp
	movl _GlobalTp, %ebx
	movb $OP_ADDR, (%ebx)
	decl %ebx
	movl %ebx, _GlobalTp
	xor %eax, %eax
	NEXT
L_spfetch:
	movl _GlobalSp, %eax
	movl %eax, %ebx
	addl $WSIZE, %eax
	movl %eax, (%ebx)
	subl $WSIZE, %ebx
	movl %ebx, _GlobalSp
	movl _GlobalTp, %ebx
	movb $OP_ADDR, (%ebx)
	decl %ebx
	movl %ebx, _GlobalTp
	xor %eax, %eax 
	NEXT
L_i:
        movl _GlobalRtp, %ebx
        movb 3(%ebx), %al
        movl _GlobalTp, %ebx
	movb %al, (%ebx)
	decl %ebx
	movl %ebx, _GlobalTp
        movl _GlobalRp, %ebx
        movl 3*WSIZE(%ebx), %eax
        movl _GlobalSp, %ebx
        movl %eax, (%ebx)
	movl $WSIZE, %eax
        subl %eax, %ebx 
	movl %ebx, _GlobalSp
        xor %eax, %eax
        NEXT
L_j:
        movl _GlobalRtp, %ebx
        movb 6(%ebx), %al
	movl _GlobalTp, %ebx
	movb %al, (%ebx)
	decl %ebx
	movl %ebx, _GlobalTp
        movl _GlobalRp, %ebx
        movl 6*WSIZE(%ebx), %eax
        movl _GlobalSp, %ebx
        movl %eax, (%ebx)
	movl $WSIZE, %eax
        subl %eax, %ebx
	movl %ebx, _GlobalSp
        xor %eax, %eax
        NEXT
L_loop:
        movl _GlobalRtp, %ebx
        incl %ebx
        movb (%ebx), %al
        cmpb $OP_ADDR, %al
        jnz loopbad
        movl _GlobalRp, %ebx	
	movl $WSIZE, %eax
	addl %eax, %ebx
	movl (%ebx), %edx
	addl %eax, %ebx
	movl (%ebx), %ecx
	addl %eax, %ebx
        movl (%ebx), %eax
        incl %eax
	cmpl %ecx, %eax	
        jz L_unloop
loop1:	
        movl %eax, (%ebx)	# set loop counter to next value
	movl %edx, %ebp		# set instruction ptr to start of loop
        xor %eax, %eax
        NEXT
L_unloop:  
	movl $WSIZE, %eax
	shll $1, %eax
	addl $WSIZE, %eax
	addl %eax, _GlobalRp  # terminal count reached, discard top 3 items
	movl $3, %eax
        addl %eax, _GlobalRtp
        xor %eax, %eax
        NEXT
loopbad:
        movl $E_RET_STK_CORRUPT, %eax
        ret	
L_plusloop:
	pushl %ebp
	movl _GlobalRtp, %ebx
        incl %ebx
        movb (%ebx), %al
        cmpb $OP_ADDR, %al
        jnz loopbad
	movl $WSIZE, %eax
	movl _GlobalSp, %ebx
	addl %eax, %ebx
	movl (%ebx), %ebp	# get loop increment 
	movl %ebx, _GlobalSp
	incl _GlobalTp		
        movl _GlobalRp, %ebx
	addl %eax, %ebx		# get ip and save in edx
	movl (%ebx), %edx
	addl %eax, %ebx
	movl (%ebx), %ecx	# get terminal count in ecx
	addl %eax, %ebx
	movl (%ebx), %eax	# get current loop count 
	addl %ebp, %eax		# increment loop count
	cmpl $0, %ebp
	jl plusloop1
	popl %ebp
	cmpl %ecx, %eax
	jge L_unloop				
	movl %eax, (%ebx)	# set loop counter to incremented value
        movl %edx, %ebp         # set instruction ptr to start of loop
        xor %eax, %eax
        NEXT
plusloop1:
	popl %ebp
	cmpl %ecx, %eax
	jl L_unloop
	movl %eax, (%ebx)
	movl %edx, %ebp
	xor %eax, %eax
	NEXT
L_jz:
	movl $WSIZE, %eax
        movl _GlobalSp, %ebx
	addl %eax, %ebx
	movl %ebx, _GlobalSp
        incl _GlobalTp
        movl (%ebx), %eax
        cmpl $0, %eax
        jz jz1
	movl $4, %eax
        addl %eax, %ebp       # do not jump
	xor %eax, %eax
        NEXT
jz1:    movl %ebp, %ebx
        incl %ebx
        movl (%ebx), %eax          # get the relative jump count
        decl %eax
        addl %eax, %ebp
	xor %eax, %eax
        NEXT
L_jnz:				# not implemented
	ret
L_jmp:
        movl %ebp, %ebx
        incl %ebx
        movl (%ebx), %eax          # get the relative jump count
        addl %eax, %ebx
        subl $2, %ebx
        movl %ebx, %ebp      # set instruction ptr
	xor %eax, %eax
        NEXT
L_count:
	movl _GlobalTp, %ebx
	movb 1(%ebx), %al
	cmpb $OP_ADDR, %al
	jnz counterror
	movb $OP_IVAL, (%ebx)
	decl _GlobalTp
	movl _GlobalSp, %ebx
	movl WSIZE(%ebx), %ebx
	xor %eax, %eax
	movb (%ebx), %al
	movl _GlobalSp, %ebx
	incl WSIZE(%ebx)
	movl %eax, (%ebx)
	movl $WSIZE, %eax
	subl %eax, _GlobalSp
	xor %eax, %eax
	ret
counterror:
	movl $E_NOT_ADDR, %eax
	ret
L_ival:
        movl %ebp, %ebx
        incl %ebx
        movl (%ebx), %eax
	addl $WSIZE-1, %ebx
	movl %ebx, %ebp
	movl _GlobalSp, %ebx
	movl %eax, (%ebx)
	subl $WSIZE, %ebx
	movl %ebx, _GlobalSp
	movl _GlobalTp, %ebx
	movb $OP_IVAL, (%ebx)
	decl %ebx
	movl %ebx, _GlobalTp
	xor %eax, %eax
	NEXT
L_addr:
	movl %ebp, %ebx
	incl %ebx
	movl (%ebx), %eax
	addl $WSIZE-1, %ebx
	movl %ebx, %ebp
	movl _GlobalSp, %ebx
	movl %eax, (%ebx)
	subl $WSIZE, %ebx
	movl %ebx, _GlobalSp
	movl _GlobalTp, %ebx
	movb $OP_ADDR, (%ebx)
	decl %ebx
	movl %ebx, _GlobalTp
	xor %eax, %eax
	NEXT
L_fval:
        movl %ebp, %ebx
        incl %ebx
        movl _GlobalSp, %ecx
        subl $WSIZE, %ecx
        movl (%ebx), %eax
	movl %eax, (%ecx)
	movl WSIZE(%ebx), %eax
	movl %eax, WSIZE(%ecx)
	subl $WSIZE, %ecx
	movl %ecx, _GlobalSp
	addl $2*WSIZE-1, %ebx
	movl %ebx, %ebp
	movl _GlobalTp, %ebx
	movb $OP_IVAL, (%ebx)
	decl %ebx
	movb $OP_IVAL, (%ebx)
	decl %ebx
	movl %ebx, _GlobalTp
	xor %eax, %eax
	NEXT
L_and:
	_AND
	NEXT
L_or:
	_OR
	NEXT
L_not:
	_NOT
	NEXT
L_xor:
	_XOR
	NEXT
L_lshift:
	movl $WSIZE, %eax
	addl %eax, _GlobalSp
	incl _GlobalTp
	movl _GlobalSp, %ebx
	movl (%ebx), %ecx
	movl WSIZE(%ebx), %eax
	shll %cl, %eax
	movl %eax, WSIZE(%ebx)
	xor %eax, %eax
	NEXT
L_rshift:
	movl $WSIZE, %eax
	addl %eax, _GlobalSp
	incl _GlobalTp
	movl _GlobalSp, %ebx
	movl (%ebx), %ecx
	movl WSIZE(%ebx), %eax
	shrl %cl, %eax
	movl %eax, WSIZE(%ebx)
	xor %eax, %eax
	NEXT
L_eq:
	REL_DYADIC sete
	NEXT
L_ne:
	REL_DYADIC setne
	NEXT
L_ult:
	REL_DYADIC setb
	NEXT
L_ugt:
	REL_DYADIC seta 
	NEXT
L_lt:
	REL_DYADIC setl
	NEXT
L_gt:
	REL_DYADIC setg
	NEXT
L_le:
	REL_DYADIC setle
	NEXT
L_ge:
	REL_DYADIC setge
	NEXT
L_zeroeq:
	REL_ZERO setz
	NEXT
L_zerone:
	REL_ZERO setnz
	NEXT
L_zerolt:
	REL_ZERO setl
	NEXT
L_zerogt:
	REL_ZERO setg
	NEXT
L_deq:
	movl _GlobalTp, %ebx
	addl $4, %ebx
	movb $OP_IVAL, (%ebx)
	decl %ebx
	movl %ebx, _GlobalTp
	movl _GlobalSp, %ebx
	addl $WSIZE, %ebx
	movl (%ebx), %edx
	addl $WSIZE, %ebx
	movl (%ebx), %ecx
	addl $WSIZE, %ebx
	movl %ebx, _GlobalSp
	movl (%ebx), %eax
	subl %edx, %eax
	addl $WSIZE, %ebx
	movl (%ebx), %edx
	subl %ecx, %edx
	orl %edx, %eax
	cmpl $0, %eax
	movl $0, %eax
	setz %al
	negl %eax
	movl %eax, (%ebx)
	xorl %eax, %eax
	NEXT
L_dzeroeq:
	movl _GlobalTp, %ebx
	addl $2, %ebx
	movb $OP_IVAL, (%ebx)
	decl %ebx
	movl %ebx, _GlobalTp
	movl _GlobalSp, %ebx
	addl $WSIZE, %ebx
	movl %ebx, _GlobalSp
	movl (%ebx), %eax
	addl $WSIZE, %ebx
	orl (%ebx), %eax
	cmpl $0, %eax
	movl $0, %eax
	setz %al
	negl %eax
	movl %eax, (%ebx)
	xorl %eax, %eax
	NEXT
L_dlt:	
	DLT
	NEXT
L_dult:	# b = (d1.hi u< d2.hi) OR ((d1.hi = d2.hi) AND (d1.lo u< d2.lo)) 
	movl $WSIZE, %ecx
	xorl %edx, %edx
	movl _GlobalSp, %ebx
	addl %ecx, %ebx
	movl (%ebx), %eax
	cmpl %eax, 2*WSIZE(%ebx)
	sete %dl
	setb %dh
	addl %ecx, %ebx
	movl (%ebx), %eax
	addl %ecx, %ebx
	movl %ebx, _GlobalSp
	addl %ecx, %ebx
	cmpl %eax, (%ebx)
	setb %al
	andb %al, %dl
	orb  %dh, %dl
	xorl %eax, %eax
	movb %dl, %al
	negl %eax
	movl %eax, (%ebx)
	movl _GlobalTp, %eax
	addl $4, %eax
	movb $OP_IVAL, (%eax)
	decl %eax
	movl %eax, _GlobalTp
	xorl %eax, %eax
	NEXT 	
L_querydup:
	movl _GlobalSp, %ebx
	movl WSIZE(%ebx), %eax
	cmpl $0, %eax
	je L_querydupexit
	movl %eax, (%ebx)
	movl _GlobalTp, %ebx
	movb 1(%ebx), %al
	movb %al, (%ebx)
	decl _GlobalTp
	movl $WSIZE, %eax
	subl %eax, _GlobalSp
	xor %eax, %eax
L_querydupexit:
	ret	
L_drop:
	DROP
        NEXT
L_dup:
	DUP
        NEXT
L_swap:
	SWAP
        NEXT
L_over:
	OVER
        NEXT
L_rot:
	pushl %ebp
	movl _GlobalSp, %ebx
	movl $WSIZE, %eax
	addl %eax, %ebx
	movl %ebx, %ebp
	addl %eax, %ebx
	addl %eax, %ebx
	movl (%ebx), %ecx
	movl (%ebp), %edx
	movl %ecx, (%ebp)
	addl %eax, %ebp
	movl (%ebp), %ecx
	movl %edx, (%ebp)
	movl %ecx, (%ebx)
        movl _GlobalTp, %ebx
        incl %ebx
	movl %ebx, %ebp
	movw (%ebx), %cx
	addl $2, %ebx
	movb (%ebx), %al
	movb %al, (%ebp)
	incl %ebp
	movw %cx, (%ebp)
	xor %eax, %eax
	popl %ebp
	NEXT
L_minusrot:
	movl _GlobalSp, %ebx
	movl WSIZE(%ebx), %eax
	movl %eax, (%ebx)
	addl $WSIZE, %ebx
	movl WSIZE(%ebx), %eax
	movl %eax, (%ebx)
	addl $WSIZE, %ebx
	movl WSIZE(%ebx), %eax
	movl %eax, (%ebx)
	movl -2*WSIZE(%ebx), %eax
	movl %eax, WSIZE(%ebx)
	movl _GlobalTp, %ebx
	movb 1(%ebx), %al
	movb %al, (%ebx)
	incl %ebx
	movw 1(%ebx), %ax
	movw %ax, (%ebx)
	movb -1(%ebx), %al
	movb %al, 2(%ebx)
	xor %eax, %eax
	NEXT
L_nip:
        SWAP
	movl $WSIZE, %eax
        addl %eax, _GlobalSp
        incl _GlobalTp
	xor %eax, %eax
        NEXT
L_tuck:
        SWAP
        OVER
        NEXT
L_pick:
	movl _GlobalSp, %ebx
	movl WSIZE(%ebx), %eax
	addl $2, %eax
	imul $WSIZE, %eax
	addl %eax, %ebx
	movl (%ebx), %eax
	movl _GlobalSp, %ebx
	movl %eax, (%ebx)
	movl WSIZE(%ebx), %eax
	addl $2, %eax
	movl _GlobalTp, %ebx
	addl %eax, %ebx
	movb (%ebx), %al
	movl _GlobalTp, %ebx
	movb %al, 1(%ebx)
	movl _GlobalSp, %ebx
	movl (%ebx), %eax
	movl %eax, WSIZE(%ebx)
	xor %eax, %eax
	NEXT
L_roll:
	movl $WSIZE, %eax
	addl %eax, _GlobalSp
	incl _GlobalTp
	movl _GlobalSp, %ebx 
	movl (%ebx), %eax
	incl %eax
	pushl %eax
	pushl %eax
	pushl %eax
	pushl %ebx
	imul $WSIZE, %eax
	addl %eax, %ebx		# addr of item to roll
	movl (%ebx), %eax
	popl %ebx
	movl %eax, (%ebx)
	popl %eax		# number of cells to copy
	movl %eax, %ecx
	imul $WSIZE, %eax
	addl %eax, %ebx
	movl %ebx, %edx		# dest addr
	subl $WSIZE, %ebx	# src addr
rollloop:
	movl (%ebx), %eax
	sub $WSIZE, %ebx
	xchgl %ebx, %edx
	movl %eax, (%ebx)
	sub $WSIZE, %ebx
	xchgl %ebx, %edx
	loop rollloop

	popl %eax		# now we have to roll the typestack
	mov _GlobalTp, %ebx	
	addl %eax, %ebx
	movb (%ebx), %al
	movl _GlobalTp, %ebx
	movb %al, (%ebx)
	popl %eax
	movl %eax, %ecx
	addl %eax, %ebx
	movl %ebx, %edx
	decl %ebx
rolltloop:
	movb (%ebx), %al
	decl %ebx
	xchgl %ebx, %edx
	movb %al, (%ebx)
	decl %ebx
	xchgl %ebx, %edx
	loop rolltloop
	xor %eax, %eax
	ret
_L_depth:
	movl _GlobalSp, %ebx
	movl _BottomOfStack, %eax
	subl %ebx, %eax
	movl $WSIZE, (%ebx)
	movl $0, %edx
	idivl (%ebx)
	movl %eax, (%ebx)
	movl $WSIZE, %eax
	subl %eax, _GlobalSp
	movl _GlobalTp, %ebx
	movb $OP_IVAL, (%ebx)
	decl _GlobalTp
	xor %eax, %eax
        ret
L_2drop:
	FDROP
        NEXT
L_2dup:
	FDUP
        NEXT
L_2swap:
	FSWAP	
        NEXT
L_2over:
	FOVER
        NEXT
L_2rot:
	movl _GlobalSp, %ebx
	addl $WSIZE, %ebx
	movl %ebx, %ecx
	movl (%ebx), %edx
	addl $WSIZE, %ebx
	movl (%ebx), %eax
	addl $WSIZE, %ebx
	xchgl %edx, (%ebx)
	addl $WSIZE, %ebx
	xchgl %eax, (%ebx)
	addl $WSIZE, %ebx
	xchgl %edx, (%ebx)
	addl $WSIZE, %ebx
	xchgl %eax, (%ebx)
	movl %ecx, %ebx
	movl %edx, (%ebx)
	addl $WSIZE, %ebx
	movl %eax, (%ebx)
	movl _GlobalTp, %ebx
	incl %ebx
	movl %ebx, %ecx
	movw (%ebx), %ax
	addl $2, %ebx
	xchgw %ax, (%ebx)
	addl $2, %ebx
	xchgw %ax, (%ebx)
	movl %ecx, %ebx
	movw %ax, (%ebx)
	xor %eax, %eax
        NEXT
L_question:
	call L_fetch
	cmpl $0, %eax
	jnz questionexit
	call _CPP_dot
questionexit:	
	ret	
L_fetch:
	movl _GlobalSp, %edx
	movl _GlobalTp, %ebx
	incl %ebx
        movb (%ebx), %al
        cmpb $OP_ADDR, %al
        jnz fetcherror
        movb $OP_IVAL, (%ebx)	
	addl $WSIZE, %edx
        movl (%edx), %ebx	
        movl (%ebx), %eax
	movl %eax, (%edx)
	xor %eax, %eax
	ret
fetcherror:
        movl $E_NOT_ADDR, %eax
        ret
L_store:
        movl _GlobalTp, %ebx
	incl %ebx
        movb (%ebx), %al
        cmpb $OP_ADDR, %al
        jnz fetcherror
	movl $WSIZE, %eax
	movl _GlobalSp, %ebx
        addl %eax, %ebx
        movl (%ebx), %ecx	# address to store to in ecx
	addl %eax, %ebx
	movl (%ebx), %edx	# value to store in edx
	movl %ebx, _GlobalSp
	movl %edx, (%ecx)
	movl _GlobalTp, %ebx
	addl $2, %ebx
	movl %ebx, _GlobalTp
	xor %eax, %eax
	NEXT
L_afetch:
	movl _GlobalSp, %edx
	movl _GlobalTp, %ebx
	incl %ebx
        movb (%ebx), %al
        cmpb $OP_ADDR, %al
        jnz fetcherror
        movb $OP_ADDR, (%ebx)
	addl $WSIZE, %edx
	movl (%edx), %ebx
	movl (%ebx), %eax
	movl %eax, (%edx)
	xor %eax, %eax
	NEXT
L_cfetch:
	movl _GlobalTp, %ebx
	incl %ebx
	movb (%ebx), %al
	cmpb $OP_ADDR, %al
	jnz fetcherror
	movb $OP_IVAL, (%ebx)
	xor %eax, %eax
	movl _GlobalSp, %ebx
	addl $WSIZE, %ebx
	movl (%ebx), %ecx
	movb (%ecx), %al
	movl %eax, (%ebx)
	xor %eax, %eax
        NEXT
L_cstore:
	movl _GlobalTp, %edx
	incl %edx
	movb (%edx), %al
	cmpb $OP_ADDR, %al
	jnz fetcherror
	movl _GlobalSp, %ebx
	addl $WSIZE, %ebx
	movl (%ebx), %ecx	# address to store
	addl $WSIZE, %ebx
	movl (%ebx), %eax	# value to store
	movb %al, (%ecx)
	movl %ebx, _GlobalSp
	incl %edx
	movl %edx, _GlobalTp
	xor %eax, %eax
	NEXT	
L_wfetch:
	movl _GlobalTp, %ebx
	movb 1(%ebx), %al
	cmpb $OP_ADDR, %al
	jnz fetcherror
	movb $OP_IVAL, 1(%ebx)
	movl _GlobalSp, %ebx
	movl WSIZE(%ebx), %ebx
	movw (%ebx), %ax
	cwde
	movl _GlobalSp, %ebx
	movl %eax, WSIZE(%ebx)
	xor %eax, %eax
        NEXT
L_wstore:
	movl $WSIZE, %eax
	addl %eax, _GlobalSp
	incl _GlobalTp
	movl _GlobalTp, %ebx
	movb (%ebx), %al
	cmpb $OP_ADDR, %al
	jnz fetcherror
	movl _GlobalSp, %ebx
	movl (%ebx), %eax
	pushl %eax
	addl $WSIZE, %ebx
	movl (%ebx), %eax
	pop %ebx
	movw %ax, (%ebx)
	movl $WSIZE, %eax
	addl %eax, _GlobalSp
	incl _GlobalTp
	xor %eax, %eax
        NEXT
L_sffetch:
	movl $WSIZE, %eax
        addl %eax, _GlobalSp
        incl _GlobalTp
        movl _GlobalTp, %ebx
        movb (%ebx), %al
        cmpb $OP_ADDR, %al
        jnz fetcherror
        movb $OP_IVAL, (%ebx)
        decl %ebx
        movb $OP_IVAL, (%ebx)
        decl _GlobalTp
	decl _GlobalTp
        movl _GlobalSp, %ebx
        movl (%ebx), %ebx
        flds (%ebx)
	movl $WSIZE, %eax
        subl %eax, _GlobalSp
        movl _GlobalSp, %ebx
        fstpl (%ebx)
        subl %eax, _GlobalSp
	xor %eax, %eax
        NEXT
L_sfstore:
	movl $WSIZE, %eax
        addl %eax, _GlobalSp
        incl _GlobalTp
        movl _GlobalTp, %ebx
        movb (%ebx), %al
        cmpb $OP_ADDR, %al
        jnz fetcherror
        movl _GlobalSp, %ebx
        addl $WSIZE, %ebx
        fldl (%ebx)              # load the f number into NDP
        subl $WSIZE, %ebx
        movl (%ebx), %ebx          # load the dest address
        fstps (%ebx)             # store as single precision float
	movl $WSIZE, %eax
	sall $1, %eax
        addl %eax, _GlobalSp
	incl _GlobalTp
	incl _GlobalTp
	xor %eax, %eax
        NEXT
L_dffetch:	
        movl _GlobalTp, %ebx
	incl %ebx
        movb (%ebx), %al
        cmpb $OP_ADDR, %al
        jnz fetcherror
	movb $OP_IVAL, (%ebx)
	decl %ebx
	movb $OP_IVAL, (%ebx)
	decl %ebx
	movl %ebx, _GlobalTp 
	movl _GlobalSp, %ebx
	movl %ebx, %edx
	addl $WSIZE, %ebx
	movl (%ebx), %ecx
	movl (%ecx), %eax
	movl %eax, (%edx)
	addl $WSIZE, %ecx
	movl (%ecx), %eax
	movl %eax, (%ebx)
	subl $WSIZE, %edx
	movl %edx, _GlobalSp
	xor %eax, %eax
	NEXT
L_dfstore:
        movl _GlobalTp, %ebx
	incl %ebx
        movb (%ebx), %al
        cmpb $OP_ADDR, %al
        jnz fetcherror
	addl $2, %ebx
	movl %ebx, _GlobalTp
	movl _GlobalSp, %ebx
	movl $WSIZE, %edx
	addl %edx, %ebx
	movl %ebx, %eax
	movl (%ebx), %ebx  # address to store
	addl %edx, %eax
	movl (%eax), %ecx
	movl %ecx, (%ebx)
	addl %edx, %eax
	addl %edx, %ebx
	movl (%eax), %ecx
	movl %ecx, (%ebx)
	movl %eax, _GlobalSp
	xor %eax, %eax
	NEXT
L_inc:
        movl _GlobalSp, %ebx
        incl WSIZE(%ebx)
        NEXT
L_dec:
        movl _GlobalSp, %ebx
        decl WSIZE(%ebx)
        NEXT
L_twoplus:
	movl _GlobalSp, %ebx
	incl WSIZE(%ebx)
	incl WSIZE(%ebx)
	NEXT
L_twominus:
	movl _GlobalSp, %ebx
	decl WSIZE(%ebx)
	decl WSIZE(%ebx)
	NEXT
L_abs:
	_ABS
        NEXT
L_neg:
	movl _GlobalSp, %ebx
	addl $WSIZE, %ebx
	movl (%ebx), %eax
	negl %eax
	movl %eax, (%ebx)
	xor %eax, %eax
        NEXT
L_max:
	movl $WSIZE, %eax
	addl %eax, _GlobalSp
	movl _GlobalSp, %ebx
	movl (%ebx), %eax
	movl WSIZE(%ebx), %ebx
	cmpl %eax, %ebx
	jl max1
	movl %ebx, %eax
	movl _GlobalSp, %ebx
	movl %eax, WSIZE(%ebx)
	jmp maxexit
max1:
	movl _GlobalSp, %ebx
	movl %eax, WSIZE(%ebx)
maxexit:
	incl _GlobalTp
	xor %eax, %eax
        ret
L_min:
	movl $WSIZE, %eax
	addl %eax, _GlobalSp
	movl _GlobalSp, %ebx
	movl (%ebx), %eax
	movl WSIZE(%ebx), %ebx
	cmpl %eax, %ebx
	jg min1
	movl %ebx, %eax
	movl _GlobalSp, %ebx
	movl %eax, WSIZE(%ebx)
	jmp minexit
min1:
	movl _GlobalSp, %ebx
	movl %eax, WSIZE(%ebx)
minexit:
	incl _GlobalTp
	xor %eax, %eax
        ret
L_dmax:
	FOVER
	FOVER
	DLT
	incl _GlobalTp
	movl _GlobalSp, %ebx
	addl $WSIZE, %ebx
	movl (%ebx), %eax
	movl %ebx, _GlobalSp
	cmpl $0, %eax
	jne dmin1
	FDROP
	xorl %eax, %eax
	NEXT
L_dmin:
	FOVER
	FOVER
	DLT
	incl _GlobalTp
	movl $WSIZE, %ecx
	movl _GlobalSp, %ebx
	addl %ecx, %ebx
	movl (%ebx), %eax
	movl %ebx, _GlobalSp
	cmpl $0, %eax
	je dmin1
	FDROP
	xorl %eax, %eax
	NEXT
dmin1:
	FSWAP
	FDROP
	xorl %eax, %eax
	NEXT
L_twostar:
	movl _GlobalSp, %ebx
	sall $1, WSIZE(%ebx)
	NEXT
L_twodiv:
	movl _GlobalSp, %ebx
	sarl $1, WSIZE(%ebx)
	NEXT
L_add:
	movl $WSIZE, %eax
	movl _GlobalSp, %ebx
	addl %eax, %ebx
	movl (%ebx), %eax
	addl %eax, WSIZE(%ebx)
	movl %ebx, _GlobalSp
	movl _GlobalTp, %ebx
	incl %ebx
	movl %ebx, _GlobalTp
	movw (%ebx), %ax
	andb %ah, %al		# and the two types to preserve addr type
	incl %ebx
	movb %al, (%ebx)
        xor %eax, %eax
        NEXT
L_sub:
	movl $WSIZE, %eax
	movl _GlobalSp, %ebx
	addl %eax, %ebx
	movl (%ebx), %eax
	subl %eax, WSIZE(%ebx)
	movl %ebx, _GlobalSp
        incl _GlobalTp		# result will have type of first operand
        xor %eax, %eax
        NEXT
L_mul:
	movl $WSIZE, %ecx
	movl _GlobalSp, %ebx
	addl %ecx, %ebx
	movl %ebx, _GlobalSp
	movl (%ebx), %eax
	addl %ecx, %ebx
	imull (%ebx)
	movl %eax, (%ebx)
	incl _GlobalTp
	xor %eax, %eax
        NEXT
L_div:
	movl $WSIZE, %eax
        addl %eax, _GlobalSp
        incl _GlobalTp
        movl _GlobalSp, %ebx
        movl (%ebx), %eax
        cmpl $0, %eax
        jnz div1
        movl $E_DIV_ZERO, %eax
        jmp divexit
div1:	
	addl $WSIZE, %ebx
        movl (%ebx), %eax
	cdq
        idivl -WSIZE(%ebx)
        movl %eax, (%ebx)
	xor %eax, %eax
divexit:
        ret
L_mod:
	call L_div
	cmpl $0, %eax
	jnz  divexit
	movl %edx, (%ebx)
	NEXT
L_slashmod:
	call L_div
	cmpl $0, %eax
	jnz  divexit
	subl $WSIZE, %ebx
	movl %edx, (%ebx)
	subl $WSIZE, %ebx
	movl %ebx, _GlobalSp
	decl _GlobalTp
	SWAP
	NEXT
L_starslash:
	STARSLASH	
	NEXT
L_starslashmod:
	STARSLASH
	movl %edx, (%ebx)
	subl $WSIZE, %ebx
	movl %ebx, _GlobalSp
	decl _GlobalTp
	SWAP
	ret
L_plusstore:
	movl _GlobalTp, %ebx
	movb 1(%ebx), %al
	cmpb $OP_ADDR, %al
	jnz fetcherror
	movl _GlobalSp, %ebx
	push %ebx
	push %ebx
	push %ebx
	movl WSIZE(%ebx), %ebx
	movl (%ebx), %eax
	pop %ebx
	movl 2*WSIZE(%ebx), %ebx
	addl %ebx, %eax
	pop %ebx
	movl WSIZE(%ebx), %ebx
	movl %eax, (%ebx)
	pop %ebx
	movl $WSIZE, %eax
	sall $1, %eax
	addl %eax, %ebx
	movl %ebx, _GlobalSp
	incl _GlobalTp
	incl _GlobalTp
	xor %eax, %eax
	NEXT
_L_dabs:
	movl _GlobalSp, %ebx
	addl $WSIZE, %ebx
	movl (%ebx), %ecx
	movl %ecx, %eax
	cmpl $0, %eax
	jl dabs_go
	xor %eax, %eax
	ret
dabs_go:
	addl $WSIZE, %ebx
	movl (%ebx), %eax
	clc
	subl $1, %eax
	notl %eax
	movl %eax, (%ebx)
	movl %ecx, %eax
	sbbl $0, %eax
	notl %eax
	movl %eax, -WSIZE(%ebx)
	xor %eax, %eax
	ret
L_dnegate:
	DNEGATE
	NEXT	
L_dplus:
	DPLUS
	NEXT
_L_dminus:
	DMINUS
	ret
L_umstar:
	movl $WSIZE, %eax
	movl _GlobalSp, %ebx
	addl %eax, %ebx
	movl (%ebx), %ecx
	addl %eax, %ebx
	movl %ecx, %eax
	mull (%ebx)
	movl %eax, (%ebx)
	subl $WSIZE, %ebx
	movl %edx, (%ebx)
	xor %eax, %eax				
	ret
L_dsstar:
	# multiply signed double and signed to give triple length product
	movl $WSIZE, %ecx
	movl _GlobalSp, %ebx
	addl %ecx, %ebx
	movl (%ebx), %edx
	cmpl $0, %edx
	setl %al
	addl %ecx, %ebx
	movl (%ebx), %edx
	cmpl $0, %edx
	setl %ah
	xorb %ah, %al      # sign of result
	andl $1, %eax
	pushl %eax
	_ABS
	movl _GlobalSp, %ebx
	addl $WSIZE, %ebx
	movl %ebx, _GlobalSp
	incl _GlobalTp
	call _L_dabs
	movl _GlobalSp, %ebx
	subl $WSIZE, %ebx
	movl %ebx, _GlobalSp
	decl _GlobalTp
	call _L_udmstar
	popl %eax
	cmpl $0, %eax
	jne dsstar1
	NEXT
dsstar1:
	TNEG
	NEXT
L_umslashmod:
# Divide unsigned double length by unsigned single length to
# give unsigned single quotient and remainder. A "Divide overflow"
# error results if the quotient doesn't fit into a single word.
	movl _GlobalSp, %ebx
	movl $WSIZE, %eax
	addl %eax, %ebx
	movl %ebx, _GlobalSp
	movl (%ebx), %ecx
	cmpl $0, %ecx
	jnz  umslashmod0
	movl $E_DIV_ZERO, %eax
	ret
umslashmod0:	
	addl %eax, %ebx
	movl $0, %edx
	movl (%ebx), %eax
	divl %ecx
	cmpl $0, %eax
	jne umslashmod_ovflow
	movl (%ebx), %edx
	addl $WSIZE, %ebx
	movl (%ebx), %eax
	divl %ecx
	jmp umslashmod_exit
umslashmod_ovflow:
	movl $E_DIV_OVERFLOW, %eax
	ret
umslashmod_exit:	
	movl %edx, (%ebx)
	subl $WSIZE, %ebx
	movl %eax, (%ebx)
	incl _GlobalTp
	xor %eax, %eax		
	ret
L_mstar:
	movl $WSIZE, %eax
	movl _GlobalSp, %ebx
	addl %eax, %ebx
	movl (%ebx), %ecx
	addl %eax, %ebx
	movl %ecx, %eax
	imull (%ebx)
	movl %eax, (%ebx)
	subl $WSIZE, %ebx
	movl %edx, (%ebx)
	xor %eax, %eax		
	ret
L_mplus:
	STOD
	DPLUS
	NEXT
L_mslash:
	movl $WSIZE, %eax
        incl _GlobalTp
        movl _GlobalSp, %ebx
	addl %eax, %ebx
        movl (%ebx), %ecx
	incl _GlobalTp
	addl %eax, %ebx
	movl %ebx, _GlobalSp
        cmpl $0, %ecx
        je mslash1
        movl (%ebx), %edx
	addl %eax, %ebx
	movl (%ebx), %eax
        idivl %ecx
        movl %eax, (%ebx)
	xor %eax, %eax		
	ret
mslash1:			
        movl $E_DIV_ZERO, %eax
        ret
_L_udmstar:
# Multiply unsigned double and unsigned single to give 
# the triple length product.
	movl _GlobalSp, %ebx
	addl $WSIZE, %ebx
	movl (%ebx), %ecx
	addl $WSIZE, %ebx
	movl (%ebx), %eax
	mull %ecx
	movl %edx, -WSIZE(%ebx)
	movl %eax, (%ebx)
	addl $WSIZE, %ebx
	movl %ecx, %eax
	mull (%ebx)
	movl %eax, (%ebx)
	subl $WSIZE, %ebx
	movl (%ebx), %eax
	subl $WSIZE, %ebx
	clc
	addl %edx, %eax
	movl %eax, WSIZE(%ebx)
	movl (%ebx), %eax
	adcl $0, %eax
	movl %eax, (%ebx)
	xor %eax, %eax 		
	ret
L_utsslashmod:
# Divide unsigned triple length by unsigned single length to
# give an unsigned triple quotient and single remainder.
	movl _GlobalSp, %ebx
	addl $WSIZE, %ebx
	movl (%ebx), %ecx		# divisor in ecx
	cmpl $0, %ecx
	jnz  utsslashmod0
	movl $E_DIV_ZERO, %eax	
	ret
utsslashmod0:	
	addl $WSIZE, %ebx
	movl (%ebx), %eax		# ut3
	movl $0, %edx
	divl %ecx			# ut3/u
	call utmslash1
	movl _GlobalSp, %ebx
	movl WSIZE(%ebx), %eax
	movl %eax, (%ebx)
	addl $WSIZE, %ebx
	movl WSIZE(%ebx), %eax
	movl %eax, (%ebx)
	addl $WSIZE, %ebx	
	movl -17*WSIZE(%ebx), %eax       # r7
	movl %eax, (%ebx)
	subl $3*WSIZE, %ebx
	movl -5*WSIZE(%ebx), %eax        # q3
	movl %eax, (%ebx)
	subl $WSIZE, %ebx
	movl %ebx, _GlobalSp
	decl _GlobalTp
	decl _GlobalTp
	xorl %eax, %eax	
	ret
L_tabs:
# Triple length absolute value (needed by L_stsslashrem, STS/REM)
        movl _GlobalSp, %ebx
        addl $WSIZE, %ebx
        movl (%ebx), %ecx
        movl %ecx, %eax
        cmpl $0, %eax
        jl tabs1
        xor %eax, %eax
        ret
tabs1:
        addl $2*WSIZE, %ebx
        movl (%ebx), %eax
        clc
        subl $1, %eax
        notl %eax
        movl %eax, (%ebx)
	subl $WSIZE, %ebx
	movl (%ebx), %eax
	sbbl $0, %eax
	notl %eax
	movl %eax, (%ebx)
        movl %ecx, %eax
        sbbl $0, %eax
        notl %eax
        movl %eax, -WSIZE(%ebx)
        xor %eax, %eax
        ret
L_stsslashrem:
# Divide signed triple length by signed single length to give a
# signed triple quotient and single remainder, according to the
# rule for symmetric division.
	movl _GlobalSp, %ebx
	addl $WSIZE, %ebx
	movl (%ebx), %ecx		# divisor in ecx
	cmpl $0, %ecx
	jnz  stsslashrem0
	movl $E_DIV_ZERO, %eax	
stsslashrem0:
	movl WSIZE(%ebx), %eax		# t3
	pushl %eax
	cmpl $0, %eax
	movl $0, %eax
	setl %al
	negl %eax
	movl %eax, %edx
	cmpl $0, %ecx
	movl $0, %eax
	setl %al
	negl %eax
	xorl %eax, %edx			# sign of quotient
	pushl %edx
	movl %ebx, _GlobalSp
	call L_tabs
	subl $WSIZE, _GlobalSp
	_ABS
	call L_utsslashmod
	popl %edx
	cmpl $0, %edx
	jz stsslashrem1
	TNEG
stsslashrem1:	
	popl %eax
	cmpl $0, %eax
	jz stsslashrem2
	movl _GlobalSp, %ebx
	addl $4*WSIZE, %ebx
	negl (%ebx)	
stsslashrem2:
	xorl %eax, %eax
	ret
_L_utmslash:
# Divide unsigned triple length by unsigned single to give 
# unsigned double quotient. A "Divide Overflow" error results
# if the quotient doesn't fit into a double word.
	movl _GlobalSp, %ebx
	addl $WSIZE, %ebx
	movl (%ebx), %ecx		# divisor in ecx
	cmpl $0, %ecx
	jnz  utmslash0
	movl $E_DIV_ZERO, %eax
	ret
utmslash0:	
	addl $WSIZE, %ebx
	movl (%ebx), %eax		# ut3
	movl $0, %edx
	divl %ecx			# ut3/u
	cmpl $0, %eax
	jz   utmslash1
	movl $E_DIV_OVERFLOW, %eax
	ret
utmslash1:	
	pushl %ebx			# keep local stack ptr
	movl _GlobalSp, %ebx
	movl %eax, -4*WSIZE(%ebx)	# q3
	movl %edx, -5*WSIZE(%ebx)	# r3
	popl %ebx
	addl $WSIZE, %ebx
	movl (%ebx), %eax		# ut2
	movl $0, %edx
	divl %ecx			# ut2/u
	pushl %ebx
	movl _GlobalSp, %ebx
	movl %eax, -2*WSIZE(%ebx)	# q2
	movl %edx, -3*WSIZE(%ebx)	# r2
	popl %ebx
	addl $WSIZE, %ebx
	movl (%ebx), %eax		# ut1
	movl $0, %edx
	divl %ecx			# ut1/u
	pushl %ebx
	movl _GlobalSp, %ebx
	movl %eax, (%ebx)		# q1
	movl %edx, -WSIZE(%ebx)		# r1
	movl -5*WSIZE(%ebx), %edx	# r3 << 32
	movl $0, %eax
	divl %ecx			# (r3 << 32)/u
	movl %eax, -6*WSIZE(%ebx)	# q4
	movl %edx, -7*WSIZE(%ebx)	# r4
	movl -3*WSIZE(%ebx), %edx	# r2 << 32
	movl $0, %eax
	divl %ecx			# (r2 << 32)/u
	movl %eax, -8*WSIZE(%ebx)	# q5
	movl %edx, -9*WSIZE(%ebx)	# r5
	movl -7*WSIZE(%ebx), %edx	# r4 << 32
	movl $0, %eax
	divl %ecx			# (r4 << 32)/u
	movl %eax, -10*WSIZE(%ebx)	# q6
	movl %edx, -11*WSIZE(%ebx)	# r6
	movl $0, %edx
	movl -WSIZE(%ebx), %eax		# r1
	addl -9*WSIZE(%ebx), %eax	# r1 + r5
	jnc   utmslash2
	incl %edx
utmslash2:			
	addl -11*WSIZE(%ebx), %eax	# r1 + r5 + r6
	jnc  utmslash3
	incl %edx
utmslash3:
	divl %ecx
	movl %eax, -12*WSIZE(%ebx)	# q7
	movl %edx, -13*WSIZE(%ebx)	# r7
	movl $0, %edx
	addl -10*WSIZE(%ebx), %eax	# q7 + q6
	jnc  utmslash4
	incl %edx
utmslash4:	
	addl -8*WSIZE(%ebx), %eax	# q7 + q6 + q5
	jnc  utmslash5
	incl %edx
utmslash5:	
	addl (%ebx), %eax		# q7 + q6 + q5 + q1
	jnc utmslash6
	incl %edx
utmslash6:
	popl %ebx
	movl %eax, (%ebx)
	subl $WSIZE, %ebx
	pushl %ebx
	movl _GlobalSp, %ebx
	movl -2*WSIZE(%ebx), %eax	# q2
	addl -6*WSIZE(%ebx), %eax	# q2 + q4
	addl %edx, %eax
	popl %ebx
	movl %eax, (%ebx)
	subl $WSIZE, %ebx
	movl %ebx, _GlobalSp
	addl $2, _GlobalTp
	xorl %eax, %eax
	ret
L_mstarslash:
	movl _GlobalSp, %ebx
	addl $WSIZE, %ebx
	addl $WSIZE, %ebx
	movl (%ebx), %eax
	addl $WSIZE, %ebx
	xorl (%ebx), %eax
	shrl $31, %eax
	pushl %eax	# keep sign of result -- negative is nonzero
	movl _GlobalSp, %ebx
	addl $WSIZE, %ebx
	movl %ebx, _GlobalSp
	incl _GlobalTp
	_ABS
	movl _GlobalSp, %ebx
	addl $WSIZE, %ebx
	movl %ebx, _GlobalSp
	incl _GlobalTp
	call _L_dabs
	movl _GlobalSp, %ebx
	subl $WSIZE, %ebx
	movl %ebx, _GlobalSp
	decl _GlobalTp
	call _L_udmstar
	movl _GlobalSp, %ebx
	subl $WSIZE, %ebx
	movl %ebx, _GlobalSp
	decl _GlobalTp
	call _L_utmslash	
	popl %eax
	cmpl $0, %eax
	jnz mstarslash_neg
	xor %eax, %eax
	ret
mstarslash_neg:
	DNEGATE
	xor %eax, %eax
	ret		
L_fmslashmod:
	movl _GlobalSp, %ebx
	movl $WSIZE, %eax
	addl %eax, %ebx
	movl %ebx, _GlobalSp
	movl (%ebx), %ecx
	cmpl $0, %ecx
	jnz  fmslashmod0
	movl $E_DIV_ZERO, %eax
	ret
fmslashmod0:
	addl %eax, %ebx
	movl (%ebx), %edx
	addl %eax, %ebx
	movl (%ebx), %eax
	idivl %ecx
	movl %edx, (%ebx)
	subl $WSIZE, %ebx
	movl %eax, (%ebx)
	incl _GlobalTp
	cmpl $0, %ecx
	jg fmslashmod2
	cmpl $0, %edx
	jg fmslashmod3
	xor %eax, %eax
	ret
fmslashmod2:		
	cmpl $0, %edx
	jge fmslashmodexit
fmslashmod3:	
	decl %eax		# floor the result
	movl %eax, (%ebx)
	addl $WSIZE, %ebx
	addl %ecx, (%ebx)
fmslashmodexit:
	xor %eax, %eax
	ret
L_smslashrem:
	movl _GlobalSp, %ebx
	movl $WSIZE, %eax
	addl %eax, %ebx
	movl %ebx, _GlobalSp
	movl (%ebx), %ecx
	cmpl $0, %ecx
	jnz smslashrem0
	movl $E_DIV_ZERO, %eax
	ret
smslashrem0:	
	addl %eax, %ebx
	movl (%ebx), %edx
	addl %eax, %ebx
	movl (%ebx), %eax
	idivl %ecx
	movl %edx, (%ebx)
	subl $WSIZE, %ebx
	movl %eax, (%ebx)
	incl _GlobalTp
	xor %eax, %eax		
	ret
L_stod:
	STOD
	NEXT
L_stof:
	movl $WSIZE, %eax
        addl %eax, _GlobalSp
        incl _GlobalTp
        movl _GlobalSp, %ebx
        fildl (%ebx)
        movl _GlobalTp, %ebx
        movb $OP_IVAL, (%ebx)
        decl %ebx
        movb $OP_IVAL, (%ebx)
	decl _GlobalTp
	decl _GlobalTp
        movl _GlobalSp, %ebx
	movl $WSIZE, %eax
        subl %eax, %ebx
        fstpl (%ebx)
	sall $1, %eax
        subl %eax, _GlobalSp
	xor %eax, %eax
        NEXT
L_dtof:
	movl $WSIZE, %eax
        movl _GlobalSp, %ebx
	addl %eax, %ebx
	movl (%ebx), %eax
	xchgl WSIZE(%ebx), %eax
	movl %eax, (%ebx)
        fildq (%ebx)
        fstpl (%ebx)
	xor %eax, %eax	
	NEXT	
L_froundtos:
	movl $WSIZE, %eax
        addl %eax, _GlobalSp
        movl _GlobalSp, %ebx
        fldl (%ebx)
        addl %eax, %ebx
        fistpl (%ebx)
        incl _GlobalTp
        movl _GlobalTp, %ebx
        incl %ebx
        movb $OP_IVAL, (%ebx)
	xor %eax, %eax
        NEXT
L_ftrunctos:
	movl $WSIZE, %eax
	addl %eax, _GlobalSp
	movl _GlobalSp, %ebx
	fldl (%ebx)
	fnstcw (%ebx)
	movl (%ebx), %ecx	# save NDP control word		
	movl %ecx, %edx	
	movb $12, %dh
	movl %edx, (%ebx)
	fldcw (%ebx)
	addl %eax, %ebx	
	fistpl (%ebx)
	subl %eax, %ebx
	movl %ecx, (%ebx)
	fldcw (%ebx)		# restore NDP control word
	incl _GlobalTp
	movl _GlobalTp, %ebx
	incl %ebx
	movb $OP_IVAL, (%ebx)
	xor %eax, %eax	
	NEXT	
L_ftod:
	movl $WSIZE, %eax
	movl _GlobalSp, %ebx
	addl %eax, %ebx
	fldl (%ebx)
	subl %eax, %ebx
	fnstcw (%ebx)
	movl (%ebx), %ecx	# save NDP control word	
	movl %ecx, %edx
	movb $12, %dh		
	movl %edx, (%ebx)
	fldcw (%ebx)
	addl %eax, %ebx	
	fistpq (%ebx)
	subl %eax, %ebx
	movl %ecx, (%ebx)
	fldcw (%ebx)		# restore NDP control word
	addl %eax, %ebx 
	movl (%ebx), %eax
	xchgl WSIZE(%ebx), %eax
	movl %eax, (%ebx)
	xor %eax, %eax	
	NEXT	
L_degtorad:
	fldl FCONST_180
        movl _GlobalSp, %ebx
        addl $WSIZE, %ebx
        fldl (%ebx)
        fdivp %st, %st(1)
        fldpi
        fmulp %st, %st(1)
        fstpl (%ebx)
        NEXT
L_radtodeg:
        movl _GlobalSp, %ebx
        addl $WSIZE, %ebx
        fldl (%ebx)
        fldpi
	fxch
        fdivp %st, %st(1)
        fldl FCONST_180
        fmulp %st, %st(1)
        fstpl (%ebx)
        NEXT
L_fne:
	FREL_DYADIC xorb $64 setnz
	NEXT
L_feq:
	FREL_DYADIC andb $64 setnz
	NEXT
L_flt:
	FREL_DYADIC andb $65 setz
	NEXT
L_fgt:
	FREL_DYADIC andb $1 setnz
	NEXT	
L_fle:
	FREL_DYADIC xorb $1 setnz
	NEXT
L_fge:
	FREL_DYADIC andb $65 setnz
	NEXT
L_fzeroeq:
	movl $WSIZE, %eax
	movl _GlobalSp, %ebx
	addl %eax, %ebx
	movl (%ebx), %ecx
	movl %ebx, _GlobalSp
	addl %eax, %ebx
	movl (%ebx), %eax
	shll $1, %eax
	orl %ecx, %eax
	movl $0, %eax
	setz %al
	negl %eax
	movl %eax, (%ebx)
frelzero:
	movl _GlobalTp, %ebx
	incl %ebx
	movl %ebx, _GlobalTp
	incl %ebx
	movb $OP_IVAL, (%ebx)
	xorl %eax, %eax
	NEXT
L_fzerolt:
	movl $WSIZE, %eax
	movl _GlobalSp, %ebx
	addl %eax, %ebx
	movl %ebx, _GlobalSp
	fldl (%ebx)
	addl %eax, %ebx
	fldz
	fcompp	
	fnstsw %ax
	andb $69, %ah
	movl $0, %eax
	setz %al
	negl %eax
	movl %eax, (%ebx)
	jmp frelzero
L_fzerogt:
	movl $WSIZE, %eax
	movl _GlobalSp, %ebx
	addl %eax, %ebx
	movl %ebx, _GlobalSp
	fldz
	fldl (%ebx)
	addl %eax, %ebx
	fucompp	
	fnstsw %ax
	sahf 
	movl $0, %eax
	seta %al
	negl %eax
	movl %eax, (%ebx)
	jmp frelzero				
L_fadd:
	movl $WSIZE, %eax
        addl %eax, _GlobalSp
        movl _GlobalSp, %ebx
        fldl (%ebx)
	sall $1, %eax
        addl %eax, %ebx
        faddl (%ebx)
        fstpl (%ebx)
	movl $WSIZE, %eax
        addl %eax, _GlobalSp
        incl _GlobalTp
	incl _GlobalTp
	xor %eax, %eax
        NEXT
L_fsub:
	movl $WSIZE, %eax
	sall $1, %eax
	addl $WSIZE, %eax
        addl %eax, _GlobalSp
        movl _GlobalSp, %ebx
        fldl (%ebx)
	subl $WSIZE, %eax
        subl %eax, %ebx
        fsubl (%ebx)
        addl %eax, %ebx
        fstpl (%ebx)
	movl $WSIZE, %eax
        subl %eax, _GlobalSp
        incl _GlobalTp
	incl _GlobalTp
	xor %eax, %eax
        NEXT
L_fmul:
	movl $WSIZE, %eax
        addl %eax, _GlobalSp
        movl _GlobalSp, %ebx
        fldl (%ebx)
	sall $1, %eax
        addl %eax, %ebx
        fmull (%ebx)
        fstpl (%ebx)
	sarl $1, %eax
        addl %eax, _GlobalSp
        incl _GlobalTp
	incl _GlobalTp
	xor %eax, %eax
        NEXT
L_fdiv:
	movl $WSIZE, %eax
        addl %eax, _GlobalSp
        movl _GlobalSp, %ebx
        fldl (%ebx)
        ftst
        jnz fdiv1
        subl %eax, _GlobalSp
        movl $E_DIV_ZERO, %eax
        jmp fdivexit
fdiv1:  addl $WSIZE, %ebx
	addl $WSIZE, %ebx
        fdivrl (%ebx)
        fstpl (%ebx)
        addl %eax, _GlobalSp
        incl _GlobalTp
	incl _GlobalTp
fdivexit:
	xor %eax, %eax
	ret
L_fabs:
        movl _GlobalSp, %ebx
        addl $WSIZE, %ebx
        fldl (%ebx)
        fabs
        fstpl (%ebx)
        NEXT
L_fneg:
        movl _GlobalSp, %ebx
        addl $WSIZE, %ebx
        fldl (%ebx)
        fchs
        fstpl (%ebx)
        NEXT
L_floor:
	movl _GlobalSp, %ebx
	addl $WSIZE, %ebx
	movl WSIZE(%ebx), %eax
	pushl %eax
	movl (%ebx), %eax
	pushl %eax
	call _floor
	addl $8, %esp
	fstpl (%ebx)
	xor %eax, %eax		
	NEXT
L_fround:
	movl _GlobalSp, %ebx
	addl $WSIZE, %ebx
	fldl (%ebx)
	frndint
	fstpl (%ebx)
	NEXT
L_ftrunc:
	movl _GlobalSp, %ebx
	addl $WSIZE, %ebx
	fldl (%ebx)
	fnstcw (%ebx)
	movl (%ebx), %ecx	# save NDP control word
	movl %ecx, %edx
	movb $12, %dh
	movl %edx, (%ebx)
	fldcw (%ebx)
	frndint
	movl %ecx, (%ebx)
	fldcw (%ebx)		# restore NDP control word	
	fstpl (%ebx)		
	xor %eax, %eax
	NEXT
L_fsqrt:
	movl _GlobalSp, %ebx
	fldl WSIZE(%ebx)
	fsqrt
	fstpl WSIZE(%ebx)
	NEXT
L_fcos:
	movl _GlobalSp, %ebx
	fldl WSIZE(%ebx)
	fcos
	fstpl WSIZE(%ebx)
	NEXT
L_fsin:
	movl _GlobalSp, %ebx
	fldl WSIZE(%ebx)
	fsin
	fstpl WSIZE(%ebx)
	NEXT
L_fatan2:
	movl _GlobalSp, %ebx
	addl $2*WSIZE, %ebx
	fldl WSIZE(%ebx)
	fldl -WSIZE(%ebx)
	fpatan
	fstpl WSIZE(%ebx)
	movl %ebx, _GlobalSp
	incl _GlobalTp
	incl _GlobalTp
	NEXT
L_fsincos:
	movl _GlobalSp, %ebx
	fldl WSIZE(%ebx)
	fsincos
	fstpl -WSIZE(%ebx)
	fstpl WSIZE(%ebx)
	subl $2*WSIZE, %ebx
	movl %ebx, _GlobalSp
	movl _GlobalTp, %ebx
	movb $OP_IVAL, (%ebx)
	decl %ebx
	movb $OP_IVAL, (%ebx)
	decl %ebx
	movl %ebx, _GlobalTp	
	NEXT
#

	.comm _GlobalSp,4
	.comm _GlobalTp,4
	.comm _GlobalIp,4
	.comm _GlobalRp,4
	.comm _GlobalRtp,4
	.comm _BottomOfStack,4
	.comm _BottomOfReturnStack,4
	.comm _BottomOfTypeStack,4
	.comm _BottomOfReturnTypeStack,4
	.comm _vmEntryRp,4
	.comm _Base,4
	.comm _State,4
	.comm _pTIB,4
	.comm _TIB,256
	.comm _WordBuf,256
	.comm _NumberCount,4
	.comm _NumberBuf,256

	
