// vm-fast.s
//
// The kForth Virtual Machine (fast version)
//
// Copyright (c) 1998--2009 Krishna Myneni, Creative Consulting for
//   Research and Education
//
// This software is provided under the terms of the General Public License.
//
// Usage from C++
//
//       extern "C" int vm (byte* ip);
//       ecode = vm(ip);
//
// Originally written for the A386 assembler
// Revisions:
//	3-21-98
//	8-25-1998 ported to GNU assembler under Linux
//	9-8-1998  additional functions added: c@, c!, f=, f<>
//	9-10-1998 added: and, or, not, xor
//	9-11-1998 added: = , <>, <, >, <=, >=
//	9-15-1998 added: +!, fsqrt, fsin, fcos, -rot, sp@, rp@
//	9-17-1998 added: a@
//	9-27-1998 added: mod, 2*, 2/
//	10-1-1998 added: pick
//	10-4-1998 fixed signed integer division and mod; added /MOD
//	10-6-1998 added ?dup
//	10-14-1998 added count
//	10-16-1998 fixed L_div error
//	10-19-1998 added 0<, 0=, 0>, true, false
//	10-20-1998 added 2+, 2-
//	02-09-1999 added execute
//	03-01-1999 added open, lseek, close, read, write
//	03-02-1999 added ioctl
//	03-03-1999 added usleep	
//	03-07-1999 added fill, cmove
//	03-09-1999 interchanged meaning of execute and call
//		   to be consistent with ANS Forth
//	03-27-1999 added +loop, unloop
//	03-29-1999 added roll
//	03-31-1999 added cmove>, key
//	05-05-1999 fixed +loop
//	05-06-1999 added fround
//	05-15-1999 added floor
//	05-26-1999 added fatan2, lshift, rshift
//	05-27-1999 added u<, quit, base
//	06-02-1999 added */, */mod
//	06-09-1999 call CPP functions from vm
//	07-18-1999 added find
//	09-06-1999 added pTIB, word, tick
//	09-12-1999 added system
//	10-04-1999 added create, variable, fvariable as intrinsic words
//	10-06-1999 added constant, fconstant as intrinsic words
//	10-07-1999 added chdir
//      10-08-1999 added erase, brackettick
//	10-09-1999 added time&date, ms, question, bl
//      10-10-1999 added char, forget, cold
//      10-20-1999 added >file, console
//	10-28-1999 added key?
//	12-24-1999 added hooks for f0=, f0<, f0>, u>, s>d, d>f, f>d, 
//                   um*, um/mod, m*, m+, m/, m*/
//	12-25-1999 added cells, cell+, dfloats, dfloat+
//	12-27-1999 added bye
//	01-08-2000 fixed f0<, increased vm loop efficiency
//	01-13-1999 added ?allot, fixed f0=
//	01-22-2000 modified + to remove ordering sensitivity for address arithmetic
//	01-23-2000 added 0<>, [char], .r, u.r, changed opcodes for 
//	             relational operators
//	01-24-2000 added CPP_literal, CPP_cquote, CPP_squote, CPP_dotquote
//	02-04-2000 implemented m*, um*
//	02-26-2000 fixed fm/mod, added sm/rem, um/mod
//	03-02-2000 modified QUIT to clear the return stacks, added CPP_do
//	03-05-2000 added CPP_begin, CPP_while, CPP_repeat, CPP_until, CPP_again,
//	             CPP_leave, CPP_if, CPP_else, CPP_then, CPP_lparen
//	05-17-2000 added CPP_does
//	05-18-2000 fix L_plusloop for negative increments
//	06-04-2000 fix L_roll to roll the typestack as well
//	06-11-2000 added CPP_case, CPP_endcase, CPP_of, and CPP_endof
//	06-15-2000 added CPP_querydo, CPP_abortquote
//	09-05-2000 handled overflow condition on um/mod; added code for m/;
//		     added m+	;  hooks for CPP_lbracket, CPP_rbracket, CPP_ddot
//	11-30-2000 added L_dabs, L_dnegate, L_udmstar, L_mstarslash
//	01-26-2001 changed jumptable address for usleep to C_usec
//	04-01-2001 changed jmpl to jmp for L_fzerolt
//	04-22-2001 added L_dplus, L_dminus
//	04-24-2001 added L_twopush, L_twopop, L_tworfetch
//      05-07-2001 added L_definition
//	05-08-2001 improved vm execution efficiency by about 10%
//	05-13-2001 added CPP_dotparen, L_dlt, L_dzeroeq, L_deq
//	05-20-2001 added CPP_sharp, CPP_bracketsharp, CPP_sharpbracket, 
//	             CPP_sharps, CPP_sign, CPP_hold; fixed problem with L_dnegate
//	07-12-2001 fixed type-stack alignment problem with L_utmslash
//	08-26-2001 updated L_fval to avoid routing fp through NDP, changed all
//	             typestack settings of OP_FVAL to OP_IVAL
//	09-02-2001 added L_tobody
//      12-08-2001 added CPP_evaluate
//      07-29-2002 fixed L_ret to set GlobalRp and GlobalRtp to bottom upon
//                   return from vm.
//	07-31-2002 fixed L_cmovefrom for zero count argument; added
//		     CPP_backslash; fixed bug in L_fmslashmod for divisor < 0
//	08-01-2002 added C_msfetch, C_search, C_compare
//      09-11-2002 modified vm and L_ret to make vm re-entrant
//	09-29-2002 restored L_usleep since it provides a different function 
//	             than L_ms, added CPP_immediate
//	04-11-2003 renamed L_ftos to L_froundtos
//	04-15-2003 added L_ftrunc and L_ftrunctos
//	04-18-2003 fixed L_ftrunctos typestack pointer update
//	06-15-2003 recoded L_rot, L_swap, and L_over for increased efficiency
//	01-01-2004 modified threading mode to a hybrid of indirect jump 
//                   threading and indirect call threading ;  results in
//		     significant speedup of kForth (about 1.7--2x)
//	01-31-2004 extended JumpTable past 256 entries,
//                   added CPP_include, CPP_source, CPP_refill, and
//                   CPP_nondeferred, CPP_state, CPP_colon, CPP_semicolon
//	02-09-2004 added CPP_allocate and CPP_free
//	03-18-2004 added L_fsincos, C_fsinh, C_fcosh, C_ftanh, C_fasinh,
//                   C_facosh, C_fatanh
//      03-21-2004 implemented L_fsincos
//      03-30-2004 modified L_ret to remove unneeded instructions
//      04-08-2004 added CPP_postpone and CPP_compilecomma
//      04-09-2004 added CPP_semicolon
//	04-26-2004 exported L_udmstar and L_utmslash for use by ForthVM.cpp
//      06-20-2004 modified L_call to prevent loss of return error code
//      09-05-2004 added C_forth_signal, C_raise, C_setitimer, C_getitimer
//      09-10-2004 moved CPP_spstore and CPP_rpstore to opcodes below 256
//      09-28-2005 changed some comment markers to // to prevent GNU as from quirky errors
//	02-10-2006 The EBP register is now used to hold GlobalIp for the jump-threaded portion
//                   of the VM. This results in 10--20% speedup for typical apps.
//                   Any additions to code in vm.s which needs to make use of the EBP
//	             register should save and restore this register using "pushl %ebp" at 
//	             the beginning and "popl %ebp" before returning (ret) or using NEXT.
//	             Also, calls from vm.s to external words (C/C++) that reference
//		     GlobalIp should update GlobalIp prior to the call:	"movl %ebp, GlobalIp"
//
//      04-11-2006 added stubs for L_dult, L_utsslashmod, L_stsslashrem, L_udmstar, L_cputest,
//                   L_dsstar, L_dmax, and L_dmin
//
//	04-19-2006 fixed bug in L_mslash, and added C_syscall
//
//	04-23-2006 recoded relational operators L_eq, L_neq, L_ult, L_ugt, L_lt, L_gt, L_le, L_ge,
//                   L_zeroeq, L_zerone, L_zerolt, L_zerogt. Added macros REL_DYADIC and REL_ZERO.
//                   Added macro LOGIC_DYADIC and used it for recoding _AND, _OR, and _XOR macros.
//
//      04-24-2006 recoded L_dequal, L_dlt, and L_dzeroeq to remove conditional 
//	           branching.
//
//      04-27-2006 made macro STOD and fixed L_mplus by using STOD and DPLUS macros; 
//                 coded L_dult, recoded L_dlt to fix problem caused by using DMINUS; 
//                 simplified _NOT macro and used it in L_fne; recoded L_feq and 
//                 L_flt to remove conditional branches.
// 
//	04-29-2006 coded L_dmin and L_dmax; added following macros: 
//                  DLT FDUP FSWAP FDROP
//
//	05-30-2006 rename L_cmove to L_move, and coded correct L_cmove
//
//	07-17-2006 recoded all macros, the VM, and primitives to keep the stack
//	           pointer in the EBX register. GlobalSp and EBX are synchronized
//	           on every cycle of the VM. The following rules must now
//	           be observed for coding primitive Forth words in this file:
//
//	           1.If a primitive in this file calls an external function,  
//                   it is required to copy the stack ptr in EBX to GlobalSp 
//		     prior to the CALL, and to copy the value of GlobalSp 
//                   back into the EBX register following the CALL:
//
//			   movl %ebx, GlobalSp
//			   call ...
//			   movl GlobalSp, %ebx
//
//		  2. If a primitive in this file terminates with a RET instruction,
//		     it must copy the contents of the EBX register into GlobalSp 
//                   prior to RET:
//
//			   movl %ebx, GlobalSp
//			   ret
//
//		  3. If a primitive may be called externally, i.e. the primitive is 
//                   declared in the .global section of this file, it must copy the 
//		     contents of GlobalSp into the EBX register upon entry. For such 
//	             a primitive, requirement 2) also applies since it must terminate 
//	             with RET.
//
//		  Primitives in this file which terminate with NEXT need not 
//                reference GlobalSp at all, although in some instances it may be 
//                convenient to use GlobalSp as a temporary variable for holding 
//                the stack ptr, for example, when EBX needs to be used temporarily 
//                for another purpose.
//
//	07-22-2006 made various fixes to words to abide by above rules governing
//                 use of EBX and GlobalSp, and made revisions to reflect current
//                 version of vm.s.
//
//      11-02-2006 recoded L_fzeroeq and L_fzerolt to remove conditional branching. Added
//                   L_fzerogt.
//
//      02-19-2007 fixed L_utmslash (now passes divtest.4th) and added division by zero
//                   and division overflow checks in this word ;  added proper checks and handling
//                   for division overflow and div by zero in L_umslashmod ;  added division by
//                   zero checks in L_fmslashmod and L_smslashrem ;  division by zero error 
//	             checks added for L_mod and L_slashmod
//
//      08-01-2007 implement UTS/MOD and STS/REM ;  updated L_utmslash to match
//                   corresponding code in vm.s
//	10-05-2007 replaced memcpy with memmove in L_move
//      03-15-2008 implemented L_calladdr
//      05-03-2009 fixed bug in L_calladdr, which was clobbering the stack ptr (EBX reg)
//      09-19-2009 implemented L_blank
//      09-26-2009 implemented L_within, L_starslash
//      09-28-2009 temporarily make L_dnegate and L_dplus words which ret instead of NEXT
//      11-22-2009 commented out code for L_tick, which has been replaced by CPP_tick
//      11-26-2009 recoded the macro FDROP; implemented L_dtwostar and L_dtwodiv

.set __FAST__, -1
.include "vm-common.s"

.macro SWAP
	movl %ebx, %eax
        addl $WSIZE, %eax
	movl (%eax), %ecx
	addl $WSIZE, %eax
	movl (%eax), %edx
	movl %ecx, (%eax)
	subl $WSIZE, %eax
	movl %edx, (%eax)
	xorl %eax, %eax
.endm


.macro OVER
        movl 2*WSIZE(%ebx), %ecx
        movl %ecx, (%ebx)
	subl $WSIZE, %ebx
.endm
	
.macro FDUP
	movl %ebx, %ecx
	addl $WSIZE, %ebx
	movl (%ebx), %edx
	addl $WSIZE, %ebx
	movl (%ebx), %eax
	movl %ecx, %ebx
	movl %eax, (%ebx)
	subl $WSIZE, %ebx
	movl %edx, (%ebx)
	subl $WSIZE, %ebx
	xor %eax, %eax
.endm
	
.macro FDROP
	addl $2*WSIZE, %ebx
.endm
	
.macro FSWAP
	movl $WSIZE, %ecx
	addl %ecx, %ebx
	movl (%ebx), %edx
	addl %ecx, %ebx
	movl (%ebx), %eax
	addl %ecx, %ebx
	xchgl %edx, (%ebx)
	addl %ecx, %ebx
	xchgl %eax, (%ebx)
	subl %ecx, %ebx
	subl %ecx, %ebx
	movl %eax, (%ebx)
	subl %ecx, %ebx
	movl %edx, (%ebx)
	subl %ecx, %ebx
	xorl %eax, %eax
.endm
		
.macro FOVER
	movl %ebx, %ecx
	addl $3*WSIZE, %ebx
	movl (%ebx), %edx
	addl $WSIZE, %ebx
	movl (%ebx), %eax
	movl %ecx, %ebx
	movl %eax, (%ebx)
	subl $WSIZE, %ebx
	movl %edx, (%ebx)
	subl $WSIZE, %ebx
	xor %eax, %eax
.endm
	
.macro _PUSH
	movl $WSIZE, %eax
	addl %eax, %ebx	
	movl (%ebx), %ecx
	movl GlobalRp, %edx
	movl %ecx, (%edx)
	subl %eax, %edx
	movl %edx, GlobalRp
        xor %eax, %eax
.endm

.macro _POP
	movl $WSIZE, %eax
	movl GlobalRp, %edx
	addl %eax, %edx
	movl %edx, GlobalRp
	movl (%edx), %ecx
	movl %ecx, (%ebx)
	subl %eax, %ebx
	xor %eax, %eax
.endm

.macro FETCH
	movl %ebx, %edx	
	addl $WSIZE, %edx
        movl (%edx), %eax	
        movl (%eax), %eax
	movl %eax, (%edx)
	xor %eax, %eax
.endm
	
.macro STORE
	movl $WSIZE, %eax
        addl %eax, %ebx
        movl (%ebx), %ecx	# address to store to in ecx
	addl %eax, %ebx
	movl (%ebx), %edx	# value to store in edx
	movl %edx, (%ecx)
	xor %eax, %eax
.endm

// Dyadic Logic operators 
	
.macro LOGIC_DYADIC op
	movl $WSIZE, %ecx
	addl %ecx, %ebx
	movl (%ebx), %eax
	addl %ecx, %ebx
	\op (%ebx), %eax
	movl %eax, (%ebx)
	subl %ecx, %ebx
	xorl %eax, %eax 
.endm
	
.macro _AND
	LOGIC_DYADIC andl
.endm

.macro _OR
	LOGIC_DYADIC orl
.endm

.macro _XOR
	LOGIC_DYADIC xorl
.endm


// use algorithm from DNW's vm-osxppc.s
.macro _ABS	
	addl $WSIZE, %ebx
	movl (%ebx), %ecx
	xorl %eax, %eax
	cmpl %eax, %ecx
	setl %al
	negl %eax
	movl %eax, %edx
	xorl %ecx, %edx
	subl %eax, %edx
	movl %edx, (%ebx)
	subl $WSIZE, %ebx
	xorl %eax, %eax
.endm

// Dyadic relational operators (single length numbers) 
	
.macro REL_DYADIC setx
	movl $WSIZE, %ecx
	addl %ecx, %ebx
	movl (%ebx), %eax
	addl %ecx, %ebx
	cmpl %eax, (%ebx)
	movl $0, %eax
	\setx %al
	negl %eax
	movl %eax, (%ebx)
	subl %ecx, %ebx
	xorl %eax, %eax
.endm

	
// Relational operators for zero (single length numbers)
	
.macro REL_ZERO setx
	addl $WSIZE, %ebx
	movl (%ebx), %eax
	cmpl $0, %eax
	movl $0, %eax
	\setx %al
	negl %eax
	movl %eax, (%ebx)
	subl $WSIZE, %ebx
	xorl %eax, %eax
.endm

.macro FREL_DYADIC logic arg set
	movl $WSIZE, %ecx
	addl %ecx, %ebx
	fldl (%ebx)
	addl %ecx, %ebx
	addl %ecx, %ebx
	fcompl (%ebx)
	fnstsw %ax
	andb $65, %ah
	\logic \arg, %ah
	movl $0, %eax
	\set %al
	negl %eax
	addl %ecx, %ebx
	movl %eax, (%ebx)
	subl %ecx, %ebx
	xorl %eax, %eax
.endm

	
.macro STOD
	movl $WSIZE, %ecx
	movl WSIZE(%ebx), %eax
	cdq
	movl %edx, (%ebx)
	subl %ecx, %ebx
	xor %eax, %eax
.endm


	# b = (d1.hi < d2.hi) OR ((d1.hi = d2.hi) AND (d1.lo u< d2.lo))
.macro DLT
	movl $WSIZE, %ecx
	xorl %edx, %edx
	addl %ecx, %ebx
	movl (%ebx), %eax
	cmpl %eax, 2*WSIZE(%ebx)
	sete %dl
	setl %dh
	addl %ecx, %ebx
	movl (%ebx), %eax
	addl %ecx, %ebx
	addl %ecx, %ebx
	cmpl %eax, (%ebx)
	setb %al
	andb %al, %dl
	orb  %dh, %dl
	xorl %eax, %eax
	movb %dl, %al
	negl %eax
	movl %eax, (%ebx)
	subl %ecx, %ebx	
	xorl %eax, %eax
.endm

		
.macro DNEGATE
	addl $WSIZE, %ebx
	movl %ebx, %ecx
	addl $WSIZE, %ebx
	movl (%ebx), %eax
	notl %eax
	clc
	addl $1, %eax
	movl %eax, (%ebx)
	movl %ecx, %ebx
	movl (%ebx), %eax
	notl %eax
	adcl $0, %eax
	movl %eax, (%ebx)
	subl $WSIZE, %ebx
	xor %eax, %eax
.endm
	
.macro DPLUS
	addl $2*WSIZE, %ebx
	movl (%ebx), %eax
	clc
	addl 2*WSIZE(%ebx), %eax
	movl %eax, 2*WSIZE(%ebx)
	movl WSIZE(%ebx), %eax
	adcl -WSIZE(%ebx), %eax
	movl %eax, WSIZE(%ebx)
	xor %eax, %eax
.endm


.macro DMINUS
	addl $2*WSIZE, %ebx
	movl 2*WSIZE(%ebx), %eax
	clc
	subl (%ebx), %eax
	movl %eax, 2*WSIZE(%ebx)
	movl WSIZE(%ebx), %eax
	sbbl -WSIZE(%ebx), %eax
	movl %eax, WSIZE(%ebx)
	xor %eax, %eax
.endm

.macro STARSLASH
	movl $WSIZE, %eax
	sall $1, %eax
        addl %eax, %ebx
        movl WSIZE(%ebx), %eax
        imull (%ebx)
	idivl -WSIZE(%ebx)
	movl %eax, WSIZE(%ebx)
	xor %eax, %eax
.endm

.macro TNEG
	movl $WSIZE, %eax
	addl %eax, %ebx
	movl (%ebx), %edx
	addl %eax, %ebx
	movl (%ebx), %ecx
	addl %eax, %ebx
	movl (%ebx), %eax
	notl %eax
	notl %ecx
	notl %edx
	clc
	addl $1, %eax
	adcl $0, %ecx
	adcl $0, %edx
	movl %eax, (%ebx)
	movl $WSIZE, %eax
	subl %eax, %ebx
	movl %ecx, (%ebx)
	subl %eax, %ebx
	movl %edx, (%ebx)
	movl GlobalSp, %ebx
	xor %eax, %eax	
.endm

// VIRTUAL MACHINE 
					
.global vm
	.type	vm,@function
vm:	
        pushl %ebp
        pushl %ebx
        pushl %ecx
        pushl %edx
	pushl GlobalIp
	pushl vmEntryRp
        movl %esp, %ebp
        movl 28(%ebp), %ebp     # load the Forth instruction pointer
        movl %ebp, GlobalIp
	movl GlobalRp, %eax
	movl %eax, vmEntryRp
	xor %eax, %eax
	movl GlobalSp, %ebx
next:
        movb (%ebp), %al         # get the opcode
	movl JumpTable(,%eax,4), %ecx	# machine code address of word
	xor %eax, %eax          # clear error code
	call *%ecx		# call the word
	movl GlobalSp, %ebx
	movl GlobalIp, %ebp
	incl %ebp		 # increment the Forth instruction ptr
	movl %ebp, GlobalIp
	cmpb $0, %al		 # check for error
	jz next        
exitloop:
        cmpl $OP_RET, %eax         # return from vm?
        jnz vmexit
        xor %eax, %eax            # clear the error
vmexit:
	pop vmEntryRp
	pop GlobalIp
	pop %edx
        pop %ecx
        pop %ebx
        pop %ebp
        ret

L_ret:
	movl vmEntryRp, %eax		# Return Stack Ptr on entry to VM
	movl GlobalRp, %ecx
	cmpl %eax, %ecx
	jl ret1
        movl $OP_RET, %eax             # exhausted the return stack so exit vm
        ret
ret1:
	addl $WSIZE, %ecx
        movl %ecx, GlobalRp
ret2:   movl (%ecx), %eax
	movl %eax, GlobalIp		# reset the instruction ptr
        xorl %eax, %eax
retexit:
        ret
L_tobody:
	addl $WSIZE, %ebx
	movl (%ebx), %ecx	# code address
	incl %ecx		# the data address is offset by one
	movl (%ecx), %ecx
	movl %ecx, (%ebx)
	subl $WSIZE, %ebx
	movl %ebx, GlobalSp
	ret

#
# For precision delays, use MS instead of USLEEP
# Use USLEEP when task can be put to sleep and reawakened by OS
#
L_usleep:
	movl $WSIZE, %eax
	addl %eax, %ebx
	pushl %ebx
	movl (%ebx), %eax
	pushl %eax
	call usleep
	popl %eax
	popl %ebx
	movl %ebx, GlobalSp
	xorl %eax, %eax
	ret
L_ms:
	movl WSIZE(%ebx), %eax
	imull $1000, %eax
	movl %eax, WSIZE(%ebx)
	movl %ebx, GlobalSp
	call C_usec
	ret
L_fill:
	SWAP
	movl $WSIZE, %eax
	addl %eax, %ebx
	movl (%ebx), %ecx
	pushl %ecx
	addl %eax, %ebx
	movl (%ebx), %ecx
	pushl %ecx
	addl %eax, %ebx
	movl %ebx, GlobalSp
	movl (%ebx), %ecx
	pushl %ecx
	call memset
	addl $3*WSIZE, %esp
	xorl %eax, %eax
fillexit:	
	ret
L_erase:
	movl $0, (%ebx)
	subl $WSIZE, %ebx
	call L_fill
	ret
L_blank:
	movl $32, (%ebx)
	subl $WSIZE, %ebx
	call L_fill
	ret			
L_move:
	movl $WSIZE, %eax
	addl %eax, %ebx
	movl (%ebx), %ecx
	pushl %ecx
	SWAP
	movl $WSIZE, %eax
	addl %eax, %ebx
	movl (%ebx), %ecx
	pushl %ecx
	addl %eax, %ebx
	movl (%ebx), %ecx
	pushl %ecx
	movl %ebx, GlobalSp
	call memmove
	addl $12, %esp
	xorl %eax, %eax				
	ret		
L_cmove:
	movl $WSIZE, %eax
	addl %eax, %ebx
	movl (%ebx), %ecx		# nbytes in ecx
	cmpl $0, %ecx
	jnz  cmove1
	addl $2*WSIZE, %ebx
	xorl %eax, %eax
	NEXT		
cmove1:	addl %eax, %ebx
	movl (%ebx), %edx		# dest addr in edx
	movl $WSIZE, %eax
	addl %eax, %ebx
	movl (%ebx), %eax		# src addr in eax
	movl %ebx, GlobalSp
	movl %eax, %ebx			# src addr in ebx
cmoveloop: movb (%ebx), %al
	movb %al, (%edx)
	incl %ebx
	incl %edx
	loop cmoveloop
	movl GlobalSp, %ebx
	xor %eax, %eax				
	NEXT			
L_cmovefrom:
	movl $WSIZE, %eax
	addl %eax, %ebx
	movl (%ebx), %ecx	# load count register
	addl %eax, %ebx
cmovefrom2:
	movl (%ebx), %edx
	addl %ecx, %edx         
	decl %edx               # dest addr in %edx
	addl %eax, %ebx
cmovefrom3:
	movl %ebx, GlobalSp
	movl %ecx, %eax
	cmpl $0, %eax
	jnz cmovefrom4
	ret
cmovefrom4:
	movl (%ebx), %ebx	
	decl %eax
	addl %eax, %ebx		# src addr in %ebx
cmovefromloop:	
	movb (%ebx), %al
	decl %ebx
	xchgl %ebx, %edx
	movb %al, (%ebx)
	decl %ebx
	xchgl %ebx, %edx
	loop cmovefromloop	
	xor %eax, %eax
	ret
L_slashstring:
	addl $WSIZE, %ebx
	movl (%ebx), %ecx
	addl $WSIZE, %ebx
	subl %ecx, (%ebx)
	addl $WSIZE, %ebx
	addl %ecx, (%ebx)
	subl $2*WSIZE, %ebx
	NEXT
L_call:	
	movl $WSIZE, %eax
	addl %eax, %ebx
	movl %ebx, GlobalSp
	xorl %eax, %eax
	call *(%ebx)
	movl GlobalSp, %ebx
	ret
L_push:
	_PUSH
        NEXT
L_pop:
	_POP
	NEXT
L_twopush:
	addl $WSIZE, %ebx
	movl (%ebx), %edx
	addl $WSIZE, %ebx
	movl (%ebx), %eax
	movl GlobalRp, %ecx
	movl %eax, (%ecx)
	subl $WSIZE, %ecx
	movl %edx, (%ecx)
	subl $WSIZE, %ecx
	movl %ecx, GlobalRp
	xor %eax, %eax
	NEXT
L_twopop:
	movl GlobalRp, %ecx
	addl $WSIZE, %ecx
	movl (%ecx), %edx
	addl $WSIZE, %ecx
	movl (%ecx), %eax
	movl %ecx, GlobalRp
	movl %eax, (%ebx)
	subl $WSIZE, %ebx
	movl %edx, (%ebx)
	subl $WSIZE, %ebx
	xor %eax, %eax				
	NEXT
L_puship:
        movl %ebp, %eax
        movl GlobalRp, %ecx
        mov %eax, (%ecx)
	movl $WSIZE, %eax
        subl %eax, GlobalRp
        xor %eax, %eax
        NEXT
L_execute:	
        movl %ebp, %ecx
        movl GlobalRp, %edx
        movl %ecx, (%edx)
	movl $WSIZE, %eax 
        subl %eax, %edx 
	movl %edx, GlobalRp
        addl %eax, %ebx
        movl (%ebx), %eax
	decl %eax
	movl %eax, %ebp
        xorl %eax, %eax
        NEXT
L_definition:
        movl %ebp, %eax
	incl %eax
	movl (%eax), %ecx # address to execute
	addl $3, %eax
	movl %eax, %edx
	movl GlobalRp, %eax
	movl %edx, (%eax)
	subl $WSIZE, %eax
	movl %eax, GlobalRp
	decl %ecx
	movl %ecx, %ebp
        xorl %eax, %eax	
	NEXT
L_rfetch:
        movl GlobalRp, %ecx
	movl $WSIZE, %eax
        addl %eax, %ecx
        movl (%ecx), %ecx
        movl %ecx, (%ebx)
        subl %eax, %ebx
        xorl %eax, %eax
	NEXT
L_tworfetch:
	movl GlobalRp, %ecx
	movl $WSIZE, %eax
	addl %eax, %ecx
	movl (%ecx), %edx
	addl %eax, %ecx
	movl (%ecx), %ecx
	movl %ecx, (%ebx)
	subl %eax, %ebx
	movl %edx, (%ebx)
	subl %eax, %ebx
	xorl %eax, %eax				
	NEXT
L_rpfetch:
	movl GlobalRp, %ecx
	movl $WSIZE, %eax
	addl %eax, %ecx
	movl %ecx, (%ebx)
	subl %eax, %ebx
	xorl %eax, %eax
	NEXT
L_spfetch:
	movl %ebx, %ecx
	movl $WSIZE, %eax
	addl %eax, %ecx
	movl %ecx, (%ebx)
	subl %eax, %ebx
	xorl %eax, %eax 
	NEXT	
L_i:
        movl GlobalRp, %ecx
        movl 3*WSIZE(%ecx), %eax
        movl %eax, (%ebx)
	movl $WSIZE, %eax
        subl %eax, %ebx 
        xorl %eax, %eax
        NEXT
L_j:
        movl GlobalRp, %ecx
        movl 6*WSIZE(%ecx), %eax
        movl %eax, (%ebx)
	movl $WSIZE, %eax
        subl %eax, %ebx
        xorl %eax, %eax
        NEXT	
L_loop:
        movl GlobalRp, %ebx	
	movl $WSIZE, %eax
	addl %eax, %ebx
	movl (%ebx), %edx
	addl %eax, %ebx
	movl (%ebx), %ecx
	addl %eax, %ebx
        movl (%ebx), %eax
        incl %eax
	cmpl %ecx, %eax	
        jz L_unloop
loop1:	
        movl %eax, (%ebx)	# set loop counter to next value
	movl %edx, %ebp		# set instruction ptr to start of loop
	movl GlobalSp, %ebx
        xorl %eax, %eax
        NEXT
L_unloop:  
	movl $WSIZE, %eax
	shll $1, %eax
	addl $WSIZE, %eax
	addl %eax, GlobalRp  # terminal count reached, discard top 3 items
	movl GlobalSp, %ebx
        xorl %eax, %eax
        NEXT
L_plusloop:
	pushl %ebp
	movl $WSIZE, %eax
	addl %eax, %ebx
	movl (%ebx), %ebp	# get loop increment 
	movl %ebx, GlobalSp		
        movl GlobalRp, %ebx
	addl %eax, %ebx		# get ip and save in edx
	movl (%ebx), %edx
	addl %eax, %ebx
	movl (%ebx), %ecx	# get terminal count in ecx
	addl %eax, %ebx
	movl (%ebx), %eax	# get current loop count 
	addl %ebp, %eax		# increment loop count
	cmpl $0, %ebp
	jl plusloop1
	popl %ebp
	cmpl %ecx, %eax
	jge L_unloop				
	movl %eax, (%ebx)	# set loop counter to incremented value
        movl %edx, %ebp		# set instruction ptr to start of loop
	movl GlobalSp, %ebx
        xorl %eax, %eax
        NEXT
plusloop1:
	popl %ebp
	cmpl %ecx, %eax
	jl L_unloop
	movl %eax, (%ebx)
	movl %edx, %ebp
	movl GlobalSp, %ebx
	xorl %eax, %eax
	NEXT
L_count:
	movl WSIZE(%ebx), %ecx
	xorl %eax, %eax
	movb (%ecx), %al
	incl WSIZE(%ebx)
	movl %eax, (%ebx)
	movl $WSIZE, %eax
	subl %eax, %ebx
	movl %ebx, GlobalSp
	xorl %eax, %eax
	ret
L_ival:
        movl %ebp, %ecx
        incl %ecx
        movl (%ecx), %eax
	addl $WSIZE-1, %ecx
	movl %ecx, %ebp
	movl %eax, (%ebx)
	subl $WSIZE, %ebx
	xorl %eax, %eax
	NEXT
L_addr:
	movl %ebp, %ecx
	incl %ecx
	movl (%ecx), %eax
	addl $WSIZE-1, %ecx
	movl %ecx, %ebp
	movl %eax, (%ebx)
	subl $WSIZE, %ebx
	xorl %eax, %eax
	NEXT
L_fval:
        movl %ebp, %ecx
        incl %ecx
        subl $WSIZE, %ebx
        movl (%ecx), %eax
	movl %eax, (%ebx)
	movl WSIZE(%ecx), %eax
	movl %eax, WSIZE(%ebx)
	subl $WSIZE, %ebx
	addl $2*WSIZE-1, %ecx
	movl %ecx, %ebp
	xorl %eax, %eax
	NEXT
L_and:
	_AND
	NEXT
L_or:
	_OR
	NEXT
L_not:
	_NOT
	NEXT
L_xor:
	_XOR
	NEXT

L_eq:
	REL_DYADIC sete
	NEXT
L_ne:
	REL_DYADIC setne
	NEXT
L_ult:
	REL_DYADIC setb
	NEXT
L_ugt:
	REL_DYADIC seta
	NEXT
L_lt:
	REL_DYADIC setl
	NEXT	
L_gt:
	REL_DYADIC setg
	NEXT
L_le:
	REL_DYADIC setle
	NEXT
L_ge:
	REL_DYADIC setge
	NEXT
L_zeroeq:
	REL_ZERO setz
	NEXT
L_zerone:
	REL_ZERO setnz
	NEXT
L_zerolt:
	REL_ZERO setl
	NEXT
L_zerogt:
	REL_ZERO setg
	NEXT
L_within:                          # stack: a b c
        movl 2*WSIZE(%ebx), %ecx   # ecx = b
	movl WSIZE(%ebx), %eax     # eax = c
	subl %ecx, %eax            # eax = c - b
	addl $WSIZE, %ebx     
	addl $WSIZE, %ebx
	movl WSIZE(%ebx), %edx     # edx = a
        subl %ecx, %edx            # edx = a - b
	cmpl %eax, %edx
	movl $0, %eax
	setb %al
	negl %eax
	movl %eax, WSIZE(%ebx)
	xorl %eax, %eax      
        NEXT
L_deq:
	addl $WSIZE, %ebx
	movl (%ebx), %edx
	addl $WSIZE, %ebx
	movl (%ebx), %ecx
	addl $WSIZE, %ebx
	movl %ebx, GlobalSp
	movl (%ebx), %eax
	subl %edx, %eax
	addl $WSIZE, %ebx
	movl (%ebx), %edx
	subl %ecx, %edx
	orl %edx, %eax
	cmpl $0, %eax
	movl $0, %eax
	setz %al
	negl %eax
	movl %eax, (%ebx)
	movl GlobalSp, %ebx
	xorl %eax, %eax
	NEXT
L_dzeroeq:
	addl $WSIZE, %ebx
	movl %ebx, %ecx
	movl (%ebx), %eax
	addl $WSIZE, %ebx
	orl (%ebx), %eax
	cmpl $0, %eax
	movl $0, %eax
	setz %al
	negl %eax
	movl %eax, (%ebx)
	movl %ecx, %ebx
	xorl %eax, %eax
	NEXT
L_dlt:
	DLT
	NEXT
L_dult:	# b = (d1.hi u< d2.hi) OR ((d1.hi = d2.hi) AND (d1.lo u< d2.lo))
	movl $WSIZE, %ecx
	xorl %edx, %edx
	addl %ecx, %ebx
	movl (%ebx), %eax
	cmpl %eax, 2*WSIZE(%ebx)
	sete %dl
	setb %dh
	addl %ecx, %ebx
	movl (%ebx), %eax
	addl %ecx, %ebx
	movl %ebx, GlobalSp
	addl %ecx, %ebx
	cmpl %eax, (%ebx)
	setb %al
	andb %al, %dl
	orb  %dh, %dl
	xorl %eax, %eax
	movb %dl, %al
	negl %eax
	movl %eax, (%ebx)
	movl GlobalSp, %ebx
	xorl %eax, %eax
	NEXT
L_querydup:
	movl WSIZE(%ebx), %eax
	cmpl $0, %eax
	je L_querydupexit
	movl %eax, (%ebx)
	movl $WSIZE, %eax
	subl %eax, %ebx
	xorl %eax, %eax
L_querydupexit:
	movl %ebx, GlobalSp
	ret

L_swap:
	SWAP
        NEXT
L_over:
	OVER
        NEXT
L_rot:
	pushl %ebp
	movl $WSIZE, %eax
	addl %eax, %ebx
	movl %ebx, %ebp
	addl %eax, %ebx
	addl %eax, %ebx
	movl (%ebx), %ecx
	movl (%ebp), %edx
	movl %ecx, (%ebp)
	addl %eax, %ebp
	movl (%ebp), %ecx
	movl %edx, (%ebp)
	movl %ecx, (%ebx)
	xorl %eax, %eax
	popl %ebp
	movl GlobalSp, %ebx
	NEXT
L_minusrot:
	movl WSIZE(%ebx), %eax
	movl %eax, (%ebx)
	addl $WSIZE, %ebx
	movl WSIZE(%ebx), %eax
	movl %eax, (%ebx)
	addl $WSIZE, %ebx
	movl WSIZE(%ebx), %eax
	movl %eax, (%ebx)
	movl -2*WSIZE(%ebx), %eax
	movl %eax, WSIZE(%ebx)
	movl GlobalSp, %ebx
	xorl %eax, %eax
	NEXT
L_nip:
        SWAP
	movl $WSIZE, %eax
        addl %eax, %ebx
	xor %eax, %eax
        NEXT
L_tuck:
        SWAP
        OVER
        NEXT	
L_pick:                        
	movl %ebx, %ecx
	movl WSIZE(%ebx), %eax
	addl $2, %eax
	imul $WSIZE, %eax
	addl %eax, %ecx
	movl (%ecx), %eax
	movl %eax, WSIZE(%ebx)
	xorl %eax, %eax
	NEXT
L_roll:
	movl $WSIZE, %eax
	addl %eax, %ebx
	movl %ebx, GlobalSp
	movl (%ebx), %eax
	incl %eax
	pushl %eax
	pushl %ebx
	imul $WSIZE, %eax
	addl %eax, %ebx		# addr of item to roll
	movl (%ebx), %eax
	popl %ebx
	movl %eax, (%ebx)
	popl %eax		# number of cells to copy
	movl %eax, %ecx
	imul $WSIZE, %eax
	addl %eax, %ebx
	movl %ebx, %edx		# dest addr
	subl $WSIZE, %ebx	# src addr
rollloop:
	movl (%ebx), %eax
	sub $WSIZE, %ebx
	xchgl %ebx, %edx
	movl %eax, (%ebx)
	sub $WSIZE, %ebx
	xchgl %ebx, %edx
	loop rollloop

	movl GlobalSp, %ebx
	xorl %eax, %eax
	ret
L_depth:
	movl GlobalSp, %ebx
	movl BottomOfStack, %eax
	subl %ebx, %eax
	movl $WSIZE, (%ebx)
	movl $0, %edx
	idivl (%ebx)
	movl %eax, (%ebx)
	movl $WSIZE, %eax
	subl %eax, GlobalSp
	xorl %eax, %eax
        ret
L_2drop:
	FDROP
        NEXT
L_2dup:
	FDUP
        NEXT
L_2swap:
	FSWAP	
        NEXT
L_2over:
	FOVER
        NEXT
L_2rot:
	addl $WSIZE, %ebx
	movl %ebx, %ecx
	movl (%ebx), %edx
	addl $WSIZE, %ebx
	movl (%ebx), %eax
	addl $WSIZE, %ebx
	xchgl %edx, (%ebx)
	addl $WSIZE, %ebx
	xchgl %eax, (%ebx)
	addl $WSIZE, %ebx
	xchgl %edx, (%ebx)
	addl $WSIZE, %ebx
	xchgl %eax, (%ebx)
	movl %ecx, %ebx
	movl %edx, (%ebx)
	addl $WSIZE, %ebx
	movl %eax, (%ebx)
	movl GlobalSp, %ebx
	xorl %eax, %eax
        NEXT
L_question:
	FETCH
	movl %ebx, GlobalSp
	cmpl $0, %eax
	jnz questionexit
	call CPP_dot
questionexit:	
	ret	
L_fetch:
	FETCH
	NEXT
L_store:
	STORE
	NEXT
L_afetch:
	FETCH
	NEXT
L_cfetch:
	xorl %eax, %eax
	movl %ebx, %edx
	addl $WSIZE, %edx
	movl (%edx), %ecx
	movb (%ecx), %al
	movl %eax, (%edx)
	xorl %eax, %eax
        NEXT
L_cstore:
	movl $WSIZE, %eax
	addl %eax, %ebx
	movl (%ebx), %ecx	# address to store
	addl %eax, %ebx
	movl (%ebx), %eax	# value to store
	movb %al, (%ecx)
	xorl %eax, %eax
	NEXT	
L_wfetch:
	movl WSIZE(%ebx), %ecx
	movw (%ecx), %ax
	cwde
	movl %eax, WSIZE(%ebx)
	xorl %eax, %eax
        NEXT
L_wstore:
	movl $WSIZE, %eax
	addl %eax, %ebx
	movl (%ebx), %ecx
	addl %eax, %ebx
	movl (%ebx), %edx
	movw %dx, (%ecx)
	xorl %eax, %eax
        NEXT
L_sffetch:
	movl $WSIZE, %eax
	movl %ebx, %ecx
        addl %eax, %ecx
        movl (%ecx), %ecx
        flds (%ecx)
        fstpl (%ebx)
        subl %eax, %ebx
	xorl %eax, %eax
        NEXT
L_sfstore:
	movl $WSIZE, %eax
        addl %eax, %ebx
        addl %eax, %ebx
        fldl (%ebx)              # load the f number into NDP
	movl %ebx, %edx
        subl %eax, %ebx
        movl (%ebx), %ebx          # load the dest address
        fstps (%ebx)             # store as single precision float
	movl %edx, %ebx
        addl %eax, %ebx
	xorl %eax, %eax
        NEXT
L_dffetch:
	movl %ebx, %edx
	addl $WSIZE, %ebx
	movl (%ebx), %ecx
	movl (%ecx), %eax
	movl %eax, (%edx)
	addl $WSIZE, %ecx
	movl (%ecx), %eax
	movl %eax, (%ebx)
	subl $WSIZE, %edx
	movl %edx, %ebx
	xorl %eax, %eax
	NEXT
L_dfstore:
	movl $WSIZE, %edx
	addl %edx, %ebx
	movl %ebx, %eax
	movl (%ebx), %ebx  # address to store
	addl %edx, %eax
	movl (%eax), %ecx
	movl %ecx, (%ebx)
	addl %edx, %eax
	addl %edx, %ebx
	movl (%eax), %ecx
	movl %ecx, (%ebx)
	movl %eax, %ebx
	xorl %eax, %eax
	NEXT
L_abs:
	_ABS
	NEXT

L_max:
	movl $WSIZE, %eax
	addl %eax, %ebx
	movl (%ebx), %eax
	movl WSIZE(%ebx), %ecx
	cmpl %eax, %ecx
	jl max1
	movl %ecx, WSIZE(%ebx)
	jmp maxexit
max1:
	movl %eax, WSIZE(%ebx)
maxexit:
	movl %ebx, GlobalSp
	xorl %eax, %eax
        ret
L_min:
	movl $WSIZE, %eax
	addl %eax, %ebx
	movl (%ebx), %eax
	movl WSIZE(%ebx), %ecx
	cmpl %eax, %ecx
	jg min1
	movl %ecx, WSIZE(%ebx)
	jmp minexit
min1:
	movl %eax, WSIZE(%ebx)
minexit:
	movl %ebx, GlobalSp
	xorl %eax, %eax
        ret
L_dmax:
	FOVER
	FOVER
	DLT
	addl $WSIZE, %ebx
	movl (%ebx), %eax
	cmpl $0, %eax
	jne dmin1
	FDROP
	xorl %eax, %eax
	NEXT	
L_dmin:
	FOVER
	FOVER
	DLT
	movl $WSIZE, %ecx
	addl %ecx, %ebx
	movl (%ebx), %eax
	cmpl $0, %eax
	je dmin1
	FDROP
	xorl %eax, %eax
	NEXT
dmin1:
	FSWAP
	FDROP
	xorl %eax, %eax
	NEXT

#  L_dtwostar and L_dtwodiv are valid for two's-complement systems 
L_dtwostar:
        addl $WSIZE, %ebx
        movl WSIZE(%ebx), %eax
        movl %eax, %ecx
        sall $1, %eax
        movl %eax, WSIZE(%ebx)
        shrl $31, %ecx
        movl (%ebx), %eax
        sall $1, %eax
        orl  %ecx, %eax
        movl %eax, (%ebx)
        subl $WSIZE, %ebx
        xorl %eax, %eax
        NEXT
L_dtwodiv:
	addl $WSIZE, %ebx
	movl (%ebx), %eax
        movl %eax, %ecx
        sarl $1, %eax
        movl %eax, (%ebx)
        shll $31, %ecx
        movl WSIZE(%ebx), %eax
        shrl $1, %eax
        orl %ecx, %eax
        movl %eax, WSIZE(%ebx)
        subl $WSIZE, %ebx
        xorl %eax, %eax
        NEXT
L_add:
	movl $WSIZE, %eax
	addl %eax, %ebx
	movl (%ebx), %eax
	addl %eax, WSIZE(%ebx)
        xorl %eax, %eax
        NEXT

L_div:
	movl $WSIZE, %eax
        addl %eax, %ebx
        movl (%ebx), %eax
        cmpl $0, %eax
        jnz div1
        movl $E_DIV_ZERO, %eax
        jmp divexit
div1:	
	addl $WSIZE, %ebx
        movl (%ebx), %eax
	cdq
        idivl -WSIZE(%ebx)
        movl %eax, (%ebx)
	subl $WSIZE, %ebx
	xorl %eax, %eax
divexit:
	movl %ebx, GlobalSp
        ret
L_mod:
	call L_div
	cmpl $0, %eax
	jnz  divexit
	movl %edx, WSIZE(%ebx)
	NEXT
L_slashmod:
	call L_div
	cmpl $0, %eax
	jnz  divexit
	movl %edx, (%ebx)
	subl $WSIZE, %ebx
	SWAP
	NEXT
L_starslash:
	STARSLASH	
	NEXT
L_starslashmod:
	STARSLASH
	movl %edx, (%ebx)
	subl $WSIZE, %ebx
	SWAP
	movl %ebx, GlobalSp
	ret
L_plusstore:
	movl $WSIZE, %edx
	addl %edx, %ebx
	movl (%ebx), %ecx
	movl (%ecx), %eax
	addl %edx, %ebx
	movl (%ebx), %edx
	addl %edx, %eax
	movl %eax, (%ecx)
	xorl %eax, %eax
	NEXT
L_dabs:
	movl GlobalSp, %ebx
	movl %ebx, %edx
	addl $WSIZE, %ebx
	movl (%ebx), %ecx
	movl %ecx, %eax
	cmpl $0, %eax
	jl dabs_go
	movl %edx, GlobalSp
	xorl %eax, %eax
	ret
dabs_go:
	addl $WSIZE, %ebx
	movl (%ebx), %eax
	clc
	subl $1, %eax
	notl %eax
	movl %eax, (%ebx)
	movl %ecx, %eax
	sbbl $0, %eax
	notl %eax
	movl %eax, -WSIZE(%ebx)
	movl %edx, GlobalSp
	xorl %eax, %eax
	ret
L_dnegate:
	movl GlobalSp, %ebx
	DNEGATE
#	NEXT
	movl %ebx, GlobalSp
	ret
L_dplus:
        movl GlobalSp, %ebx
	DPLUS
#	NEXT
        movl %ebx, GlobalSp
        ret
L_dminus:
	movl GlobalSp, %ebx
	DMINUS
	movl %ebx, GlobalSp
	ret
L_umstar:
	movl $WSIZE, %eax
	addl %eax, %ebx
	movl (%ebx), %ecx
	addl %eax, %ebx
	movl %ecx, %eax
	mull (%ebx)
	movl %eax, (%ebx)
	subl $WSIZE, %ebx
	movl %edx, (%ebx)
	movl GlobalSp, %ebx
	xorl %eax, %eax				
	NEXT
L_dsstar:
	# multiply signed double and signed to give triple length product
	movl $WSIZE, %ecx
	addl %ecx, %ebx
	movl (%ebx), %edx
	cmpl $0, %edx
	setl %al
	addl %ecx, %ebx
	movl (%ebx), %edx
	cmpl $0, %edx
	setl %ah
	xorb %ah, %al      # sign of result
	andl $1, %eax
	pushl %eax
	movl GlobalSp, %ebx
	_ABS
	addl $WSIZE, %ebx
	movl %ebx, GlobalSp
	call L_dabs
	movl GlobalSp, %ebx
	subl $WSIZE, %ebx
	movl %ebx, GlobalSp
	call L_udmstar
	movl GlobalSp, %ebx
	popl %eax
	cmpl $0, %eax
	jne dsstar1
	NEXT
dsstar1:
	TNEG
	NEXT
L_umslashmod:
# Divide unsigned double length by unsigned single length to
# give unsigned single quotient and remainder. A "Divide overflow"
# error results if the quotient doesn't fit into a single word.
	movl $WSIZE, %eax
	addl %eax, %ebx
	movl %ebx, GlobalSp
	movl (%ebx), %ecx
	cmpl $0, %ecx
	jnz  umslashmod0
	movl $E_DIV_ZERO, %eax
	ret
umslashmod0:	
	addl %eax, %ebx
	movl $0, %edx
	movl (%ebx), %eax
	divl %ecx
	cmpl $0, %eax
	jne umslashmod_ovflow
	movl (%ebx), %edx
	addl $WSIZE, %ebx
	movl (%ebx), %eax
	divl %ecx
	jmp umslashmod_exit
umslashmod_ovflow:
	movl $E_DIV_OVERFLOW, %eax
	ret
umslashmod_exit:	
	movl %edx, (%ebx)
	subl $WSIZE, %ebx
	movl %eax, (%ebx)
	xorl %eax, %eax		
	ret
L_mstar:
	movl $WSIZE, %eax
	addl %eax, %ebx
	movl (%ebx), %ecx
	addl %eax, %ebx
	movl %ecx, %eax
	imull (%ebx)
	movl %eax, (%ebx)
	subl $WSIZE, %ebx
	movl %edx, (%ebx)
	movl GlobalSp, %ebx
	xorl %eax, %eax		
	NEXT
L_mplus:
	STOD
	DPLUS
	NEXT
L_mslash:
	movl $WSIZE, %eax
        movl GlobalSp, %ebx
	addl %eax, %ebx
        movl (%ebx), %ecx
	addl %eax, %ebx
	movl %ebx, GlobalSp
        cmpl $0, %ecx
        je mslash1
        movl (%ebx), %edx
	addl %eax, %ebx
	movl (%ebx), %eax
        idivl %ecx
        movl %eax, (%ebx)
	xor %eax, %eax		
	ret
mslash1:			
        movl $E_DIV_ZERO, %eax
        ret
L_udmstar:
# Multiply unsigned double and unsigned single to give 
# the triple length product.
	movl GlobalSp, %ebx
	addl $WSIZE, %ebx
	movl (%ebx), %ecx
	addl $WSIZE, %ebx
	movl (%ebx), %eax
	mull %ecx
	movl %edx, -WSIZE(%ebx)
	movl %eax, (%ebx)
	addl $WSIZE, %ebx
	movl %ecx, %eax
	mull (%ebx)
	movl %eax, (%ebx)
	subl $WSIZE, %ebx
	movl (%ebx), %eax
	subl $WSIZE, %ebx
	clc
	addl %edx, %eax
	movl %eax, WSIZE(%ebx)
	movl (%ebx), %eax
	adcl $0, %eax
	movl %eax, (%ebx)
	xor %eax, %eax
	ret
L_utsslashmod:
# Divide unsigned triple length by unsigned single length to
# give an unsigned triple quotient and single remainder.
	addl $WSIZE, %ebx
	movl (%ebx), %ecx		# divisor in ecx
	cmpl $0, %ecx
	jnz  utsslashmod0
	movl $E_DIV_ZERO, %eax	
	ret
utsslashmod0:	
	addl $WSIZE, %ebx
	movl (%ebx), %eax		# ut3
	movl $0, %edx
	divl %ecx			# ut3/u
	call utmslash1
	movl GlobalSp, %ebx
	movl WSIZE(%ebx), %eax
	movl %eax, (%ebx)
	addl $WSIZE, %ebx
	movl WSIZE(%ebx), %eax
	movl %eax, (%ebx)
	addl $WSIZE, %ebx	
	movl -17*WSIZE(%ebx), %eax       # r7
	movl %eax, (%ebx)
	subl $3*WSIZE, %ebx
	movl -5*WSIZE(%ebx), %eax        # q3
	movl %eax, (%ebx)
	subl $WSIZE, %ebx
	movl %ebx, GlobalSp
	xorl %eax, %eax	
	ret
L_tabs:
# Triple length absolute value (needed by L_stsslashrem, STS/REM)
        movl WSIZE(%ebx), %ecx
        movl %ecx, %eax
        cmpl $0, %eax
        jl tabs1
        xor %eax, %eax
        ret
tabs1:
        addl $3*WSIZE, %ebx
        movl (%ebx), %eax
        clc
        subl $1, %eax
        notl %eax
        movl %eax, (%ebx)
	subl $WSIZE, %ebx
	movl (%ebx), %eax
	sbbl $0, %eax
	notl %eax
	movl %eax, (%ebx)
        movl %ecx, %eax
        sbbl $0, %eax
        notl %eax
        movl %eax, -WSIZE(%ebx)
	subl $2*WSIZE, %ebx
	movl %ebx, GlobalSp
        xor %eax, %eax
        ret
L_stsslashrem:
# Divide signed triple length by signed single length to give a
# signed triple quotient and single remainder, according to the
# rule for symmetric division.
	addl $WSIZE, %ebx
	movl (%ebx), %ecx		# divisor in ecx
	cmpl $0, %ecx
	jnz  stsslashrem0
	movl $E_DIV_ZERO, %eax
stsslashrem0:
	movl WSIZE(%ebx), %eax		# t3
	pushl %eax
	cmpl $0, %eax
	movl $0, %eax
	setl %al
	negl %eax
	movl %eax, %edx
	cmpl $0, %ecx
	movl $0, %eax
	setl %al
	negl %eax
	xorl %eax, %edx			# sign of quotient
	pushl %edx
	call L_tabs
	subl $WSIZE, %ebx
	_ABS
	movl %ebx, GlobalSp
	call L_utsslashmod
	popl %edx
	cmpl $0, %edx
	jz stsslashrem1
	TNEG
stsslashrem1:	
	popl %eax
	cmpl $0, %eax
	jz stsslashrem2
	movl GlobalSp, %ebx
	addl $4*WSIZE, %ebx
	negl (%ebx)	
stsslashrem2:
	xorl %eax, %eax
	ret
L_utmslash:
# Divide unsigned triple length by unsigned single to give 
# unsigned double quotient. A "Divide Overflow" error results
# if the quotient doesn't fit into a double word.
	movl GlobalSp, %ebx
	addl $WSIZE, %ebx
	movl (%ebx), %ecx		# divisor in ecx
	cmpl $0, %ecx
	jnz utmslash0
	movl $E_DIV_ZERO, %eax
	ret
utmslash0:
	addl $WSIZE, %ebx
	movl (%ebx), %eax		# ut3
	movl $0, %edx
	divl %ecx			# ut3/u
	cmpl $0, %eax
	jz   utmslash1
	movl $E_DIV_OVERFLOW, %eax
	ret
utmslash1:	 
	pushl %ebx			# keep local stack ptr
	movl GlobalSp, %ebx
	movl %eax, -4*WSIZE(%ebx)	# q3
	movl %edx, -5*WSIZE(%ebx)	# r3
	popl %ebx
	addl $WSIZE, %ebx
	movl (%ebx), %eax		# ut2
	movl $0, %edx
	divl %ecx			# ut2/u
	pushl %ebx
	movl GlobalSp, %ebx
	movl %eax, -2*WSIZE(%ebx)	# q2
	movl %edx, -3*WSIZE(%ebx)	# r2
	popl %ebx
	addl $WSIZE, %ebx
	movl (%ebx), %eax		# ut1
	movl $0, %edx
	divl %ecx			# ut1/u
	pushl %ebx
	movl GlobalSp, %ebx
	movl %eax, (%ebx)		# q1
	movl %edx, -WSIZE(%ebx)		# r1
	movl -5*WSIZE(%ebx), %edx	# r3 << 32
	movl $0, %eax
	divl %ecx			# (r3 << 32)/u
	movl %eax, -6*WSIZE(%ebx)	# q4
	movl %edx, -7*WSIZE(%ebx)	# r4
	movl -3*WSIZE(%ebx), %edx	# r2 << 32
	movl $0, %eax
	divl %ecx			# (r2 << 32)/u
	movl %eax, -8*WSIZE(%ebx)	# q5
	movl %edx, -9*WSIZE(%ebx)	# r5
	movl -7*WSIZE(%ebx), %edx	# r4 << 32
	movl $0, %eax
	divl %ecx			# (r4 << 32)/u
	movl %eax, -10*WSIZE(%ebx)	# q6
	movl %edx, -11*WSIZE(%ebx)	# r6
	movl $0, %edx
	movl -WSIZE(%ebx), %eax		# r1
	addl -9*WSIZE(%ebx), %eax	# r1 + r5
	jnc  utmslash2
	incl %edx
utmslash2:
	addl -11*WSIZE(%ebx), %eax	# r1 + r5 + r6
	jnc  utmslash3
	incl %edx
utmslash3:
	divl %ecx
	movl %eax, -12*WSIZE(%ebx)      # q7
	movl %edx, -13*WSIZE(%ebx)      # r7	
	movl $0, %edx
	addl -10*WSIZE(%ebx), %eax	# q7 + q6
	jnc  utmslash4
	incl %edx
utmslash4:	
	addl -8*WSIZE(%ebx), %eax	# q7 + q6 + q5
	jnc  utmslash5
	incl %edx
utmslash5:	
	addl (%ebx), %eax		# q7 + q6 + q5 + q1
	jnc  utmslash6
	incl %edx
utmslash6:	
	popl %ebx
	movl %eax, (%ebx)
	subl $WSIZE, %ebx
	pushl %ebx
	movl GlobalSp, %ebx
	movl -2*WSIZE(%ebx), %eax	# q2
	addl -6*WSIZE(%ebx), %eax	# q2 + q4
	addl %edx, %eax
	popl %ebx
	movl %eax, (%ebx)
	subl $WSIZE, %ebx
	movl %ebx, GlobalSp
	xorl %eax, %eax
	ret
L_mstarslash:
	movl GlobalSp, %ebx
	addl $WSIZE, %ebx
	addl $WSIZE, %ebx
	movl (%ebx), %eax
	addl $WSIZE, %ebx
	xorl (%ebx), %eax
	shrl $31, %eax
	pushl %eax	# keep sign of result -- negative is nonzero
	movl GlobalSp, %ebx
	addl $WSIZE, %ebx
	movl %ebx, GlobalSp
	_ABS
	addl $WSIZE, %ebx
	movl %ebx, GlobalSp
	call L_dabs
	movl GlobalSp, %ebx
	subl $WSIZE, %ebx
	movl %ebx, GlobalSp
	call L_udmstar
	movl GlobalSp, %ebx
	subl $WSIZE, %ebx
	movl %ebx, GlobalSp
	call L_utmslash
	movl GlobalSp, %ebx
	popl %eax
	cmpl $0, %eax
	jnz mstarslash_neg
	xor %eax, %eax
	ret
mstarslash_neg:
	DNEGATE
	movl %ebx, GlobalSp
	xorl %eax, %eax
	ret		
L_fmslashmod:
	movl $WSIZE, %eax
	addl %eax, %ebx
	movl %ebx, GlobalSp
	movl (%ebx), %ecx
	cmpl $0, %ecx
	jnz  fmslashmod0
	movl $E_DIV_ZERO, %eax
	ret
fmslashmod0:	
	addl %eax, %ebx
	movl (%ebx), %edx
	addl %eax, %ebx
	movl (%ebx), %eax
	idivl %ecx
	movl %edx, (%ebx)
	subl $WSIZE, %ebx
	movl %eax, (%ebx)
	cmpl $0, %ecx
	jg fmslashmod2
	cmpl $0, %edx
	jg fmslashmod3
	xor %eax, %eax
	ret
fmslashmod2:		
	cmpl $0, %edx
	jge fmslashmodexit
fmslashmod3:	
	decl %eax		# floor the result
	movl %eax, (%ebx)
	addl $WSIZE, %ebx
	addl %ecx, (%ebx)
fmslashmodexit:
	xorl %eax, %eax
	ret
L_smslashrem:
	movl $WSIZE, %eax
	addl %eax, %ebx
	movl %ebx, GlobalSp
	movl (%ebx), %ecx
	cmpl $0, %ecx
	jnz  smslashrem0
	movl $E_DIV_ZERO, %eax
	ret
smslashrem0:	
	addl %eax, %ebx
	movl (%ebx), %edx
	addl %eax, %ebx
	movl (%ebx), %eax
	idivl %ecx
	movl %edx, (%ebx)
	subl $WSIZE, %ebx
	movl %eax, (%ebx)
	xorl %eax, %eax		
	ret
L_stod:
	STOD
	NEXT
L_stof:
	movl $WSIZE, %eax
	movl %ebx, %ecx
        addl %eax, %ecx
        fildl (%ecx)
        fstpl (%ebx)
        subl %eax, %ebx
	xorl %eax, %eax
        NEXT
L_dtof:
	movl $WSIZE, %eax
	movl %ebx, %ecx
	addl %eax, %ecx
	movl (%ecx), %eax
	xchgl WSIZE(%ecx), %eax
	movl %eax, (%ecx)
        fildq (%ecx)
        fstpl (%ecx)
	xorl %eax, %eax	
	NEXT	
L_froundtos:
	movl $WSIZE, %eax
        addl %eax, %ebx
	movl %ebx, %ecx
        fldl (%ecx)
        addl %eax, %ecx
        fistpl (%ecx)
	xorl %eax, %eax
        NEXT
L_ftrunctos:
	movl $WSIZE, %eax
	addl %eax, %ebx
	fldl (%ebx)
	fnstcw (%ebx)
	movl (%ebx), %ecx	# save NDP control word		
	movl %ecx, %edx	
	movb $12, %dh
	movl %edx, (%ebx)
	fldcw (%ebx)
	addl %eax, %ebx	
	fistpl (%ebx)
	subl %eax, %ebx
	movl %ecx, (%ebx)
	fldcw (%ebx)		# restore NDP control word
	xorl %eax, %eax	
	NEXT	
L_ftod:
	movl $WSIZE, %eax
	addl %eax, %ebx
	fldl (%ebx)
	subl %eax, %ebx
	fnstcw (%ebx)
	movl (%ebx), %ecx	# save NDP control word	
	movl %ecx, %edx
	movb $12, %dh		
	movl %edx, (%ebx)
	fldcw (%ebx)
	addl %eax, %ebx	
	fistpq (%ebx)
	subl %eax, %ebx
	movl %ecx, (%ebx)
	fldcw (%ebx)		# restore NDP control word
	addl %eax, %ebx 
	movl (%ebx), %eax
	xchgl WSIZE(%ebx), %eax
	movl %eax, (%ebx)
	movl GlobalSp, %ebx
	xorl %eax, %eax	
	NEXT	
L_degtorad:
	fldl FCONST_180
        movl %ebx, %ecx
        addl $WSIZE, %ebx
        fldl (%ebx)
        fdivp %st, %st(1)
        fldpi
        fmulp %st, %st(1)
        fstpl (%ebx)
	movl %ecx, %ebx
        NEXT
L_radtodeg:
	movl %ebx, %ecx
        addl $WSIZE, %ebx
        fldl (%ebx)
        fldpi
	fxch
        fdivp %st, %st(1)
        fldl FCONST_180
        fmulp %st, %st(1)
        fstpl (%ebx)
	movl %ecx, %ebx
        NEXT
L_fne:
	FREL_DYADIC xorb $64 setnz	
	NEXT
L_feq:
	FREL_DYADIC andb $64 setnz
	NEXT
L_flt:
	FREL_DYADIC andb $65 setz
	NEXT	
L_fgt:
	FREL_DYADIC andb $1 setnz
	NEXT		
L_fle:
	FREL_DYADIC xorb $1 setnz
	NEXT
L_fge:
	FREL_DYADIC andb $65 setnz
	NEXT
L_fzeroeq:
	movl $WSIZE, %eax
	addl %eax, %ebx
	movl (%ebx), %ecx
	movl %ebx, %edx
	addl %eax, %ebx
	movl (%ebx), %eax
	shll $1, %eax
	orl %ecx, %eax
	movl $0, %eax
	setz %al
	negl %eax
	movl %eax, (%ebx)
	movl %edx, %ebx
	xorl %eax, %eax
	NEXT
L_fzerolt:
	movl $WSIZE, %eax
	addl %eax, %ebx
	movl %ebx, %ecx 
	fldl (%ebx)
	add %eax, %ebx
	fldz
	fcompp	
	fnstsw %ax
	andb $69, %ah
	movl $0, %eax
	setz %al
	negl %eax
	movl %eax, (%ebx)
	movl %ecx, %ebx
	xorl %eax, %eax
	NEXT
L_fzerogt:
	movl $WSIZE, %eax
	addl %eax, %ebx
	movl %ebx, %ecx
	fldz
	fldl (%ebx)
	addl %eax, %ebx
	fucompp	
	fnstsw %ax
	sahf 
	movl $0, %eax
	seta %al
	negl %eax
	movl %eax, (%ebx)
	movl %ecx, %ebx
	xorl %eax, %eax
	NEXT				
L_fadd:
	movl $WSIZE, %eax
        addl %eax, %ebx
        fldl (%ebx)
	sall $1, %eax
        addl %eax, %ebx
        faddl (%ebx)
        fstpl (%ebx)
        subl $WSIZE, %ebx
	xorl %eax, %eax
        NEXT
L_fsub:
	movl $WSIZE, %eax
	sall $1, %eax
	addl $WSIZE, %eax
        addl %eax, %ebx
        fldl (%ebx)
	subl $WSIZE, %eax
        subl %eax, %ebx
        fsubl (%ebx)
        addl %eax, %ebx
        fstpl (%ebx)
        subl $WSIZE, %ebx
	xorl %eax, %eax
        NEXT
L_fmul:
	movl $WSIZE, %eax
        addl %eax, %ebx
        fldl (%ebx)
        addl %eax, %ebx
	movl %ebx, %ecx
	addl %eax, %ebx
        fmull (%ebx)
        fstpl (%ebx)
        movl %ecx, %ebx
	xorl %eax, %eax
        NEXT
L_fdiv:
	movl $WSIZE, %eax
        addl %eax, %ebx
        fldl (%ebx)
        ftst
        jnz fdiv1
        subl %eax, %ebx
        movl $E_DIV_ZERO, %eax
        jmp fdivexit
fdiv1:  addl %eax, %ebx
	movl %ebx, %ecx
	addl %eax, %ebx
        fdivrl (%ebx)
        fstpl (%ebx)
        movl %ecx, %ebx
fdivexit:
	movl %ebx, GlobalSp
	xorl %eax, %eax
	ret
L_floor:
	addl $WSIZE, %ebx
	movl WSIZE(%ebx), %eax
	pushl %ebx
	pushl %eax
	movl (%ebx), %eax
	pushl %eax
	call floor
	addl $8, %esp
	popl %ebx
	fstpl (%ebx)
	subl $WSIZE, %ebx
	xorl %eax, %eax		
	NEXT
L_fround:
	movl %ebx, %ecx
	addl $WSIZE, %ecx
	fldl (%ecx)
	frndint
	fstpl (%ecx)
	NEXT
L_ftrunc:
	movl %ebx, %eax
	addl $WSIZE, %ebx
	fldl (%ebx)
	fnstcw (%ebx)
	movl (%ebx), %ecx	# save NDP control word
	movl %ecx, %edx
	movb $12, %dh
	movl %edx, (%ebx)
	fldcw (%ebx)
	frndint
	movl %ecx, (%ebx)
	fldcw (%ebx)		# restore NDP control word	
	fstpl (%ebx)
	movl %eax, %ebx		
	xorl %eax, %eax
	NEXT
L_fatan2:
	addl $2*WSIZE, %ebx
	fldl WSIZE(%ebx)
	fldl -WSIZE(%ebx)
	fpatan
	fstpl WSIZE(%ebx)
	NEXT
L_fsincos:
	fldl WSIZE(%ebx)
	fsincos
	fstpl -WSIZE(%ebx)
	fstpl WSIZE(%ebx)
	subl $2*WSIZE, %ebx	
	NEXT

