\ ========================================================
\                                             00-04-26 Wil Baden
\
\    Multiple-Precision Arithmetic, Part 1
\    *************************************
\
\    The Classical Algorithms: + - *
\    -------------------------------
\
\    Multiple-precision integers are represented as arrays
\    of unsigned twos-complement integers in little-endian
\    order.  The address and precision are given in the
\    invocation of operations.  Additional space of at least
\    one cell should be alloted for arrays that may be
\    modified.  Applications must declare arrays that will be
\    big enough.  (This has been no problem, as my applications
\    use only a few multiple-precision integers with prescribed
\    precisions.)
\
\ MP+                 &u &v &w n -- &w n
\    Adds v to u, storing in w.  w may be u or v or different.
\
\ MP-                 &u &v &w n -- &w n
\    Subtracts v from u, storing in w.  w may be u or v or
\    different.
\
\ MP*                 &u m &v n &w -- &w m+n
\    Multiplies u by v storing in w. The precision of w will
\    be m+n.  This is essentially pencil and paper
\    multiplication with addition and single digit
\    multiplication combined.  w must be different from
\    u and v.
\
\ --------------------------------------------------------------
\
\    In the real world, characters have 8 bits.  They just do.
\    All integral arithmetic is done in twos-complement
\    binary.  It just is.  Several simple "get real"
\    assumptions like these make our work possible.
\
\ ======================================================== 

\ --------------------------------------------------------
\    Off the shelf tools.  Comment out redundancy.  These should
\    be CODE definitions.
\ -------------------------------------------------------- 

: THIRD   ( x y z -- x y z x )  2 PICK ;
: FOURTH  ( x y z w -- x y z w x )  3 PICK ;
: OFF     ( addr -- )  0 SWAP ! ;
: TH      ( addr i -- addr+i*cells )  CELLS + ;

VARIABLE Carry-Digit

: MP+                                    ( &u &v &w n -- &w n)
    Carry-Digit OFF
    DUP >R
    0 DO                                           ( &u &v &w)
        THIRD I TH @  THIRD I TH @         ( . . &w u[i] v[i])
        0 TUCK D+                          ( . . &w sum carry)
        Carry-Digit @ M+ Carry-Digit !           ( . . &w sum)
        OVER I TH ( &w[i]) !                       ( &u &v &w)
    LOOP
    NIP NIP R>                                         ( &w n)
    ;

: MP-                                   ( &u &v &w n -- &w n )
    Carry-Digit OFF
    DUP >R
    0 DO                                           ( &u &v &w)
        THIRD I TH @  THIRD I TH @         ( . . &w u[i] v[i])
        0 TUCK D-                        ( . . &w diff borrow)
        Carry-Digit @ M+ Carry-Digit !          ( . . &w diff)
        OVER I TH ( &w[i]) !                       ( &u &v &w)
    LOOP
    NIP NIP R>                                         ( &w n)
    ;

\ --------------------------------------------------------
\ Mp-Multiply-Add      &u m v[j] &w[j] --
\    Multiplies the m "digits" of u by the j-th "digit" of v
\    and adds to w beginning at the j-th "digit".  A "digit"
\    is a cell as an unsigned twos-complement integer.  The
\    carry is put on top of the sub-array, where it will be
\    used in the next round.  Used in `MP*`.
\ -------------------------------------------------------- 

: Mp-Multiply-Add                   ( &u m v[j] &w[j] -- )
        Carry-Digit OFF
        THIRD ( m) 0 DO
            FOURTH I TH @  THIRD      ( . . . &w[j] u[i] v[j])
            UM*                       ( . . . &w[j] product .)
            THIRD I TH ( &w[j+i]) DUP >R  @ 0 D+
            Carry-Digit @ 0 D+ Carry-Digit !
                R> ( &w[j+i]) !             ( &u m v[j] &w[j])
        LOOP
        NIP                                      ( &u m &w[j])
        SWAP TH ( &w[j+m]) Carry-Digit @ SWAP !          ( &u)
        DROP ;

: MP*                               ( &u m &v n &w -- &w m+n )
    dup 10000 U< abort" Missing Product Address "
    SWAP                                       ( &u m &v &w n)
    FOURTH >R 2DUP R> + ( ... &w n+m ) DUP >R  CELLS ERASE
    0 DO                                         ( &u m &v &w)
        OVER I TH @  OVER I TH        ( &u m &v &w v[j] &w[j])
        2>R 2OVER 2R>                   ( ... &u m v[j] &w[j])
        Mp-Multiply-Add                          ( &u m &v &w)
    LOOP
    NIP NIP NIP R> ;

\ --------------------------------------------------------
\    Multiple Precision Tell and Show
\
\ MP!                 x[n-1] ... x[0] addr n --
\    Sets <addr n> to  <x[n-1]> ... <x[0]>.
\ MP.                 addr n --
\    Displays the "digits" of <addr n>.
\ -------------------------------------------------------- 

: MP!                             ( x[n-1] ... x[0] addr n -- )
    0 ?DO  TUCK !  CELL+  LOOP DROP ;

: MP.                                            ( addr n -- )
    CR ." \ "
    0 SWAP 1- DO
        DUP I TH @ .
    -1 +LOOP DROP ;

(
-- 
Wil Baden   Costa Mesa, California   WilBaden@Netcom.com
)

