C     MATHLIB.FOR
C
C     Math Library Functions
C
C     14 January 1990
C
C
C     Updated: 18 July 90; added polynomial fitting routines -- KM
C     Updated: 16 Aug 90; added subroutine INDEX_INTERVAL -- KM
C     Updated:  6 Sep 90; modified PARAB_AREA and added
C	                  LINEAR_TRANSFORMATION  -- KM
C     Updated: 11 Sep 90; added BINOMIAL_COEFFICIENT, FACTORIAL, and
C	                  TRANSFORM_COEFFICIENTS  -- KM
C	
C ---------------------------------------------------------------------
C
C       SMOOTH
C
C       Purpose: Perform a three-point smooth by a weighted average
C                over adjacent points (see Bevington, p. 260).
C       Syntax: CALL SMOOTH (X, Y, NLOW, NHIGH)
C       Input Parameters:
C         X, Y   -- data arrays
C         NLOW   -- beginning index in arrays
C         NHIGH  -- ending index
C       Output Parameters: NONE
C       Author: P.R. Bevington; slightly modified by J. Kielkopf
C       Date:
C
        SUBROUTINE SMOOTH (X, Y, NLOW, NHIGH)


        DIMENSION X(*), Y(*)

        IMAX = NHIGH-1
        IMIN = NLOW+1
        YIM1 = Y(NLOW)
        Y(NLOW)=(3.*Y(NLOW)+Y(IMIN))/4.
  
        DO I = IMIN, IMAX
          YNEW = (YIM1 + 2.*Y(I) + Y(I+1))/4.  
          YIM1 = Y(I)
          Y(I) = YNEW
        END DO   

        Y(NHIGH) = (Y(IMAX)+3.*Y(NHIGH))/4.  
        RETURN
        END
C
C ---------------------------------------------------------------------
C
C       JFIND
C
C       Purpose: Find index of nearest number in an array, with specified
C                start and stop search indices. This routine is
C                based on the subroutine LOCATE, in Numerical Recipes,
C                by Press, et al.
C       Syntax: J = JFIND (X, XX, JS, JE)
C       Input Parameters:
C         X    -- array to be searched. Array should be monotonically
C                 increasing or decreasing.
C         XX   -- number to be found.
C         JS   -- starting search index in array.
C         JE   -- ending search index in array.
C       Returns:
C         J    -- index of closest number in array.
C       Author: Press, et al.; modified by K. Myneni
C       Date: 13 Feb 90
C
	INTEGER FUNCTION JFIND (X, XX, JS, JE)

        DIMENSION X(*)

        LOGICAL INCREASING_ORDER

        INCREASING_ORDER = X(JE).GT.X(JS)
        JL = JS - 1
        JU = JE + 1

        DO WHILE ((JU - JL).GT.1)
           JM = (JU + JL)/2
           IF (INCREASING_ORDER.EQV.(XX.GT.X(JM))) THEN
              JL = JM
           ELSE
              JU = JM
           END IF
        END DO

        IF (INCREASING_ORDER) THEN
           IF (JL.LT.JE) JL = JL + 1
        ELSE
           IF (JL.LT.JS) JL = JL + 1
        END IF
 
        JFIND = JL

        RETURN
        END
C
C --------------------------------------------------------------------
C
C       INDEX_LIMITS
C
C       Purpose: Determine the indices of points falling in
C         the domain (x1, x2).
C       Syntax: CALL INDEX_LIMITS (X, X1, X2, JS, JE, J1, J2, IERR)
C       Input Parameters:
C         X     -- array of X values
C         X1    -- lower limit of domain
C         X2    -- upper limit of domain
C         JS    -- lower index limit in X array
C         JE    -- upper index limit in X array
C       Output Parameters:
C         J1    -- lower index limit of domain
C         J2    -- upper index limit of domain
C         IERR  -- error code has the following meaning:
C                    0 -- no error
C                    1 -- (X1, X2) interval does not overlap
C                         interval (X(JS), X(JE))
C       Author: K. Myneni
C       Date: 16 August 90
C       Notes:
C         The X array values between indices JS and JE may be in
C         increasing or decreasing order. The output indices, J1
C         and J2 will always be returned with J1 less than J2,
C         regardless of the ordering of x values.
C
        SUBROUTINE INDEX_LIMITS (X, X1, X2, JS, JE, J1, J2, IERR)

	DIMENSION X(*)
	LOGICAL INCREASING_ORDER

	IERR = 0

C	Make sure x limits are used in order of minimum, maximum.

	IF (X2.LT.X1) THEN
   	   XX1 = X2
	   XX2 = X1
	ELSE
	   XX1 = X1
	   XX2 = X2
	END IF


	INCREASING_ORDER = (X(JE).GT.X(JS))

C	Search for index of lower limit

	IF (INCREASING_ORDER) THEN
	   DO J = JS, JE
	      IF (X(J).GE.XX1) THEN
	         J1 = J
 	         GO TO 100
	      END IF
	   END DO			
	ELSE
	   DO J = JE, JS, -1
	      IF (X(J).GE.XX1) THEN
	         J1 = J
		 GO TO 100
	      END IF
	   END DO
	END IF

	! x limits outside domain of data

	IERR = 1
	RETURN

 100	CONTINUE

C	Search for index of upper limit

	IF (INCREASING_ORDER) THEN
	   DO J = JE, JS, -1
	      IF (X(J).LE.XX2) THEN
	         J2 = J
		 GO TO 200
	      END IF
	   END DO
	ELSE
	   DO J = JS, JE
	      IF (X(J).LE.XX2) THEN
	         J2 = J
		 GO TO 200
	      END IF
	   END DO
	END IF

	! x limits outside domain of data

	IERR = 1
	RETURN

 200	CONTINUE

C	Put indices in order

	IF (J2.LT.J1) THEN
	   J = J1
	   J1 = J2
	   J2 = J
	END IF

	RETURN
        END
C
C -----------------------------------------------------------------------
C
C       DETERM
C	
C       Purpose: Calculate the determinant of a square matrix.
C                This routine is taken from Bevington, p. 294.
C       Syntax: DET = DETERM (ARRAY, NORDER)
C       Input Parameters:
C         ARRAY  -- two dimensional, double precision, square
C                   matrix.
C         NORDER -- order of matrix (maximum is 10).
C       Returns:
C         DET will be a double precision number containing
C         the determinant.
C       Author: P.R. Bevington; re-written in VAX FORTRAN by K. Myneni.
C       Date: 17 July 90
C
        DOUBLE PRECISION FUNCTION DETERM (ARRAY, NORDER)

        DOUBLE PRECISION ARRAY(10, 10), SAVE

        DETERM = 1.D0

        DO K = 1, NORDER
C
C  Interchange columns if diagonal element is zero
C
           IF (ARRAY(K, K).EQ.0.D0) THEN

              DO J = K, NORDER
                 IF (ARRAY(K, J).EQ.0.D0) GO TO 10
              END DO

 10           CONTINUE

              IF (ARRAY(K, J).EQ.0.D0) THEN
                 DETERM = 0.D0
                 RETURN
              END IF

              DO I = K, NORDER
                 SAVE = ARRAY(I, J)
                 ARRAY(I, J) = ARRAY(I, K)
                 ARRAY(I, K) = SAVE
              END DO

              DETERM = -DETERM

           END IF
C
C  Subtract row K from lower rows to get diagonal matrix
C
           DETERM = DETERM * ARRAY(K, K)

           IF (K.LT.NORDER) THEN
              K1 = K + 1

              DO I = K1, NORDER
                 DO J = K1, NORDER

                    ARRAY(I, J) = ARRAY(I, J) - ARRAY(I, K) *
     +                            ARRAY(K, J) / ARRAY(K, K)

                 END DO
              END DO

           END IF

        END DO

        RETURN
        END
C
C ------------------------------------------------------------------------
C
C	HORNER
C
C       Purpose: Compute polynomial using Horner's algorithm.
C       Syntax: P = HORNER (A, NTERMS, X)
C       Input Parameters:
C         A      -- array of polynomial coefficients (A(1) is 0th order).
C         NTERMS -- number of terms in polynomial (order + 1).
C         X      -- real x value at which polynomial is evaluated.
C       Returns:
C         P is a real number containing the value of the polynomial
C         function.
C       Author: Horner
C       Date: 17 July 90
C
        FUNCTION HORNER (A, NTERMS, X)

        DIMENSION A(*)
        DOUBLE PRECISION P, XX

C  Perform calculation in double precision. Final result is single
C  precision.

        XX = DBLE(X)
        P = DBLE(A(NTERMS))

        DO J = NTERMS - 1, 1, -1
           P = DBLE(A(J)) + P * XX
        END DO

        HORNER = SNGL(P)

        RETURN
        END
C
C ----------------------------------------------------------------------
C
C	LINEAR_INTERP
C
C       Purpose: Perform linear interpolation to obtain Y, given X.
C       Syntax: CALL LINEAR_INTERP (X1, Y1, X2, Y2, X, Y, IERR)
C       Input Parameters:
C         X1, Y1 -- x, y coordinates of first endpoint.
C         X2, Y2 -- x, y coordinates of second endpoint.
C         X      -- x coordinate at which interpolation is desired.
C       Output Parameters:
C         IERR -- integer error code set to 1 when X1 = X2; 0 otherwise.
C         Y   -- interpolated y value for given x.
C       Author: K. Myneni
C       Date: 17 July 90
C
        SUBROUTINE LINEAR_INTERP (X1, Y1, X2, Y2, X, Y, IERR)


        IF (X2.EQ.X1) THEN
          IERR = 1
        ELSE
          Y = (Y2 * X - Y2 * X1 - Y1 * X + Y1 * X2) / (X2 - X1)
          IERR = 0
        END IF

        RETURN
        END
C
C --------------------------------------------------------------------
C
C 	POLFIT
C
C       Purpose: Perform a least-squares fit to data with a polynomial.
C       Syntax: CALL POLFIT (X, Y, J1, J2, NTERMS, A, CHISQR)
C       Input Parameters:
C         X     -- real array of x values.
C         Y     -- real array of y values.
C         J1    -- integer index of starting point for fit.
C         J2    -- integer index of ending point for fit.
C         NTERMS -- number of coefficients to be fitted
C                   (i.e. polynomial order + 1).
C       Output Parameters:
C         A     -- real array of fitted coefficients (A(1) is 0th order).
C         CHISQR -- real*16 value of reduced chi-square for fit.
C       Author: P.R. Bevington; see pp 140-141 in his book; translated
C               to VAX FORTRAN by K. Myneni.
C       Date: 17 July 90
C
        SUBROUTINE POLFIT (X, Y, J1, J2, NTERMS, A, CHISQR)

        REAL X(*), Y(*), A(*)
        DOUBLE PRECISION ARRAY(10, 10), DELTA, CHISQ
        DOUBLE PRECISION SUMX(19), SUMY(10), XTERM, YTERM
C
C  Accumulate weighted sums
C
        NPTS = J2 - J1 + 1
        NMAX = 2 * NTERMS - 1

        DO N = 1, NMAX
           SUMX(N) = 0.D0
        END DO

        DO J = 1, NTERMS
           SUMY(J) = 0.D0
        END DO

        CHISQ = 0.D0

        DO I = J1, J2
           XI = X(I)
           YI = Y(I)
C
C   No weighting is used.
C
           XTERM = 1.D0

           DO N = 1, NMAX
              SUMX(N) = SUMX(N) + XTERM
              XTERM = XTERM * DBLE(XI)
           END DO

           YTERM = YI

           DO N = 1, NTERMS
              SUMY(N) = SUMY(N) + YTERM
              YTERM = YTERM * DBLE(XI)
           END DO

           CHISQ = CHISQ + YI ** 2.
        END DO
C
C  Construct matrices and calculate coefficients.
C
        DO J = 1, NTERMS
           DO K = 1, NTERMS
              N = J + K - 1
              ARRAY(J, K) = SUMX(N)
           END DO
        END DO

        DELTA = DETERM(ARRAY, NTERMS)

        IF (DELTA.EQ.0.D0) THEN
           CHISQR = 0.

           DO J = 1, NTERMS
              A(J) = 0.
           END DO
        ELSE
           DO L = 1, NTERMS
              DO J = 1, NTERMS
                 DO K = 1, NTERMS
                    N = J + K - 1
                    ARRAY(J, K) = SUMX(N)
                 END DO
                 ARRAY(J, L) = SUMY(J)
              END DO

              A(L) = SNGL(DETERM(ARRAY, NTERMS) / DELTA)

           END DO
C
C   Calculate chi square.
C
           DO J = 1, NTERMS
              CHISQ = CHISQ - 2.*A(J)*SUMY(J)

              DO K = 1, NTERMS
                 N = J + K - 1

                 CHISQ = CHISQ + A(J)*A(K)*SUMX(N)

              END DO
           END DO

           FREE = FLOAT(NPTS - NTERMS)

           IF (FREE.EQ.0.) THEN
              CHISQR = 0.
           ELSE
              CHISQR = SNGL(CHISQ) / FREE
           END IF
        END IF

        RETURN
        END
C
C ------------------------------------------------------------------------
C
C	PARAB_AREA
C
C       Purpose: Find the area under a set of points using local
C	           parabolic fitting.
C	Syntax: CALL PARAB_AREA (X, Y, JS, JE, AREA)
C	Input Parameters:
C	  X, Y	 -- X and Y data arrays
C	  JS, JE -- starting and ending indices of points under which
C		      to compute area.
C	Output Parameters:
C	  AREA   -- double precision computed area.
C	Author: K. Myneni
C	Date: March 1990
C	Note: Translated from QuickBasic to VAX FORTRAN on 20 July 90.
C
        SUBROUTINE PARAB_AREA (X, Y, JS, JE, AREA)

        DIMENSION X(*), Y(*)
	DIMENSION XP(3), XPP(3), YP(3), A(4)
        DOUBLE PRECISION AREA, AREA_ELEMENT

        AREA = 0.D0

        XP(1) = X(JS)
        YP(1) = Y(JS)
        XP(2) = X(JS + 1)
        YP(2) = Y(JS + 1)
        XP(3) = X(JS + 2)
        YP(3) = Y(JS + 2)

        J = JS + 2

        DO WHILE (J.LE.JE)

C  Subtract offset from each x value to keep numbers small compared
C  to width of interval.

	  X_OFFSET = - XP(1)

	  DO K = 1, 3
	    XPP(K) = XP(K) + X_OFFSET
	  END DO
C
C  Perform parabolic fit to 3 points
C
           
           CALL POLFIT (XPP, YP, 1, 3, 3, A, CHISQR)
           
           IF (J.EQ.JE) THEN
              XM = XPP(3)
           ELSE
              XM = (XPP(3) + XPP(2)) / 2.
           END IF

           AREA_ELEMENT = A(1) * DBLE (XM) + A(2) * XM ** 2 / 2.D0
     +                    + A(3) * XM ** 3 / 3.D0
C
C  Accumulate area
C
           AREA = AREA + AREA_ELEMENT
C
C  Choose next 3 points for parabolic fit; first point is the
C    midpoint between last two fitting points.
C
           XP(1) = XM - X_OFFSET
           YP(1) = A(1) + A(2) * XM + A(3) * XM ** 2.

           XP(2) = X(J)
           YP(2) = Y(J)

           IF (J.LT.JE) THEN
              XP(3) = X(J + 1)
              YP(3) = Y(J + 1)
           END IF

           J = J + 1

        END DO

C  If data is ordered in decreasing x, then change sign of area.

	IF (X(JE).LT.X(JS)) THEN
	   AREA = -AREA
	END IF

        RETURN
        END
C
C ------------------------------------------------------------------------
C
C	DERIVATIVE                                      
C
C  	Purpose: Compute the numerical derivative of a data set.
C  	Syntax: CALL DERIVATIVE (X, Y, J1, J2, J3, IERR)        
C  	Input Parameters:                                       
C  	  X, Y    -- data arrays.                     
C         J1, J2  -- starting and ending indices of data set in  
C                      X, Y arrays.         
C         J3      -- starting index to store derivative.           
C    	Output Parameters:                                      
C         X, Y    -- contain derivative starting at index J3.
C         IERR    -- integer error code has the following meaning:      
C                      0  -- no error                 
C                      1  -- two points have same x value                                                           
C  	Author: Krishna Myneni                                  
C  	Date: 30 Dec 89                                         
C  	Note: Translated from QuickBasic to VAX FORTRAN on 30 Aug 90.
C                                                            

	SUBROUTINE DERIVATIVE (X, Y, J1, J2, J3, IERR)


	DIMENSION X(*), Y(*)
	REAL DX, DX1, DX2

        NPTS = J2 - J1 + 1

        DO J = 1, NPTS
           X(J3 + J - 1) = X(J1 + J - 1)
        END DO

        J4 = J3 + NPTS - 1
C
C  Calculate derivative at endpts with just two points.
C
        DX = X(J1 + 1) - X(J1)

        IF (DX.EQ.0.) THEN
           IERR = 1
           RETURN
        END IF

        Y(J3) = (Y(J1 + 1) - Y(J1)) / DX
        DX = X(J2) - X(J2 - 1)

        IF (DX.EQ.0.) THEN
           IERR = 1
           RETURN
        END IF

        Y(J4) = (Y(J2) - Y(J2 - 1)) / DX
C
C  Calculate derivative everywhere else with forward and backward points.
C
        DO J = 2, NPTS - 1
           JO = J1 + J - 1
           JN = J3 + J - 1
           DX1 = X(JN + 1) - X(JN)
           DX2 = X(JN) - X(JN - 1)

           IF ((DX1.EQ.0.).OR.(DX2.EQ.0.)) THEN
              IERR = 0
              RETURN
           END IF

           Y(JN) = ((Y(JO + 1) - Y(JO)) / DX1 +
     1              (Y(JO) - Y(JO - 1)) / DX2) / 2.
        END DO
C
C  Derivative computed successfully.
C
        IERR = 0

	RETURN
	END
C
C ---------------------------------------------------------------------
C
C	LINEAR_TRANSFORMATION
C
C	Purpose: Perform a linear transformation on an array
C	         of values to bring them within specified limits.
C
C	         x' = A*x + B
C
C	Syntax: CALL LINEAR_TRANSFORMATION (X, J1, J2, X1, X2, A, B)
C	Input Parameters:
C	  X	-- array of values to be transformed.
C	  J1	-- index of starting value.
C	  J2	-- index of ending value.
C	  X1	-- lower limit of transformation interval.
C	  X2    -- upper limit of transformation interval.
C	Output Parameters:
C	  X	-- transformed values between indices J1 and J2
C	  A	-- multiplicative constant in linear transformation.
C	  B	-- additive constant in linear transformation.
C
	SUBROUTINE LINEAR_TRANSFORMATION (X, J1, J2, X1, X2, A, B)

	DIMENSION X(*)

	IF (J2.LT.J1) THEN
	   JS = J1
	   J1 = J2
	   J2 = JS
	END IF

	XMIN = X(J1)
	XMAX = X(J1)

	DO J = J1, J2
	  IF (X(J).LT.XMIN) XMIN = X(J)
	  IF (X(J).GT.XMAX) XMAX = X(J)
	END DO

	IF (XMIN.EQ.XMAX) THEN
	  A = 1.                       ! no transformation
	  B = 0.
	  RETURN
	END IF

C  Determine transformation constants A and B.

	A = (X2 - X1)/(XMAX - XMIN)
	B = X1 - A*XMIN

C  Apply transformation

	DO J = J1, J2
	  X(J) = A*X(J) + B
	END DO

	RETURN
	END
C
C --------------------------------------------------------------------
C
C	BINOMIAL_COEFFICIENT
C
C	Purpose: Return the binomial coefficient (N L).
C	Syntax: BC = BINOMIAL_COEFFICIENT (N, L)
C	Arguments:
C	  N, L  -- positive integer arguments
C	Returns:
C	  Floating point binomial coefficient
C	Author: K. Myneni
C	Date: 11 Sep 90
C
	REAL FUNCTION BINOMIAL_COEFFICIENT (N, L)

	IF (N.LT.L) THEN
	  BINOMIAL_COEFFICIENT = 0.
	ELSE IF (N.EQ.L) THEN
	  BINOMIAL_COEFFICIENT = 1.
	ELSE
	  FNUMERATOR = FACTORIAL (N)
	  FDENOMINATOR = FACTORIAL (L) * FACTORIAL (N - L)
	  BINOMIAL_COEFFICIENT = FNUMERATOR / FDENOMINATOR
	END IF

	RETURN
	END
C
C -------------------------------------------------------------------
C
C	FACTORIAL
C
C	Purpose: Return factorial of a positive integer.
C	Syntax: F = FACTORIAL (J)
C	Arguments:
C	  J  -- positive integer
C	Returns:
C	  Floating point factorial
C	Author: K. Myneni
C	Date: 11 Sep 90
C	
	REAL FUNCTION FACTORIAL (J)

	IF (J.LT.0) THEN
	  FACTORIAL = 0.
	ELSE IF (J.LT.2) THEN
	  FACTORIAL = 1.
        ELSE
	  F = 1.
	  DO K = 2, J
	    F = F * SNGL(K)
	  END DO
	  FACTORIAL = F
	END IF

	RETURN
	END
C
C -------------------------------------------------------------------
C
C	TRANSFORM_COEFFICIENTS
C
C	Purpose: Transform the coefficients of a polynomial that
C	         has undergone a linear transformation in its
C	         X values.
C	Syntax: CALL TRANSFORM_COEFFICIENTS (A, AP, N, C1, C2)
C	Input Parameters:
C	  AP	 --  array of polynomial coefficients in the 
C	             transformed space.
C	  N	 --  polynomial order
C	  C1, C2 --  constants of the linear transformation:
C			X' = C1*X + C2
C	Output Parameters:
C	  A      --  array of polynomial coefficients in normal space.
C	Author: K. Myneni
C	Date: 11 Sep 90
C
	SUBROUTINE TRANSFORM_COEFFICIENTS (A, AP, N, C1, C2)

	REAL A(*), AP(*)

	IF (N.LT.0) RETURN

	IF (C1.EQ.0.) THEN
	  DO K = 0, N
	    A(K + 1) = 0.
	  END DO
	ELSE
	  CP = C2 / C1

	  DO J = 0, N
	    SUM = 0.
	    DO L = J, N
	      AJ = AP(L + 1) * BINOMIAL_COEFFICIENT(L, J)

	      ! Trap the case where CP is 0. (i.e., when C2 is 0.)
	      ! to avoid 0. ** 0

	      IF (L.EQ.J) THEN
	        CPFACTOR = 1.
	      ELSE
	        CPFACTOR = CP ** (L - J)
	      END IF

	      AJ = AJ * (C1 ** L) * CPFACTOR
	      SUM = SUM + AJ
	    END DO
	    A(J + 1) = SUM
	  END DO

        END IF

	RETURN
	END
C
C -------------------------------------------------------------------------
C	
