C **********************************************************************
C
C       VAXFILE
C
C         This file contains a collection of subroutines for reading
C       and writing VAXPLOT style data files. The documentation
C       for these modules may be found in the file, VAXFILE.DOC.
C
C         Modules in this file may be used in FORTRAN programs by
C       linking with VAXFILE.OBJ.
C
C       The following modules are included:
C
C         DELETE_FILE
C         INPUT_HEADER
C         NEXT_AVAILABLE_UNIT
C         READ_BLANK_LINES
C         READ_HEADER
C         TYPE_HEADER
C         TYPE_XYREAD_ERROR
C         WRITE_HEADER
C         WRITE_XYFILE
C         XYREAD
C         XYWRITE
C         String Utility Modules
C
C       Created: 19 July 90, K. Myneni
C       Modified: 19 July 90 -- KM
C       Modified: 20 July 90 -- KM
C       Modified: 22 July 90 -- KM; added DELETE_FILE and
C                                   TYPE_XYREAD_ERROR
C       Modified: 23 July 90 -- KM; added INPUT_HEADER
C       Modified: 24 July 90 -- KM; added WRITE_XYFILE
C
C **********************************************************************
C
C       NEXT_AVAILABLE_UNIT
C
C       Purpose: Determine next available logical unit number.
C                Calling this function to determine the unit number
C                before executing an OPEN statement will avoid
C                potential conflicts with files left open by
C                other programs.
C       Syntax: ILU = NEXT_AVAILABLE_UNIT()
C       Input Parameters:
C         NONE
C       Returns:
C         An integer specifying the logical unit number to be used
C         in a subsequent OPEN statement.
C       Author: K. Myneni
C       Date: 19 July 90
C
        INTEGER FUNCTION NEXT_AVAILABLE_UNIT

        LOGICAL UNIT_IS_OPEN, UNIT_IS_CLOSED

        NEXT_AVAILABLE_UNIT = 0

C   Start search at unit 22; max unit number is 99.

        DO J = 22, 99
           INQUIRE (J, OPENED = UNIT_IS_OPEN)
           UNIT_IS_CLOSED = .NOT.UNIT_IS_OPEN

           IF (UNIT_IS_CLOSED) THEN
              NEXT_AVAILABLE_UNIT = J
              RETURN
           END IF

        END DO

        RETURN
        END
C
C -----------------------------------------------------------------------
C
C       DELETE_FILE
C
C       Purpose: Delete an existing file.
C       Syntax: CALL DELETE_FILE (FILENAME, IERR)
C       Input Parameters:
C         FILENAME  -- character substring containing name of file
C                      to be deleted.
C       Output Parameters:
C         IERR  --  error code has the following meaning:
C                       0  -- no error
C                       1  -- file does not exist
C                       2  -- unable to delete file
C       Author: K. Myneni
C       Date: 22 July 90
C
        SUBROUTINE DELETE_FILE (FILENAME, IERR)

        CHARACTER*(*) FILENAME
        LOGICAL FILE_EXISTS

        IERR = 0

C       Check to see if file exists.

        INQUIRE (FILE = FILENAME(1:), EXIST = FILE_EXISTS)
        IF (.NOT.FILE_EXISTS) THEN
           IERR = 1
           RETURN
        END IF

C       Open and delete file.

        ILU = NEXT_AVAILABLE_UNIT()
        OPEN (UNIT = ILU, FILE = FILENAME(1:), ERR = 100,
     +        STATUS = 'UNKNOWN', DISPOSE = 'DELETE')

C       Successfull delete upon closing unit.

        CLOSE (ILU)
        RETURN

 100    CONTINUE

C       Error occured; Unable to open/delete the file.

        IERR = 2

        RETURN
        END
C
C -----------------------------------------------------------------------
C
C
C     XYREAD
C
C     Purpose: Read a VAXPLOT data file.  
C     Syntax: CALL XYREAD (X, Y, JSTART, JMAX, NPTS, IERR, FILENAME, HEADER)
C
C     Input Parameters:
C
C        X, Y                  -- data arrays
C        JSTART                -- starting index at which x, y
C                                 values are stored
C        JMAX                  -- maximum index allowed
C        FILENAME              -- name of input file
C
C     Output Parameters:
C
C        X, Y                  -- data arrays containing new data
C        NPTS                  -- number of points read from input
C                                 file
C        HEADER                -- text header of input file
C        IERR                  -- error number has the following
C                                 values:
C
C             0                -- no error
C             1                -- error opening file
C             2                -- no data found in file
C             3                -- first line of data not found
C             4                -- too many columns of data
C             5                -- too many points; index past JMAX
C             6                -- file does not exist
C
C     Authors: K. Myneni and J. Kielkopf
C     Date: 8 Jan 90
C     Modified: 20 July 90 to check for file existence -- KM
C
        SUBROUTINE XYREAD (X, Y, JSTART, JMAX, NPTS,
     +                     IERR, FILENAME, HEADER)

        DIMENSION X(*), Y(*)
        CHARACTER*(*) HEADER
        CHARACTER*(*) FILENAME
        CHARACTER*80 CLINE
        LOGICAL FILE_EXISTS
C
C  Check to see that file exists.
C
        INQUIRE (FILE = FILENAME(1:), EXIST = FILE_EXISTS)
        IF (.NOT.FILE_EXISTS) THEN
           IERR = 6
           RETURN
        END IF
C
C  Open next available unit.
C
        ILU = NEXT_AVAILABLE_UNIT()

        OPEN (UNIT = ILU, FILE = FILENAME(1:), READONLY,
     +        STATUS = 'OLD', ERR = 1000)
C
C     Read any empty lines at beginning of file
C
        CALL READ_BLANK_LINES (ILU, N, IERR)
        IF (IERR.NE.0) GO TO 150
C
C     File pointer at first non-blank line
C
        CALL READ_HEADER (ILU, HEADER)
C
C     Read any empty lines between header and first line of data
C
        CALL READ_BLANK_LINES (ILU, N, IERR)
        IF (IERR.NE.0) GO TO 150
C
C     How many columns of numbers in file?
C
        READ(ILU, 100) (CLINE(J:J), J = 1, 80)
 100    FORMAT(80A1)
        BACKSPACE(ILU)

        NCOLS = NNUMS(CLINE)

        IF (NCOLS.EQ.0) THEN
C
C        First line of data was not found.  
C
           IERR = 3
           GO TO 210

        ELSEIF (NCOLS.EQ.1) THEN
C
C        Read y values, and make x a running index.                      
C
           DO J = JSTART, JMAX
              READ (ILU, *, END = 200) Y(J)
              X(J) = FLOAT (J - J1 + 1)
           END DO

        ELSEIF (NCOLS.EQ.2) THEN
C
C        Read x, y pairs.                                                
C
           DO J = JSTART, JMAX
              READ (ILU, *, END = 200) X(J), Y(J)
           END DO

        ELSE
C
C        Too many columns of data.  
C
           IERR = 4
           GO TO 210

        ENDIF
C
C     Too many points. Buffer exceeded.  
C
        NPTS = J - JSTART
        IERR = 5
        GO TO 210
C
C     No data in file
C
 150    CLOSE(ILU)
        IERR = 2
        RETURN
C
C     File successfully read
C
 200    IERR = 0
        NPTS = J - JSTART
C
C     Close file and return
C
 210    CLOSE (ILU)
        RETURN
C
C     Cannot open file.  
C
 1000   IERR = 1
        NPTS = 0

        RETURN
        END
C
C --------------------------------------------------------------------
C
C       TYPE_XYREAD_ERROR
C
C       Purpose: Display an error message given by subroutine
C                XYREAD
C       Syntax: CALL TYPE_XYREAD_ERROR (IERR)
C       Input Parameters:
C         IERR -- integer error code given by XYREAD
C       Output Parameters: NONE
C       Author: K. Myneni
C       Date: 22 July 90
C
        SUBROUTINE TYPE_XYREAD_ERROR (IERR)

        IF (IERR.EQ.0) THEN
           RETURN
        END IF

        GO TO (100, 200, 300, 400, 500, 600) IERR

        TYPE *, 'Undiagnosed error in module XYREAD.'
        GO TO 1000

 100    TYPE *, 'Error opening file.'
        GO TO 1000
 200    TYPE *, 'No data found in file.'
        GO TO 1000
 300    TYPE *, 'First line of data not found.'
        GO TO 1000
 400    TYPE *, 'Too many columns of data.'
        GO TO 1000
 500    TYPE *, 'Too many points in file for given input array.'
        GO TO 1000
 600    TYPE *, 'Input file does not exist.'
        GO TO 1000

 1000   CONTINUE

        RETURN
        END
C
C ---------------------------------------------------------------------
C
C       XYWRITE
C
C       Purpose: Write the X, Y data between specified indices
C                and a text header to an output file.
C       Syntax: CALL XYWRITE (X, Y, JS, JE, HEADER, FNAME, IERR)
C       Input Parameters:
C         X, Y    -- data arrays
C         JS, JE  -- starting and ending indices
C         HEADER  -- character string containing header text
C         FNAME   -- output filename
C       Output Parameters:
C         IERR    -- error code has the following meaning.
C                      0  -- no error
C                      1  -- file already exists
C       Authors: J. Kielkopf and K. Myneni
C       Date: 12 Feb  90
C       Modified: 19 July 90 (original module name was COPY_TO_FILE) -- KM
C       Modified: 20 July 90 to check for file existence -- KM
C
        SUBROUTINE XYWRITE (X, Y, JS, JE, HEADER, FNAME, IERR)

        CHARACTER*(*) FNAME
        CHARACTER*(*) HEADER     
        DIMENSION X(*), Y(*)
        LOGICAL FILE_EXISTS
C
C  Check to see if output file already exists; if so, return with error.
C
        INQUIRE (FILE = FNAME(1:), EXIST = FILE_EXISTS)
        IF (FILE_EXISTS) THEN
           IERR = 1
           RETURN
        END IF

        ILU = NEXT_AVAILABLE_UNIT()
        OPEN(UNIT = ILU, FILE = FNAME(1:), STATUS = 'NEW')

        CALL WRITE_HEADER (ILU, HEADER)
C
C       Write x,y pairs in exponential format
C
        DO J = JS, JE
          WRITE(ILU, 1030) X(J), Y(J)
 1030     FORMAT(1X, 2E)
        END DO

        CLOSE(ILU)
        IERR = 0

        RETURN
        END
C
C ---------------------------------------------------------------------
C
C       WRITE_XYFILE
C
C       Purpose: Write a VAXPLOT style x, y data file with
C                error checking for existing file.
C       Syntax: CALL WRITE_XYFILE (X, Y, J1, J2, HEADER, PROMPT)
C       Input Parameters:
C         X, Y          -- x and y data arrays
C         J1, J2        -- starting and ending indices of data
C                          to be output to file.
C         HEADER        -- character substring containing header
C                          text to be output to file; must be
C                          terminated with null character (CHAR(0)).
C         PROMPT        -- character substring containing prompt
C                          for output filename.
C       Output Parameters: NONE
C       Author: K. Myneni
C       Date: 24 July 90
C
        SUBROUTINE WRITE_XYFILE (X, Y, J1, J2, HEADER, PROMPT)

        DIMENSION X(*), Y(*)
        CHARACTER*(*) HEADER, PROMPT
        CHARACTER*80 FNAME
        CHARACTER*4 ANSW

        IRP = NRTRIM (PROMPT)

 50     CONTINUE

        WRITE(6, 100) PROMPT(1:IRP + 1)
 100	FORMAT(1X, A, $)

        READ (5, 200) (FNAME(J:J), J = 1, 80)
 200    FORMAT(80A1)

 300    CONTINUE

        CALL XYWRITE (X, Y, J1, J2, HEADER, FNAME, IERR)

C       Check to see if file already exists error occured

        IF (IERR.EQ.1) THEN
           WRITE (6, 400)
           READ (5, 500) (ANSW(J:J), J = 1, 4)
           CALL STR_UPPER_CASE (ANSW)

           IF (ANSW(1:1).EQ.'Y') THEN
              CALL DELETE_FILE (FNAME, IERR)
              IF (IERR.NE.0) THEN
                 TYPE *, 'Error deleting old file.'
                 TYPE *, 'Please enter a new file name.'
                 GO TO 50
              END IF
              GO TO 300
           ELSE
              TYPE *, 'Please enter a new file name.'
              GO TO 50
           END IF
        END IF

 400   FORMAT(' Output file already exists. Overwrite? '$)
 500   FORMAT(4A1)

        RETURN
        END
C
C --------------------------------------------------------------------
C
C       INPUT_HEADER
C
C       Purpose: Allow user to input header text from keyboard.
C       Syntax: CALL INPUT_HEADER (HEADER)
C       Input Parameters:
C         NONE
C       Output Parameters:
C         HEADER  --  character substring containing header text.
C       Author: K. Myneni
C       Date: 23 July 90
C
        SUBROUTINE INPUT_HEADER (HEADER)

        CHARACTER*(*) HEADER
        CHARACTER*80 CLINE
        CHARACTER*5 END_TEST

C       Request a file header

        CALL FILL_STRING (HEADER, ' ')
        CALL FILL_STRING (CLINE, '-')
        TYPE *, 'Please enter the header text.'
        TYPE *, 'Enter "/EXIT" on a new line when finished.'
        TYPE *, CLINE(1:80)
        CALL FILL_STRING (CLINE, ' ')

        K = 1

        DO WHILE (.TRUE.)
           READ (5, 100) (CLINE(J:J), J = 1, 80)

C          Test for end of input

           END_TEST(1:5) = CLINE(1:5)
           CALL STR_UPPER_CASE (END_TEST)
           IF (END_TEST(1:5).EQ.'/EXIT') GO TO 500

           IRC = NRTRIM (CLINE)

C          Copy each input line into header

           DO L = 1, IRC
             LK1 = L + K - 1
             HEADER(LK1:LK1) = CLINE(L:L)
           END DO

           K1 = K + IRC
           K2 = K + IRC + 1

C          Add CR, LF at end of each line

           HEADER(K1:K1) = CHAR(13)
           HEADER(K2:K2) = CHAR(10)

           K = K2 + 1
        END DO

 100    FORMAT (80A1)

 500    CONTINUE

C       End of input; add null character to end of header

        HEADER(K:K) = CHAR(0)

        RETURN
        END
C
C --------------------------------------------------------------------
C
C     READ_HEADER
C
C     Purpose: Read the text header of a VAXPLOT file.  
C     Syntax: CALL READ_HEADER (ILU, HEADER)
C     Input Parameters:
C       ILU -- integer specifying logical unit number of input file.
C     Output Parameters:
C       HEADER -- contains the header text as a null terminated
C                 character substring.
C     Author: K. Myneni
C     Date: 8 Jan 90
C     Updated: 22 Feb 90
C     Modified: 19 July 90 -- KM
C
        SUBROUTINE READ_HEADER (ILU, HEADER)

        CHARACTER*(*) HEADER
        CHARACTER*80 CLINE

        L = LEN (HEADER)

        ! initialize header to spaces

        CALL FILL_STRING (HEADER, ' ')

        J = 0
 100    READ (ILU, 110) (CLINE (K : K), K = 1, 80)
 110    FORMAT (80A1)
        IPOS = INDEX (CLINE, '/*')
        IF (IPOS.NE.0) THEN
C
C        Read a VAXPLOT header
C
           DO WHILE (.TRUE.)
              READ (ILU, 110) (CLINE (K : K), K = 1, 80)
              IPOS = INDEX (CLINE, '*/')
              IF (IPOS.EQ.0) THEN
                 L = NRTRIM (CLINE)
                 DO K = 1, L
                    M = J + K
                    HEADER (M:M) = CLINE(K:K)
                 END DO
                 J = J + L + 1
                 HEADER(J:J) = CHAR(13)
                 J = J + 1
                 HEADER(J:J) = CHAR(10)
              ELSE
C
C              End of Header
C
                 HEADER(J:J) = CHAR(0)
                 RETURN
              ENDIF
           END DO
        ELSE
C
C        File does not contain a VAXPLOT header
C
           HEADER(1:1) = CHAR(0)
           BACKSPACE (ILU)
           RETURN
        ENDIF

        GOTO 100
        END
C
C ---------------------------------------------------------------------
C
C     READ_BLANK_LINES
C
C     Purpose: Read blank lines from file.  
C     Syntax: CALL READ_BLANK_LINES (ILU, N, IERR)
C     Input Parameters:
C        ILU -- integer specifying logical unit number of input file.
C     Output Parameters:
C        N      -- number of blank lines read.
C        IERR   -- error code:
C                    0 -- no error
C                    1 -- EOF before finding non-blank line
C        The file pointer is positioned at the first non-blank line
C        encountered.                                                    
C
C     Author: K. Myneni
C     Date: 11 Jan 90
C     Modified: 19 July 90 -- KM
C
      SUBROUTINE READ_BLANK_LINES (ILU, N, IERR)

      LOGICAL BLANK_LINE
      CHARACTER*80 CLINE

      N = 0
      IERR = 0

      DO WHILE (.TRUE.)
         READ(ILU, 100, END = 40)(CLINE(J:J), J = 1, 80)
         IF (.NOT.BLANK_LINE(CLINE)) THEN
            BACKSPACE(ILU)
            RETURN
         END IF
         N = N + 1
      END DO

C
C     End of File reached
C
 40   IERR = 1
      RETURN

 100  FORMAT(80A1)

      END
C
C ---------------------------------------------------------------------
C
C       TYPE_HEADER
C
C       Purpose: Display a VAXPLOT file header on terminal.
C       Syntax: CALL TYPE_HEADER (HEADER)
C       Input Parameters:
C          HEADER -- null terminated character substring containing
C                    header text.
C       Output Parameters: NONE
C       Author: K. Myneni
C       Date: 19 July 90
C	Modified: 20 July 90 -- KM
C
        SUBROUTINE TYPE_HEADER (HEADER)

        CHARACTER*(*) HEADER
      
        L = NTSLEN(HEADER)
        IF (L.EQ.0) GO TO 600

        K = 1

        DO WHILE (.TRUE.)
           NK = INDEX(HEADER(K:L), CHAR(13))
           IF (NK.EQ.0) GO TO 500
           MK = NK + K - 1
           WRITE(6, *) HEADER(K:MK-1)
           K = MK + 2
           IF (K.GT.L) GO TO 600
        END DO

 500    WRITE(6,*) HEADER(K:L)
 600    CONTINUE

        RETURN
        END
C
C ------------------------------------------------------------------------
C
C       WRITE_HEADER
C
C       Purpose: Write a VAXPLOT header to output file.
C       Syntax: CALL WRITE_HEADER (ILU, HEADER)
C       Input Parameters:
C         ILU     -- integer specifying logical unit number of
C                    output file.
C         HEADER  -- null terminated character substring containing
C                    header text.
C       Output Parameters:
C         NONE
C       Author: K. Myneni
C       Date: 19 July 90
C
        SUBROUTINE WRITE_HEADER (ILU, HEADER)

        CHARACTER*(*) HEADER

        L = NTSLEN(HEADER)
C                              
C       Write header to output file
C
        WRITE(ILU, 1000)
        
        K = 1

        DO WHILE (.TRUE.)
           NK = INDEX(HEADER(K:L), CHAR(13))
           IF (NK.EQ.0) GO TO 500
           MK = NK + K - 1
           WRITE(ILU, *) HEADER(K:MK - 1)
           K = MK + 2
           IF (K.GE.L) GO TO 600
        END DO

 500    WRITE(ILU, *) HEADER(K + 2:L)
 600    CONTINUE 

        WRITE(ILU, 1010)
 1000   FORMAT(' /*')
 1010   FORMAT(' */')

        RETURN
        END
C
C ***********************************************************************
C
C
C     String Utility Functions
C
C     14 January 1990
C
C
C     Last Update: 22 Feb 90
C                        
C *********************************************************************
C
C     FFORMAT
C
C     Purpose: Format a real number as a string with specified
C              number of decimal places.  
C     Syntax: CALL FFORMAT (X, N, F)
C     Input Parameters:
C        X    -- real number
C        N    -- number of decimal places
C     Output Parameters:
C        F    -- 13 character substring
C     Author: J. Kielkopf
C     Date:
C
        SUBROUTINE FFORMAT (X, N, F)

        CHARACTER*13 F
        DIMENSION IX(7)

        F(1:13)='             '

        IF(N.GT.7.OR.N.LT.0) RETURN
        XMAX=10.**(7-N)
        IF(ABS(X).GT.XMAX) RETURN
        IXP=ABS(X)*10.**N
        DO J=1,7
          IP=10**(7-J)
          IX(J)=IXP/IP
          IXP=IXP-IX(J)*IP
        END DO

        IF(X.LT.0.) THEN
          F(1:1)='-'
        ELSE
          F(1:1)=' '
        END IF

        ND=9-N
        NDL1=ND-1
        NDP1=ND+1
        I=0
        DO J=2,NDL1
          I=I+1
          IC=IX(I)+48
          F(J:J)=CHAR(IC)
        END DO

        F(ND:ND)='.'

        IF(NDP1.GT.9) RETURN
        DO J=NDP1,9
          I=I+1
          IC=IX(I)+48
          F(J:J)=CHAR(IC)
        END DO

        RETURN
        END
C
C ---------------------------------------------------------------------
C
        SUBROUTINE EFORMAT (X, N, F)

C        Formats the real number x as a exponential in string f

        CHARACTER*13 F

        IF(X.EQ.0.) THEN
          F(1:13)=' 000000000E+00'
          RETURN
        END IF

        BX=ABS(X)
        AX=ALOG10(BX)
        IE=AX
        IF(IE.LE.0)IE=IE-1
        AX=X/10.**IE
        CALL FFORMAT (AX, N, F)
        F(10:10)='E'
        F(11:11)='+'
        IF(IE.LT.0) F(11:11)='-'
        IE=IABS(IE)
        IE1=IE/10
        IE2=IE-10*IE1
        IE1=IE1+48
        IE2=IE2+48
        F(12:12)=CHAR(IE1)
        F(13:13)=CHAR(IE2)

        RETURN
        END
C
C ---------------------------------------------------------------------
C
C     NTSLEN
C
C     Purpose: Determine the length of a null-terminated string.  
C     Syntax: L = NTSLEN (STRING)
C     Input Parameters:
C        STRING -- a null terminated character substring.
C
C     Returns: number of characters (excluding null character).         
C     Author: K. Myneni
C     Date: 9 Jan 90
C
      INTEGER FUNCTION NTSLEN (STRING)

      CHARACTER*(*) STRING

      J = INDEX(STRING, CHAR(0))
      IF (J.EQ.0) THEN
         NTSLEN = 0
      ELSE
         NTSLEN = J - 1
      END IF
      RETURN
      END
C
C ---------------------------------------------------------------------
C
C     NRTRIM
C
C     Purpose: Find the last position in a string beyond which
C              all other characters are blanks.  
C
C     Syntax: L = NRTRIM (STRING)
C
C     Input Parameters:
C        STRING -- a FORTRAN character substring.
C     Returns:
C        Right-most position for trimming string.  
C     Author: K. Myneni
C     Date: 9 Jan 90
C
      INTEGER FUNCTION NRTRIM (STRING)

      CHARACTER*(*) STRING

      J = LEN (STRING)
      DO K = J, 1, -1
         IF (STRING(K:K).NE.' ') THEN
            NRTRIM = K
            RETURN
         ENDIF
      END DO

      NRTRIM = 1
      RETURN
      END
C
C ---------------------------------------------------------------------
C
C     NLTRIM
C
C     Purpose: Find the first position in a string before which
C              all other characters are blanks.  
C
C     Syntax: L = NLTRIM (STRING)
C
C     Input Parameters:
C        STRING -- a FORTRAN character substring.
C     Returns:
C        Left-most position for trimming string.  
C     Author: K. Myneni
C     Date: 10 Jan 90
C
      INTEGER FUNCTION NLTRIM (STRING)

      CHARACTER*(*) STRING

      J = LEN (STRING)
      DO K = 1, J
         IF (STRING(K:K).NE.' ') THEN
            NLTRIM = K
            RETURN
         ENDIF
      END DO

      NLTRIM = J
      RETURN
      END
C
C ---------------------------------------------------------------------
C
C     NNUMS
C
C     Purpose: Determine how many numbers are in an input string.         
C     Syntax: N = NNUMS (STRING)
C     Input Parameters:
C        STRING -- a FORTRAN character substring
C     Returns:
C        number of items separated by blanks, commas, or tabs.  
C     Author: K. Myneni
C     Date: 10 Jan 90
C     Modified: 21 Feb 90  to fix DO loop index modification error -- KM
C
      INTEGER FUNCTION NNUMS (STRING)

      CHARACTER*(*) STRING
      CHARACTER AC

      J1 = NLTRIM (STRING)
      J2 = NRTRIM (STRING)

      IF (J2.EQ.1) THEN
         NNUMS = 0
         RETURN
      ELSE
         NNUMS = 1
      ENDIF
      
      K = J1

      DO WHILE (K.LT.J2)
         AC = STRING(K:K)
         IF ((AC.EQ.' ').OR.(AC.EQ.',').OR.(AC.EQ.CHAR(9))) THEN
            NNUMS = NNUMS + 1
            DO L = K, J2
               AC = STRING(L:L)
               IF ((AC.NE.' ').AND.(AC.NE.',').AND.  
     +            (AC.NE.CHAR(9))) GOTO 100
            END DO
 100        K = L
         ELSE
            K = K + 1
         END IF
      END DO

      RETURN
      END
C
C ---------------------------------------------------------------------
C
C     BLANK_LINE
C
C     Purpose: Determine if a string is completely full of blanks.  
C     Syntax: LB = BLANK_LINE (STRING)
C     Input Parameters:
C        STRING -- a FORTRAN character substring.
C     Returns: a LOGICAL true or false (LB must be of type LOGICAL).      
C     Author: K. Myneni
C     Date: 11 Jan 90
C
      LOGICAL FUNCTION BLANK_LINE (STRING)

      CHARACTER*(*) STRING

      IF (NRTRIM (STRING).EQ.1) THEN
         BLANK_LINE = .TRUE. 
      ELSE
         BLANK_LINE = .FALSE. 
      END IF

      RETURN
      END
C
C ---------------------------------------------------------------------
C
C     STRVAL
C
C     Purpose: Return the numeric value of a character substring
C              representation of a number.                               
C     Syntax: X = STRVAL (F)
C     Input Parameters:
C        F    --  character substring (number may be written in
C                 exponential notation).                                 
C     Returns:
C        real number.  
C     Author: K. Myneni
C     Date: 15 Jan 90
C
      REAL FUNCTION STRVAL (F)

      CHARACTER*(*) F
      CHARACTER AC
      LOGICAL ISIGN

      X = 0.                   ! initialize value to 0.                  
      XE = 0.                  ! initialize exponent to 0.               
      ISIGN = .TRUE.           ! initialize sign to be positive.         

      JS = NLTRIM (F)
      JE = NRTRIM (F)

      K = 0

      DO J = JE, JS, -1
         AC = F (J:J)
         I = ICHAR(AC)
         IF (AC.EQ.'-') THEN
            ISIGN = .FALSE.                                           
         ELSE IF (AC.EQ.'+') THEN
            ISIGN = .TRUE.                                           
         ELSE IF ((I.GT.47).AND.(I.LT.58)) THEN
            X = X + FLOAT(I - 48)*(10.**K)
            K = K + 1
         ELSE IF (AC.EQ.'.') THEN
            X = X / 10.**K
            K = 0
         ELSE IF ((AC.EQ.'E').OR.(AC.EQ.'D')) THEN
            XE = X                                  ! set exponent,
            IF (.NOT.ISIGN) XE = -XE                ! and re-initialize
            X = 0.                                  ! to read mantissa
            K = 0
            ISIGN = .TRUE.  
         ELSE
            STRVAL = 0.        ! illegal character in string
            RETURN             ! return 0. value
         END IF
      END DO

      IF (.NOT.ISIGN) X = -X
      STRVAL = X * (10.**XE)

      RETURN
      END
C
C ---------------------------------------------------------------------
C
C     STR_UPPER_CASE
C
C     Purpose: Convert a character substring to upper case.  
C     Syntax: CALL STR_UPPER_CASE (STRING)
C     Input Parameters:
C        STRING -- character substring
C     Output Parameters:
C        STRING -- all lower case characters (a-z) converted to
C                  upper case (all other characters unchanged).  
C     Author: K. Myneni
C     Date: 17 Jan 90
C
      SUBROUTINE STR_UPPER_CASE (STRING)

      CHARACTER*(*) STRING

      L = LEN (STRING)

      DO J = 1, L
         IC = ICHAR (STRING(J:J))
         IF ((IC.GE.97).AND.(IC.LE.122)) THEN
            STRING(J:J) = CHAR (IC - 32)
         END IF  
      END DO

      RETURN
      END
C
C ---------------------------------------------------------------------
C
C     ADD_NTSTRINGS
C
C     Purpose: Add one null terminated string to another.                
C     Syntax: CALL  ADD_NTSTRINGS (STR1, STR2, I)
C     Input Parameters:
C        STR1 -- null terminated string.                   
C        STR2 -- null terminated string to be added to STR1.           
C     Output Parameters:
C        STR1 -- new STR1.                          
C        I -- null terminated string length of modified STR1.
C     Author: K. Myneni
C     Date: 20 Jan 90
C
      SUBROUTINE ADD_NTSTRINGS (STR1, STR2, I)

      CHARACTER*(*) STR1
      CHARACTER*(*) STR2

      L1 = LEN (STR1)
      N1 = NTSLEN (STR1)

      N2 = NTSLEN (STR2)
      NT = N1 + N2

      IF (N2.EQ.0) THEN
         I = N1
      ELSE IF (N2.GE.L1) THEN
         STR1(1:) = STR2(1:)
         STR1(L1:L1) = CHAR(0)
         I = L1 - 1
      ELSE
         CALL SCROLL_STRING (STR1, N2, ' ')
         STR1(1:N2) = STR2(1:N2)
         STR1(L1:L1) = CHAR(0)
         I = NTSLEN (STR1)
      END IF

      RETURN
      END
C
C ---------------------------------------------------------------------
C
C     SCROLL_STRING
C
C     Purpose: scroll a character substring by a given number of
C              characters, filling the empty positions with a
C              fill character.  
C     Syntax: CALL SCROLL_STRING (STRING, N, FILL_CHAR)
C     Input Parameters:
C        STRING -- character substring.  
C        N -- number of positions to scroll;
C             positive N scrolls to the right; negative scrolls to left.          
C        FILL_CHAR -- character to fill vacant positions.  
C     Output Parameters:
C        STRING -- scrolled string.  
C     Author: K. Myneni
C     Date: 20 Jan 90
C
      SUBROUTINE SCROLL_STRING (STRING, N, FILL_CHAR)

      CHARACTER*(*) STRING
      CHARACTER FILL_CHAR

      L = LEN (STRING)

      IF (N.EQ.0) RETURN

      IF (ABS(N).GE.L) THEN
         DO J = 1, L
            STRING(J:J) = FILL_CHAR
         END DO
         RETURN
      END IF

      IF (N.GT.0) THEN
         DO J = L - N, 1, -1
            K = J + N
            STRING(K:K) = STRING(J:J)
         END DO
         DO J = 1, N
            STRING(J:J) = FILL_CHAR
         END DO
      ELSE
         M = ABS(N)
         DO J = M + 1, L
            K = J - M
            STRING(K:K) = STRING(J:J)
         END DO
         DO J = L - M + 1, L
            STRING(J:J) = FILL_CHAR
         END DO
      END IF

      RETURN
      END
C
C ---------------------------------------------------------------------
C
C     STRIP_LEADING_ZEROS
C
C     Purpose: Strip leading zeros from the output character string
C              produced by subroutines EFORMAT and FFORMAT.
C     Syntax: CALL STRIP_LEADING_ZEROS (STRING)
C     Input Parameters:
C       STRING  -- character substring containing number.
C     Output Parameters:
C       STRING  -- leading zeros are replaced by spaces.
C     Author: K. Myneni
C     Date:
C   
      SUBROUTINE STRIP_LEADING_ZEROS (STRING)

      CHARACTER*(*) STRING
      CHARACTER AC
      CHARACTER SC
      LOGICAL INSERT_SIGN
                         
      INSERT_SIGN = .FALSE.

      L = LEN(STRING)
      
      DO J = 1, L
         AC = STRING(J:J)
         IF (.NOT.((AC.EQ.' ').OR.(AC.EQ.'0').OR.(AC.EQ.'-')
     +             .OR.(AC.EQ.'+'))) GO TO 100
         IF ((AC.EQ.'-').OR.(AC.EQ.'+')) THEN
            SC = AC
            INSERT_SIGN = .TRUE.
            STRING(J:J) = ' '
         END IF

         IF (AC.EQ.'0') STRING(J:J) = ' '
      END DO

 100  CONTINUE

      ! Reached first non-zero or non-blank or non-sign character.

      IF (J.EQ.1) RETURN

      IF (INSERT_SIGN) THEN
         K = J - 1
         STRING(K:K) = SC
      END IF

      RETURN
      END

C
C --------------------------------------------------------------------------
C
C     LEFT_TRIM
C
C     Purpose: Trim blank spaces on the left side of a character string.
C     Syntax: CALL LEFT_TRIM (STRING)
C     Input Parameters: 
C       STRING -- character substring.
C     Output Parameters:
C       STRING -- left trimmed string. Vacant positions on the right
C                 are filled with blanks.
C     Author: K. Myneni
C     Date: 5 Feb 90
C
      SUBROUTINE LEFT_TRIM (STRING)
                          
      CHARACTER*(*) STRING

      J = NLTRIM(STRING)
      IF (J.EQ.1) RETURN
      CALL SCROLL_STRING(STRING, -(J-1), ' ')

      RETURN
      END
C
C --------------------------------------------------------------------
C
C     RIGHT_TRIM
C
C     Purpose: Trim blank spaces on the right of a character string.
C     Syntax: CALL RIGHT_TRIM (STRING)
C     Input Parameters: 
C       STRING -- character substring
C     Output Parameters:
C       STRING -- right trimmed string
C     Author: K. Myneni
C     Date: 5 Feb 90
C
      SUBROUTINE RIGHT_TRIM (STRING)

      CHARACTER*(*) STRING
      
      L = LEN(STRING)
      J = NRTRIM(STRING)
      IF (J.EQ.1) RETURN
      CALL SCROLL_STRING(STRING, L - J, ' ')

      RETURN
      END
C
C -----------------------------------------------------------------
C			
C     FILL_STRING
C
C     Purpose: Fill a character substring with specified character.
C              This is useful for clearing a string by filling it
C              with blank characters.
C     Syntax: CALL FILL_STRING (STRING, AC)
C     Input Parameters:
C       AC -- character used to fill string.
C     Output Parameters:
C       STRING -- filled string.
C     Author: K. Myneni
C     Date: 13 Feb 90  
C     
      SUBROUTINE FILL_STRING (STRING, AC)

      CHARACTER*(*) STRING
      CHARACTER AC

      DO J = 1, LEN(STRING)
	 STRING(J:J) = AC
      END DO

      RETURN
      END
C
C ---------------------------------------------------------------------
C        
