/*
XYPLOT.CPP

Copyright (c) 1995--1999 Krishna Myneni and John Kielkopf 
Creative Consulting for Research & Education

Provided under the terms of the GNU General Public License.

Written by:
        Krishna Myneni
Contributions by:
        John Kielkopf
        Bryan Frazar

Created: 2-1-94
Last Update: 4-11-2000

System: Linux/X Windows
*/

#include <math.h>
#include <stdlib.h>
#include <unistd.h>
#include "fbc.h"
#include "ForthCompiler.h"
#include "ForthVM.h"
#include "xyplot.h"

void ExitCB (Widget, caddr_t, XmAnyCallbackStruct*);
void ReDrawCB (Widget, caddr_t, XmAnyCallbackStruct*);
void FileMenuCB (Widget, caddr_t, XmAnyCallbackStruct*);
void DataMenuCB (Widget, caddr_t, XmAnyCallbackStruct*);
void PlotMenuCB (Widget, caddr_t, XmAnyCallbackStruct*);
void MathMenuCB (Widget, caddr_t, XmAnyCallbackStruct*);
void AboutCB (Widget, caddr_t, XmAnyCallbackStruct*);
void ForthCB (Widget, caddr_t, XmAnyCallbackStruct*);
void InputCB (Widget, caddr_t, XmAnyCallbackStruct*);

void CoordinatesEH (Widget, XtPointer, XEvent*, Boolean*);
void ButtonEH (Widget, XtPointer, XEvent*, Boolean*);
void KeysEH (Widget, XtPointer, XEvent*, Boolean*);

extern int* GlobalSp;
extern byte* GlobalTp;

Widget TopLevel;

extern Widget FileWidgets[];
extern Widget EditWidgets[];
extern Widget DataWidgets[];
extern Widget PlotWidgets[];
extern Widget MathWidgets[];
extern Widget HelpWidgets[];

int debug = False;                   // global variable for debug mode
volatile int InputData;

// Forth interface constant function pointers

const void* FN_GET_ACTIVE = get_active;
const void* FN_GET_OPERAND = get_operand;
const void* FN_GET_DS = get_ds;
const void* FN_SET_DS_EXTREMA = set_ds_extrema;
const void* FN_MAKE_DS = make_ds;
const void* FN_GET_PLOT = get_plot;
const void* FN_MAKE_PLOT = make_plot;
const void* FN_SET_PLOT_SYMBOL = set_plot_symbol;
const void* FN_SET_PLOT_COLOR = set_plot_color;
const void* FN_DRAW_PLOT = draw_plot;
const void* FN_SET_GRID_TICS = set_grid_tics;
const void* FN_SET_GRID_LINES = set_grid_lines;
const void* FN_CLEAR_WINDOW = clear_window;
const void* FN_DRAW_WINDOW = draw_window;
const void* FN_RESET_WINDOW = reset_window;
const void* FN_GET_WINDOW_LIMITS = get_window_limits;
const void* FN_SET_WINDOW_LIMITS = set_window_limits;
const void* FN_ADD_MENU_ITEM = add_menu_item;
const void* FN_SET_BACKGROUND = set_background;
const void* FN_SET_FOREGROUND = set_foreground;
const void* FN_GET_INPUT = get_input;

vector<char*> ForthMenuCommandList;

CPlotWindow* pMainWnd;

void main(int argc, char* argv[])
{

  TopLevel = XtInitialize(argv[0],"",NULL,0,&argc,argv);
  XtRealizeWidget(TopLevel);
  
  // Create the main window

  pMainWnd = new CPlotWindow(argc, argv);

  // Initialize the Forth environment

  OpenForth();       
  InitForthInterface();

  // Load initialization file (xyplot.4th in user's .xyplot directory)

  LoadInitializationFile();

  // Load any files specified on the command line

  vector<char*> FileList = GetStartupFileList(argc, argv);
  if (FileList.size())
    {
        vector<char*>::iterator i;

        for (i = FileList.begin(); i < FileList.end(); i++)
        {
            pMainWnd->LoadFile(*i);
        }
    }
  else
    pMainWnd->WriteStatusMessage ("Ready!");

  // Setup callbacks and event handlers

  XtAddCallback (pMainWnd->m_nPlotWindow, XmNexposeCallback, 
    ReDrawCB, PL_REDRAW);
  XtAddCallback (pMainWnd->m_nPlotWindow, XmNresizeCallback,
    ReDrawCB, PL_RESIZE);
  XtAddEventHandler (pMainWnd->m_nPlotWindow, PointerMotionMask,
    False, CoordinatesEH, NULL);
  XtAddEventHandler (pMainWnd->m_nPlotWindow, ButtonPressMask,
    False, ButtonEH, NULL);
  XtAddEventHandler (pMainWnd->m_nPlotWindow, KeyPressMask,
    False, KeysEH, NULL);
  XtAddCallback (pMainWnd->m_nFileOpenDialog, XmNokCallback,
    FileMenuCB, PL_FILE_OPEN);
  XtAddCallback (pMainWnd->m_nFileOpenDialog, XmNcancelCallback,
    FileMenuCB, PL_FILE_CANCEL);
  XtAddCallback (pMainWnd->m_nFileSaveDialog, XmNokCallback,
    FileMenuCB, PL_FILE_SAVE);
  XtAddCallback (pMainWnd->m_nFileSaveDialog, XmNcancelCallback,
    FileMenuCB, PL_SAVE_CANCEL);
  XtAddCallback (HelpWidgets[ID_ABOUT], XmNactivateCallback,
    AboutCB, NULL);
  XtAddCallback (FileWidgets[ID_FILE_NEW],  XmNactivateCallback, 
    FileMenuCB, PL_NEW);
  XtAddCallback (FileWidgets[ID_FILE_OPEN], XmNactivateCallback,
    FileMenuCB, PL_OPEN);
  XtAddCallback (FileWidgets[ID_FILE_SAVE], XmNactivateCallback,
    FileMenuCB, PL_SAVE);
  XtAddCallback (FileWidgets[ID_FILE_PRINT], XmNactivateCallback,
    FileMenuCB, PL_PRINT);
  XtAddCallback (FileWidgets[ID_APP_EXIT], XmNactivateCallback,
    ExitCB, NULL);
  XtAddCallback (DataWidgets[ID_DATA_HDR], XmNactivateCallback,
    DataMenuCB, PL_HEADER);
  XtAddCallback (pMainWnd->m_nHeaderMessage, XmNokCallback,
    DataMenuCB, PL_HEADER_OK);
  XtAddCallback (pMainWnd->m_nHeaderMessage, XmNcancelCallback,
    DataMenuCB, PL_HEADER_CANCEL);
  XtAddCallback (DataWidgets[ID_DATA_DEL], XmNactivateCallback,
    DataMenuCB, PL_DELETE);
  XtAddCallback (DataWidgets[ID_DATA_DUP], XmNactivateCallback,
    DataMenuCB, PL_DUPLICATE);
  XtAddCallback (PlotWidgets[ID_PLOT_PICK], XmNactivateCallback,
    PlotMenuCB, PL_PICK);
  XtAddCallback (pMainWnd->m_nPickDialog, XmNokCallback,
    PlotMenuCB, PL_PICK_DATA);
  XtAddCallback (PlotWidgets[ID_PLOT_EXTREMA], XmNactivateCallback,
    PlotMenuCB, PL_EXTREMA_INPUT);
  XtAddCallback (PlotWidgets[ID_PLOT_DROP], XmNactivateCallback,
    PlotMenuCB, PL_DROP);
  XtAddCallback (MathWidgets[ID_MATH_EXPRESSION], XmNactivateCallback,
    MathMenuCB, PL_EXPRESSION_INPUT);

  XtAddCallback (pMainWnd->m_nForthShell, XmNvalueChangedCallback,
    ForthCB, NULL);

  // Disable menu functions not yet implemented

  XtSetSensitive (FileWidgets[ID_SAVE_OPTS], False);
  XtSetSensitive (EditWidgets[ID_EDIT_COPY], False);
  XtSetSensitive (PlotWidgets[ID_PLOT_GRID], False);
  XtSetSensitive (PlotWidgets[ID_PLOT_COLOR], False);
  XtSetSensitive (PlotWidgets[ID_PLOT_SYMBOL], False);

  XtMainLoop();

}
//------------------------------------------------------------------------

void ExitCB (Widget w, caddr_t client_data,
  XmAnyCallbackStruct* call_data)
{
  delete pMainWnd;
  CloseForth();

  vector<char*>* v = &ForthMenuCommandList;
  v->erase(v->begin(), v->end());

  exit(0);
}

void ReDrawCB (Widget w, caddr_t client_data, 
  XmAnyCallbackStruct* call_data)
{
  int i = (int) client_data;
  switch (i)
    {
    case PL_REDRAW:
      pMainWnd->OnPaint();
      break;
    case PL_RESIZE:
      pMainWnd->OnResize();
      break;
    default:
      ;
    }
}
//--------------------------------------------------------------

void CoordinatesEH (Widget w, XtPointer client_data,
  XEvent* event, Boolean* continue_to_dispatch)
{
  CPoint p(event->xmotion.x, event->xmotion.y);
  pMainWnd->OnMouseMove(0, p);
}

void ButtonEH (Widget w, XtPointer client_data,
  XEvent* event, Boolean* continue_to_dispatch)
{
  CPoint p(event->xmotion.x, event->xmotion.y);
  pMainWnd->OnLButtonDown(0, p);
}

void KeysEH (Widget w, XtPointer client_data,
  XEvent* event, Boolean* continue_to_dispatch)
{
  KeySym key;
  XComposeStatus compose;
  char cl [64];

  if (event->type == KeyPress)
    {
      XLookupString (&event->xkey, cl, 64, &key, &compose);
      switch (key)
	{
	case 'D': case 'd':
	  pMainWnd->OnDrop();
	  break;
	case 'E': case 'e':
	  pMainWnd->OnExtremaInput();
	  break;
	case 'F': case 'f':
	  pMainWnd->OnFileOpen();
	  break;
	case 'H': case 'h':
	  pMainWnd->OnHeader();
	  break;
	case 'I': case 'i':
	  pMainWnd->OnPick();
	  break;
	case 'R': case 'r':
	  pMainWnd->OnExpressionInput();;
	  break;
	case XK_Escape:
	  pMainWnd->OnDeleteView();
	  break;
	case XK_period:
	  pMainWnd->OnSymbolToggle();
	  break;
	case XK_F9:
	  pMainWnd->OnFileSave();
	  break;
	default:
	  ;
	}
    }  
}
//--------------------------------------------------------------

void AboutCB (Widget w, caddr_t client_data,
  XmAnyCallbackStruct* call_data)
{
  pMainWnd->OnAbout();
}
//-------------------------------------------------------------

void FileMenuCB (Widget w, caddr_t client_data,
  XmAnyCallbackStruct* call_data)
{
  XmFileSelectionBoxCallbackStruct* sel;
  XmSelectionBoxCallbackStruct* sel2;
  char* filename;
  char* s;
  int i = (int) client_data;

  switch (i)
    {
    case PL_NEW:
      pMainWnd->OnFileNew();
      break;
    case PL_OPEN:
      pMainWnd->OnFileOpen();
      break;
    case PL_FILE_OPEN:
      XtUnmanageChild(pMainWnd->m_nFileOpenDialog);
      sel = (XmFileSelectionBoxCallbackStruct *) call_data;
      XmStringGetLtoR(sel->value, XmSTRING_DEFAULT_CHARSET, &filename);
      strcpy (pMainWnd->m_pFileName, filename);
      pMainWnd->LoadFile(filename);
      break;
    case PL_FILE_LOAD:
      XtUnmanageChild (pMainWnd->m_nInputDialog);
      sel2 = (XmSelectionBoxCallbackStruct *) call_data;
      XmStringGetLtoR(sel2->value, XmSTRING_DEFAULT_CHARSET, &s);
      if (*(pMainWnd->m_pFileName))
	pMainWnd->LoadDatasetFile(pMainWnd->m_pFileName, s);
      break;
    case PL_LOAD_CANCEL:
      XtUnmanageChild(pMainWnd->m_nInputDialog);
      break;
    case PL_FILE_CANCEL:
      XtUnmanageChild(pMainWnd->m_nFileOpenDialog);
      break;
    case PL_SAVE:
      pMainWnd->OnFileSave();
      break;
    case PL_FILE_SAVE:
      XtUnmanageChild (pMainWnd->m_nFileSaveDialog);
      sel = (XmFileSelectionBoxCallbackStruct *) call_data;
      XmStringGetLtoR(sel->value, XmSTRING_DEFAULT_CHARSET, &filename);
      pMainWnd->SaveFile(filename);
      break;
    case PL_SAVE_CANCEL:
      XtUnmanageChild (pMainWnd->m_nFileSaveDialog);
      break;
    case PL_PRINT:
      pMainWnd->OnPrint();
      break;
    default:
      break;
    }
}
//------------------------------------------------------------------

void DataMenuCB (Widget w, caddr_t client_data,
  XmAnyCallbackStruct* call_data)
{
  int i = (int) client_data;

  switch (i)
    {
    case PL_HEADER:
      pMainWnd->OnHeader();
      break;
    case PL_HEADER_OK:
      XtUnmanageChild(pMainWnd->m_nHeaderDialog);
      pMainWnd->OnSetHeader(XmTextGetString(pMainWnd->m_nHeaderText));
      break;
    case PL_HEADER_CANCEL:
      XtUnmanageChild(pMainWnd->m_nHeaderDialog);
      break;
    case PL_DELETE:
      pMainWnd->OnDelete();
      break;
    case PL_DUPLICATE:
      pMainWnd->OnDuplicate();
      break;
    default:
      break;
    }
}
//-------------------------------------------------------------

void PlotMenuCB (Widget w, caddr_t client_data,
  XmAnyCallbackStruct* call_data)
{
  char *s;
  XmSelectionBoxCallbackStruct *selection;

  int i = (int) client_data;
  switch(i)
    {
    case PL_EXTREMA_INPUT:
      pMainWnd->OnExtremaInput();
      break;
    case PL_EXTREMA:
      selection=(XmSelectionBoxCallbackStruct *) call_data;
      XmStringGetLtoR(selection->value, XmSTRING_DEFAULT_CHARSET, &s);
      XtRemoveCallback (pMainWnd->m_nInputDialog, XmNokCallback,
        PlotMenuCB, PL_EXTREMA);
      XtRemoveCallback (pMainWnd->m_nInputDialog, XmNcancelCallback,
	PlotMenuCB, PL_EXTREMA_CANCEL);
      XtUnmanageChild (pMainWnd->m_nInputDialog);
      pMainWnd->OnExtrema(s);
      break;
    case PL_EXTREMA_CANCEL:
      XtRemoveCallback (pMainWnd->m_nInputDialog, XmNokCallback,
        PlotMenuCB, PL_EXTREMA);
      XtRemoveCallback (pMainWnd->m_nInputDialog, XmNcancelCallback,
	PlotMenuCB, PL_EXTREMA_CANCEL);
      XtUnmanageChild (pMainWnd->m_nInputDialog);
      break;
    case PL_DROP:
      pMainWnd->OnDrop();
      break;
    case PL_PICK:
      pMainWnd->OnPick();
      break;
    case PL_PICK_DATA:
      pMainWnd->OnPickData();
      break;
    default:
      break;
    }
}
//----------------------------------------------------------------

void MathMenuCB (Widget w, caddr_t client_data,
  XmAnyCallbackStruct* call_data)
{

  char *s;
  XmSelectionBoxCallbackStruct *selection;

  int i = (int) client_data;
  switch(i)
    {
    case PL_EXPRESSION_INPUT:
      pMainWnd->OnExpressionInput();
      break;
    case PL_EXPRESSION:
      selection=(XmSelectionBoxCallbackStruct *) call_data;
      XmStringGetLtoR(selection->value, XmSTRING_DEFAULT_CHARSET, &s);
      XtRemoveCallback (pMainWnd->m_nInputDialog, XmNokCallback, 
	MathMenuCB, PL_EXPRESSION);
      XtRemoveCallback (pMainWnd->m_nInputDialog, XmNcancelCallback, 
	MathMenuCB, PL_EXPRESSION_CANCEL);  
      XtUnmanageChild (pMainWnd->m_nInputDialog);
      pMainWnd->OnExpression(s);
      break;
    case PL_EXPRESSION_CANCEL:
      XtRemoveCallback (pMainWnd->m_nInputDialog, XmNokCallback, 
	MathMenuCB, PL_EXPRESSION);
      XtRemoveCallback (pMainWnd->m_nInputDialog, XmNcancelCallback, 
	MathMenuCB, PL_EXPRESSION_CANCEL);  
      XtUnmanageChild (pMainWnd->m_nInputDialog);
      break;
    default:
      break;
    }
}
//---------------------------------------------------------------

void ForthCB (Widget wSrc, caddr_t client_data,
  XmAnyCallbackStruct* call_data)
{
  Widget w = pMainWnd->m_nForthShell;
  char s[256];
  int nError, nLine, nEsrc, nPos, nLen;
  *s = '\0';

  nPos = XmTextGetLastPosition(w);
  static bool ignore = False;

  if (ignore)
    {
      ignore = False;
      return;
    }

  char* fs = XmTextGetString(w); 

  if (client_data)
    {
      strcpy (s, (char*) client_data);
    }
  else
    {
      if (nPos)
	{
	  --nPos;
	  if (fs[nPos] == '\n')  // was Enter pressed?
	    {
	      char* cp = fs + nPos;
	      while (nPos)              // find the beginning of this line
		{
		  --cp; --nPos;
		  if (*cp == '\n') break;	      
		}
	      strcpy (s, cp);           // copy the last line and execute
	    }
	  else
	    {
	      return;
	    }
	}
      else
	return;
    }

  char out_s[256];
  memset (out_s, 0, 256);
  ostrstream* pSS = new ostrstream(out_s, 255);
  nEsrc = ExecuteForthExpression (s, pSS, &nError, &nLine);
  delete pSS;
  if (nEsrc == 1)
      sprintf (s, "Compiler Error %d\n", nError);
  else if (nEsrc == 2)
      sprintf (s, "VM Error %d\n", nError);
  else
      strcpy (s, " ok\n");

  strcat (out_s, s);
  nLen = strlen(fs);
  XtFree(fs);
  ignore = True;

  XmTextInsert (w, nLen, out_s);

  nPos = XmTextGetLastPosition (w);
  XmTextShowPosition(w, nPos);
  XmTextSetInsertionPosition(w, nPos);

} 
//---------------------------------------------------------------

int BlankLine (char* s)
{
// Return TRUE if the line is blank

    char* p = s;
    while (*p != '\0')
    {
        if ((*p != ' ') && (*p != '\t') && (*p != '\n')) return 0;
        ++p;
    }

    return 1;
}
//---------------------------------------------------------------

int NumberParse (float* nums, char* s)
{
// Parse a string into array of floats; Return number of values

    const char d1 = 44; // comma
    const char d2 = 9;  // tab
    const char d3 = 32; // space
    char* p = s, *d, *ep = s + strlen(s);

    if (*s == '\0' || *s == '\r' || *s == 27) return 0;
	if (BlankLine(s)) return 0;

	int n = 0;

	do
	{
      d = strchr (p, d1);
      if (!d)
      {
        d = strchr (p, d2);
        if (!d)
        {
          d = strchr (p, d3);
        }
      }

	  if (!d)       // last delimited value
	  {
        nums [n] = atof (p);
        ++n;
        break;
	  }

      if (d == p)
      {
        if (*p == d1 || *p == d2) ++n;
        ++p;
      }
      else
      {
        nums [n] = atof (p);
        ++n;
        p = d + 1;
      }

	} while (p < ep) ;

    return n;

}
//---------------------------------------------------------------

void SortRect (CRect* prect)
{
	CPoint p1 = prect->TopLeft();
	CPoint p2 = prect->BottomRight();

	int i;

	if (p2.x < p1.x)
	{
		i = p1.x;
		p1.x = p2.x;
		p2.x = i;
	}
	if (p2.y < p1.y)
	{
		i = p1.y;
		p1.y = p2.y;
        p2.y = i;
	}
	prect->SetRect (p1.x, p1.y, p2.x, p2.y);
}
//---------------------------------------------------------------------

char* LabelFormat (float x1, float x2, char axis)
{
// Determine label output format for range of numbers between x1 and x2.

    static char format[255];

    float diff = fabs (x2 - x1);
    int j = (x1 < 0.) || (x2 < 0.);

	switch (axis)
	{
	  case 'Y':
        if ((diff < .001f) || (diff >= 100000.f))
          strcpy (format, "%9.1e");
        else if ((diff >= .001f) && (diff < 1.f))
          strcpy (format, "%6.4f");
        else if ((diff >= 1.f) && (diff < 10.f))
          strcpy (format, "%6.3f");
        else if ((diff >= 10.f) && (diff < 100.f))
          strcpy (format, "%6.2f");
        else if ((diff >= 100.f) && (diff < 1000.f))
          strcpy (format, "%6.1f");
        else
          strcpy (format, "%6.0f");
	    break;
	  case 'X':
        strcpy (format, DisplayFormat(x1, x2));
        break;
      default:
        ;
    }

    return format;

}
//----------------------------------------------------------------

char* DisplayFormat (float x1, float x2)
{
// Determine display format for range of numbers between x1 and x2.

    static char format[255];
    float diff = fabs ((x2 - x1));
    float a1 = fabs (x1);
    float a2 = fabs (x2);
    float vmax = (a2 > a1) ? a2 : a1;
    int field_width = 6;
    int precision = 4;

	if ((diff < .01f) || (diff >= 100000.f))
	  field_width += 6;
	else if ((diff >= .01f) && (diff < .1f))
	  ++precision;
	else if ((diff >= .1f) && (diff < 1.f))
	  ;
	else if ((diff >= 1.f) && (diff < 10.f))
	  --precision;
	else if ((diff >= 10.f) && (diff < 100.f))
	  precision -= 2;
	else if ((diff >= 100.f) && (diff < 1000.f))
	  precision -= 3;
	else
	  {
	    field_width = 8;
	    precision = 0;
	  }

    if (diff != 0.)
    {
        float a3 = vmax/diff;
        if (a3 >= 10. && a3 <= 1.e6)
        {
            int i = (int) (log(a3)/log(10.)) - 1;
            if ((x1 < 0.) || (x2 < 0.)) i++;
            if (i < 10) field_width += i;;
	}
    }
    
    if (field_width > 8)
      sprintf (format, "%c%2d.%1de", '%', field_width, precision);
    else
      sprintf (format, "%c%1d.%1df", '%', field_width, precision);

    return format;
}
//--------------------------------------------------------------

void LoadInitializationFile ()
{
// Load the Forth file xyplot.4th, residing in the
// user's HOME/.xyplot directory. This file can
// load other forth files containing initialization scripts.

  char start_dir [512];
  getcwd (start_dir, 511);

  char* home_dir = getenv("HOME");
  char path[512];
  strcpy (path, home_dir);
  strcat (path, "/.xyplot");
  if (chdir(path) == 0)
    {
      LoadForthFile("xyplot.4th");
    }
  chdir(start_dir);

}    
//--------------------------------------------------------------

vector<char*> GetStartupFileList (int argc, char* argv[])
{
// Parse the command line into a vector of filenames.
//  The first element in the vector is the program pathname.

    vector<char*> filelist;

    int i = 1;

    while (i < argc)
      {
	if (*argv[i] != '-') filelist.push_back(argv[i]);
	i++;
      }

    return filelist;
}
//--------------------------------------------------------------

int AddToHeader (char* text, char* hdr, bool prefix)
{
  // Add text to the beginning of a header string
  // Return non-zero if length exceeded.

  int ecode = 0;
  int new_length = strlen(text) + strlen(hdr);
  if (prefix) new_length += strlen(ANNOTATION_PREFIX);

  if (new_length < HEADER_LENGTH)
    {
      char* temp_str = new char[HEADER_LENGTH];
      if (prefix) 
	{
	  strcpy (temp_str, ANNOTATION_PREFIX);
	  strcat (temp_str, text);
	}
      else
	{
	  strcpy (temp_str, text);
	}
      strcat (temp_str, hdr);
      strcpy (hdr, temp_str);
      delete [] temp_str;
    }
  else
    {
      ecode = -1;  // new header will exceed the HEADER_LENGTH
    }

  return ecode;
}
//-------------------------------------------------------------

int ExecuteForthExpression (char* s, ostrstream* pOutput, int* pError, int* pLine)
{
  // Return zero if no error, 1 if compiler error, 2 if VM error

  istrstream* pSS = new istrstream (s);
  int* sp;
  byte* tp;
  vector<byte> op;

  *pLine = 0;

  SetForthInputStream(*pSS);
  SetForthOutputStream(*pOutput);
  *pError = ForthCompiler (&op, pLine);
  delete pSS;

  if (*pError) return 1;

  if (op.size())
    {
      SetForthInputStream(cin);
      *pError = ForthVM (&op, &sp, &tp);
      if (*pError) return 2;
    }

  return 0;
}
//-------------------------------------------------------------

void InitForthInterface ()
{
  char fs[4096], s[64];
  char* cs = " constant ";
  int esrc, ec, lnum;

  sprintf (s, "%lu%sFN_GET_ACTIVE\n", FN_GET_ACTIVE, cs);
  strcpy (fs, s);
  sprintf (s, "%lu%sFN_GET_OPERAND\n", FN_GET_OPERAND, cs);
  strcat (fs, s);
  sprintf (s, "%lu%sFN_GET_DS\n", FN_GET_DS, cs);
  strcat (fs, s);
  sprintf (s, "%lu%sFN_SET_DS_EXTREMA\n", FN_SET_DS_EXTREMA, cs);
  strcat (fs, s);
  sprintf (s, "%lu%sFN_MAKE_DS\n", FN_MAKE_DS, cs);
  strcat (fs, s);
  sprintf (s, "%lu%sFN_GET_PLOT\n", FN_GET_PLOT, cs);
  strcat (fs, s);
  sprintf (s, "%lu%sFN_MAKE_PLOT\n", FN_MAKE_PLOT, cs);
  strcat (fs, s);
  sprintf (s, "%lu%sFN_SET_PLOT_SYMBOL\n", FN_SET_PLOT_SYMBOL, cs);
  strcat (fs, s);
  sprintf (s, "%lu%sFN_SET_PLOT_COLOR\n", FN_SET_PLOT_COLOR, cs);
  strcat (fs, s);
  sprintf (s, "%lu%sFN_DRAW_PLOT\n", FN_DRAW_PLOT, cs);
  strcat (fs, s);
  sprintf (s, "%lu%sFN_SET_GRID_TICS\n", FN_SET_GRID_TICS, cs);
  strcat (fs, s);
  sprintf (s, "%lu%sFN_SET_GRID_LINES\n", FN_SET_GRID_LINES, cs);
  strcat (fs, s);
  sprintf (s, "%lu%sFN_CLEAR_WINDOW\n", FN_CLEAR_WINDOW, cs);
  strcat (fs, s);
  sprintf (s, "%lu%sFN_DRAW_WINDOW\n", FN_DRAW_WINDOW, cs);
  strcat (fs, s);
  sprintf (s, "%lu%sFN_RESET_WINDOW\n", FN_RESET_WINDOW, cs);
  strcat (fs, s);
  sprintf (s, "%lu%sFN_GET_WINDOW_LIMITS\n", FN_GET_WINDOW_LIMITS, cs);
  strcat (fs, s);
  sprintf (s, "%lu%sFN_SET_WINDOW_LIMITS\n", FN_SET_WINDOW_LIMITS, cs);
  strcat (fs, s);
  sprintf (s, "%lu%sFN_ADD_MENU_ITEM\n", FN_ADD_MENU_ITEM, cs);
  strcat (fs, s);
  sprintf (s, "%lu%sFN_SET_BACKGROUND\n", FN_SET_BACKGROUND, cs);
  strcat (fs, s);
  sprintf (s, "%lu%sFN_SET_FOREGROUND\n", FN_SET_FOREGROUND, cs);
  strcat (fs, s);
  sprintf (s, "%lu%sFN_GET_INPUT\n", FN_GET_INPUT, cs);
  strcat (fs, s);
  sprintf (s, "%lu%sMN_FILE\n", pMainWnd->m_nFileMenu, cs);
  strcat (fs, s);
  sprintf (s, "%lu%sMN_EDIT\n", pMainWnd->m_nEditMenu, cs);
  strcat (fs, s);
  sprintf (s, "%lu%sMN_DATA\n", pMainWnd->m_nDataMenu, cs);
  strcat (fs, s);
  sprintf (s, "%lu%sMN_PLOT\n", pMainWnd->m_nPlotMenu, cs);
  strcat (fs, s); 
  sprintf (s, "%lu%sMN_MATH\n", pMainWnd->m_nMathMenu, cs);
  strcat (fs, s);
  sprintf (s, "%lu%sMN_HELP\n", pMainWnd->m_nHelpMenu, cs);
  strcat (fs, s);

  char out_s[256];
  memset (out_s, 0, 256);
  ostrstream* pSS = new ostrstream(out_s, 255);

  esrc = ExecuteForthExpression(fs, pSS, &ec, &lnum);
  if (esrc) 
    {
      pMainWnd->MessageBox(out_s);
    }
  delete pSS;
}
//-----------------------------------------------------------------

int LoadForthFile(char* fname)
{
  char s[256], out_s[256];
  int esrc, ec, lnum;

  strcpy (s, "include ");
  strcat (s, fname);

  memset (out_s, 0, 256);
  ostrstream* pSS = new ostrstream(out_s, 255);
  esrc = ExecuteForthExpression (s, pSS, &ec, &lnum);

  if (esrc)
    {
      pMainWnd->MessageBox (out_s);
    }
  delete pSS;

  return ec;
}

//----------------------------------------------------------------
// WARNING:
//
//  DO NOT CALL the following functions from C, C++, or
//  assembly code. They are intended to be called from the
//  Forth environment only.
//
//----------------------------------------------------------------

int get_active ()
{
  // Return the active set index number on top of the Forth stack
  // Return a -1 if error.

  CDataset* d = pMainWnd->m_pDi->GetActiveSet();
  CDatabase* pDb = pMainWnd->m_pDb;

  int nSet = pDb->IndexOf(d);

  // Place the set number on top of the stack

  *GlobalSp = nSet;
  --GlobalSp;
  *GlobalTp = OP_IVAL;
  --GlobalTp;

  return 0;
}
//-----------------------------------------------------------------

int get_operand ()
{
  // Return the operand set index on top of the Forth stack; -1 if error

  CDataset* d = pMainWnd->m_pDi->GetOperandSet();
  CDatabase* pDb = pMainWnd->m_pDb;
  int nSet = pDb->IndexOf(d);
  *GlobalSp-- = nSet; *GlobalTp-- = OP_IVAL;
  return 0;
}
//-----------------------------------------------------------------

int get_ds ()
{
  // Return information about the specified dataset

  ++GlobalSp; ++GlobalTp;
  if (*GlobalTp == OP_ADDR)
    {
      // obtain the pointer to the dataset info structure
      unsigned int* ds_info = *((int**)GlobalSp);

      ++GlobalSp;
      int nSet = *GlobalSp;
      CDatabase* pDb = pMainWnd->m_pDb;
      int n = pDb->Nsets();

      if (nSet >= 0 && nSet < n)
	{
	  // Fill in the dataset info structure
	  CDataset* d = (*pDb)[nSet];
	  *ds_info++ = (unsigned int) d->m_szName;
	  *ds_info++ = (unsigned int) d->m_szHeader;
	  *ds_info++ = 0;  // set zero for the dataset type
	  *ds_info++ = d->NumberOfPoints();
	  *ds_info++ = d->SizeOfDatum();
	  *ds_info = (unsigned int) d->begin();
	  *GlobalSp = 0;
	}
      else
	{
	  *GlobalSp = -1;
	}
      --GlobalSp;
    }
  else
    {
      pMainWnd->MessageBox("Invalid parameter for get_ds");
    }
  return 0;
}
//-----------------------------------------------------------------

int set_ds_extrema ()
{
  // Reset extrema for specified dataset

  ++GlobalSp; ++GlobalTp;
  int i = *GlobalSp;
  if (i >= 0 && i < pMainWnd->m_pDb->Nsets())
    {
      CDataset* d = (*(pMainWnd->m_pDb))[i];
      if (d) d->SetExtrema();
    }
  return 0;
}
//-----------------------------------------------------------------

int make_ds ()
{
  // Make a dataset from the dataset info structure passed
  // on top of the Forth stack.

  char s[256];

  ++GlobalSp; ++GlobalTp;
  if (*GlobalTp == OP_ADDR)
    {
      int* ds_info = *((int**) GlobalSp);
      char* name = *((char**) ds_info);
      CDataset* d = pMainWnd->m_pDb->MakeDataset(ds_info);
      if (d)
	{
	  sprintf (s, "Created new dataset %s", name);
	  pMainWnd->WriteStatusMessage(s);
	}
      else
	{
	  sprintf (s, "Cannot create dataset %s", name);
	  pMainWnd->MessageBox(s);
	}
    }
  else
    {
      pMainWnd->MessageBox("Invalid parameter for make_ds");
    }
  return 0;
}
//-----------------------------------------------------------------

int get_plot ()
{
  // Return information about the specified plot

  ++GlobalSp; ++GlobalTp;
  if (*GlobalTp == OP_ADDR)
    {
      int* pl_info = *((int**)GlobalSp);
      ++GlobalSp;
      int nPlot = *GlobalSp;
      CPlotDisplay* pDi = pMainWnd->m_pDi;
      CDatabase* pDb = pMainWnd->m_pDb;
      int n = pDi->Nplots();

      if (nPlot >= 0 && nPlot < n)
	{
	  // Fill in the plot info structure
	  CPlot* p = (*pDi)[nPlot];
	  if (p)
	    {
	      CDataset* d = p->GetSet();
	      int nSet = pDb->IndexOf(d);
	      *pl_info++ = nSet;  // dataset index number
	      *pl_info++ = 0;     // set zero for the plot type
	      *pl_info++ = p->GetSymbol();  // plot symbol
	      *pl_info = p->GetColor();   // plot color
	      *GlobalSp = 0;
	    }
	  else
	    *GlobalSp = -1;
	}
      else
	{
	  *GlobalSp = -1;
	}
      --GlobalSp;
    }
  else
    {
      pMainWnd->MessageBox("Invalid parameter for get_plot");
    }
  return 0;
} 
//-----------------------------------------------------------------

int make_plot ()
{
  // Make a new plot

  ++GlobalSp; ++GlobalTp;
  if (*GlobalTp == OP_ADDR)
    {
      CDatabase* pDb = pMainWnd->m_pDb;
      int* pl_info = *((int**) GlobalSp);
      int nSet = *pl_info++;
      if ((nSet >= 0) && (nSet < pDb->Nsets()))
	  {
	    int nType = *pl_info++;
	    CDataset* d = (*pDb)[nSet];
	    if (d)
	      {
		CPlot* p = pMainWnd->m_pDi->MakePlot(d, nType);
		if (p)
		  {
		    Symbol nSym = *pl_info++;
		    p->SetSymbol(nSym);
		    int nColor = *pl_info;
		    p->SetColor(nColor);
		  }
	      }
	  }
    }
  else
    {
      pMainWnd->MessageBox("Invalid parameter for make_plot");
    }
  return 0;
}
//------------------------------------------------------------------

int set_plot_symbol ()
{
  ++GlobalSp; ++GlobalTp;
  if (*GlobalTp == OP_ADDR)
    {
      char* symbol_name = *((char**)GlobalSp);
      ++symbol_name;
      Symbol nSym = pMainWnd->GetSymbol(symbol_name);
      pMainWnd->m_pDi->GetActivePlot()->SetSymbol(nSym);
      pMainWnd->Invalidate();
    }
  else
    {
      pMainWnd->MessageBox("Invalid parameter for set_plot_symbol");
    }
  return 0;
}
//------------------------------------------------------------------

int set_plot_color ()
{
  ++GlobalSp; ++GlobalTp;
  if (*GlobalTp == OP_ADDR)
    {
      char* color_name = *((char**)GlobalSp);
      ++color_name;
      unsigned c = pMainWnd->GetColor(color_name);
      pMainWnd->m_pDi->GetActivePlot()->SetColor(c);
      pMainWnd->Invalidate();
    }
  else
    {
      pMainWnd->MessageBox("Invalid parameter for set_plot_color");
    }
  return 0;
}
//------------------------------------------------------------------

int draw_plot ()
{
  ++GlobalSp; ++GlobalTp;
  int vis = *GlobalSp;
  CPlot* p = pMainWnd->m_pDi->GetActivePlot();
  unsigned c = p->GetColor();
  if (! vis)
    {
      unsigned bkg = pMainWnd->GetBackgroundColor();
      p->SetColor(bkg);
    }
  CDeviceContext* pDc = pMainWnd->GetDC();
  p->Draw(pDc);
  XFlush(XtDisplay(pMainWnd->m_nPlotWindow));
  XSync(XtDisplay(pMainWnd->m_nPlotWindow), False);
  p->SetColor(c);
  return 0;
}
//------------------------------------------------------------------

int set_grid_tics ()
{
  ++GlobalSp; ++GlobalTp;
  int ny = *GlobalSp++; ++GlobalTp;
  int nx = *GlobalSp;

  pMainWnd->m_pDi->m_pGrid->SetTics(nx, ny);
  pMainWnd->Invalidate();
  return 0;
}

int set_grid_lines ()
{
  ++GlobalSp; ++GlobalTp;
  bool by = *GlobalSp++; ++GlobalTp;
  bool bx = *GlobalSp;
  pMainWnd->m_pDi->m_pGrid->SetLines(bx, by);
  pMainWnd->Invalidate();
  return 0;
}

//------------------------------------------------------------------

int clear_window ()
{
  // Clear the plot window

  pMainWnd->Clear();
  return 0;
}

int reset_window ()
{
  // Reset the plot window

  pMainWnd->OnReset();
  return 0;
}

int draw_window ()
{
  // Force drawing of the plot window

  pMainWnd->Invalidate();
  pMainWnd->OnPaint();

  return 0;
}
//-----------------------------------------------------------------

int get_window_limits ()
{

  // Return x1, y1, x2, y2 on top of the stack

  vector<float> e = pMainWnd->m_pDi->GetExtrema();

  --GlobalSp; --GlobalTp;
  *((double*)GlobalSp) = (double) e[0];
  *GlobalTp = OP_FVAL;
  GlobalSp -= 2; GlobalTp -= 2;
  *((double*)GlobalSp) = (double) e[2];
  *GlobalTp = OP_FVAL;
  GlobalSp -= 2; GlobalTp -= 2;
  *((double*)GlobalSp) = (double) e[1];
  *GlobalTp = OP_FVAL;
  GlobalSp -= 2; GlobalTp -= 2;
  *((double*)GlobalSp) = (double) e[3];
  *GlobalTp = OP_FVAL;
  --GlobalSp; --GlobalTp;

  return 0;
}
//-----------------------------------------------------------------

int set_window_limits ()
{
  // Set the plot window limits

  vector<float> e(4);
  ++GlobalSp;
  e[3] = *((double*)GlobalSp);
  GlobalSp += 2;
  e[1] = *((double*)GlobalSp);
  GlobalSp += 2;
  e[2] = *((double*)GlobalSp);
  GlobalSp += 2;
  e[0] = *((double*)GlobalSp);
  ++GlobalSp;
  GlobalTp += 8;

  pMainWnd->m_pDi->CreateView(CARTESIAN, e);
  pMainWnd->Invalidate();

  return 0;
}
//---------------------------------------------------------------

int add_menu_item ()
{
  // Add item to the specified menu.
  // Stack: ( menu item_name forth_command -- )
  //   menu is the menu widget id (e.g. MN_MATH)
  //   item_name is the address of a Forth counted string
  //   forth_command is the address of a Forth counted string

  char name[256];
  char* emsg1 = "add_menu_item: Invalid parameter";
  unsigned char *cp;
  Widget menu, new_item;

  ++GlobalSp; ++GlobalTp;
  if (*GlobalTp != OP_ADDR)
    {
      pMainWnd->MessageBox(emsg1);
      return 0;
    }
  cp = *((unsigned char**) GlobalSp);
  int nCount = (int) *cp;

  char* fc = new char[256];  // allocate a new command string
                              
  strncpy (fc, cp+1, nCount);
  fc[nCount] = '\0';
  ++GlobalSp; ++GlobalTp;
  if (*GlobalTp != OP_ADDR)
    {
      pMainWnd->MessageBox(emsg1);
      return 0;
    }
  cp = *((unsigned char**) GlobalSp);
  nCount = (int) *cp;
  strncpy (name, cp+1, nCount);
  name[nCount] = '\0';
  ++GlobalSp; ++GlobalTp;
  menu = (Widget) *GlobalSp;
  
  //  if (menu == pMainWnd->m_nMathMenu)
  //  {
      new_item = pMainWnd->AddMenuItem (menu, name, MathWidgets);
      if (new_item)
	{
	  ForthMenuCommandList.push_back(fc);
	  XtAddCallback (new_item, XmNactivateCallback,
			 ForthCB, fc);
	}
  //  }
  // else
  //  {
  //    pMainWnd->MessageBox ("Unknown menu");
  //  }
  return 0;
}
//---------------------------------------------------------------

int set_background ()
{
  // Stack: ( addr -- | counted string contains color name )

  char color_name[256];

  ++GlobalSp; ++GlobalTp;
  if (*GlobalTp == OP_ADDR)
    {
      char* name = *((char**) GlobalSp);
      unsigned nLen = (unsigned) *name;
      ++name;
      strncpy (color_name, name, nLen);
      color_name[nLen] = '\0';
      pMainWnd->SetBackgroundColor(color_name);
    }
  else
    {
      pMainWnd->MessageBox("Invalid parameter for set_background");
    }

  return 0;
}
//----------------------------------------------------------------

int set_foreground ()
{
  // Stack: ( addr -- | counted string contains color name )

  char color_name[256];

  ++GlobalSp; ++GlobalTp;
  if (*GlobalTp == OP_ADDR)
    {
      char* name = *((char**) GlobalSp);
      unsigned nLen = (unsigned) *name;
      ++name;
      strncpy (color_name, name, nLen);
      color_name[nLen] = '\0';
      pMainWnd->SetForegroundColor(color_name);
    }
  else
    {
      pMainWnd->MessageBox("Invalid parameter for set_foreground");
    }

  return 0;
}
//----------------------------------------------------------------

int get_input ()
{
  ++GlobalSp; ++GlobalTp;
  if (*GlobalTp == OP_ADDR)
    {
      char* prompt = *((char**) GlobalSp);
      ++prompt;
      Arg al[4];
      int ac = 0;
      XtSetArg(al[ac], XmNselectionLabelString, 
	       XmStringCreateLtoR(prompt, 
				  XmSTRING_DEFAULT_CHARSET));  
      ac++;
      XtSetArg(al[ac], XmNtextString, XmStringCreateLtoR("",
							 XmSTRING_DEFAULT_CHARSET));
      ac++;
      XtSetValues(pMainWnd->m_nInputDialog, al, ac);

      XtAddCallback (pMainWnd->m_nInputDialog, XmNokCallback,
      	     InputCB, TRUE);
      XtAddCallback (pMainWnd->m_nInputDialog, XmNcancelCallback,
      	     InputCB, FALSE);
      XtManageChild (pMainWnd->m_nInputDialog);
      XtPopup(XtParent(pMainWnd->m_nInputDialog), XtGrabNone);

      InputData = FALSE;

      while ((!InputData) || XtPending()) 
	XtProcessEvent(XtIMAll);
    }
  else
    {
      pMainWnd->MessageBox("Invalid parameter for get_input");
    }

  return 0;
}
//----------------------------------------------------------------
//
// The following callback is for use by get_input only

void InputCB (Widget w, caddr_t client_data,
  XmAnyCallbackStruct* call_data)
{
  static char rs[64];  // counted response string

  int i = (int) client_data, n;
  switch (i)
    {
    case TRUE:
      // Ok button has been pressed

      XmSelectionBoxCallbackStruct *selection;
      char* s;
      selection=(XmSelectionBoxCallbackStruct *) call_data;
      XmStringGetLtoR(selection->value, XmSTRING_DEFAULT_CHARSET, &s);
      n = strlen(s);
      strcpy (rs + 1, s);
      *rs = n;
      break;

    case FALSE:
      // Cancel button has been pressed
      *rs = 0;
      break;
    default:
      ;
    }
  *GlobalSp-- = (int) rs;
  *GlobalTp-- = OP_ADDR;
  *GlobalSp-- = i;
  *GlobalTp-- = OP_IVAL;

  XtRemoveCallback (pMainWnd->m_nInputDialog, XmNokCallback,
        InputCB, TRUE);
  XtRemoveCallback (pMainWnd->m_nInputDialog, XmNcancelCallback,
	InputCB, FALSE);
  XtUnmanageChild (pMainWnd->m_nInputDialog);
  InputData = TRUE;
}
//---------------------------------------------------------------





